


/*
 *
 *
 * Explanation: State of slice registering:
 * 
 * 
 * 1. Find free slice
 * 2. Put in side_slices registry
 * 3. Put in fss->slice_bits (with BITS_UNSENT)
 * 
 * 4. Wait 
 * 
 * 5. send_queue will send something and clear BITS_UNSENT
 *     uids go to srr
 * 
 * 6. Wait
 * 
 * 7. upon successful reception at another station it will acknowledge the slice
 * 
 * 8. When receiving the acknowledge, we set BITS_SEEN
 * 
 * 
 * 
 * 
 */







#ifndef FS_SLICES_C
#define FS_SLICES_C

#include <stdint.h>
#include <stdbool.h>



#ifndef NOCODE

#define NOCODE
#include "FrameSwitch.c"
#include "multihw.c"
#undef FS_SLICES_C
#include "FS-slice-helper.c"
#define FS_SLICES_C
#undef NOCODE

#define i8 int8_t
#define u8 uint8_t

#else

#include "multihw.c"

#endif



bool slice_our_identity(MHI_ARG    uint8_t slice)
#ifndef NOCODE
{
	if (is_main_slice(MHI_USE_arg     slice))
		return true;
	
	u8 i = 1;
	while (global(side_slices)[++i] != SIDE_SLICES_FINISH) {
		if (global(side_slices)[i] == slice)
			return true;
	}
	
	return false;
}
#endif
;


bool is_main_slice(MHI_ARG    uint8_t slice)
#ifndef NOCODE
{
	if (slice == gl main_slice)
		return true;
	
	if (slice == gl side_slices[0])
		return true;
	
	if (slice == gl side_slices[1])
		return true;
	
	return false;
}
#endif
;



#define SOLE false
#define CHECK_TABLE true

void print_side_slices(MHI_ARG    bool check_table)
#ifndef NOCODE
{
	ulog_str_L("Side slices: ");
	
	uint64_t actual_time = get_systime(MHI_USE_single);
	bool old;
	
	u8 i = 0;
	while (global(side_slices)[i] != SIDE_SLICES_FINISH) {
		old = false;
		if (global(fss->slice_sack[global(side_slices)[i]])
					+ SRR_TIME_TO_LIVE < actual_time)
			old = true;
		
		log_str_num_BASE_str_P(old ? "x":"", global(side_slices)[i], DEC, " ");
		i++;
	}
	log_str_L("\n");
	
	if (!check_table)
		return;
	
	frame_slice_table(MHI_USE_arg    COMPACT);
	
	// extra check
	u8 t = 0xff;
	while (++t < 30) {
		if (global(fss->slice_addr)[t] != global(my_address))
			continue;
		
		if (global(main_slice) == t)
			continue;
		if (global(side_slices)[0] == t)
			continue;
		if (global(side_slices)[1] == t)
			continue;
		
		u8 l = 1;
		while (global(side_slices)[++l] != SIDE_SLICES_FINISH) {
			if (global(side_slices)[l] == t)
				break;
		}
		
		if (global(side_slices)[l] == SIDE_SLICES_FINISH) {
			// This must not happen. When changed the code, it may be a display
			// error and may be cured with the next slice management invokation.
			// Please make this error never appear!
			printf("__Error: slice %u not present in side slice list!\n", t);
			//assert(0);
		}
	}
	if (  global(fss->slice_addr)[global(main_slice)    ] != global(my_address)
		||global(fss->slice_addr)[global(side_slices)[0]] != global(my_address)
		||global(fss->slice_addr)[global(side_slices)[1]] != global(my_address))
		// This must not happen. When changed the code, it may be a display
		// error and may be cured with the next slice management invokation.
		// Please make this error never appear!
		printf("__Error: main slices not present in frame state table!\n");
	u8 l = 1;
	while (global(side_slices)[++l] != SIDE_SLICES_FINISH) {
		if (global(fss->slice_addr)[global(side_slices)[l]] != global(my_address))
			// This must not happen. When changed the code, it may be a display
			// error and may be cured with the next slice management invokation.
			// Please make this error never appear!
			printf("__Error: slice list element %u not present in frame state table!\n", global(side_slices)[l]);
	}
	
	
}
#endif // -ifndef NOCODE
;



#ifndef NOCODE
bool OK(MHI_ARG   u8 slice) {
	if (slice == SIDE_SLICES_FINISH)
		return false;
	
	u8 bits = global(fss->slice_bits[slice]);
	if (bits & BITS_UNSENT)
		// Have to send once
		return true;
	
	if (gl fss->slice_sack[slice]+SRR_TIME_TO_LIVE>get_systime(MHI_USE_single))
		// There should be an srr entry which timeouts so this side slice
		// entry will be removed somewhen.
		return false;
	
	
	if (bits & BITS_SEEN)
		// Have got ACK from others
		return true;
	
	// We have sent once and do wait now.
	return false;
}
#endif


void update__next_slice(MHI_ARG     bool print_debug)
#ifndef NOCODE
{
	
	u8 sl = global(next_slice);
	
	
	// main_slice list 0-1
	u8 alt = SIDE_SLICES_FINISH;
	
	if (OK(MHI_USE_arg   global(side_slices[1]))
		&& global(side_slices[1]) > sl)
		alt = global(side_slices[1]);
	
	if (OK(MHI_USE_arg   global(side_slices[0]))
		&& global(side_slices[0]) > sl
		&& global(side_slices[0]) < alt
		&& global(side_slices[0]) != SIDE_SLICES_FINISH)
		alt = global(side_slices[0]);
	
	//sl = SIDE_SLICES_FINISH;
	
	//printf("OK(side slice %u) = %c\n", gl side_slices[1], OK(MHI_USE_arg   gl side_slices[1])?'Y':'N');
	//printf("OK(side slice %u) = %c\n", gl side_slices[0], OK(MHI_USE_arg   gl side_slices[0])?'Y':'N');
	
	// other list items 2-30
	u8 i = 1;
	bool found = false;
	while (global(side_slices[++i]) != SIDE_SLICES_FINISH) {
		if (sl == global(side_slices[i])) {
			sl =  global(side_slices[i+1]);
			//printf("OK(sl %u) = %c\n", sl, OK(MHI_USE_arg   sl)?'Y':'N');
			if (!OK(MHI_USE_arg    sl))
				continue;
			found = true;
			break;
		}
		if (sl < global(side_slices[i])) {
			sl = global(side_slices[i]);
			//printf("OK(sl %u) = %c\n", sl, OK(MHI_USE_arg   sl)?'Y':'N');
			if (!OK(MHI_USE_arg    sl))
				continue;
			found = true;
			break;
		}
	}
	
	if (!found || alt < sl)
		sl = alt;
	
#ifdef DEBUG_SLICE_MGMT
	ulog_str_num_BASE_str_L("switching next_slice from ",
							global(next_slice), DEC, " to ");
	log_num_BASE(sl, DEC);
	log_str_L(".\n");
	//printf("(alt=%d)\n", alt);
	bool check_table = print_debug ? CHECK_TABLE : SOLE;
	print_side_slices(MHI_USE_arg    check_table);
#endif
	global(next_slice) = sl;
}
#endif // -ifndef NOCODE
;


uint8_t which_side_slice(MHI_ARG    uint8_t slice)
#ifndef NOCODE
{
	u8 retv = 1;
	
	while (global(side_slices)[++retv] != slice) {
		if (global(side_slices)[retv] == SIDE_SLICES_FINISH)
			return SIDE_SLICES_FINISH;
	}
	
	return retv;
}
#endif
;


void remove_side_slice(MHI_ARG    uint8_t slice_to_free, bool adjust_capacity,
		bool print_debug)
#ifndef NOCODE
{
	// Find list item
	u8 i = 1;
	
	while (global(side_slices)[++i] != SIDE_SLICES_FINISH) {
		//printf("__Info: global(side_slices)[%u] = %u.\n", i, global(side_slices)[i]);
		if (global(side_slices)[i] == slice_to_free) {
			//printf("%d\n", i);
			remove_side_slice_list_item(MHI_USE_arg     i, adjust_capacity, print_debug);
			return;
		}
	}
	
	ulog_str_L("WARNING: Couldn't remove: not present\n");
}
#endif
;


void remove_side_slice_list_item(MHI_ARG    uint8_t list_index_to_free, bool adjust_capacity, bool print_debug)
#ifndef NOCODE
{
	//print_side_slices(MHI_USE_single);
	
	u8 slnum = gl side_slices[list_index_to_free];
	
#ifdef DEBUG_SLICE_MGMT
	printf("RM slice %u (pos %u) in the face of next_slice %u.\n",
		   slnum, list_index_to_free, gl next_slice);
#endif
	
	assert (slnum < 30);
	
	if (   slnum == global(main_slice)
		|| slnum == global(side_slices[0])
		|| slnum == global(side_slices[1])) {
		
		// Cann't remove ourselves.
		return;
	}
	
	while (global(side_slices)[list_index_to_free++] != SIDE_SLICES_FINISH) {
		   global(side_slices)[list_index_to_free-1]
		 = global(side_slices)[list_index_to_free];
	}
	
	if (slnum == global(next_slice)) {
		update__next_slice(MHI_USE_arg    print_debug);
	}
	
	if (capacity_deficit(MHI_USE_single))
		gl slice_need--;
	else if (adjust_capacity)
		gl slice_need--;
	
	if (print_debug)
		print_side_slices(MHI_USE_arg    SOLE);
}
#endif
;



// if possible keep gl slice_need so the "capacity statement" gets resolved
#define CAP_UNTOUCHED false

// if possible move gl slice_need so the "capacity statement" remains
#define CAP_ADJUST     true

void add_side_slice(MHI_ARG     u8 new_slice, bool adjust_capacity)
#ifndef NOCODE
{
	i8 extra = 0;
	if (capacity_deficit(MHI_USE_single))     extra =  1;
	if (capacity_spill_over(MHI_USE_single))  extra = -1;
	
	u8 i = gl slice_need + extra;
	while (i > 2) {
		gl side_slices[i] = gl side_slices[i-1];
		
		if (gl side_slices[i] < new_slice)
			break;
		
		i--;
	}
	gl side_slices[i] = new_slice;
	
	if (extra == -1) {
		gl slice_need++; // We must adjust
	}
	else if (adjust_capacity) {
		gl slice_need++;
	}
}
#endif // -ifndef NOCODE
;






#define COMPACT true // for frame_slice_table
#define LONG false   // for frame_slice_table

void frame_slice_table(MHI_ARG     bool compact_table)
#ifndef NOCODE
{
	//print_side_slices(MHI_USE_single);
	
#ifdef DEBUG_ALGO
	ulog_str_L("====== Frame state table  ======\n");
	
	if (!compact_table) {
		ulog_str_num_BASE_str_L("   frame_align = ",
								global(frame_align), DEC, "   \n");
	}
	
	if (compact_table) {
		log_str_L("         __  __  __  __  __  --  --  --  --  --  __  __  __  __  __\nState:   ");
		for (int i = 0; i < 30; i++) 
			log_num_BASE(global(fss->slice_bits)[i], HEXBYTE);
		log_str_L("\nAddress: ");
	}
	
	for (int i = 0; i < 30; i++) {
		if (!compact_table) {
			ulog_str_num_BASE_str_L("Slice ", i, HEXBYTE, ": state ");
			log_num_BASE(global(fss->slice_bits)[i], HEXBYTE);
		}
		
		uint8_t addr = global(fss->slice_addr[i]);
		
		//if (!(global(fss->slice_bits)[i] & 0x02)) {
		if (addr == PSEUDO_ADDRESS_FREE) {
			log_str_P(compact_table ? " ." : "   free      ");
		}
		else if (addr == PSEUDO_ADDRESS_CONFLICT) {
			log_str_P(compact_table ? "C!" : "   conflict  ");
		}
		else if (addr == PSEUDO_ADDRESS_OVERLAP) {
			// Not in use so far
			log_str_P(compact_table ? "O-" : "   overlap   ");
		}
		else {
			if(!compact_table)
				log_str_num_BASE_str_L(" with addr=", addr, DEC, " ");
			if (addr == global(my_address)) {
				log_str_P(compact_table ? "us" : "(that's us) ");
			}
			else if (compact_table) {
				if (addr < 10) {
					log_str_L(" ");
				}
				log_num_BASE(addr, DEC)
			}
		}
		
		if (!compact_table) {
			log_str_L("\n");
		}
	}
	if (compact_table);
		log_str_L("       \n");
#endif // -ifdef DEBUG_ALGO
}
#endif // -ifndef NOCODE
;


void update_slice_bits(MHI_ARG    uint8_t slice, uint8_t bits)
#ifndef NOCODE
{
	
	global(fss->slice_bits[slice]) <<= 2;
	global(fss->slice_bits[slice]) |= bits & 3;
	
}
#endif
;

// Will rotate the slice_bits and set the address.
void update_slice(MHI_ARG    uint8_t slice, uint8_t bits, uint8_t addr)
#ifndef NOCODE
{
	assert(slice < 30);
	
	update_slice_bits(MHI_USE_arg    slice, bits);
	
#ifdef DEBUG_SLICE_MGMT
	ulog_str_num_BASE_str_L("updating slice ", slice, DEC, " to addr ");
	log_num_BASE(addr, DEC);
	log_str_L(".\n");
#endif
	
	if (gl fss->slice_addr[slice] != addr)
		gl fss->slice_bits[slice] |= 0xA8;
		// Of course this is not fully correct. If a station ocurres again
		// after a short interlude by another station, the old occurences
		// anyway will be marked as unknown by this operation.
	
	global(fss->slice_addr[slice]) = addr;
	
	if (addr == PSEUDO_ADDRESS_FREE)
		global(fss->slice_sack[slice]) = 0;
	else
		global(fss->slice_sack[slice]) = get_systime(MHI_USE_single);
	
}
#endif // ifndef NOCODE
;




bool capacity_spill_over(MHI_ARG_single)
#ifndef NOCODE
{
	if (       gl side_slices[gl slice_need]   != SIDE_SLICES_FINISH
		   &&  gl side_slices[gl slice_need+1] != SIDE_SLICES_FINISH) {
		// "Reduced pointer" in alternative terminology.
		return true;
	}
	return false;
}
#endif
;

bool capacity_deficit(MHI_ARG_single)
#ifndef NOCODE
{
	if (       gl side_slices[gl slice_need]   == SIDE_SLICES_FINISH) {
		// "Enriched pointer" in alternative terminology.
		return true;
	}
	return false;
}
#endif
;



// Reduces the pointer. If the pointer is already reduced once, take action.
void request_fewer_slices(MHI_ARG_single)
#ifndef NOCODE
{
	u8 slptr = global(slice_need);
	
	if (!slptr) {
		printf("slice_need == 0    -- this cann't happen!\n");
	}
	
	if (slptr == 1)
		// We've got 1 main slice and 2 side slices - don't take fewer.
		return;
	
	
	if (capacity_deficit(MHI_USE_single)) {
		// Enriched pointer. Reduce it -> normal pointer
		global(slice_need)--;
		return;
	}
	
	if (!capacity_spill_over(MHI_USE_single)) {
		// Normal pointer. Reduce it -> reduced pointer
		global(slice_need)--;
		return;
	}
	
	// Reduced pointer. Choose some slice to free
	// binary backoff -> first count side slices, then choose halfs and halfs
	// of halfs.
	// [Could do this with random numbers but want to stabilize the simulation
	//  reproducability]
	
	
	// really?
	if (!hop(MHI_USE_single)) {
#ifdef DEBUG_SLICE_MGMT
		ulog_str_L("Not yet freeing one slice.\n");
#endif
		return;
	}
	
	
#ifdef DEBUG_SLICE_MGMT
	ulog_str_L("Freeing one slice.\n");
#endif
	
	// First count
	u8 cnt = 2;
	while(global(side_slices[cnt++] != SIDE_SLICES_FINISH)) {}
	
	
	bool halfbit = hop(MHI_USE_single);
	
	// Then choose
	u8 left = 2;
	u8 right = cnt - 2;
	while (left < right) {
		u8 center = (left + right) / 2;
		if (cnt % 2) {
			if (halfbit)
				center++;
			halfbit = !halfbit;
		}
		
		if (hop(MHI_USE_single))
			left = center;
		else
			right = center;
	}
	
	u8 idx_to_free = left;
	
	
	update_slice(MHI_USE_arg     gl side_slices[idx_to_free], 0, PSEUDO_ADDRESS_FREE);
	
	remove_side_slice_list_item(MHI_USE_arg    idx_to_free, CAP_UNTOUCHED, false);
	
	
#ifdef DEBUG_ALGO
	print_side_slices(MHI_USE_arg    CHECK_TABLE);
#endif
	
	// Slice pointer is correct after this operation.
}
#endif // -ifndef NOCODE
; // request_fewer_slices()






void request_more_slices(MHI_ARG_single)
#ifndef NOCODE
{
	u8 slptr = global(slice_need);
	
	if (slptr >= MAX_NUM_SIDE_SLICES-1) { // Safety
	//if (slptr == MAX_NUM_SIDE_SLICES-1) {
		// No, more is not possible.
		return;
	}
	
	if (slptr == 1) {
		// enriching pointer
		global(slice_need) = 2;
		//global(side_slices[0]) = SIDE_SLICES_FINISH;
		return;
	}
	
	if (!capacity_deficit(MHI_USE_single)) {
		// Normal or reduced pointer.
		// Enrich it -> enriched or normal pointer
		global(slice_need)++;
		return;
	}
	
	
	// Enriched pointer. Choose some slice to add.
	
	
	// ... really?
	if (!hop(MHI_USE_single)) {
#ifdef DEBUG_SLICE_MGMT
		ulog_str_L("Not yet adding one slice.\n");
#endif
		return;
	}
	
	
#ifdef DEBUG_SLICE_MGMT
	ulog_str_L("Adding one slice.\n");
#endif
	
	
	
	uint8_t fold_offset;
	uint8_t fold_tail;
	int8_t * fold_arr;
	
	
	
	fold_offset = 1; // 1 field before A
	fold_tail   = 7; // 2 fields after A
//	fold_arr    =  (i8[]) {1, -3, 5, 1, 0, 0, 1, 5, 1};
	//   that's us  ->         ^ 
	//   there we want to go  ->  ^
	//   or that                             ->  ^
	
	fold_arr    =  (i8[]) {1, 5, 1, 0, 0, 0, 1, -3, 1, 0};
	//     
	if (hop(MHI_USE_single) && fold_arr[7] == -3) {
		fold_arr[9] = fold_arr[8];
		fold_arr[8] = fold_arr[7];
		fold_arr[7] = fold_arr[6];
	}
	
	
	struct fafres fafres = fold_and_find(MHI_USE_arg
								 fold_offset, fold_tail, fold_arr, false);
	u8 new_slice = choose_after_fold(MHI_USE_arg    fafres);
	
	if (new_slice == SIDE_SLICES_FINISH) {
		ulog_str_L("Did meet our own slices -> hardly possible to find something free.\n");
		return;
	}
	
#ifdef DEBUG_SLICE_MGMT
	ulog_str_num_BASE_str_L("Taking slice ", new_slice, DEC, " which had address ");
	log_num_BASE(global(fss->slice_addr[new_slice]), HEXBYTE);
	log_str_num_BASE_str_L(" and status ",
	             global(fss->slice_bits[new_slice]), HEXBYTE, ".\n");
#endif
	
	// Invalidate old uids, if any
	invalidate_uids(MHI_USE_arg    new_slice);
	
	// Register slice in frame slice table and side slice list.
	// It will get active when it was sent once and received an ack.
	update_slice(MHI_USE_arg    new_slice, BITS_UNSENT, gl my_address);
	// Remove protection for SRR_TIME_TO_LIVE
	gl fss->slice_sack[new_slice] = get_systime(MHI_USE_single);
	
	global(fss->slice_cnfl)[new_slice] = 0; // outdated?
	
	//print_side_slices(MHI_USE_single);
	
	add_side_slice(MHI_USE_arg    new_slice, CAP_UNTOUCHED);
	
	
	// global(slice_need)++; // Don't increment; let's look how it develops.
	//   (original intention was protecting against fast removing)
	//   (but this would lead to fast reserving which is not good either)
	
#ifdef DEBUG_ALGO
	print_side_slices(MHI_USE_arg    CHECK_TABLE);
#endif
}
#endif // -ifndef NOCODE
; // request_more_slices


// Looks what uids are hanging around and whether they are too old
void invalidate_uids(MHI_ARG    uint8_t slice)
#ifndef NOCODE
{
	
 	uint64_t old_time = get_systime(MHI_USE_single);
 	if (old_time < SRR_TIME_TO_LIVE)
 		old_time = 0;
 	else
 		old_time -= SRR_TIME_TO_LIVE;
	
	
	
	
	bool do_remove = true;
	u32 i;
	// Check side slices uid
	for (i = 0; i < SLICE_RECEPTION_REGISTRY; i++) {
		struct slice_reception_registry_entry srre = global(srr)[i];
		
		if (srre.slice == 0xFF)
			continue;
		/* New version
		if (srre.time_sent <= old_time)
			gl srr[i].slice = 0xFF; // Free outdated srr entry
		else
		if (srre.slice == slice)
				do_remove = false;
		
		----*/
		
		// Old version
		//TODO: keep?
		if (srre.slice != slice)
			continue;
		
		if (srre.time_sent > old_time)
			do_remove = false;
		else
			gl srr[i].slice = 0xFF; // Free outdated srr entry
	}
	
	
	u16 verylast = i;
	u16 offset = 0;
	for (i = 0; i < verylast; i++) {
		if (global(srr)[i].slice == 0xFF) {
			offset++;
			continue;
		}
		if (!offset)
			continue;
		global(srr)[i - offset] = global(srr)[i];
	}
	while (offset)
		global(srr)[verylast - offset--].slice = 0xFF;
	
#if 0
	i = 0;
	while (i < verylast) {
		ulog_str_num_BASE_str_L("srr[", i, DEC, "] =   @");
		log_num_BASE(gl srr[i].time_sent, DEC);
		log_str_num_BASE_str_L("   sl ", gl srr[i].slice, DEC, "   ");
		log_num_BASE(gl srr[i].fuid, HEXBYTE);
		log_str_L("\n");
		i++;
	}
#endif
	
	
	if (global(fss->slice_addr[slice]) != global(my_address)) {
		return; // Don't need this.
		// TODO: Remove this return if slice_sack is set for foreign slices!
		
		//ulog_str_L("invalidate_uids(): WARNING: Don't own this slice!\n");
	}
	
	
	
	
	if (gl fss->slice_sack[slice] <= old_time)
		// Have some recent ack present.
		do_remove = false;
	
	
	if (do_remove) {
		if (   slice == gl main_slice
			|| slice == gl side_slices[0]
			|| slice == gl side_slices[1]) {
			
			think_about_a_new_main_slice(MHI_USE_single);
		}
		else {
			update_slice     (MHI_USE_arg    slice, 0, PSEUDO_ADDRESS_FREE);
			remove_side_slice(MHI_USE_arg    slice, CAP_UNTOUCHED, false);
#ifdef DEBUG_ALGO_DETAIL
			print_side_slices(MHI_USE_arg    CHECK_TABLE);
#endif
		}
	}
	

}
#endif // -ifndef NOCODE
;





#endif
