
// (c) Jonas Bechtel 2018

#ifndef FRAMESWITCH_C
#define FRAMESWITCH_C


#define AMRLOAD_FILE "kombi.s3.amr"

// 107 kbit/s - see requirement analysis for 1 high quality speech channel.
#define FRAMESWITCH_BITRATE 107000
// -> 1 full FIFO packet will take 5 ms

#define UNFINISHED_KEEP_TIME 30 MSEC


#define APP_GLOBAL_VARS(extern) \
	extern struct route_instance * rti; \
	extern struct frameswitch_state * fss; \
	extern unsigned long rand_state; \
	extern uint32_t recent_random; \
	extern uint8_t hop_state; \
	extern uint64_t frame_align; /* See comment */  \
	extern uint8_t uid; \
	extern struct old_uid old[NUM_UIDS]; \
	extern uint8_t my_address; \
	extern uint8_t aloha_slice_unreq; /*unrequited*/ \
	extern bool     main_slice_invalidated; \
	extern uint8_t  main_slice; \
	extern uint8_t side_slices[30]; \
	extern uint8_t slice_need; \
	extern uint8_t next_slice; \
	extern uint8_t   last_reg_slice; \
	extern uint16_t  last_reg_count; \
	extern bool      last_reg_main_present; \
	/*extern uint8_t last_reg_addr;*/ \
	extern struct slice_reception_registry_entry srr[SLICE_RECEPTION_REGISTRY];\
	extern uint8_t pbuf_1st[PACKET_BUFFER]; \
	/*extern uint8_t pbuf_2nd[2*PACKET_BUFFER];*/ \
	extern uint8_t pbuf_2nd[PACKET_BUFFER]; \
	/*extern uint8_t pbuf_2nd[PACKET_BUFFER/3];*/ \
	extern uint64_t unfinished_packet; \
	extern struct amrload amr; \
	extern uint8_t shadow_fifo[66]; \
	extern uint8_t shadow_fifo_pos; \
	extern uint64_t shadow_fifo_touch; \
	//              extern uint8_t amrfold_state;

#define MAX_NUM_SIDE_SLICES 29 // includes the two "main side slices"
#define SRR_TIME_TO_LIVE 5000 MSEC
#define SLICE_RECEPTION_REGISTRY (30*NUM_UIDS*2*20) // huge, assuming 20 p./sl.
#define NUM_UIDS (2*3) // 2 frames, assuming that 3 uids will be used in 1 frame
/* UIDs
 * TODO: bla bla
 */


/* FRAME ALIGN 
 * 
 * global(frame_align) acts also as slice_align.
 * 
 * You may not alter it except for small corrections or with the
 * floor_frame_align() and ceil_frame_align() functions.
 */
/* SIDE SLICES
 * 
 * global(side_slices) contains the side slices which this station will use for
 * data transmission.
 * 
 * side slices array is sorted and it's finish is denoted by 0xFF
 * 
 * slice_need points to the finish entry of side_slices iff one more slice
 *   is needed. It points to the last valid entry of side_slices
 *   (or 0xFF if there's no side slice) if the numer of side slices is right.
 *   It points to an entry before the last valid entry to demark that the
 *   number of slices which fits after the pointer is to be freed.
 */
#define SIDE_SLICES_FINISH 0xFF
/* SLICE NEED
 * How many side slices (+ main slice) we need.
 */

//#define PACKET_BUFFER 1500
#define PACKET_BUFFER 200
/* pbuf / PACKET BUFFER 
 * 
 * This packet buffer holds 2 seconds of AMR MR475 speech (4,75 kbit/s) which
 * will take 2*50*19 + 2*50*1 bytes with header (->Bitreduktionsplanung.ods)
 * and length byte. [TODO: Use AMR_SHORT header with 16 bytes instead of 19.]
 * 
 * A length byte == 0x00 demarks the end of the queue.
 */

/* UNFINISHED PACKET
 * 
 * When we discovered a packet on FIFO but the FIFO hasn't as much bytes
 * received as needed, we set this variable to the time we discoverd the
 * packet.
 */




// TODO: Eliminate all these magic numbers
// -> #define ...

#define CBNUM_FRAMEFIND 1


#define REASON_FRAMEFIND_START     1
#define REASON_FRAMEFIND_TRY       4
#define REASON_FRAMEFIND_TRYWAIT   5
#define REASON_FRAMEFIND_KEEPALIVE 7

#define REASON_GPSPULSE          5
#define REASON_WAIT_FOR_AIR      6
#define REASON_WAIT_FOR_AIR_END 26
#define REASON_TX_START         30

#define REASON_SLEEP_TILL_TRY (BIGGEST_REASON +  1)

#define REASON_TIMER1         (BIGGEST_REASON +  5)
#define REASON_TIMER_END      (BIGGEST_REASON + 25)


#if (defined request_DEBUG_ALGO)        && !(defined DEBUG_ALGO)
#define DEBUG_ALGO
#endif
#if (defined request_DEBUG_ALGO_DETAIL) && !(defined DEBUG_ALGO_DETAIL)
#define DEBUG_ALGO_DETAIL
#endif

#if (defined DEBUG_ALGO) && !(defined DEBUG_ALGO_PACKET)
#define DEBUG_ALGO_PACKET
#endif


#ifndef NOCODE
#define NOCODE
#undef FRAMESWITCH_C
#include "FrameSwitch.c"
#define FRAMESWITCH_C
#include "FS-slice-helper.c"
#undef NOCODE

#endif


#include <stdint.h>

struct route_instance; // for APP_GLOBAL_VARS
struct frameswitch_state; // for APP_GLOBAL_VARS

#ifdef NOCODE // (!)
struct old_uid {
	uint8_t uid;
	uint64_t time;
};

struct slice_reception_registry_entry {
	uint64_t time_sent;
	uint32_t fuid; // foreign uid, given by routing algorithm 
	uint8_t slice; // 0xFF if this entry is invalid.
};
#endif // -ifdef(!) NOCODE



#ifndef NOCODE
#define NOCODE
#include "cassert.h"
#include <math.h>
// present in libm.so, but not in math.h:
float exp10f(float);



#define  u8  uint8_t
#define u16 uint16_t
#define  i8   int8_t
#define i16  int16_t
#define BOOL_UNITIALIZED true
#define U8_UNINITIALIZED 0xFF


#include "checksum.c"
#include "amrload.c"
#include "only.c"
#include "RFM69trancv/trancv.c"
#include "RFM69trancv/convenience.c"
#define CN_PORT
#include "AODVroute/AODVintern.h" // Only for numerical number from aodvuudd
	// -> In future the numerical address should be implemented as station var.
#include "common/route-user.h"
//#include "AODVroute/subobject.h"
#ifndef SIMUL
#define SIMUL
#include "packetizer.c"
#undef SIMUL
#else
#include "packetizer.c"
#endif // -ifndef SIMUL
#include "debug.c"
#include "cli-dev.c"
#undef NOCODE
#include "RFM69trancv/RFM69-reg.h"
#include "RFM69trancv/RFM69-reghelper.h"


#else // -ifndef NOCODE
#include "amrload.c"
#include "only.c"
//#include "RFM69trancv/trancv.c"
#endif // -ifndef NOCODE -else



#include "all-app.c"


#ifdef SIMUL
#define TI_USE_single (MHI_USE_single->trancv)
//#define TI_USE_single mh_station_arg -> trancv
#define TI_USE_arg    TI_USE_single,
//#define TI_USE_access MHI_USE_single->trancv->
#else
#define TI_USE_single
#define TI_USE_arg
//#define TI_USE_access
#endif


#ifdef NOCODE // (!)
struct frameswitch_state {
	time_int state; // This may change
	
	// slice_bits: MSB abcdefgh LSB with
	// h is set if some packet/activity was seen last time.
	// g is set if this packet was scrumbled or the station is not the station
				// in slice_addr
	// f is set if some packet was seen the last but two time
	// e is set if scrumbled/not the saved addr.
	// d, c is like f, e but one time more in past
	// b, a is like d, c but one time more in past
#define BITS_SEEN 1
#define BITS_WRONG_STATION 2 // shared
#define BITS_UNSENT        2 // shared
	uint8_t  slice_bits[30]; // This may change
	uint8_t  slice_cnfl[30];
	uint64_t slice_sack[30]; // last seen ack
	uint8_t  slice_addr[30]; // This may change
	// slice_addr = 0xC0 for long overlap (so much part of the frame
	//   is taken by the last frame's contents which probably have started
	//   very late in last frame. 
	// slice_addr = 0xD0 for nothing (free) but don't rely on that.
	// slice_addr = 0xE0 conflict.
#define PSEUDO_ADDRESS_OVERLAP 0xC0 //[TODO: No overlap detection so far]
#define PSEUDO_ADDRESS_FREE 0xD0
#define PSEUDO_ADDRESS_NOADDR 0xD1
#define PSEUDO_ADDRESS_CONFLICT 0xE0
	
	
	uint32_t slice_time_offset;
	
#define CAP_ACK 10
	// 16 bit & 0x8000: invalid
	// 16 bit & 0x0f00: station_addr
	// 16 bit & 0x00ff: uid
	uint16_t to_acknowledge[CAP_ACK];
};
#endif // -ifdef (!) NOCODE


// Looks ~ random.
uint8_t inc_unique(uint8_t in)
#ifndef NOCODE
{
#define bit_from_to(in, from, to) ( ((in >> from)&1) << to)
	in =   bit_from_to(in, 4, 2) | bit_from_to(in, 5, 5)
	     | bit_from_to(in, 2, 3) | bit_from_to(in, 7, 4)
	     | bit_from_to(in, 6, 6) | bit_from_to(in, 1, 7)
		 | bit_from_to(in, 3, 0) | bit_from_to(in, 0, 1);
	in++;
	
	return bit_from_to(in, 2, 4) | bit_from_to(in, 5, 5)
	     | bit_from_to(in, 3, 2) | bit_from_to(in, 4, 7)
	     | bit_from_to(in, 6, 6) | bit_from_to(in, 7, 1)
		 | bit_from_to(in, 0, 3) | bit_from_to(in, 1, 0);
}
#endif // -ifndef NOCODE
;


uint8_t get_new_uid(MHI_ARG_single)
#ifndef NOCODE
{
	gl uid = inc_unique(gl uid);
	
	uint8_t i = 0;
	uint8_t oldest = 0;
	while (i < NUM_UIDS) {
		if (gl old[i].time == 0) {
			oldest = i;
			break;
		}
		
		if (gl old[i].time < gl old[oldest].time) {
			oldest = i;
		}
		i++;
	}
	
	
	gl old[oldest].time = get_systime(MHI_USE_single);
	gl old[oldest].uid = gl uid;
	
	return gl uid;
}
#endif // -ifndef NOCODE
;


// Sets 'frame_align' to a value which is higher than 'around' yet 
// exactly keeping the alignment.
// Does nothing if 'frame_align' is higher than 'around'
void ceil_frame_align(MHI_ARG    uint64_t around)
#ifndef NOCODE
{
	if (around > global(frame_align)) {
		uint64_t add = (around - global(frame_align)) / (3 SEC);
		global(frame_align) += (add + 1) * 3 SEC;
	}
}
#endif
;


// Sets 'frame_align' to the highest possible value which is lower than
// 'around' while exactly keeping the alignment.
// Does nothing if 'frame_align' is higher than 'around'
void floor_frame_align(MHI_ARG    uint64_t around)
#ifndef NOCODE
{
	if (around > global(frame_align) + 3 SEC) {
		uint64_t add = (around - global(frame_align)) / (3 SEC);
		global(frame_align) += add * 3 SEC;
	}
}
#endif
;



u8 deduce_slice(uint64_t frame_align, uint64_t actual_time)
#ifndef NOCODE
{
	if (actual_time < frame_align)
		frame_align -= 6 SEC;
		// This must be sufficient. - Who would set frame_align higher?
	
	uint64_t tmp = actual_time - frame_align;
	
	tmp /= 100 MSEC;
	
	return tmp % 30;
}
#endif
;

u32 deduce_offset(uint64_t frame_align, uint64_t actual_time)
#ifndef NOCODE
{
	while (actual_time < frame_align)
		frame_align -= 3 SEC;
	
	uint64_t retv = (actual_time - frame_align) % (100 MSEC);
	return retv;
}
#endif // -ifndef NOCODE
;



#ifndef NOCODE
// This method is adapted from Debian GNU/Linux manpage for rand/rand_r/srand
// You can see: this will give you 15 bit.
uint32_t random_number(MHI_ARG_single) {
	gl rand_state = gl rand_state * 1103515245 + 12345;
	return ((unsigned)(gl rand_state/65536) % 32768);
}





#endif


// Returns some "random" bit (0 for "keep" and 1 for "hop")
// It's actually not random but based on the stations' address
bool hop(MHI_ARG_single)
#ifndef NOCODE
{
	if (gl hop_state == 0) {
		gl recent_random = random_number(MHI_USE_single);
#ifdef DEBUG_ALGO_DETAIL
		ulog_str_num_BASE_str_L("New random number: ", gl recent_random, HEX, "\n");
#endif
	}
	
	bool retv = (gl recent_random >> global(hop_state)) & 1;
	
	global(hop_state) = (global(hop_state)+1) % 15;
#ifdef DEBUG_ALGO
	ulog_str_num_BASE_str_L("Hop -> ", retv, DEC, "\n");
#endif
	return retv;
}
#endif // -ifndef NOCODE
;



// give selfconception = "packet_end - (selfconc_bits * (100 MSEC / 16))"
// frame_alignoffset = calc_slicetimeoffset() + 100 MSEC * return_slice
uint32_t calc_slicetimeoffset(MHI_ARG    
			time_int packet_start, uint32_t selfconception,
			uint8_t * return_slice)
#ifndef NOCODE
{
	
	time_int lstart = global(frame_align);
	while (packet_start >= lstart + 3 SEC) {
		lstart += 3 SEC;
	}
	
	uint32_t start_diff = packet_start - lstart;
	uint32_t retval = start_diff;
	
	if (return_slice) {
		// will floor start_diff
		start_diff /= 100 MSEC;
		*return_slice = start_diff;
		//printf("calc_slicetimeoffset(%'lu) <- %'lu\n",
		//		packet_start, global(frame_align));
	}
	
	return retval % 100 MSEC;
}
#endif // -ifndef NOCODE
;




// Assuming packet is in FIFO
// TODO: use packet_end information (for double-use of slices)
void register_station(MHI_ARG
		uint8_t slice_number,
		uint8_t sender, uint8_t sender_uid, bool is_sender_main,
		bool shortpacket,
		uint32_t slicetimeoffset)
#ifndef NOCODE
{
	
	
	// TODO: Don't need to be so careful with ancient data because 
	// ancient data hase been wiped out 90 lines ago. [<<---- TODO: ???]
	
	// Don't need this - it happens with update function.
	//global(fss->slice_bits)[slice_number] <<= 2;
	
#ifdef DEBUG_ALGO
	ulog_str_num_BASE_str_L("register_station(): sender=", sender, DEC, ", slice=");
	 log_num_BASE(slice_number, DEC);
	 log_str_L("\n");
#endif
	
	
	u8 sl_addr = global(fss->slice_addr)[slice_number];
	
#define NO 0
#define CONFLICT 1
#define YES 2
	u8 do_ack = shortpacket ? NO : YES;
	
	if (sl_addr == gl my_address) {
#ifdef DEBUG_ALGO
		ulog_str_L("register_station(): WARNING: conflicting with us!\n");
#endif
		global(fss->slice_cnfl)[slice_number] ++;
		global(last_reg_slice) = slice_number;
		gl last_reg_main_present = true;
		return;
	}
	
	
	if (global(last_reg_slice) == slice_number) {
		if (sl_addr == PSEUDO_ADDRESS_CONFLICT) {
			// Don't acknowledge conflicting stations
			do_ack = CONFLICT;
		}
		else if (sl_addr != sender) {
			// Conflict
#ifdef DEBUG_ALGO
			ulog_str_num_BASE_str_L("register_station(): slice ", slice_number, DEC, ": st. ");
			log_num_BASE(sender, DEC);
			log_str_num_BASE_str_L(" competes ", sl_addr, DEC, " (conflict)\n");
#endif
			update_slice(MHI_USE_arg
					slice_number, BITS_WRONG_STATION, PSEUDO_ADDRESS_CONFLICT);
			
			// Don't acknowledge conflicting stations
			do_ack = CONFLICT;
			// If there were two senders perfectly interleaving themselfs
			// unfortunately we wouldn't give them acks.
			// We simply assume only one sender per slice.
		}
		else {
			gl last_reg_count++;
			// Do nothing. (allow ack)
		}
		
	}
	else {
		if (sl_addr != sender) {
#ifdef DEBUG_ALGO_DETAIL
			ulog_str_num_BASE_str_L("register_station(): slice ", slice_number, DEC, ": replace ");
			log_num_BASE(global(fss->slice_addr)[slice_number], DEC);
			log_str_num_BASE_str_L("->", sender, DEC, "\n");
#endif
		}
		
		//TODO: Leave this activated?
		invalidate_uids(MHI_USE_arg    slice_number);
	
		// Update slice in any case.
		update_slice(MHI_USE_arg    slice_number, BITS_SEEN, sender);
		
		
		gl last_reg_slice = slice_number;
		gl last_reg_count = 0;
		gl last_reg_main_present = !shortpacket && is_sender_main;
		// Remark: the last_reg_slice mechansim relies on active slice
		// switching. If only one slice is used for several frames, it
		// will fail.
	}
	
	
	if (do_ack == YES) {
		u32 expres = 1;
		while (expres - 1 < gl last_reg_count) {
			expres *= 2;
		}
		if (expres - 1 != gl last_reg_count) {
#ifdef DEBUG_ALGO_PACKET
			ulog_str_num_BASE_str_L("Would acknowledge but count=", gl last_reg_count, DEC, " doesn't fit.\n");
#endif
			do_ack = NO;
		}
	}
	
	if (do_ack == CONFLICT) {
		// Vorrang für main-slices:
		if (!shortpacket && is_sender_main && ! gl last_reg_main_present) {
			gl last_reg_main_present = true;
			do_ack = NO;
		}
	}
	
	
	if (!shortpacket && (do_ack == YES || do_ack == CONFLICT)) {
		u16 add = (do_ack == CONFLICT) ? 0x4000 : 0;
		
		// Acknowledge it  [= put it in the acknowledge list]
		int i = 0;
		while (i < CAP_ACK) {
			if (global(fss->to_acknowledge)[i] & 0x8000) {
				if (i + 1 < CAP_ACK)
					global(fss->to_acknowledge)[i+1] = 0x8000;
				else {
#ifdef DEBUG_ALGO
					ulog_str_L("ERROR: no space left on acknowledgement queue!\n");
#endif
					// Override the oldest and make space for some new newest.
					//         (delete one from the center)
					
					i = CAP_ACK / 2;
					while (i + 1 < CAP_ACK) {
						global(fss->to_acknowledge)[i]
								= global(fss->to_acknowledge)[i+1];
						i++;
					}
					global(fss->to_acknowledge)[i] = 0x8000;
					
					i = 0; // = override oldest
				}
				global(fss->to_acknowledge)[i] = add | (sender<<8) | sender_uid;
#ifdef DEBUG_ALGO_DETAIL
				ulog_str_num_BASE_str_L("Acknowledge @pos ", i, DEC, ".\n");
				ulog_str_num_BASE_str_L("              -> ", global(fss->to_acknowledge)[i], HEXBYTE, ".\n");
#endif
				break;
			}
			i++;
		}
		// This silently fails if there's no space left on acknowledge array
		
		frame_slice_table(MHI_USE_arg    COMPACT);
	}
	
	
	// Update slice time offset
	
	uint64_t gsto = global(fss->slice_time_offset);
	uint8_t cnt = 50;
	gsto *= cnt;
	
	if (slicetimeoffset > 31 MSEC && slicetimeoffset < 69 MSEC) {
		
		// The station we see is completely out of sync.
		// We use some "random" to decide whether to approach the station
		// or to keep on track.
		
		if (hop(MHI_USE_single)) {
			gsto += 10* slicetimeoffset;
			cnt += 9;
		}
		else
			gsto += slicetimeoffset;
	}
	else if (slicetimeoffset <= 31 MSEC)
		gsto += slicetimeoffset;
	else
		gsto -= (100 MSEC - slicetimeoffset);
	cnt++;
	
	gsto /= cnt;
	global(fss->slice_time_offset) = gsto;
	
	
}
#endif // -ifndef NOCODE
;

#define SUB_TYPE_MASK_MAIN_SIDE 1
#define SUB_TYPE_MAIN           0
#define SUB_TYPE_SIDE           1
#define SUB_TYPE_MASK_PACKETS 2
#define SUB_TYPE_WO_PACKETS   0
#define SUB_TYPE_WITH_PACKETS 2


void send_queue(MHI_ARG    bool is_main_slice)
#ifndef NOCODE
{
	uint64_t actual_time = get_systime(MHI_USE_single);
	uint8_t slice = deduce_slice(global(frame_align), actual_time);
	
	
#ifdef DEBUG_ALGO
	ulog_str_num_BASE_str_L("Sending @slice ", slice, DEC, "\n");
#endif
	
	
	
	// At first fetch the amr packets which have arrived so far.
	// They will get queued.
	// TODO: Ablaufzeit definieren: darf nur 2 Sekunden in der queue sein oder so?
	// (Wäre im Realbetrieb wichtig, weil AODV die Pakete zurückhält, bis es
	// eine Route gibt.)
	
	// Station 5 (letter O) sends packets to Station 1 (W)
	if (global(my_address) == 5 && actual_time > 10 SEC) {
#ifdef VARIANT_AMR_BROADCAST
		u8 dst = PACKED_BROADCAST;
#else
		u8 dst = 1;
#endif
		struct mr475_frame afr;
		while (gl rti->outbound_raw(gl rti, NULL, 19, 0)) {
			
			if (!get_amr_frame(MHI_USE_arg    &global(amr), &afr))
				break;
			
			global(rti)->outbound_unpacked(global(rti), afr.data, 12, dst);
			
			
		};
	}
	
	// Go to TX mode.
	
	REGFILE_DO_WRITE(TI_USE_single, cbMode, RFM_MODE_TX);
	global(unfinished_packet) = 0;	// TX mode clears fifo if coming from
	//  RX mode. (unfinished_packet belongs to RX mode)
	
	// We really have to choose a very high value as simulated FIFO sends
	// out too fast. [simulator internal lacking feature]
	// FIFO is 66 byte long
	// But not all 66 bytes cann't be used -> take 65 (error in trancv.c?)
	REGFILE_DO_WRITE(TI_USE_single, cbFifoThreshold, 65); // 66 = 0x42
	//REGFILE_DO_WRITE(TI_USE_single, cbFifoThreshold, 20);
	
	
	multitimer_delay_coarse(MHI_USE_arg    800 USEC); // TX Startup
	
	REGFILE_DO_WRITE(TI_USE_single, cbTxStartCondition, COND_FifoThreshold);// 0
	// Assuming FifoThreshold is 15, as default -> bigger than our data.
	
	// Which data to send?
	
	// from what_is_frameswitch.txt:
	// PPSSLADDDDDD[...]DDDDDDDppSS1a
	//      (P = Preamble, S = Sync word, L = Length indicator, A = Addr,
	//       D = Data, p = fake preamble, 1 = 0x01, a = address/selfconception)
	// from packetizer.c:
	//      station_addr:4
	//      nexthop_addr:4
	//      ttl:4
	//      header_chksum:3
	//      top_type:2
	
	// Sync between .txt and .c:  station_addr:4 and nexthop_addr:4 = A byte
	
	// -> Our packet looks like this:
	// P preamble:8 (makes RFM)
	// P preamble:8 (makes RFM)
	// S sync:8 = 'C' (makes RFM)
	// S sync:8 = 'N' (makes RFM)
	// L length:8 = ... (to FIFO)
	//   station_addr:4
	//   uid_msb:4
	//   uid_lsb:4
	//   header_chksum:3 (ignore so far)
	//   top_type:1 = 1
	//   top_type:1 = 1
	//   sub_type:2 = 00 (Main slice) or 01 (Side slice)
	//                10 (Main slice with packets following)
	//                                or 11 (Side slice with packets following)
	//   pcnt:3 = ... (how many packets follow directly after fill)
	//                (if sub_type = 10 or 11 -> confirmation)
	//   is_main_slice:1
	//   fill:1
	//             so far: 3 Bytes (L not included in calculation, see p.52)
	// <-- include packets here (preceeded with 1 length byte) --> (1+x B each)
	// <-- include acknowledgements here --> (2 B each)
	// p fake preamble:8 = 0xAA
	// p fake preamble:8 = 0xAA
	// S fake sync:8 = 'C'
	// S fake sync:8 = 'N'
	// 1 fake length:8 = 1
	// a:
	//   station_addr:4
	//   selfconception_bits:4
	//             finally: 9 Bytes
	
	
	uint8_t ack_cnt = 0;
	while(ack_cnt < CAP_ACK) {
		if (global(fss->to_acknowledge)[ack_cnt] & 0x8000)
			break;
#ifdef DEBUG_ALGO_DETAIL
		ulog_str_num_BASE_str_L("ack = ", global(fss->to_acknowledge)[ack_cnt], HEXBYTE, "\n");
#endif
		ack_cnt++;
	}
	
	// TODO: Make a global variable out of this rather than use the Regfile
	uint8_t fifo_size_check = REGFILE_DO_READ(TI_USE_single, cbFifoThreshold);
	// +1 because im unsure about off-by-one
	assert(fifo_size_check >= 11+1); // allow at least one ack
	fifo_size_check -= 9+1;//9+1: packet length w/len indic. w/o any payload/ack
	uint8_t acks_to_send = fifo_size_check / 2; // floors
	if (acks_to_send > ack_cnt)
		acks_to_send = ack_cnt;
	
	
	uint8_t packet_bytes_to_send;
	if(is_main_slice)
		packet_bytes_to_send = fifo_size_check / 2; // = 28, 1/2 of the payload
	else
		packet_bytes_to_send =  2 + fifo_size_check * 2 / 3;
			// = 40, ~ 2/3th of the payload size, but fits 2 AMR packets
	
	// But more if there's no conflict
	uint8_t free_bytes = fifo_size_check - 2 * acks_to_send;
#ifdef DEBUG_ALGO_DETAIL
	ulog_str_num_BASE_str_L("packet_bytes_to_send = max(", packet_bytes_to_send, DEC, ","); log_num_BASE(free_bytes, DEC); log_str_L(")\n");
#endif
	if (free_bytes > packet_bytes_to_send)
		packet_bytes_to_send = free_bytes;
	
	// Now look whether there are packets
	u16 bufptr = 0;
	u8 packets_to_send_1st = 0;
	u8 packets_to_send_2nd = 0;
	u8 actual_packet_bytes = 0;
	while (true) {
		assert(bufptr < sizeof(gl pbuf_1st));
		
		if (!gl pbuf_1st[bufptr])
			break;
		
		u16 size = gl pbuf_1st[bufptr];
#ifdef DEBUG_ALGO_DETAIL
		ulog_str_num_BASE_str_L("Check1: Packet size = ", size, DEC, "\n");
		printf("%02x|", gl pbuf_1st[bufptr]);
		u16 i;
		for(i = bufptr + 1; i < bufptr + 4 + 1; i++) {
			printf(" %02x", gl pbuf_1st[i]);
		}
		printf(" |");
		for(i = bufptr + 4 + 1; i < bufptr + size + 1; i++) {
			//fprintf(stderr, " _%d_ ", i);
			printf(" %02x", gl pbuf_1st[i]);
		}
		printf(" |(%02x)\n", gl pbuf_1st[i]);
#endif
		
		if (actual_packet_bytes + 1 + size - 4 > packet_bytes_to_send)
			break;
		
		actual_packet_bytes += 1 + size - 4;
		bufptr += 1 + size;
		packets_to_send_1st++;
	}
	bufptr = 0xCAFE;
	bufptr = 0;
	while (true) {
		assert(bufptr < sizeof(gl pbuf_2nd));
		
		if (!gl pbuf_2nd[bufptr])
			break;
		
		u16 size = gl pbuf_2nd[bufptr];
#ifdef DEBUG_ALGO_DETAIL
		ulog_str_num_BASE_str_L("Check2: Packet size = ", size, DEC, "\n");
		printf("%02x|", gl pbuf_2nd[bufptr]);
		u16 i;
		for(i = bufptr + 1; i < bufptr + 4 + 1; i++) {
			printf(" %02x", gl pbuf_2nd[i]);
		}
		printf(" |");
		for(i = bufptr + 4 + 1; i < bufptr + size + 1; i++) {
			//fprintf(stderr, " _%d_ ", i);
			printf(" %02x", gl pbuf_2nd[i]);
		}
		printf(" |(%02x)\n", gl pbuf_2nd[i]);
#endif
		
		if (actual_packet_bytes + 1 + size - 4 > packet_bytes_to_send)
			break;
		
		actual_packet_bytes += 1 + size - 4;
		bufptr += 1 + size;
		packets_to_send_2nd++;
	}
	bufptr = 0xCAFE;
	u8 packets_to_send = packets_to_send_1st + packets_to_send_2nd;
	
	
	// If you would want to speed it up, you could use the FIFO burst access
	// so write address only once and then much of data.
	
	// Length
	uint8_t send_len = 9 + actual_packet_bytes + acks_to_send * 2;
	REGFILE_DO_WRITE(TI_USE_single, cbFifo, send_len);
	
	
	uint8_t uid = get_new_uid(MHI_USE_single);
	
	// Register uid
	//   1. find free slice_reception_registry_entry
	u32 free; 
	for (free = 0; free < SLICE_RECEPTION_REGISTRY; free++) {
		if (global(srr)[free].slice == 0xFF)
			break;
	}
	if (free >= SLICE_RECEPTION_REGISTRY)
		// Very lazy - take the first entry if there's no free.
		free = 0;
	
	//   2. set tracking mark
	global(srr)[free].time_sent = actual_time;
	global(srr)[free].fuid = uid;
	global(srr)[free].slice = slice;
	
	
#ifdef DEBUG_ALGO_DETAIL
	ulog_str_num_BASE_str_L("uid = ", uid, HEXBYTE, "\n");
#endif
	uint8_t tmp1 = (global(my_address) << 4) + (uid >> 4);
	uint8_t tmp2 = (uid << 4) + 1; // 1 for first bit of top_type (=3)
	uint8_t tmp3 =   0x80 // second bit of top_type (=3)
	               | (is_main_slice ? SUB_TYPE_MAIN : SUB_TYPE_SIDE) << 5
	               | (packets_to_send ? SUB_TYPE_WITH_PACKETS
	                                  : SUB_TYPE_WO_PACKETS) << 5
	               | (packets_to_send ? packets_to_send : 1) << 2
	               | (is_main_slice ? 2 : 0)
	               | 0; // 0 for fill
	uint32_t single_step = 100 MSEC / 16;
	uint32_t ptime = get_bit_length(TI_USE_single) * 10 * 8;
	uint32_t selfconception_bits = 
		deduce_offset(global(frame_align), actual_time) + ptime / single_step;
	uint8_t final_info = (tmp1 & 0xf0) | (selfconception_bits & 0x0f);
	
	// checksum
	tmp2 |= chkH(tmp1, tmp2, tmp3, final_info) << 1;
	
	
	REGFILE_DO_WRITE(TI_USE_single, cbFifo, tmp1)
	REGFILE_DO_WRITE(TI_USE_single, cbFifo, tmp2)
	REGFILE_DO_WRITE(TI_USE_single, cbFifo, tmp3)
	
	
	// PACKETs
	bufptr = 0;
	while (packets_to_send_1st--) {
		u8 size = gl pbuf_1st[bufptr];
		//printf("size %u = %u - %u.\n", size, global(pbuf)[bufptr], bufptr);
		REGFILE_DO_WRITE(TI_USE_single, cbFifo, size - 4);
		u8 i;
		for (i = bufptr + 5; i < 1 + bufptr + size; i++) {
#ifdef DEBUG_FOR_VALGRIND
			fprintf(stderr, "write1: [%d] -> 0x%02x.\n", i, gl pbuf_1st[i]);
#endif
			REGFILE_DO_WRITE(TI_USE_single, cbFifo, gl pbuf_1st[i]);
		}
		u32 fuid; memcpy(&fuid, &(gl pbuf_1st[bufptr+1]), 4);
#ifdef DEBUG_ALGO_DETAIL
		ulog_str_num_BASE_str_L("Subpacket of 1 + ", size - 4, DEC,
								" B written to FIFO.\n");
		ulog_str_num_BASE_str_L("    With tracking mark ", fuid, HEX, "\n");
#endif
		
#ifdef DEBUG_FLOW
		struct host_info* hi = &gl rti->aodvuudd;
		pretty_packed(hi, " S     ", & gl pbuf_1st[bufptr+5], size-4, OUTBOUND);
#endif
		
		bufptr = 1 + bufptr + size;
	}
	
	if (bufptr) {
		u16 fin = bufptr;
		while (gl pbuf_1st[fin] && fin + 1 + gl pbuf_1st[fin] <= PACKET_BUFFER)
			fin += 1 + gl pbuf_1st[fin];
#ifdef DEBUG_FOR_VALGRIND
		printf("fin1 = %u\n", fin);
#endif
		
		for (i16 i = 0; i < fin           - bufptr; i++) {
			gl pbuf_1st[i] = gl pbuf_1st[i+bufptr];
			
		}
		// Just in case ...
		gl pbuf_1st[fin-bufptr] = 0;
	}
	
	
	bufptr = 0;
	while (packets_to_send_2nd--) {
		u8 size = gl pbuf_2nd[bufptr];
		//printf("size %u = %u - %u.\n", size, global(pbuf)[bufptr], bufptr);
		REGFILE_DO_WRITE(TI_USE_single, cbFifo, size - 4);
		u8 i;
		for (i = bufptr + 5; i < 1 + bufptr + size; i++) {
#ifdef DEBUG_FOR_VALGRIND
			fprintf(stderr, "write2: [%d] -> 0x%02x.\n", i, gl pbuf_2nd[i]);
#endif
			REGFILE_DO_WRITE(TI_USE_single, cbFifo, gl pbuf_2nd[i]);
		}
		u32 fuid; memcpy(&fuid, &(gl pbuf_2nd[bufptr+1]), 4);
#ifdef DEBUG_ALGO_DETAIL
		ulog_str_num_BASE_str_L("Subpacket of 1 + ", size - 4, DEC,
								" B written to FIFO.\n");
		ulog_str_num_BASE_str_L("    With tracking mark ", fuid, HEX, "\n");
#endif
		
#ifdef DEBUG_FLOW
		struct host_info* hi = &gl rti->aodvuudd;
		pretty_packed(hi, " S     ", & gl pbuf_2nd[bufptr+5], size-4, OUTBOUND);
#endif
		
		bufptr = 1 + bufptr + size;
		
/*		// Add tracking mark
		//   1. find free slice_reception_registry_entry
		u32 free; 
		for (free = 0; free < SLICE_RECEPTION_REGISTRY; free++) {
			if (global(srr)[free].slice == 0xFF)
				break;
		}
		if (free >= SLICE_RECEPTION_REGISTRY)
			// Very lazy - take the first entry if there's no free.
			free = 0;
		
		//   2. set tracking mark
		global(srr)[free].time_sent = actual_time;
		global(srr)[free].fuid = fuid;
		global(srr)[free].slice = slice;
		*/
		
	}
	if (bufptr) {
		// Move the rest of the packet buffer to the front.
		
		u16 fin = bufptr;
		while (gl pbuf_2nd[fin] && fin + 1 + gl pbuf_2nd[fin] <= PACKET_BUFFER)
			fin += 1 + gl pbuf_2nd[fin];
#ifdef DEBUG_FOR_VALGRIND
		printf("fin2 = %u\n", fin);
#endif
		
		for (i16 i = 0; i < fin           - bufptr; i++) {
			gl pbuf_2nd[i] = gl pbuf_2nd[i+bufptr];
			
		}
		// Just in case ...
		gl pbuf_2nd[fin-bufptr] = 0;
	}
	
	
	
	
	
	
	// ACKs
	uint8_t new_ptr = 0;
	if(ack_cnt)
		new_ptr = ack_cnt - 1;
	uint8_t old_ptr = 0;
	while(acks_to_send--) {
		uint8_t ptr;
		if(acks_to_send % 2)
			ptr = old_ptr++;
		else
			ptr = new_ptr--;
		// This order ensures that the newest uid will be acknowledged surely.
		//printf("ptr: %d\n", ptr);
		// Reusing tmp1, tmp2
		tmp1 = global(fss->to_acknowledge)[ptr] >> 8;
		tmp2 = global(fss->to_acknowledge)[ptr] & 0xFF;
		REGFILE_DO_WRITE(TI_USE_single, cbFifo, tmp1)
		REGFILE_DO_WRITE(TI_USE_single, cbFifo, tmp2)
	}
	if (ack_cnt)
		new_ptr++;
	uint8_t ptr = 0;
	while(old_ptr < new_ptr) {
		//printf("old_ptr=%d\n", old_ptr);
		global(fss->to_acknowledge)[ptr] = global(fss->to_acknowledge)[old_ptr];
		old_ptr++;
		ptr++;
	}
	global(fss->to_acknowledge)[ptr] = 0x8000;
	
	
	
	REGFILE_DO_WRITE(TI_USE_single, cbFifo, 0xAA)
	REGFILE_DO_WRITE(TI_USE_single, cbFifo, 0xAA)
	REGFILE_DO_WRITE(TI_USE_single, cbFifo, 'C')
	REGFILE_DO_WRITE(TI_USE_single, cbFifo, 'N')
	REGFILE_DO_WRITE(TI_USE_single, cbFifo, 1)
	REGFILE_DO_WRITE(TI_USE_single, cbFifo, final_info)
	
	// Let it send.
	REGFILE_DO_WRITE(TI_USE_single, cbTxStartCondition, COND_FifoNotEmpty); // 1
//#define PACKET_SENT_STRING "Packet sent: \"Hi!\\n\\0\"\n"
//		SDlog_dataline(PACKET_SENT_STRING, sizeof(PACKET_SENT_STRING)-1);
#ifdef DEBUG_ALGO
	ulog_str_num_BASE_str_L("Packet with length ", send_len, DEC, " finally given to tranceiver.\n");
#endif
	
	
	while (!REGFILE_DO_READ(TI_USE_single, cbPacketSent)) {
		// wait 2/3 of the packets' length
		multitimer_delay_coarse(MHI_USE_arg    (send_len+2+2+1+1) * 8 * get_bit_length(TI_USE_single) / 3 * 2);
	}
	
	// TODO: What about resetting TxStartCondition? Maybe it hinders a correct
	// sending.
	REGFILE_DO_WRITE(TI_USE_single, cbTxStartCondition, COND_FifoThreshold);// 0
	
	// Please note that we are in TX mode and will emit the preamble bits
	// till mode changes!
	
	// (at least in real life)
	
	// And we do change mode in order to reset PacketSent.
	REGFILE_DO_WRITE(TI_USE_single, cbMode, RFM_MODE_STANDBY);
	
}
#endif // -ifndef NOCODE
;


void aloha_payload_inbound(MHI_ARG    uint8_t* data, uint8_t len, float pwr_mW)
#ifndef NOCODE
{
	global(rti)->inbound_raw(
				global(rti), get_systime(MHI_USE_single), data, len, pwr_mW);
}
#endif 
;



bool aloha_payload_outbound(MHI_ARG    uint8_t* data, uint8_t len)
#ifndef NOCODE
{
	
	return true;
}
#endif
;



// Strategic part of file begins to begin here.


#define ANALY_NOFIFO  0x00
#define ANALY_CHKFAIL 0x01
#define ANALY_OK      0x02
#define ANALY_BADFORM 0x03 // Bad format. Packet too long for FIFO
#define ANALY_AGAIN   0x04 // Packet not yet received fully
// TODO TODO TODO: respect these five values.
// TODO: Make the radio receiving work.

// if *shortpacket = true, following are invalid: sender_uid, is_sender_main, valid_ack_for_us
uint8_t analyze_packet(MHI_ARG    uint8_t* errors, uint8_t *sender_addr, uint8_t* sender_uid, bool* is_sender_main, bool* valid_ack_for_us, bool* shortpacket, uint8_t* selfconception_bits, float pwr_mW) 
#ifndef NOCODE
{
	if (!REGFILE_DO_READ(TI_USE_single, cbFifoNotEmpty))
		return ANALY_NOFIFO;
	
	if (shortpacket) *shortpacket = false;
	
	// TODO: When we ensured that analyze_packet is invoked very frequently
	//  we could dramatically decrease inavlidation time (2-3 ms for high
	// bitrates)
	uint64_t invalidation_time = 100 MSEC;
	
	if (gl shadow_fifo_touch + invalidation_time < get_systime(MHI_USE_single))
		// Clear shadow FIFO when it is too old.
		gl shadow_fifo_pos = 0;
	
	
	uint8_t plen;
	if (REGFILE_DO_READ(TI_USE_single, cbPacketFormat)
				== PACKET_FORMAT_FIXEDlength) {
		plen = REGFILE_DO_READ(TI_USE_single, cbPayloadLength);
//			ulog_str_num_BASE_str_L("plen from register: ", plen, DEC, ".\n");
		assert(0); // Fixed length packets seem not to be too practical.
	}
	else {
		if (gl shadow_fifo_pos) {
			plen = gl shadow_fifo[0];
//			ulog_str_num_BASE_str_L("plen from shadow: ", plen, DEC, " (have ");
//			log_num_BASE(gl shadow_fifo_pos, DEC);
//			log_str_L(" bytes in shadow).\n");
		}
		else {
			plen = REGFILE_DO_READ(TI_USE_single, cbFifo);
//			ulog_str_num_BASE_str_L("plen from FIFO: ", plen, DEC, ".\n");
		}
	}
	
	if (plen > FIFO_SIZE) {
		// -1 as I assume that otherwise the treshold mechanism won't work.
	
#ifdef DEBUG_ALGO
#define stb(fifo_size) #fifo_size
#define st(fifo_size) stb(fifo_size)
		ulog_str_num_BASE_str_L("Warning: cann't handle packet of length ",
					plen, DEC, " (fifo has " st(FIFO_SIZE) " byte space).\n");
#endif 
		// We don't resolve that. When enough data arrives, [the collisions
		// or] RX<->TX mode changes will lead to FIFO clearings.
// 		
		// TODO: In real life we could draw data out of FIFO piece by piece.
		return ANALY_BADFORM;
	}
	
	
	REGFILE_DO_WRITE(TI_USE_single,cbFifoThreshold, plen-gl shadow_fifo_pos-1);
	// To update FifoLevel we have to perform a read/write operation
#define UNUSED __attribute__ ((unused))
	{
	REGFILE_DO_WRITE(TI_USE_single, cbFifo, 42);
	u8 UNUSED trash42 = REGFILE_DO_READ(TI_USE_single, cbFifo);
	}
	
	
	if (!(REGFILE_DO_READ(TI_USE_single, cbFifoLevel))) {
#ifdef DEBUG_ALGO_DETAIL
		ulog_str_L("Warning: packet not yet in FIFO fully.\n");
#endif
		
		if (!gl shadow_fifo_pos
				&& REGFILE_DO_READ(TI_USE_single, cbPacketFormat)
					== PACKET_FORMAT_VARIABLElength) {
			// We put back length byte.
			REGFILE_DO_WRITE(TI_USE_single, cbFifo, plen);
		}
		
		// 1+ because it shouldn't be 0 because 0 means inactive
		gl unfinished_packet = 1+get_systime(MHI_USE_single);
		
		return ANALY_BADFORM; // force rxrestart
		
		/*
		// Now copy the whole fragment to shadow fifo
		bool proceed = false;
		while (REGFILE_DO_READ(TI_USE_single, cbFifoNotEmpty)) {
			if (gl shadow_fifo_pos >= FIFO_SIZE)
				break; // Something is wrong?!
			
			if (gl shadow_fifo_pos >= plen) {
				// Something is wrong, we are in the packet-not-there-loop
				// so the packet mustn't be there.
				
				printf("Something is wrong: enough bytes read (didn't expect that).\n");
				
				proceed = true;
				break;
			}
			
			gl shadow_fifo[gl shadow_fifo_pos++]
					= REGFILE_DO_READ(TI_USE_single, cbFifo);
			CASSERT(sizeof(gl shadow_fifo) == FIFO_SIZE, ___DOUBLE_CHECK_YOUR_BUFFER_SIZE___);
		}
		
		gl shadow_fifo_touch = get_systime(MHI_USE_single);
		
		if (!proceed)
			return ANALY_AGAIN;*/
	}
	
	
	// OK, we have enough bytes for the packet.
	
	uint8_t flen = 0;
	
	uint8_t data[2 + FIFO_SIZE];
	
	while (gl shadow_fifo_pos||REGFILE_DO_READ(TI_USE_single, cbFifoNotEmpty)) {
		//printf("shadow_fifo_pos = %d [pre-read]\n", gl shadow_fifo_pos);
		
		if (gl shadow_fifo_pos) {
			data[flen] = gl shadow_fifo[--gl shadow_fifo_pos];
		}
		else {
			data[flen] = REGFILE_DO_READ(TI_USE_single, cbFifo);
		}
		
		flen++;
		
		if (flen >= sizeof(data))
			break; // Something is wrong?!
		
		if (flen >= plen)
			break; // Don't take a second packet if present.
	}
	
	if (flen < plen) {
		// This could mean that a packet is just arriving.
		
		return ANALY_NOFIFO; // shadow mechanism is not working somehow.
		
		// put back bytes
		while (flen--) {
			// Hopefully shadow FIFO is not full then.
			assert(gl shadow_fifo_pos < sizeof(gl shadow_fifo));
			
			gl shadow_fifo[gl shadow_fifo_pos++] = data[flen];
			printf("shadow_fifo_pos = %d [putback]\n", gl shadow_fifo_pos);
		
			//REGFILE_DO_WRITE(TI_USE_single, cbFifo, data[flen]);
		}
		
		return ANALY_NOFIFO;
	}
	
	if (plen == 1) {
#ifdef DEBUG_ALGO
		ulog_str_L("analyze_packet(): 1 Byte Packet\n");
#endif
		
		if (shortpacket) *shortpacket = true;
		if (sender_addr) *sender_addr = data[0] >> 4;
		if (selfconception_bits) *selfconception_bits = data[0] & 0xF;
		
		return ANALY_OK;
	}
	
	if (plen < 9) {
		
#ifdef DEBUG_ALGO
		ulog_str_num_BASE_str_L("analyze_packet(): Too short (",
								plen, DEC, "<9)\n");
#endif
		
		// TODO: in real life it could have arrived some further byte
		// while executing code from the beginning of this method till now.
		/*if (REGFILE_DO_READ(TI_USE_single, cbFifoNotEmpty)) {
			global(unfinished_packet) = 0;
			
			// There are more bytes to read
		}
		else*/
			return ANALY_BADFORM; // There cann't be any packet with len < 9.
	}
	
#ifdef DEBUG_ALGO
	ulog_str_num_BASE_str_L("Analyzing packet of len ", plen, DEC, " (");
	log_num_BASE(flen, DEC);
	log_str_L(" byte read from FIFO).\n");
#endif
	
#ifdef DEBUG_ALGO_DETAIL
	printf("analyze_packet(): data[%2d] = 0x%2x.\n", 0, data[0]);
	printf("analyze_packet(): data[%2d] = 0x%2x.\n", 1, data[1]);
	printf("analyze_packet(): data[%2d] = 0x%2x.\n", 2, data[2]);
#endif
	uint8_t sender = data[0] >>4;
	uint8_t uid_reconstructed = ((data[0] & 0x0f) <<4) | (data[1] >>4);
	uint8_t chk = (data[1] & 0x0e) >> 1;
	uint8_t top_type = (data[1] <<1) + (data[2] >>7);
	top_type &= 3;
	uint8_t sub_type = (data[2] >>5) & 3;
	uint8_t pcnt = (data[2] >>2) & 7;
	bool is_main_slice = (data[2] >> 1) & 1;
	uint8_t UNUSED fill = data[2] & 1;
	
	data[1] &= 0xF1;
	if (chk != chkH(data[0], data[1], data[2], data[flen - 1])) {
#ifdef DEBUG_ALGO
		ulog_str_L("WARNING: Checksum fail. Dropping packet.\n");
#endif
#ifdef DEBUG_ALGO_DETAIL
		printf("chksum: expected %x, calculated %x.\n", chk,
				chkH(data[0], data[1], data[2], data[flen - 1]));
#endif
		return ANALY_CHKFAIL; 
	}
	
	
	bool found = false;
	uint8_t got_uid;
	
	u8 bufptr = 3;
	
	if ((sub_type & SUB_TYPE_MASK_PACKETS) != SUB_TYPE_WITH_PACKETS)
		pcnt = 0;
	
	// PACKET in
	while (pcnt-- && bufptr < flen - 5) {
#ifdef DEBUG_ALGO
		ulog_str_num_BASE_str_L("Fetching packet ", pcnt+1, DEC, ".\n");
#endif
		
		u8 size = data[bufptr];
		if (size > flen - 9) size = flen - 9; 
#ifdef DEBUG_ALGO_DETAIL
		u8 i = 0xFF;
		while (++i < size + 1)
			printf("%02x ", data[bufptr + i]);
		printf("\n");
#endif
		
		if (bufptr + size > flen - 5) {
#ifdef DEBUG_ALGO_PACKET
			ulog_str_L("Warning: cutting length of packet.!\n");
#endif
			size = flen - bufptr - 5;
		}
		
		aloha_payload_inbound(MHI_USE_arg
					&(data[bufptr+1]), size, pwr_mW);
		bufptr += size + 1;
	}
	pcnt++; // while does decrease on time too much.
	
	if (pcnt) {
#ifdef DEBUG_ALGO_PACKET
		ulog_str_num_BASE_str_L("Warning: too few packets received: ", pcnt, DEC, " missing.\n");
#endif
	}
	
	// ACKs in
	while (5+bufptr+1 < flen) {
		uint16_t cb = data[bufptr] <<8 | data[bufptr+1];
		if (cb & 0x8000)
			break;
		bufptr += 2;
		
#ifdef DEBUG_ALGO
		ulog_str_num_BASE_str_L("Got acknowledge ", cb, HEXBYTE, ".");
#endif
		
		if ((cb&0x3FFF)>>8 == global(my_address)) {
			
#ifdef DEBUG_ALGO
			log_str_L(" That's us!");
#endif
			
			got_uid=cb & 0xFF;
			
			// check main slice uids.
			bool uid_is_old = true;
			uint64_t old_time = get_systime(MHI_USE_single);
			if (old_time < SRR_TIME_TO_LIVE)
				old_time = 0;
			else
				old_time -= SRR_TIME_TO_LIVE;
			
			
			u32 i = 0;
			while (i < NUM_UIDS) {
				if (gl old[i].uid == got_uid) {
					if (gl old[i].time > old_time)
						// If only one recent UID matches the received,
						// we're happy.
						uid_is_old = false;
					found = true;
				}
				i++;
			}
			
			if (!found) {
#ifdef DEBUG_ALGO
				ulog_fin_line();
				ulog_str_num_BASE_str_L("WARNING: sent main uids (e.g. ",
										global(uid), HEXBYTE, "");
				log_str_num_BASE_str_L(") but got uid ",
									   got_uid, HEXBYTE, ".");
#endif
			}
			else {
#ifdef DEBUG_ALGO
				log_str_L(" - correct UID");
#endif
				
				if(uid_is_old || cb & 0x4000) {
#ifdef DEBUG_ALGO
					if (uid_is_old) {
						log_str_L(", but too old");
					}
#endif
					if (cb & 0x4000) {
#ifdef DEBUG_ALGO
						log_str_L(", CONFLICT");
#endif
						// TODO: Too direct for one single conflict bit?
						gl main_slice_invalidated = true;
					}
#ifdef DEBUG_ALGO
					log_str_L("!");
#endif
					found = false;
				}
			}
			
			if (found) {
				gl aloha_slice_unreq = 0;
			}
			
#ifdef DEBUG_ALGO
			log_str_L("\n");
			
			ulog_str_num_BASE_str_L("                ", cb, HEXBYTE, " ");
#endif
			// Check side slices uid
			for (i = 0; i < SLICE_RECEPTION_REGISTRY; i++) {
				struct slice_reception_registry_entry srre = global(srr)[i];
				
				if (srre.slice == 0xFF)
					continue;
				
				if (srre.fuid == got_uid) {
#ifdef DEBUG_ALGO
					log_str_L(" - correct UID");
#endif
					if (srre.time_sent <= old_time) {
#ifdef DEBUG_ALGO
						log_str_L(", but too old!\n");
#endif
						gl srr[i].slice = 0xFF; // Free srre
						continue;
					}
					if (cb & 0x4000) {
						// TODO: Too direct for one single conflict bit?
#ifdef DEBUG_ALGO
						log_str_L(", CONFLICT\n");
#endif
						if (    srre.slice != gl main_slice
						     && srre.slice != gl side_slices[0]
						     && srre.slice != gl side_slices[1]) {
							
							// Possibly find a replacement
							request_more_slices(MHI_USE_single);
							// and invalidate the conflicted/conflicting slice.
							gl fss->slice_sack[srre.slice] = 0;
						}
						else {
							// Main slices are already handled above.
						}
						continue;
					}
#ifdef DEBUG_ALGO
					log_str_L("\n");
#endif
					if (gl fss->slice_addr[srre.slice] != gl my_address) {
						// We some time had reserved that but don't use it.
						
						if (!capacity_deficit(MHI_USE_single)) {
							// Ignore ack.
							continue;
						}
						else {
							// OK, let's re-try it.
							add_side_slice(MHI_USE_arg    srre.slice,
									CAP_UNTOUCHED);
									// untouched means = we resolve deficite
							// update_slice comes below:
						}
					}
					
					update_slice(MHI_USE_arg    
							srre.slice, BITS_SEEN, gl my_address);
					
					if (global(fss->slice_cnfl[srre.slice]))
						global(fss->slice_cnfl[srre.slice])--;
						
					
					
					if (    srre.slice == gl main_slice
					     || srre.slice == gl side_slices[0]
					     || srre.slice == gl side_slices[1]) {
						
						// This may unblock an invalidated state.
						think_about_a_new_main_slice(MHI_USE_single);
					}
				}
			}
			
		}
		else {
			// TODO: Register the received stations as precursors?
#ifdef DEBUG_ALGO
			log_str_L("\n");
#endif
		}
		
	}
	
	bufptr = flen - 6;
	
#ifdef DEBUG_ALGO_DETAIL
#define check_common(condition, data_descr) \
	if (condition) { \
		ulog_str_L("OK  " data_descr "\n"); \
	} \
	else { \
		if (errors) (*errors)++; \
		ulog_str_L("ERR " data_descr "\n"); \
	}
#else
#define check_common(condition, data_descr) \
	if(!(condition) && errors) (*errors)++;
#endif
	
#ifdef DEBUG_ALGO_DETAIL
	// So now check
	printf("analyze_packet(): data[%2d] = 0x%2x.\n", bufptr  , data[bufptr]);
	printf("analyze_packet(): data[%2d] = 0x%2x.\n", bufptr+1, data[bufptr+1]);
	printf("analyze_packet(): data[%2d] = 0x%2x.\n", bufptr+2, data[bufptr+2]);
	printf("analyze_packet(): data[%2d] = 0x%2x.\n", bufptr+3, data[bufptr+3]);
	printf("analyze_packet(): data[%2d] = 0x%2x.\n", bufptr+4, data[bufptr+4]);
#endif
	check_common(data[bufptr]   == 0xAA, "preamble:8 1");
	check_common(data[bufptr+1] == 0xAA, "preamble:8 2");
	check_common(data[bufptr+2] == 'C',  "sync:8 1");
	check_common(data[bufptr+3] == 'N',  "sync:8 2");
	check_common(data[bufptr+4] == 1,    "length:8 == 1?");
	
#ifdef DEBUG_ALGO_DETAIL
	printf("analyze_packet(): data[%2d] = 0x%2x.\n", bufptr+5, data[bufptr+5]);
#endif
	uint8_t sender_final = data[bufptr+5] >>4;
	uint8_t selfconception = data[bufptr+5] & 0x0f;
	
	check_common(sender == sender_final, "sender == sender?");
	
	
	if (sender_addr)      *sender_addr = sender;
	if (sender_uid)       *sender_uid = uid_reconstructed;
	if (is_sender_main)   *is_sender_main = is_main_slice;
	
	if (valid_ack_for_us) *valid_ack_for_us = found;
	if (selfconception_bits)
	                      *selfconception_bits = selfconception;
	
	return ANALY_OK; // Despite of possible errors all is fine.
}
#endif
; // analyze_packet()



#ifdef NOCODE
struct fafres {u8 best; u8 second_best; u8 third_best;};
#endif


// uint8_t fold_offset = 1; // 1 field before A
// uint8_t fold_tail = 2; // 2 fields after A
// int8_t fold_arr[] = {2, 5, 2, -1};
// fold offset:            ^           <- there's A
// -1: Try to "glue" before existing used slice. (with 1 slice space)
struct fafres fold_and_find(MHI_ARG
			uint8_t fold_offset, uint8_t fold_tail, int8_t fold_arr[],
			bool condensed)
#ifndef NOCODE
{
	
	// Fold
	uint8_t combined[40];
	uint8_t end = condensed ? 10 : 30;
	u8 mya = global(my_address);
	if (condensed) {
		for (int i = 0; i < 10; i++) {
			// If someone else was recently sending here.
			// (and if this is our slice and it is acknowledged.)
			combined[i+1] = 0;
			
			if (global(fss->slice_addr)[i   ]!=mya)
				combined[i+1] +=   (global(fss->slice_bits)[i   ] & BITS_SEEN);
			if (global(fss->slice_addr)[i+10]!=mya)
				combined[i+1] +=   (global(fss->slice_bits)[i+10] & BITS_SEEN);
			if (global(fss->slice_addr)[i+10]!=mya)
				combined[i+1] +=   (global(fss->slice_bits)[i+20] & BITS_SEEN);
			
			// If this is already our slice.
			combined[i+1] +=  (global(fss->slice_addr)[i   ]==mya)
							+ (global(fss->slice_addr)[i+10]==mya)
							+ (global(fss->slice_addr)[i+20]==mya);
		}
		// Data has an offset of 1 B for future folding.
		//{J', A, B, C, D, E, F, G, H, I, J, A', B', C', D', E', F', G', H', I'}
	}
	else {
		for (int i = 0; i < 30; i++) {
			
			// If this is already our slice.
			combined[i+1] =      global(fss->slice_addr)[i]==mya;
			// If someone else was recently sending here.
			if (global(fss->slice_addr)[i] != mya)
				combined[i+1] =  global(fss->slice_bits)[i] & BITS_SEEN;
		}
		// Data has an offset of 1 B for future folding.
	}
	
	
	// Copiing the data to all locations.
	combined[0] = combined[end];
	for (int i = end+1; i < end+10; i++) {
		combined[i] = combined[i-end];
	}
	
	
	assert(fold_offset <= 1);
	assert(fold_offset + fold_tail <= 9);
	
	int16_t fold_res[30];
	u8 min_pos = SIDE_SLICES_FINISH;
	u8 secondary = SIDE_SLICES_FINISH;
	u8 ternary = SIDE_SLICES_FINISH;
	for (int i = 0; i < end; i++) {
		fold_res[i] = 0;
		for (int j = 0; j < fold_offset + 1 + fold_tail; j++) {
			//fprintf(stderr, "fold_arr[%d] = %d    ", j, fold_arr[j]);
			//fprintf(stderr, "%d\n", (int16_t)combined[1 - fold_offset + i+j]);
			fold_res[i] += fold_arr[j] * (int16_t)combined[1 - fold_offset + i+j];
		}
		//fprintf(stderr, "i=%d, min_pos=%d\n", i, min_pos);
		//fprintf(stderr, "                     fold_res[%d]=%d\n", i, fold_res[i]);
		//fprintf(stderr, "                                       fold_res[%d]=%d\n", min_pos, fold_res[min_pos]);
		if (min_pos == SIDE_SLICES_FINISH || fold_res[i] < fold_res[min_pos]) {
			ternary = secondary;
			secondary = min_pos;
			min_pos = i;
		}
		else if (    secondary == SIDE_SLICES_FINISH
			      || fold_res[i] < fold_res[secondary]) {
			ternary = secondary;
			secondary = i;
		}
		else if (    ternary == SIDE_SLICES_FINISH
			      || fold_res[i] < fold_res[ternary]) {
			ternary = i;
		}
	}
	
#ifdef DEBUG_ALGO_DETAIL
	for (int i = 0; i < end; i++) {
		ulog_str_num_BASE_str_L("fold_res[", i, DEC, "] = ");
		log_num_BASE(fold_res[i], DEC);
		if (i == min_pos) {
			log_str_L(" <--- suggested slice (minimal value)\n");
		}
		else if (i == secondary) {
			log_str_L(" <---     second best suggestion\n");
		}
		else if (i == ternary) {
			log_str_L(" <---     third best suggestion\n");
		}
		else {
			log_str_L("\n");
		}
	}
#elif (defined DEBUG_ALGO)
	ulog_str_num_BASE_str_L("fold_res = ", min_pos, DEC, "/");
	log_num_BASE(secondary, DEC);
	log_str_num_BASE_str_L("/", ternary, DEC, " (out of 0-");
	log_num_BASE(end, DEC);
	log_str_L(")\n");
#endif
	
	
	return (struct fafres)
			{.best = min_pos, .second_best = secondary, .third_best = ternary};
}
#endif // -ifndef NOCODE
;



u8 choose_after_fold(MHI_ARG     struct fafres fafres)
#ifndef NOCODE
{
	u8 new_slice = fafres.best;
	if (slice_our_identity(MHI_USE_arg    new_slice)) {
		new_slice = fafres.second_best;
		if (slice_our_identity(MHI_USE_arg    new_slice)) {
			new_slice = fafres.third_best;
			if (slice_our_identity(MHI_USE_arg    new_slice)) {
				return SIDE_SLICES_FINISH;
			}
		}
	}
	return new_slice;
}
#endif // -ifndef NOCODE
;







 
// The text of the definition on the top here in graphical manner:
 
// OK:               3    5   11   17  255
//                                  ^
// Need more:        3    5   11   17  255
//                                       ^
// Need less:        3    5   11   17  255
//                             ^

// OK:             255    
//                                           ---^



#ifndef NOCODE




uint32_t unnotified_time(MHI_ARG_single) {
	// Considering there was some header time (2 byte preamble and
	// ~two byte sync word)
	uint8_t unnotified_bytes = 2; // Preamble
	unnotified_bytes += REGFILE_DO_READ(TI_USE_single, cbSyncSize);
	uint32_t unnotified_time =
				unnotified_bytes * 8 * get_bit_length(TI_USE_single);
	if (get_systime(MHI_USE_single) < unnotified_time) // sanity
		// This happens at the very begin of the simulation
		// Happens seldomly
		unnotified_time = get_systime(MHI_USE_single);
	return unnotified_time;
}


#endif // -ifndef NOCODE






void think_about_a_new_main_slice(MHI_ARG_single)
#ifndef NOCODE
{
	
	u8 p = global(main_slice); // primary
	u8 s = global(side_slices)[0]; // secondary
	u8 t = global(side_slices)[1];
	
	
	
	uint64_t old_time = get_systime(MHI_USE_single);
	if (old_time < SRR_TIME_TO_LIVE)
		old_time = 0;
	else
		old_time -= SRR_TIME_TO_LIVE;
	
	
	bool serious = false;
	if (   global(fss->slice_sack)[p] <= old_time
		|| global(fss->slice_sack)[s] <= old_time)
		// First and second slice are serious. (So I say)
		serious = true;
	
	if (   global(fss->slice_sack)[t] <= old_time && !serious) {
		// It doesn't matter if the third slice is somewhat old.
		gl main_slice_invalidated = false;
		return;
	}
	
	gl main_slice_invalidated = true;
	
}
#endif
;













// Listens for a while (max_time) and informs about the next found station.
// Clears the slice bits of the slices where no station was found.
// send_extra: set it to true if managing and using side slices is wanted.
// Returns true if a packet was found and false on timeout
// MAY ALTER packet_start or packet_end even if there's no actual packet.
bool listen_for_station(MHI_ARG    time_int fin_time,
		bool continuing, bool send_extra,
		// Return values
		float* pwr_mW, time_int* packet_start, time_int* packet_end)
#ifndef NOCODE
{
	assert(packet_start); assert(packet_end);
	
	bool packet_arriving = false;
	bool    rssi_start_set = false;
	uint8_t rssi_start = U8_UNINITIALIZED; // -Werror=maybe-uninitialized
	
#ifdef DEBUG_ALGO_DETAIL
	ulog_str_num_BASE_str_L("listen_for_station(->", fin_time, TIME, ")\n");
#endif
	
	time_int startup = get_systime(MHI_USE_single);
	
	if (global(unfinished_packet)) {
		if (global(unfinished_packet) < startup - UNFINISHED_KEEP_TIME) {
			global(unfinished_packet) = 0;
#ifdef DEBUG_ALGO_DETAIL
			ulog_str_L("listen_for_station(): clear FIFO.\n");
#endif
			REGFILE_DO_WRITE(TI_USE_single, cbFifoOverrun, 1);
			assert(!REGFILE_DO_READ(TI_USE_single, cbFifoNotEmpty));
		}
	}
	
#ifdef DEBUG_ALGO_DETAIL
	if (!continuing) {
		ulog_str_L("listen_for_station(): fresh call.\n");
	}
	else {
		ulog_str_num_BASE_str_L("listen_for_station(): continuing, unfinished_packet=",gl unfinished_packet,DEC,"\n");
	}
#endif
	
	if (continuing && ! global(unfinished_packet)){
		// Need free FIFO for sensing.
		
		packet_arriving =    REGFILE_DO_READ(TI_USE_single, cbFifoNotEmpty)
		                  || gl shadow_fifo_pos;
		
		
		if (packet_arriving) {
#ifdef DEBUG_ALGO_DETAIL
			ulog_str_L("listen_for_station(): There's a packet which already arrived!\n");
#endif
		
			*packet_start = startup - unnotified_time(MHI_USE_single);
		}
		
		/*
		ulog_str_num_BASE_str_L("fifo_free = ", fifo_free, DEC, "\n");
		if (!fifo_free) {
			ulog_str_L("listen_for_station(): ERROR: need free FIFO!\n");
			
			multitimer_delay_coarse(MHI_USE_arg    100 USEC); // for simulation
			
			return false;
		}*/
	}
	
	// Sense for packets on channel
	
	if (!continuing) {
		continuing = true;
		
		// -70 dBm -> RssiThreshold=35
		REGFILE_DO_WRITE(TI_USE_single, cbRssiThreshold, 35);
		REGFILE_DO_WRITE(TI_USE_single, cbMode, RFM_MODE_RX);
		
		multitimer_delay_coarse(MHI_USE_arg    1500 USEC); // RX Startup
		
#ifdef DEBUG_ALGO_DETAIL
		uint8_t rxready = REGFILE_DO_READ(TI_USE_single, cbRxReady);
		if (!rxready) {
			ulog_str_L("listen_for_station(): WARNING: Receiver not ready!\n");
		}
#endif
	}
	
	
#define LISTEN_TIME_SEGMENT 2500 USEC
	time_int actual_time = 0;
	bool free_time_found = false;
	while (actual_time + LISTEN_TIME_SEGMENT < fin_time) {
		
		actual_time = get_systime(MHI_USE_single);
		//ulog_str_L("Iteration: ");
		//printf("packet_arriving = %d, next_slice = %u.>\n", packet_arriving, gl next_slice);
		if (!packet_arriving) {
			// Has Rssi raised?
			bool rssi_high = REGFILE_DO_READ(TI_USE_single, cbRssi);
			uint8_t rssi = 0xFF;
			if (rssi_high) {
				// Rssi is high, so someone else is sending. This gets logged:
				rssi = REGFILE_DO_READ(TI_USE_single, cbRssiValue);
				
				ulog_str_num_BASE_str_L("RSSI: -.5*", rssi, DEC, "\n");
				
				// Some waste of CPU time (if crossing this several times)
				rssi_start     = rssi; 
				rssi_start_set = true;
			}
			
			if (REGFILE_DO_READ(TI_USE_single, cbFifoNotEmpty)) {
				packet_arriving = true;
				
				
				
				// printf ("actual_time %'lu - unnotified_time %'u = %'lu\n");
				*packet_start = actual_time - unnotified_time(MHI_USE_single);
				
				// we look for free time.
				if (startup < *packet_start)
					free_time_found = true;
				
			}
			
		}
		
		if (send_extra) {
			// Side slices
			
			
			// -5 MSEC to start 5 milliseconds after slice align
			u8 actual_slice = deduce_slice(gl frame_align, actual_time-5 MSEC);
			
			if (slice_our_identity(MHI_USE_arg     actual_slice))
				// slice our identity is equal to slice==next_slice 
				// (at least in our situation)  -> could be combined
				invalidate_uids(MHI_USE_arg    actual_slice);
			
			if (actual_slice == gl next_slice) {
				if (!gl pbuf_1st[0] && !gl pbuf_2nd[0])
					// No packets to send -> we don't need this slice
					request_fewer_slices(MHI_USE_single);
				
				u8 transmitting_slice = actual_slice;
				
				// Tell everyone that we're there
				// and transmit our data ...
				do {
					send_queue(MHI_USE_arg    is_main_slice(MHI_USE_arg actual_slice));
					
					// This wastes lots of our sending time but is needed in simulaton
					// for that the receiver can take the bytes out of FIFO.
					multitimer_delay_coarse(MHI_USE_arg    
							400 USEC + get_rx_startup_time(TI_USE_single));
					
					
					// + 10 MSEC to stop 10 milliseconds before slice align
					actual_slice = deduce_slice(
								global(frame_align),
								get_systime(MHI_USE_single) + 10 MSEC);
					
					if (actual_slice != transmitting_slice)
						break;
					
				} while (gl pbuf_1st[0] || gl pbuf_2nd[0]);
				
				
				if (gl pbuf_1st[0] || gl pbuf_2nd[0]) {
					// Vorzeitiger Abbruch; nicht alle Pakete konnten gesandt werden.
					request_more_slices(MHI_USE_single);
				}
				
				
				if (gl next_slice == transmitting_slice)
					update__next_slice(MHI_USE_arg    CHECK_TABLE);
				else {}
					// Someone else did this for us, probably the ack mechanism.
				
				
				*packet_start = actual_time;
				free_time_found = true; // All the time before sending is free.
				//packet_arriving = true; // We have transmitted and now expect other's packets.
				
				REGFILE_DO_WRITE(TI_USE_single, cbMode, RFM_MODE_RX);
				multitimer_delay_coarse(MHI_USE_arg
							100 USEC + get_rx_startup_time(TI_USE_single));
			}
		}
		
		
		if (free_time_found) {
			
			floor_frame_align(MHI_USE_arg    startup);
			
			// startup falls in slice #:
			// (-1 for later ceiling)
			// (+ 2 MSEC because assuming that we would see something on air 
			//      after those 2 ms.
			uint8_t startup_slice 
						= (startup - global(frame_align) - 1 + 2 MSEC)
							/ (100 MSEC);
				
				// + 10 MSEC in order to ignore packets which are 10 ms too 
				// early with respect to their actual slice.
			uint8_t end_slice
						= (*packet_start + 10 MSEC - global(frame_align)) / (100 MSEC);
#ifdef DEBUG_ALGO_DETAIL
			printf("listen(): frame %'lu, lstart %'lu, lend %'lu\n",
						global(frame_align), startup, *packet_start);
#endif
			
			startup_slice++; // Ceil
			if (end_slice) // sanity
				end_slice--; // Floor
			
			if (startup_slice > end_slice) {
#ifdef DEBUG_ALGO_DETAIL
				ulog_str_num_BASE_str_L("No free slice from           ",
										startup_slice, DEC, "");
				log_str_num_BASE_str_L(" to ", end_slice, DEC, "\n");
#endif
			}
			else {
#ifdef DEBUG_ALGO
				ulog_str_num_BASE_str_L("Listened to free slices from ",
										startup_slice, DEC, "");
				log_str_num_BASE_str_L(" to ", end_slice, DEC, "\n");
#endif
				
				while (startup_slice <= end_slice) {
					// printf("Free slice: %u\n", startup_slice % 30);
					u8 sl = startup_slice%30;
					
					u8 addr = global(fss->slice_addr)[sl];
					u8 bits = global(fss->slice_bits)[sl];
					if (addr != global(my_address) && bits & 0x14) {
						// Last two times seen, now not seen -> but keep record
						update_slice_bits(MHI_USE_arg  sl, 0);
					}
					else if (addr != global(my_address)) {
						// At most seen once for last three times -> delete
						update_slice(MHI_USE_arg  sl, 0, PSEUDO_ADDRESS_FREE);
					}
					
					startup_slice++;
				}
			}
			
			// We make this only once.
			free_time_found = false;
			startup = get_systime(MHI_USE_single);
		}
		
		if (packet_arriving) {
			
			
			if(REGFILE_DO_READ(TI_USE_single, cbPayloadReady)) {
				packet_arriving = false;
				*packet_end = actual_time;
				
				uint8_t rssi = REGFILE_DO_READ(TI_USE_single, cbRssiValue);
				if (!rssi_start_set)
					 rssi_start = rssi;
#ifdef DEBUG_ALGO
				ulog_str_num_BASE_str_L("Packet arrived with RSSI -.5*",
										rssi, DEC, "\n");
#endif
				//printf("rssi=%d\n", rssi);
				//printf("rssi_start=%d\n", rssi_start);
				*pwr_mW = 1.0 
						/ exp10f( (((float)rssi_start/2)+((float)rssi/2))/2);
				
				return true;
			}
			
			// Payload is not ready, check unfinished_packet timer.
			
			// ---  slice 4  ----------------------  |  -------- slice 5 -------
			//        >UNFINISHED<              >---CHECK---< 
			//                   <---min 30ms--><5ms>|<-5ms->
			
			// Checking around frame change for unfinished packets.
			
			uint32_t offset = deduce_offset(global(frame_align), actual_time);
			if (offset > 95 MSEC || 5 MSEC < offset) {
				if (global(unfinished_packet)
						&& global(unfinished_packet) + UNFINISHED_KEEP_TIME
								< actual_time) {
					// clear FIFO
					REGFILE_DO_WRITE(TI_USE_single, cbFifoOverrun, 1);
					assert(!REGFILE_DO_READ(TI_USE_single, cbFifoNotEmpty));
					global(unfinished_packet) = 0;
#ifdef DEBUG_ALGO_DETAIL
					ulog_str_L("Clearing unfinished packet from FIFO.\n");
#endif 
				}
			}
			
		}
		
		
		multitimer_delay_coarse(MHI_USE_arg    LISTEN_TIME_SEGMENT);
		
		
		//actual_time = get_systime(MHI_USE_single);
	}
	
	// Yet anything there.
	return false;
}
#endif
;



#ifndef NOCODE

// Newly finds a slice.
// This will block for ~ 3 s.
// This will clear all frame/slice information.
int find_first_free (MHI_ARG_single) {
	
	uint32_t UNUSED slicetimeoffset[30];
	
	for (int i = 0; i < 30; i++) {
		global(fss->slice_addr)[i] = PSEUDO_ADDRESS_FREE;
		global(fss->slice_bits)[i] = 0;
		global(fss->slice_sack)[i] = 0;
		slicetimeoffset[i] = 0;
	}
	global(fss->slice_time_offset) = 0;
	global(fss->to_acknowledge)[0] = 0x8000; // invalid.
	
	
	global(frame_align) = get_systime(MHI_USE_single); // rather constant then
	time_int actual_time = global(frame_align);
	
	bool first = true;
	
	while (actual_time < global(frame_align) + 3 SEC) {
		time_int packet_start, packet_end;
		uint8_t selfconception_bits = 0xFF; // -Werror=maybe-uninitialized
		uint8_t sender = 0xFF; // -Werror=maybe-uninitialized
		float pwr_mW;
		
		bool success;
		success = listen_for_station(MHI_USE_arg
					(uint32_t)(global(frame_align) + 3 SEC),
					!first, false,
					&pwr_mW, &packet_start, &packet_end);
		
		if (!success) {
			first = false; // TODO: Need this here? Inhibits slice sync!
#ifdef DEBUG_ALGO
			ulog_str_L("No success.\n");
#endif
			actual_time = get_systime(MHI_USE_single);
			continue; // There will be some wait in listen_for_station.
		}
		
		u8 ares = success ? ANALY_OK : ANALY_NOFIFO;
		while(ares == ANALY_OK || ares == ANALY_CHKFAIL)
			ares = analyze_packet(MHI_USE_arg  NULL, &sender, NULL, NULL, NULL, NULL, &selfconception_bits, pwr_mW);
		/*
		if (!success) {
			actual_time = get_systime(MHI_USE_single);
			continue; // There will be some wait in listen_for_station.
		}*/
		
		
		uint32_t timediff_start = packet_start - global(frame_align);
		//uint32_t timediff_end   = packet_end   - frame_align; // unused
		
		
		// No numerical problem
		uint32_t selfconception = selfconception_bits * 100 MSEC / 16;
		
		
		uint32_t packet_duration = packet_end - packet_start;
		
		// e. g. we receive the packet with duration = 80 ms
		// self conception of sender is 70 ms
		// so for us the sender has a delayed start of 10 ms.
		
		int32_t self_start_misconception =
				(int32_t)packet_duration - (int32_t)selfconception;
		
		if (self_start_misconception > 10 MSEC) {
			ulog_str_num_BASE_str_L("Warning: self_start_misconception (",
					self_start_misconception, DEC, ") is very high!\n");
		}
		
		
		// So we adjust this to senders' view.
		timediff_start -= self_start_misconception;
		
		
		
		if (first) {
			// Adjust slice synchronization to the first (German: erstbesten)
			// packet we see.
			
			if (packet_duration > 100 MSEC) {
				ulog_str_num_BASE_str_L("Warning: first packet duration (", packet_duration, DEC, ") is strange!\n");
			}
			
			uint32_t remainder = timediff_start % (100 MSEC);
			
			// HERE WE slightly ADJUST STARTTIME
			global(frame_align) += remainder - 20 MSEC;
			// -20 MSEC in order to fit the following packets directly into
			// the right frame, assuming that they will be ~20 ms sooner or
			// later
			
			timediff_start -= remainder - 20 MSEC;
			
			first = false;
		}
		else {
			
		}
		
		// find the slice
		uint8_t slice_number = (uint32_t)timediff_start / (uint32_t)(100 MSEC);
		
		
		
		slicetimeoffset[slice_number] = timediff_start;
		
		
		// TODO: Don't need to be so careful with ancient data because 
		// ancient data hase been wiped out 90 lines ago.
		global(fss->slice_bits)[slice_number] <<= 2;
		
		if (global(fss->slice_addr)[slice_number] != sender) {
			global(fss->slice_addr)[slice_number] = sender;
			global(fss->slice_bits)[slice_number] |= 
					  (BITS_WRONG_STATION << 6)
					| (BITS_WRONG_STATION << 4)
					| (BITS_WRONG_STATION << 2);
			// Of course this is not fully correct. If a station ocurres again
			// after a short interlude by another station, the old occurences
			// anyway will be marked as unknown by this operation.
		}
		
		global(fss->slice_bits)[slice_number] |= BITS_SEEN;
		
		
		ulog_str_num_BASE_str_L("Slice ", slice_number, HEXBYTE, ": with addr ");
		log_num_BASE(global(fss->slice_addr)[slice_number], DEC);
		log_str_num_BASE_str_L(" has time offset = ", timediff_start, DEC, "\n");
		
		actual_time = get_systime(MHI_USE_single);
	}
	
	
	frame_slice_table(MHI_USE_arg    COMPACT);
	
	
	uint8_t fold_offset = 1; // 1 field before A
	uint8_t fold_tail   = 2; // 2 fields after A
	int8_t fold_arr[] = {2, 5, 2, -1};
	// fold offset:         ^           <- there's A
	// -1: Try to "glue" before existing used slice. (with 1 slice space)
	
	
	struct fafres fafres = fold_and_find(MHI_USE_arg   
				fold_offset, fold_tail, fold_arr, true);
	u8 new = fafres.best;
	
	update_slice(MHI_USE_arg    new, 1, global(my_address));
	
	global(aloha_slice_unreq) = 0;
	
	return new;
} // find_first_free





void CB_framefind(MHI_ARG    uint8_t reason) {
	
	//TODO: This looks rather like a main loop with state machine and does
	//not belong into an CB method. Oder?
#ifdef DEBUG_ALGO
	ulog_str_num_BASE_str_L("CB_framefind(reason=", reason, DEC, ")\n");
#endif
	
	if (reason == REASON_FRAMEFIND_START) {
		
		// blocks for 3 seconds
		global(main_slice) = find_first_free(MHI_USE_single);
		global(side_slices)[0] = global(main_slice) + 10;
		global(side_slices)[1] = global(main_slice) + 20;
		global(side_slices)[2] = SIDE_SLICES_FINISH;
		//print_side_slices(MHI_USE_single);
		
		gl main_slice_invalidated = false;
		
		update_slice(MHI_USE_arg    gl main_slice,  BITS_UNSENT, gl my_address);
		update_slice(MHI_USE_arg    gl main_slice+10,BITS_UNSENT,gl my_address);
		update_slice(MHI_USE_arg    gl main_slice+20,BITS_UNSENT,gl my_address);
		/*global(fss->slice_addr)[global(main_slice)] = global(my_address);
		global(fss->slice_bits)[global(main_slice)] = 0x55; // 0x55 is "perfect"
		global(fss->slice_addr)[global(main_slice)+10] = global(my_address);
		global(fss->slice_bits)[global(main_slice)+10] = 0x55;
		global(fss->slice_addr)[global(main_slice)+20] = global(my_address);
		global(fss->slice_bits)[global(main_slice)+20] = 0x55;*/
		
#ifdef DEBUG_ALGO_DETAIL
		print_side_slices(MHI_USE_arg    CHECK_TABLE);
#endif
		
		gl next_slice = gl main_slice + 10;
		
		reason = REASON_FRAMEFIND_TRY;
	}
	
	if (   reason == REASON_FRAMEFIND_TRY
	    || reason == REASON_FRAMEFIND_TRYWAIT
	    || reason == REASON_FRAMEFIND_KEEPALIVE) {
		
		uint64_t actual_time = get_systime(MHI_USE_single);
#define UPDATE_SLICE_TIME(uint64_t) \
		floor_frame_align(MHI_USE_arg    actual_time); \
		uint64_t slice_time =   global(frame_align) \
		                      + global(main_slice) * 100 MSEC; \
		if (slice_time < actual_time) \
			slice_time += 3 SEC;
		UPDATE_SLICE_TIME(uint64_t);
		
		
		bool continuing = false;
		bool ack_present = false;
		
		// TODO: It would be smart to operate on a sub-frame basis.
		// (specially if ACK come often+fast there's much better operation.)
		
		while (actual_time < slice_time) {
			time_int packet_start, packet_end;
			float pwr_mW;
			
			packet_start = slice_time; // For later
			bool res = listen_for_station(MHI_USE_arg    slice_time,
					continuing, true,
					// Return values
					&pwr_mW,
					&packet_start, &packet_end);
#ifdef DEBUG_ALGO_DETAIL
			ulog_str_L("listen_for_station() returned.\n");
#endif
			continuing = true;
			
			
			
			uint8_t sender, selfconception_bits;
			uint8_t sender_uid = U8_UNINITIALIZED;
			bool is_sender_main = BOOL_UNITIALIZED;//-Werror=maybe-uninitialized
			bool rxrestart = false;
			bool shortpacket;
			
			while (res) {
				uint8_t errors = 0;
				
				u8 ares = res ? ANALY_CHKFAIL : ANALY_NOFIFO;
				if (ares != ANALY_NOFIFO) {
					ares = analyze_packet(MHI_USE_arg  &errors,
								&sender, &sender_uid, &is_sender_main,
								&ack_present, &shortpacket,
								&selfconception_bits, pwr_mW);
					
				}
				if (         ares == ANALY_OK) {
					uint8_t slice;
					uint32_t slicetimeoffset = calc_slicetimeoffset(MHI_USE_arg
							packet_start,
							packet_end - (selfconception_bits * 100 MSEC / 16),
							&slice);
					
					
					if (sender == gl my_address) {
#ifdef DEBUG_ALGO_DETAIL
						ulog_str_L("Error: I cannot receive messages from myself!\n");
#endif
						continue;
					}
					
					if (errors <= 2)
						register_station(MHI_USE_arg     slice,
									sender, sender_uid, is_sender_main,
									shortpacket, slicetimeoffset);
					//rxrestart = true;
					res = true;
				}
				else if (    ares == ANALY_BADFORM) {
					
					rxrestart = true;
					res = false;
				}
				else if (    ares == ANALY_NOFIFO
				          || ares == ANALY_AGAIN) {
					res = false;
				}
				else if (    ares == ANALY_CHKFAIL) {
					//rxrestart = true;
					res = true;
					// Maybe next packet has better checksum?
				}
				else {
					assert(0); // Please add new modes here if you add them 
					//            to analyze_packet().
				}
				
				// There was nothing in a timespan
				// from actual_time -> packet_start (res = true)
				// from actual_time -> slice_time   (res = false)
				//    (optimization: packet_start = slice_time | res = false)
				
				
			} // while (res)
			
			if (rxrestart) {
				// Main reason to do this is to clear FIFO
				
				REGFILE_DO_WRITE(TI_USE_single, cbMode, RFM_MODE_STANDBY);
				multitimer_delay_coarse(MHI_USE_arg    50 USEC);
				
				REGFILE_DO_WRITE(TI_USE_single, cbMode, RFM_MODE_RX);
				uint64_t Tstart = get_rx_startup_time(TI_USE_single);
				//ulog_str_num_BASE_str_L("Tstart = ", Tstart, TIME, "\n");
				multitimer_delay_coarse(MHI_USE_arg    Tstart);
			}
			
			
			//multitimer_delay_coarse(MHI_USE_arg    3 MSEC);
#ifdef SIMUL
			multitimer_delay_coarse(MHI_USE_arg    150 USEC);
#endif
			
			actual_time = get_systime(MHI_USE_single);
			
			
			if (gl main_slice_invalidated) {
				
				reason = REASON_FRAMEFIND_KEEPALIVE;
				break;
			}
			
			// UPDATE_SLICE_TIME()
			//printf(">>actual_time = %'lu\n", actual_time);
			//printf(">>slice_time = %'lu\n", slice_time);
			
			
		}
		
		//print_side_slices(MHI_USE_arg    CHECK_TABLE);
		
		
		// Analyze what we found
		
		if (reason == REASON_FRAMEFIND_TRY) {
			reason =  REASON_FRAMEFIND_TRYWAIT;
			global(aloha_slice_unreq) = 0;
		}
		else {
			//if (global(fss->slice_sack(global(main_slice)))
			//			>= actual_time - SRR_TIME_TO_LIVE) {
			if (!gl main_slice_invalidated) {
#ifdef DEBUG_ALGO
				ulog_str_L("ALOHA: Main slice seems healthy.\n");
				//frame_slice_table(MHI_USE_arg     COMPACT);
#endif
				reason = REASON_FRAMEFIND_KEEPALIVE;
				global(aloha_slice_unreq) = 0;
			}
/*			if (ack_present) {
				global(aloha_slice_unreq) = 0;
				register_station(MHI_USE_arg
							global(main_slice), global(my_address), 0, 0);
#ifdef DEBUG_ALGO
				ulog_str_L("ALOHA: Ack is present.\n");
				frame_slice_table(MHI_USE_arg     COMPACT);
#endif
				reason = REASON_FRAMEFIND_KEEPALIVE;
			}*/
			else {
				// REASON_FRAMEFIND_KEEPALIVE
				// REASON_FRAMEFIND_TRYWAIT
				
				
				global(aloha_slice_unreq)++;
				
				bool do_hop = false;
				uint8_t fold_offset;
				uint8_t fold_tail;
				int8_t * fold_arr;
				uint8_t jmp; // Last resort: how many channels to jump forward.
				
				if (reason == REASON_FRAMEFIND_TRYWAIT) {
					ulog_str_num_BASE_str_L("WARNING: Tried to reserve slice ",
							   global(main_slice), DEC,
							   " but didn't get ACK so far!\n");
					
					
					// Stay at least once one slice (so that others can ack).
					if (global(aloha_slice_unreq) > 1)
						do_hop = hop(MHI_USE_single);
					
					
					fold_offset = 1; // 1 field before A
					fold_tail   = 5; // 5 fields after A
				//	fold_arr    =  (i8[]) {1, -3, 1, 0, 0, 1, 5, 1};
				//	fold_arr    = (i8[]) {0,  0, 0, 0, 0, 1, 0, 1, -1};
					fold_arr    = (i8[]) {1,  5,  2,  -2, 0, 0, 0};
					//        that's us  ->   ^ 
				// we want to go two before this   ->  ^   (keep this at -2)
					
					jmp = 4;
				}
				else if (reason == REASON_FRAMEFIND_KEEPALIVE) {
#ifdef DEBUG_ALGO
					// Don't get ACK for our reserved channel -> complain.
					ulog_str_num_BASE_str_L("WARNING: Didn't get ACK for our reserved slice ",
							   global(main_slice), DEC, "!\n");
#endif
					
					
					if (global(aloha_slice_unreq) > 2) {
						do_hop = hop(MHI_USE_single);
						// But somewhen we have to stop complaining and to hop.
					}
					
					// TODO: Tune
					fold_offset = 1; // field before us
					fold_tail   = 5; // fields after us
				//	fold_arr    = (i8[]) {1, -3, 1, 0, 0, 1, 5, 1};
				//	fold_arr    = (i8[]) {0,  0, 0, 0, 0, 1, 0, 1, -1};
					fold_arr    = (i8[]) {1,  5,  2,  -2, 0, 0, 0};
					//        that's us  ->   ^ 
				// we want to go two before this   ->  ^   (keep this at -2)
					
					jmp = 3;
				}
				
				
				if (do_hop) {
					
					u8 add_bits =   (hop(MHI_USE_single) << 1)
					               + hop(MHI_USE_single);
					if (fold_arr[3] == -2) { // sanity check
						fold_arr[3] = 0;
						fold_arr[3+add_bits] = -2;
					}
					jmp += add_bits;
					
					struct fafres fafres = fold_and_find(MHI_USE_arg
								fold_offset, fold_tail, fold_arr, true);
					u8 new_slice = choose_after_fold(MHI_USE_arg    fafres);
					
					
					if (new_slice == SIDE_SLICES_FINISH) {
						// Kind of last resort
						// Danger of going to someone other's slice.
						new_slice = (new_slice
									+ 1 + (global(hop_state))*jmp) % 10;
						
						
						ulog_str_L("All propositions are already our own -> jumping.\n");
					}
					
#ifdef DEBUG_ALGO
					ulog_str_num_BASE_str_L("Setting new main slice to ", new_slice, DEC, ".\n");
#endif
					
					
					u8 ns = new_slice;
					u8 ms = global(main_slice);
					
					
					
					update_slice(MHI_USE_arg    ms,    0, PSEUDO_ADDRESS_FREE);
					update_slice(MHI_USE_arg    ms+10, 0, PSEUDO_ADDRESS_FREE);
					update_slice(MHI_USE_arg    ms+20, 0, PSEUDO_ADDRESS_FREE);
					
					update_slice(MHI_USE_arg   ns,   BITS_UNSENT,gl my_address);
					update_slice(MHI_USE_arg   ns+10,BITS_UNSENT,gl my_address);
					update_slice(MHI_USE_arg   ns+20,BITS_UNSENT,gl my_address);
					
					
					u8  tmp = which_side_slice(MHI_USE_arg   ns);
					if (tmp != SIDE_SLICES_FINISH)
						remove_side_slice(MHI_USE_arg  ns, CAP_UNTOUCHED, false);
					tmp = which_side_slice(MHI_USE_arg   ns+10);
					if (tmp != SIDE_SLICES_FINISH)
						remove_side_slice(MHI_USE_arg  ns+10, CAP_UNTOUCHED, false);
					tmp = which_side_slice(MHI_USE_arg   ns+20);
					if (tmp != SIDE_SLICES_FINISH)
						remove_side_slice(MHI_USE_arg  ns+20, CAP_UNTOUCHED, false);
					
					
					gl main_slice = new_slice;
					gl side_slices[0] = ns+10;
					gl side_slices[1] = ns+20;
					
#ifdef DEBUG_ALGO
					print_side_slices(MHI_USE_arg    CHECK_TABLE);
#endif
					
					gl main_slice_invalidated = false;
					
					global(aloha_slice_unreq) = 0; // TODO: Do we need this?
					
					reason = REASON_FRAMEFIND_TRY;
					goto wakeup;
					
				} // if        -> do_hop = true
				
			} // if-else   -> main_slice_invalidated == true
			
		} // if-else  -> reason == TRYWAIT / reason == KEEPALIVE
		
#ifdef DEBUG_ALGO
		print_side_slices(MHI_USE_arg    CHECK_TABLE);
		//frame_slice_table(MHI_USE_arg    COMPACT);
#endif
		
		if (gl main_slice != deduce_slice(gl frame_align, actual_time))
			goto wakeup;
		
		
#if 0
//#ifdef DEBUG_ALGO_DETAIL
		// To send at least 1 mini packet right now
		ulog_str_num_BASE_str_L(
				"Sending @slice ", gl main_slice, DEC, " (probably)\n");
#endif
		assert(gl main_slice == deduce_slice(gl frame_align,
											 get_systime(MHI_USE_single)));
		
		do {
			// This will happen also in listening loop.
			send_queue(MHI_USE_arg    true);
			
			// This wastes lots of our sending time but is needed in simulaton
			// for that the receiver can take the bytes out of FIFO.
			multitimer_delay_coarse(MHI_USE_arg    
							400 USEC + get_rx_startup_time(TI_USE_single));
			
			u8 slice_now = deduce_slice(
						global(frame_align),
						get_systime(MHI_USE_single) + 10 MSEC);
			if (slice_now != global(main_slice))
				break;
		} while(gl pbuf_1st[0] || gl pbuf_2nd[0]);
		
		if (gl pbuf_1st[0] || gl pbuf_2nd[0]) {
			// Vorzeitiger Abbruch; nicht alle Pakete konnten gesandt werden.
			request_more_slices(MHI_USE_single);
		}
		
		
		global(next_slice) = global(side_slices[0]);
		if (global(side_slices[2]) < global(next_slice))
			global(next_slice) = global(side_slices[2]);
			
		
wakeup:
		
		multitimer_wakemeup(MHI_USE_arg    CBNUM_FRAMEFIND, 1 MSEC, reason);
	}
	
	
} // CB_framefind

#endif // -ifndef NOCODE
;





#ifdef BEAGLE
// cli-dev.c of aodv-uu expects these two methods without _FS:
#define inbound_unpacked_FS inbound_unpacked
#define outbound_raw_FS outbound_raw
#endif

#ifndef NOCODE
// 
void inbound_unpacked_FS(struct route_instance * ri, uint64_t time, void* data, uint8_t len, uint16_t seq, uint8_t sender, uint8_t destination, float power_in_mW) {
	//struct station* mh_station_arg = ri->station;
	// TODO: This is for us! Final treatment (e. g. display content, check data integrity).
	
	amr_prettyprint(&ri->aodvuudd, data, seq, sender);
	
	/*ulog_str_num_BASE_str_L("Got packet from ", sender, DEC16, " to ");
	if (destination == PACKED_BROADCAST) {
		log_str_L("ALL");
	}
	else {
		log_num_BASE(destination, DEC16);
	}
	
	log_str_num_BASE_str_L(" w/ seq #", seq, DEC16, ".\n");
	*/
}


// TODO: Implemented the testrun (data=NULL) in inefficient way.
bool outbound_raw_FS(struct route_instance * ri, uint8_t* data, uint8_t len, uint32_t tracking_mark) {
#ifdef SIMUL
	struct station* mh_station_arg = ri->station; // for ulog,  equiv. MHI_ARG
#endif
	
	
#ifdef DEBUG_ALGO
	printf("AODVroute -> ALOHA: Enque packet of length %u.\n", len);
#endif
	
	uint8_t * pbuf = len < 16 ? gl pbuf_1st : gl pbuf_2nd;
	
#ifdef DEBUG_FOR_VALGRIND
	ulog_str_P(pbuf == gl pbuf_1st ? "Using primary pbuf.\n"
	                               : "Using secondary pbuf.\n");
#endif
	
	// We have to walk through the buffer.
	u16 bufptr = 0;
	while (bufptr < PACKET_BUFFER && pbuf[bufptr]) {
#ifdef DEBUG_FOR_VALGRIND
//#define OUT stderr
#define OUT stdout
			fprintf(OUT, "pbuf[%u] = %2x\n", bufptr, pbuf[bufptr]);
#endif
			bufptr += pbuf[bufptr] + 1;
	}
	
#ifdef DEBUG_FOR_VALGRIND
	ulog_str_num_BASE_str_L("Next free buf pos: ", bufptr, DEC, "\n");
#endif
	
	if (PACKET_BUFFER - bufptr < 1+4+len) {
		if (data) { // This is no testrun
			ulog_str_P(pbuf == gl pbuf_1st
		             ? "FrameSwitch.c: outbound_raw_FS(): primary buffer:\n"
		             : "FrameSwitch.c: outbound_raw_FS(): secondary buffer:\n");
			ulog_str_L("                    Buffer full, packet not queued.\n");
		}
		return false; // TODO: Error status.
	}
	if (!data)
		return true; // testrun. We have enough space to buffer.
	
	pbuf[bufptr] = 4 + len;
	memcpy(pbuf + bufptr + 1, &tracking_mark, 4);
	memcpy(pbuf + bufptr + 5, data, len);
	
	if (bufptr + 4 + len < PACKET_BUFFER)
		pbuf[1+bufptr+4+len] = 0;
	
#ifdef DEBUG_FOR_VALGRIND
	fprintf(OUT, "FS @%4u:_%2x_\n", bufptr, pbuf[bufptr]);
	u16 i = bufptr + 1 -1; 
	while (++i < bufptr+5)
		fprintf(OUT, "FS @%4u: %2x#\n", i, pbuf[i]);;
	i = bufptr + 5 -1; 
	while (++i < bufptr+5+len)
		fprintf(OUT, "FS @%4u: %2x \n", i, pbuf[i]);
	fprintf(OUT, "FS @%4u:=%2x=\n", bufptr+5+len, pbuf[bufptr + 5 + len]);
#endif
	return true;
}

#endif // -ifndef NOCODE

#ifdef SIMUL
#ifndef NOCODE
#include "FRAMESWITCHapp/subintern.h"
#endif

// Thanks https://stackoverflow.com/a/28880394 (in reply to:)
// https://stackoverflow.com/questions/386220/how-can-i-hide-defined-but-not-used-warnings-in-gcc/610141#610141
void __attribute__ ((unused)) FS_destroy(struct app_instance * ai)
#ifndef NOCODE
{
	struct station* mh_station_arg = ai->station;
	
	gl rti->destroy(gl rti);
	
	
	ulog_str_num_BASE_str_L("Destroying FrameSwitch; Final stat of trancv: "
				" collected ", TI_USE_single->total_dBms, FLOAT, " dBms.\n");
	
	ai->old_destroy(ai);
}
#endif // -ifndef NOCODE
;
#endif // -ifdef SIMUL

#if !(defined SIMUL)
APP_GLOBAL_VARS()
#endif

#ifdef SIMUL
void __attribute__ ((unused)) FSapp_setup(MHI_ARG_single)
#else
void app_setup(MHI_ARG_single)
#endif
#ifndef NOCODE
{
	// put your setup code here, to run once:
	
#define SETUP_DESCR "FrameSwitch.c"
	ulog_str_L("STARTUP Sequence of " SETUP_DESCR "\n");
	
	
#ifdef SIMUL
	mh_station_arg->mgs->app = malloc(sizeof (struct app_global));
	if (mh_station_arg->mgs->app == NULL) {
		ulog_str_L("Couldn't allocate memory!\n");
		quit_with_error(MHI_USE_arg   sharedERRNO_STARTUP);
		return; // unneeded
	}
#endif
	
	// Initialize UIDs
	for (int i = 0; i < NUM_UIDS; i++) 
		gl old[i].time = 0;
	gl uid = station_name; // Anything can be chosen here.
	//printf("station_name = %c // uid=%d\n", station_name, global(uid));
	get_new_uid(MHI_USE_single); // -> scrumbles station_name
	
	
	
	// Initialize Routing
	
	
#ifdef SIMUL
	global(rti) = new_AODVroute(mh_station_arg);
#else
	global(rti) = new_AODVroute(NULL);
#endif
	int res = start_AODVroute(global(rti));
	if (res != OBJECT_START_OK) {
		ulog_str_L("Couldn't start AODVroute!\n");
		quit_with_error(MHI_USE_arg   sharedERRNO_STARTUP);
		return; // unneded
	}
	
	gl rti->inbound_unpacked = inbound_unpacked_FS;
	gl rti->outbound_raw = outbound_raw_FS;
	assert2(global(rti)->aodvuudd.signature == AODV_STRUCT_SIG, "YOUR_COMPILE_DEFINES_ARE_WRONG");
	
	global(fss) = malloc(sizeof (struct frameswitch_state));
	if (global(fss) == NULL) {
		ulog_str_L("Couldn't allocate memory! (Very improbable at this time)\n");
		quit_with_error(MHI_USE_arg   sharedERRNO_STARTUP);
		return; // unneded
	}
	memset(gl fss, 0xFF, sizeof(struct frameswitch_state));
	
	// Address
	get_addr_byte(global(my_address), 3,
				  global(rti)->aodvuudd.devs[0].ipaddr.s_addr);
	
	// Packet buffers
	gl pbuf_1st[0] = 0x00;
	gl pbuf_2nd[0] = 0x00;
	
	
	// Pseudo random
	global(hop_state) = 0;
	global(rand_state) = global(my_address) << 12; // Just something
	
	// Slices
	
	
	global(frame_align) = 0;
	
	memset(& global(srr), 0xFF, sizeof(struct slice_reception_registry_entry)*SLICE_RECEPTION_REGISTRY);
	
	
	
	
	global(amr).starting_time = get_systime(MHI_USE_single);
	
	if (global(my_address) == 5)
		init_amrload(&global(amr));
	else
		noinit_amrload(&global(amr));
	
	
	global(slice_need) = 1;
	global(side_slices)[0] = SIDE_SLICES_FINISH;
	
	global(last_reg_slice) = 0xFF;
	
	
	global(unfinished_packet) = 0;
	
	gl shadow_fifo_pos = 0;
	gl shadow_fifo_touch = 0;
	
	// Set the registers accordingly
	
	
	// Safety check: is this RFM 69 and is it working?
	/*volatile*/
	uint8_t boost = REGFILE_DO_READ(TI_USE_single, cbSensitivityBoost);
	// AH! Once we declare boost as volatile the cbSensitivityBoost variable
	//  will be detected properly. 
	// OK, we don't need boost to be volatile but the read function 
	// -> see ardu_spi_read_single().
	/*ulog_num_BASE(cbSensitivityBoost, HEX);
	log_str_L(" <- cbSensitivityBoost\n");
	ulog_num_BASE(regfile_is_read(cbSensitivityBoost), DEC);
	log_str_L(" <-   regfile_is_read?\n");
	ulog_num_BASE(regfile_name(cbSensitivityBoost), HEXBYTE);
	log_str_L(" <-   regfile_name.\n");
	CASSERT(cbSensitivityBoost == 0x0586FF00, OnlySend);
	CASSERT(regfile_name(cbSensitivityBoost) == 0x58, OnlySend);*/
	if (boost != 0x1B && boost != 0x2D) {
		ulog_str_L("This is not RFM69! Quitting.\n");
		quit_with_error(MHI_USE_arg   ERRNO_RXSTART);
		return;
	}
	
	
	// RX Part
	
	// "High Power settings MUST be turned off when using PA0, and in Receive mode"
	REGFILE_DO_WRITE(TI_USE_single, cbRegOcp, 0x17)
	REGFILE_DO_WRITE(TI_USE_single, cbRegTestPa1, 0x55)
	REGFILE_DO_WRITE(TI_USE_single, cbRegTestPa2, 0x70)
	
	
	REGFILE_DO_WRITE(TI_USE_single, cbPa1On, 0)
	REGFILE_DO_WRITE(TI_USE_single, cbPa2On, 0)
	REGFILE_DO_WRITE(TI_USE_single, cbPa0On, 1)
	// Now power mode 1 0 0, see Datasheet p. 21
	// Power Range: -18 to +13 dBm; Pout Formula -18 dBm + OutputPower
	REGFILE_DO_WRITE(TI_USE_single, cbOutputPower, 28); // (-18 + 28) dBm = 10 dBm. (= 10 mW)
	
	
	// Set frequency to 433.4 MHz
	set_frequency(TI_USE_arg    COMMON_FREQUENCY);
	set_deviation(TI_USE_arg    COMMON_DEVIATION);
	set_bitrate(TI_USE_arg   FRAMESWITCH_BITRATE);
	
	
//	// Table 14: Mant=01b, Exp=7 -> Receiver Bandwidth 3.1 kHz
//	REGFILE_DO_WRITE(TI_USE_single, cbRxBwMant, 1);
//	REGFILE_DO_WRITE(TI_USE_single, cbRxBwExp,  7);
//	// Table 14: Mant=01b, Exp=6 -> Receiver Bandwidth 6.3 kHz
//	REGFILE_DO_WRITE(TI_USE_single, cbRxBwMant, 1);
//	REGFILE_DO_WRITE(TI_USE_single, cbRxBwExp,  6);
	// Table 14: Mant=01b, Exp=1 -> Receiver Bandwidth 200 kHz
	REGFILE_DO_WRITE(TI_USE_single, cbRxBwMant, 1);
	REGFILE_DO_WRITE(TI_USE_single, cbRxBwExp,  1);
	// Table 14: Mant=00b, Exp=1 -> Receiver Bandwidth 250 kHz
	REGFILE_DO_WRITE(TI_USE_single, cbRxBwMant, 0);
	REGFILE_DO_WRITE(TI_USE_single, cbRxBwExp,  1);
//	// Table 14: Mant=10b, Exp=5 -> Receiver Bandwidth 10.4 kHz
//	REGFILE_DO_WRITE(TI_USE_single, cbRxBwMant, 2); // Recommended
//	REGFILE_DO_WRITE(TI_USE_single, cbRxBwExp,  5); // Recommended
	
	REGFILE_DO_WRITE(TI_USE_single, cbDocFreq,  2); // Recommended
	
	
	
	REGFILE_DO_WRITE(TI_USE_single, cbDataMode, PACKET_MODE); // This is default per hardware
	REGFILE_DO_WRITE(TI_USE_single, cbModulationType, MODULATION_FSK); // This is default per hardware
	
	// Packetizer details
	//REGFILE_DO_WRITE(TI_USE_single, cbPayloadLength, 5); // Hi!\n\0
	REGFILE_DO_WRITE(TI_USE_single,cbPacketFormat,PACKET_FORMAT_VARIABLElength);
	REGFILE_DO_WRITE(TI_USE_single, cbSyncOn, 1);
	// Write SyncBytes.
	REGFILE_DO_WRITE(TI_USE_single, cbSyncSize, 1); // "1" = 2 SyncBytes
	REGFILE_DO_WRITE(TI_USE_single, cbSyncValue1, 'C');
	REGFILE_DO_WRITE(TI_USE_single, cbSyncValue2, 'N');
	REGFILE_DO_WRITE(TI_USE_single, cbSyncTol, 3); // Tolerate three wrong bits in Sync word.
		// (TODO: is that per Sync Byte of per complete Sync phrase?
		//    I guess rather the latter, because SyncTol = 7 would make no
		//    sense otherwise.)
		// TODO: SyncTol = 3 is very high, actually the sync word should be
		//    constructed in a mathematical manner.
	
	
	
	//REGFILE_DO_WRITE(TI_USE_single, cbCrcOn, 1); // Let CrcOn = 1 as default (receiver is not obligated to read this)
	// REGFILE_DO_WRITE(TI_USE_single, cbAddressFiltering, ADDRESS_FILTER_none); // already default
	
	
	
	REGFILE_DO_WRITE(TI_USE_single, cbDio5Mapping, ClkOut5);
	REGFILE_DO_WRITE(TI_USE_single, cbClkOut, CLK_RC);
//	REGFILE_DO_WRITE(TI_USE_single, cbDio5Mapping, ModeReady5);
//	REGFILE_DO_WRITE(TI_USE_single, cbClkOut, CLK_OFF);
	REGFILE_DO_WRITE(TI_USE_single, cbDio2Mapping, Data2);
//	REGFILE_DO_WRITE(TI_USE_single, cbDio2Mapping, 2);
	REGFILE_DO_WRITE(TI_USE_single, cbDio1Mapping, FifoNotEmpty1);
	REGFILE_DO_WRITE(TI_USE_single, cbDio4Mapping, TxReady4);
	REGFILE_DO_WRITE(TI_USE_single, cbDio3Mapping, PllLock3);
	
	
	
	multitimer_initialize_callback(MHI_USE_arg   CBNUM_FRAMEFIND, CB_framefind, TIME_LAZY);
	multitimer_wakemeup(MHI_USE_arg
				CBNUM_FRAMEFIND, 50 MSEC, REASON_FRAMEFIND_START);
	
	ulog_str_L("END of STARTUP Sequence of " SETUP_DESCR "\n");
}
#endif // -ifndef NOCODE
;







// Combining the loop (unlike Only* theres only one loop() method which
// doesn't return. (So this goes away from Arduino.)
// (Startup, however, was called for us normally.)

#ifdef SIMUL
void __attribute__ ((unused)) FS_main(MHI_ARG_single)
#else // This is BEAGLE:
void loop(void)
#endif

#ifndef NOCODE
{
	
	// The impliementation of the routing algorithm AODV, aodv-uu, 
	// requires at most 1-second polling timeouts.
	//time_int route_time = get_systime(MHI_USE_single) + (1 SEC);
	
	
	while (true) {
		
#ifdef DEBUG_ALGO
		ulog_str_L("Main loop here.\n");
#endif
		
		
		if (global(rti)->service_task)
			global(rti)->service_task(global(rti));
		
		
		
		global(thetime) = get_systime(MHI_USE_single);
		
		
		// Polling-like mechanism:
		
		time_int least_time = global(thetime) + (1 SEC);
		
		int i;
		for (i = 0; i < NUMBER_OF_TIMERS; i++) {
			if (!MHI_USE_access timer_callback[i]) continue;
			if (MHI_USE_access nsecs_of_callback[i] == TIME_INVALID) continue;
			
			if (MHI_USE_access nsecs_of_callback[i] <= global(thetime)) {
				
				MHI_USE_access nsecs_of_callback[i] = TIME_INVALID;
				// BEFORE actual callback  (-> callback can request re-call.)
				
				
				MHI_USE_access timer_callback[i](
						MHI_USE_arg
						MHI_USE_access reason_opaque[i]);
				
				// In FrameSwitch so far, the time will have proceeded far
				// (this is a bit insane)  so we need this:
				global(thetime) = get_systime(MHI_USE_single);
				least_time = global(thetime) + (1 SEC);
			}
			
			if (MHI_USE_access nsecs_of_callback[i] < least_time) {
				least_time = MHI_USE_access nsecs_of_callback[i];
			}
		}
		
#ifdef DEBUG_LOOP_TIMER
		for(i = 0; i < NUMBER_OF_TIMERS; i++) {
			if (!MHI_USE_access timer_callback[i]) {
				printf("%02d: no timer registered.\n", i);
				continue;
			}
			printf("%02d: nsecs_of_callback = %'lu.\n", i, MHI_USE_access nsecs_of_callback[i]);
		}
		
		ulog_str_num_BASE_str_L("Main: least time = ", least_time, DEC, ".\n");
#endif
		
		
		multitimer_delay_coarse(MHI_USE_arg    least_time - global(thetime));
		
		
	} // while(true)
	
}
#endif // -ifndef NOCODE
;

#endif // #ifndef FRAMESWITCH_C

