
// (c) Jonas Bechtel 2018


#ifndef NOCODE
#define NOCODE
#include "only.c"
#include "RFM69trancv/trancv.c"
#undef NOCODE
#else
#include "only.c"
#include "RFM69trancv/trancv.c"
#endif


#include "RFM69trancv/RFM69-reg.h"
#include "RFM69trancv/RFM69-reghelper.h"

#define CBNUM_RFM 0
#define CBNUM_AFC 1
#define REASON_AFC 100

#include "all-app.c"


#ifdef SIMUL
#define TI_USE_single (MHI_USE_single->trancv)
//#define TI_USE_single mh_station_arg -> trancv
#define TI_USE_arg    TI_USE_single,
#else
#define TU_USE_single
#define TI_USE_arg
#endif

static void CB_AFC(MHI_ARG  uint8_t reason) {
	
	if (reason == REASON_AFC) {
		uint8_t tmp = REGFILE_DO_READ(TI_USE_single, cbAfcValueMsb);
		uint16_t afcval = tmp << 8;
		afcval += REGFILE_DO_READ(TI_USE_single, cbAfcValueLsb);
		if (afcval < 0x8000) {
			ulog_str_num_BASE_str_L("AfcValue is so far: ", (0xffff&afcval), DEC16, ". Triggering an AFC now.\n");
		}
		else {
			ulog_str_L("AFCAFCAFCAFC.\n");
			ulog_str_num_BASE_str_L("AfcValue is so far: -", (afcval^0xffff)+1, DEC16, ". Triggering an AFC now.\n" );
		}
		
		REGFILE_DO_WRITE(TI_USE_single, cbAfcStart, 1);
		
		multitimer_wakemeup(MHI_USE_arg  CBNUM_AFC, 400 MSEC, REASON_AFC);
	}
}




// app_loop will be called repeatedly
void app_loop(MHI_ARG   uint8_t reason) {
	// There's no callback or other complicated timing so we ignore 'reason'.
	ulog_str_L("App loop is here.\n");
	
	uint8_t rssi = REGFILE_DO_READ(TI_USE_single, cbRssi);
	if (rssi) {
		REGFILE_DO_WRITE(TI_USE_single, cbRssiStart, 1);
		multitimer_delay_coarse(MHI_USE_arg 400 USEC);
		uint8_t __attribute((unused)) unused;
		unused = REGFILE_DO_READ(TI_USE_single, cbRssiDone);
		// (assuming this is true.)
		
		rssi = REGFILE_DO_READ(TI_USE_single, cbRssiValue);
		
		ulog_str_L("RSSI is now: in dBm: -.5*");
		log_num_BASE(rssi, DEC);
		ulog_fin_line();
		
		// multitimer_delay_coarse(MHI_USE_arg   10 MSEC); // 10 ms
	}
	
	
	
	// TODO: Remove this, it isn't needed
	uint8_t __attribute((unused)) unused;
	unused = REGFILE_DO_READ(TI_USE_single,  cbMode);
	
	
	// "Wait" for PayloadReady
	uint8_t PayloadReady = REGFILE_DO_READ(TI_USE_single, cbPayloadReady);
	if (!PayloadReady) {
		//ulog_str_L("App loop: not ready.\n");
		
		
		//multitimer_delay_coarse(MHI_USE_arg   50 MSEC); // Kleine Verschnaufpause.
		
 		//// Sofort wieder versuchen
 		//multitimer_wakemeup(MHI_USE_arg   CBNUM_RFM, 1 USEC, 0xff/*No reason.*/);
		
		return;
	}
	
	uint8_t PacketLength = REGFILE_DO_READ(TI_USE_single, cbPayloadLength);
	static uint8_t buf[20];
	if (PacketLength/*+2*/ > sizeof(buf)) {
		// TODO: Error handling - we must make buf bigger or even allocate it dynamically (if this is possible?)
		// Or we could write the SD card in bunches.
		quit_with_error(MHI_USE_arg  sharedERRNO_INTERN_ASSERT);
	}
	
	while (REGFILE_DO_READ(TI_USE_single, cbFifoNotEmpty)) {
		// Read fifo
		uint8_t FifoState;
		uint8_t i;
		for(i = 0; i < PacketLength; i++) {
			FifoState = REGFILE_DO_READ(TI_USE_single, cbFifoNotEmpty);
			if (!FifoState) {
				ulog_str_L("Got some bytes out of fifo, but not as many as I wanted.\n");
				break;
			}
			
			buf[i] = REGFILE_DO_READ(TI_USE_single, cbFifo);
		}
		/*if (i == PacketLength) {
			i = 0;
			do {
				FifoState = REGFILE_DO_READ(TI_USE_single, cbFifoNotEmpty);
				if (FifoState) {
					if (i == 0)
						ulog_str_L("Got packet out of fifo, but there are too "
									"many further bytes.\n");
					REGFILE_DO_READ(TI_USE_single, cbFifo);
				}
				i ++;
			} while (FifoState);
		}
		else
			PacketLength = i;
		*/
		
		ulog_str_L("Received packet (hex: ");
		for (i = 0; i < PacketLength; i++)
			log_num_BASE(buf[i], HEXBYTE);
		log_str_L(", raw: ");
		log_str_N(buf, PacketLength);
		log_str_L(").\n");
		
	}
	
	// Toggling between LNA Modes
/*	uint8_t Zin = REGFILE_DO_READ(TI_USE_single, cbLnaZin);
	uint8_t Gain = REGFILE_DO_READ(TI_USE_single, cbLnaGainSelect);
	if (Gain == 6) {
		Zin = 1 - Zin;
		Gain = 1;
	}
	else
		Gain++;
	REGFILE_DO_WRITE(TI_USE_single, cbLnaZin, Zin);
	REGFILE_DO_WRITE(TI_USE_single, cbLnaGainSelect, Gain);*/
	
	
	//ulog_str_num_BASE_str_L("Zin*0x80+Gain = 0x", (Zin<<7)+Gain, HEX, ". App loop: fin.\n");
	
	ulog_str_L("App loop: fin.\n");
	
// (Fast) sofort wieder versuchen
	multitimer_wakemeup(MHI_USE_arg   CBNUM_RFM, 1000 USEC, 0xff/*No reason.*/);
}




// CODE DUPLICATION from OnlySend.c!
void app_setup(MHI_ARG_single) {
	// put your setup code here, to run once:
	
	
#define SETUP_DESCR "OnlyReceive.c (Packet Mode)"
	ulog_str_L("STARTUP Sequence of " SETUP_DESCR "\n");
	
	// Set the registers accordingly
	
	
	// Safety check: is this RFM 69 and is it working?
	uint8_t boost = REGFILE_DO_READ(TI_USE_single, cbSensitivityBoost);
	if (boost != 0x1B && boost != 0x2D) {
		ulog_str_L("This is not RFM69! Quitting.\n");
		quit_with_error(MHI_USE_arg  ERRNO_RXSTART);
		return;
	}
	
	// RX Part
	
	// "High Power settings MUST be turned off when using PA0, and in Receive mode"
	REGFILE_DO_WRITE(TI_USE_single, cbRegOcp, 0x17)
	REGFILE_DO_WRITE(TI_USE_single, cbRegTestPa1, 0x55)
	REGFILE_DO_WRITE(TI_USE_single, cbRegTestPa2, 0x70)
	
	
	REGFILE_DO_WRITE(TI_USE_single, cbPa1On, 0)
	REGFILE_DO_WRITE(TI_USE_single, cbPa2On, 0)
	REGFILE_DO_WRITE(TI_USE_single, cbPa0On, 1)
	// Low power mode 1 0 0, see Datasheet p. 21
	// Power Range: -18 to +13 dBm; Pout Formula -18 dBm + OutputPower
	REGFILE_DO_WRITE(TI_USE_single, cbOutputPower, 28); // (-18 + 28) dBm = 10 dBm. (= 10 mW)
	
	
	REGFILE_DO_WRITE(TI_USE_single, cbLnaZin, 1)
//	REGFILE_DO_WRITE(TI_USE_single, cbLnaGainSelect, 4)
	REGFILE_DO_WRITE(TI_USE_single, cbLnaGainSelect, 6)
	
	
	
	// Set frequency to 433.4 MHz
	set_frequency(TI_USE_arg  COMMON_FREQUENCY);
	set_deviation(TI_USE_arg  COMMON_DEVIATION);
	set_bitrate(TI_USE_arg  COMMON_BITRATE);
	
	
//	// Table 14: Mant=01b, Exp=7 -> Receiver Bandwidth 3.1 kHz
//	REGFILE_DO_WRITE(TI_USE_single, cbRxBwMant, 1);
//	REGFILE_DO_WRITE(TI_USE_single, cbRxBwExp,  7);
//	// Table 14: Mant=01b, Exp=6 -> Receiver Bandwidth 6.3 kHz
//	REGFILE_DO_WRITE(TI_USE_single, cbRxBwMant, 1);
//	REGFILE_DO_WRITE(TI_USE_single, cbRxBwExp,  6);
	// Table 14: Mant=01b, Exp=1 -> Receiver Bandwidth 200 kHz
	REGFILE_DO_WRITE(TI_USE_single, cbRxBwMant, 1);
	REGFILE_DO_WRITE(TI_USE_single, cbRxBwExp,  1);
	
	REGFILE_DO_WRITE(TI_USE_single, cbDocFreq,  2); // Recommended
	
	
	REGFILE_DO_WRITE(TI_USE_single, cbDataMode, PACKET_MODE); // This is default per hardware
	//REGFILE_DO_WRITE(TI_USE_single, cbDataMode, CONTINUOUS_MODE_WO_BIT_SYNC);
	REGFILE_DO_WRITE(TI_USE_single, cbModulationType, MODULATION_FSK); // This is default per hardware
	
	// Packetizer details
	REGFILE_DO_WRITE(TI_USE_single, cbPayloadLength, 5); // Hi!\n\0
	REGFILE_DO_WRITE(TI_USE_single, cbPacketFormat, PACKET_FORMAT_FIXEDlength);
	REGFILE_DO_WRITE(TI_USE_single, cbSyncOn, 1);
	// Write SyncBytes.
	REGFILE_DO_WRITE(TI_USE_single, cbSyncSize, 1); // "1" = 2 SyncBytes
	REGFILE_DO_WRITE(TI_USE_single, cbSyncValue1, 'C');
	REGFILE_DO_WRITE(TI_USE_single, cbSyncValue2, 'N');
	REGFILE_DO_WRITE(TI_USE_single, cbSyncTol, 3); // Tolerate three wrong bits in Sync word.
		// (TODO: is that per Sync Byte of per complete Sync phrase?
		//    I guess rather the latter, because SyncTol = 7 would make no
		//    sense otherwise.)
		// TODO: SyncTol = 3 is very high, actually the sync word should be
		//    constructed in a mathematical manner.
	
	
	//REGFILE_DO_WRITE(TI_USE_single, cbCrcOn, 1); // Let CrcOn = 1 as default (receiver is not obligated to read this)
	// REGFILE_DO_WRITE(TI_USE_single, cbAddressFiltering, ADDRESS_FILTER_none); // already default
	
	
	
	//REGFILE_DO_WRITE(TI_USE_single, cbRssiThreshold, 35);
	REGFILE_DO_WRITE(TI_USE_single, cbRssiThreshold, 0xE4);
	//REGFILE_DO_WRITE(TI_USE_single, cbRssiThreshold, 0x84);
	//REGFILE_DO_WRITE(TI_USE_single, cbRssiThreshold, 0xB4);
	REGFILE_DO_WRITE(TI_USE_single, cbMode, RFM_MODE_RX);
	
	
	REGFILE_DO_WRITE(TI_USE_single, cbDio5Mapping, ClkOut5);
	REGFILE_DO_WRITE(TI_USE_single, cbClkOut, CLK_RC);
//	REGFILE_DO_WRITE(TI_USE_single, cbDio5Mapping, ModeReady5);
//	REGFILE_DO_WRITE(TI_USE_single, cbClkOut, CLK_OFF);
	REGFILE_DO_WRITE(TI_USE_single, cbDio2Mapping, Data2);
//	REGFILE_DO_WRITE(TI_USE_single, cbDio2Mapping, 2);
	REGFILE_DO_WRITE(TI_USE_single, cbDio1Mapping, PllLock1);
	REGFILE_DO_WRITE(TI_USE_single, cbDio4Mapping, Rssi4);
	REGFILE_DO_WRITE(TI_USE_single, cbDio3Mapping, SyncAddress3);
	
	
	multitimer_initialize_callback(MHI_USE_arg  CBNUM_RFM, app_loop, TIME_LAZY);
	multitimer_initialize_callback(MHI_USE_arg  CBNUM_AFC, CB_AFC, TIME_LAZY);
	multitimer_wakemeup(MHI_USE_arg  CBNUM_AFC, 300 MSEC, REASON_AFC);
	
	ulog_str_L("END of STARTUP Sequence of " SETUP_DESCR "\n");
}




