
// (c) Jonas Bechtel 2018


#ifndef NOCODE
#define NOCODE
#include "only.c"
#include "RFM69trancv/trancv.c"
#undef NOCODE
#else
#include "only.c"
#include "RFM69trancv/trancv.c"
#endif

#include "RFM69trancv/RFM69-reg.h"
#include "RFM69trancv/RFM69-reghelper.h"

#define CBNUM_first 0

#define REASON_GPSPULSE          5
#define REASON_WAIT_FOR_AIR      6
#define REASON_WAIT_FOR_AIR_END 26
#define REASON_TX_START         30

#define REASON_SLEEP_TILL_TRY (BIGGEST_REASON +  1)

#define REASON_TIMER1         (BIGGEST_REASON +  5)
#define REASON_TIMER_END      (BIGGEST_REASON + 25)

#include "cassert.h"

#include "all-app.c"


#ifdef SIMUL
#define TI_USE_single (MHI_USE_single->trancv)
//#define TI_USE_single mh_station_arg -> trancv
#define TI_USE_arg    TI_USE_single,
#else
#define TU_USE_single
#define TI_USE_arg
#endif


void CB_rfm(MHI_ARG   uint8_t reason);

void app_setup(MHI_ARG_single) {
	// put your setup code here, to run once:
	
#define SETUP_DESCR "OnlySend.c"
	ulog_str_L("STARTUP Sequence of " SETUP_DESCR "\n");
	
	// Set the registers accordingly
	
	
	// Safety check: is this RFM 69 and is it working?
	/*volatile*/
	uint8_t boost = REGFILE_DO_READ(TI_USE_single, cbSensitivityBoost);
	// AH! Once we declare boost as volatile the cbSensitivityBoost variable
	//  will be detected properly. 
	// OK, we don't need boost to be volatile but the read function 
	// -> see ardu_spi_read_single().
	/*ulog_num_BASE(cbSensitivityBoost, HEX);
	log_str_L(" <- cbSensitivityBoost\n");
	ulog_num_BASE(regfile_is_read(cbSensitivityBoost), DEC);
	log_str_L(" <-   regfile_is_read?\n");
	ulog_num_BASE(regfile_name(cbSensitivityBoost), HEXBYTE);
	log_str_L(" <-   regfile_name.\n");
	CASSERT(cbSensitivityBoost == 0x0586FF00, OnlySend);
	CASSERT(regfile_name(cbSensitivityBoost) == 0x58, OnlySend);*/
	if (boost != 0x1B && boost != 0x2D) {
		ulog_str_L("This is not RFM69! Quitting.\n");
		quit_with_error(MHI_USE_arg   ERRNO_RXSTART);
		return;
	}
	
	
	// RX Part
	
	// "High Power settings MUST be turned off when using PA0, and in Receive mode"
	REGFILE_DO_WRITE(TI_USE_single, cbRegOcp, 0x17)
	REGFILE_DO_WRITE(TI_USE_single, cbRegTestPa1, 0x55)
	REGFILE_DO_WRITE(TI_USE_single, cbRegTestPa2, 0x70)
	
	
	REGFILE_DO_WRITE(TI_USE_single, cbPa1On, 0)
	REGFILE_DO_WRITE(TI_USE_single, cbPa2On, 0)
	REGFILE_DO_WRITE(TI_USE_single, cbPa0On, 1)
	// Now power mode 1 0 0, see Datasheet p. 21
	// Power Range: -18 to +13 dBm; Pout Formula -18 dBm + OutputPower
	REGFILE_DO_WRITE(TI_USE_single, cbOutputPower, 28); // (-18 + 28) dBm = 10 dBm. (= 10 mW)
	
	
	// Set frequency to 433.4 MHz
	set_frequency(TI_USE_arg   COMMON_FREQUENCY);
	set_deviation(TI_USE_arg   COMMON_DEVIATION);
	set_bitrate(TI_USE_arg   COMMON_BITRATE);
	
	REGFILE_DO_WRITE(TI_USE_single, cbDataMode, PACKET_MODE); // This is default per hardware
	REGFILE_DO_WRITE(TI_USE_single, cbModulationType, MODULATION_FSK); // This is default per hardware
	
	// Packetizer details
	REGFILE_DO_WRITE(TI_USE_single, cbPayloadLength, 5); // Hi!\n\0
	REGFILE_DO_WRITE(TI_USE_single, cbPacketFormat, PACKET_FORMAT_FIXEDlength);
	REGFILE_DO_WRITE(TI_USE_single, cbSyncOn, 1);
	// Write SyncBytes.
	REGFILE_DO_WRITE(TI_USE_single, cbSyncSize, 1); // "1" = 2 SyncBytes
	REGFILE_DO_WRITE(TI_USE_single, cbSyncValue1, 'C');
	REGFILE_DO_WRITE(TI_USE_single, cbSyncValue2, 'N');
	REGFILE_DO_WRITE(TI_USE_single, cbSyncTol, 3); // Tolerate three wrong bits in Sync word.
		// (TODO: is that per Sync Byte of per complete Sync phrase?
		//    I guess rather the latter, because SyncTol = 7 would make no
		//    sense otherwise.)
		// TODO: SyncTol = 3 is very high, actually the sync word should be
		//    constructed in a mathematical manner.
	
	
	REGFILE_DO_WRITE(TI_USE_single, cbFifoThreshold, 20);
	
	//REGFILE_DO_WRITE(TI_USE_single, cbCrcOn, 1); // Let CrcOn = 1 as default (receiver is not obligated to read this)
	// REGFILE_DO_WRITE(TI_USE_single, cbAddressFiltering, ADDRESS_FILTER_none); // already default
	
	
	
	REGFILE_DO_WRITE(TI_USE_single, cbDio5Mapping, ClkOut5);
	REGFILE_DO_WRITE(TI_USE_single, cbClkOut, CLK_RC);
//	REGFILE_DO_WRITE(TI_USE_single, cbDio5Mapping, ModeReady5);
//	REGFILE_DO_WRITE(TI_USE_single, cbClkOut, CLK_OFF);
	REGFILE_DO_WRITE(TI_USE_single, cbDio2Mapping, Data2);
//	REGFILE_DO_WRITE(TI_USE_single, cbDio2Mapping, 2);
	REGFILE_DO_WRITE(TI_USE_single, cbDio1Mapping, FifoNotEmpty1);
	REGFILE_DO_WRITE(TI_USE_single, cbDio4Mapping, TxReady4);
	REGFILE_DO_WRITE(TI_USE_single, cbDio3Mapping, PllLock3);
	
	
	
	multitimer_initialize_callback(MHI_USE_arg   CBNUM_first, CB_rfm, TIME_LAZY);
	
	ulog_str_L("END of STARTUP Sequence of " SETUP_DESCR "\n");
}



void CB_rfm(MHI_ARG   uint8_t reason) {
	
	if (reason == REASON_GPSPULSE) {
		ulog_str_L("Starting new listen-send-cycle.\n");
		
		// Sense for free channel
		// -70 dBm -> RssiThreshold=35
		REGFILE_DO_WRITE(TI_USE_single, cbRssiThreshold, 35);
		REGFILE_DO_WRITE(TI_USE_single, cbMode, RFM_MODE_RX);
		
		uint32_t wtime = 10 MSEC;
		//  + 40 ms (B), + 80 ms (D)
		if (station_name == 'B')
			wtime += 40 MSEC;
		else if (station_name == 'D')
			wtime += 120 MSEC;
		multitimer_wakemeup(MHI_USE_arg  CBNUM_first, wtime, REASON_WAIT_FOR_AIR);
		
		//multitimer_wakemeup(MHI_USE_arg   CBNUM_first, 10000 USEC, REASON_WAIT_FOR_AIR);
//		ulog_str_num_BASE_str_L("Registered reason: ", reason_opaque[CBNUM_first], DEC, "\n");
//		ulog_str_num_BASE_str_L("Registered time: ", nsecs_of_callback[CBNUM_first], DEC, "\n");
		return;
	}
	
	if (REASON_WAIT_FOR_AIR <= reason && reason < REASON_WAIT_FOR_AIR_END) {
		
		// Has Rssi raised?
		bool rssi = REGFILE_DO_READ(TI_USE_single, cbRssi);
		if (rssi) {
			// Rssi is high, so someone else is sending and I cann't
			//   send. This gets logged:
			uint8_t rssi = REGFILE_DO_READ(TI_USE_single, cbRssiValue);
//#define RSSI_STR "RSSI is too high to send: in dBm: -.5*"
			//SDlog_dataline(RSSI_STR, sizeof(RSSI_STR)-1);
			ulog_str_L("RSSI is too high to send: in dBm: -.5*");
			log_num_BASE(rssi, DEC);
			ulog_fin_line();
			//dataFile.println(rssi, DEC);
			
			
			// We do not re-register the interrupt.
		}
		else {
			multitimer_wakemeup(MHI_USE_arg   CBNUM_first, 2000 USEC, reason+1);
			// up to next 2 ms sleep (-> 20x 2ms = 40 ms will be executed)
		}
		return;
	}
	
	if (reason == REASON_WAIT_FOR_AIR_END) {
		
		// So now we have waited for some arbitrary time, then have sensed
		// for RSSI. There was no RSSI (no other station) -> we can send.
		
		ulog_str_L("Air is clean, entering TX mode.\n");
		
		REGFILE_DO_WRITE(TI_USE_single, cbMode, RFM_MODE_TX);
		
		multitimer_wakemeup(MHI_USE_arg   CBNUM_first, 800 USEC, REASON_TX_START);
		return;
	}
	
	if (reason == REASON_TX_START) {
		
		// Now TX must be ready
		
		if (REGFILE_DO_READ(TI_USE_single, cbTxReady) != 1) {
			quit_with_error(MHI_USE_arg   ERRNO_TXSTART);
		}
		
		// Store Packet to FIFO
		
		REGFILE_DO_WRITE(TI_USE_single, cbTxStartCondition, COND_FifoThreshold); // 0
		// Assuming FifoThreshold is 15, as default -> bigger than our data.
		
		// If you would want to speed it up, you could use the FIFO burst access
		// so write address only once and then much of data.
		REGFILE_DO_WRITE(TI_USE_single, cbFifo, 'H')
		REGFILE_DO_WRITE(TI_USE_single, cbFifo, 'i')
		REGFILE_DO_WRITE(TI_USE_single, cbFifo, '!')
		REGFILE_DO_WRITE(TI_USE_single, cbFifo, '\n')
		REGFILE_DO_WRITE(TI_USE_single, cbFifo, 0)
		
		// Let it send.
		REGFILE_DO_WRITE(TI_USE_single, cbTxStartCondition, COND_FifoNotEmpty); // 1
//#define PACKET_SENT_STRING "Packet sent: \"Hi!\\n\\0\"\n"
//		SDlog_dataline(PACKET_SENT_STRING, sizeof(PACKET_SENT_STRING)-1);
		ulog_str_L("Packet given: \"Hi!\\n\\0\"\n");
		
		// TODO: - Set mode to sleep., then RXRestart.
		
		// TODO: What about resetting TxStartCondition? Maybe it hinders a correct
		// sending.
		REGFILE_DO_WRITE(TI_USE_single, cbTxStartCondition, COND_FifoThreshold); // 0
		
		// We don't reregister as the interrupt will come from GPSPULSE.
		return;
	}
	
	// TODO: This takes too long, it is more than one second in total.
}


// app_loop will be called repeatedly
void app_loop(MHI_ARG   uint8_t reason) {
	ulog_str_num_BASE_str_L("App loop here. (reason=", reason, DEC, ")\n");
	
	if (reason == REASON_TIMEPULSE || reason == REASON_TIMEEMULATION) {
		// Wait for 100 ms (in future this will be different on different stations.)
#ifdef ARDU
		printf("usecs_to_callback (C) = %lx", usecs_to_callback[CBNUM_first]);
#else
		char* str = "nsecs_of_callback (C) = %llx.\n";
#ifdef SIMUL
		str = "nsecs_of_callback (C) = %lx.\n";
#endif
		printf(str, MHI_USE_access nsecs_of_callback[CBNUM_first]);
#endif // #ifdef ARDU #else
		
		
#ifdef ARDU
		if (usecs_to_callback[CBNUM_first] == -1) {
#elif defined BEAGLE || defined SIMUL
			// TODO: Something is wrong here: 
			// when we request == TIME_INVALID, then it shows 0xffffffffffffffff
			// when we request == 0xffffffffffffffff, then it shows TIME_INVALID
		if (MHI_USE_access nsecs_of_callback[CBNUM_first] == TIME_INVALID) {
#endif
			multitimer_wakemeup(MHI_USE_arg   CBNUM_first, 100 MSEC, REASON_GPSPULSE);
		}
#if defined BEAGLE || defined SIMUL
		str = "nsecs_of_callback (C) = %llx.\n";
#ifdef SIMUL
		str = "nsecs_of_callback(C) = %lx.\n";
#endif
		printf(str, MHI_USE_access nsecs_of_callback[CBNUM_first]);
#endif
		return;
	}
	
}

