
// (c) Jonas Bechtel 2018


#define NOCODE
#include "only.c"
#undef NOCODE

#include "all-app.c"

#define CBNUM_first

static File soundFile;


// CODE DUPLICATION from OnlySend.c!
void app_setup() {
	// put your setup code here, to run once:
	
	
#define SETUP_DESCR "SD-Read-Write.c"
	ulog_str_L("STARTUP Sequence of " SETUP_DESCR "\n");
	
	
	soundFile = SD.open("MADONNA.AMR", FILE_READ);
	if (!soundFile) {
		quit_with_error(sharedERRNO_2nd_fileopen);
	}
	
	multitimer_initialize_callback(CBNUM_first, app_loop, TIME_LAZY);
	
	ulog_str_L("END of STARTUP Sequence of " SETUP_DESCR "\n");
}


// CAUTION: mem-Block must have one additional byte which contains '\n'!
static void rwcycle(uint8_t * mem, uint16_t size)  {
	
	/*dataFile_log_selected = false;
	dataFile.flush();
	dataFile.close();
	dataFile = SD.open("MADONNA.AMR", FILE_READ);
	if (!dataFile) {
		quit_with_error(sharedERRNO_2nd_fileopen);
	}
#ddefine soundFile dataFile
	*/
	int16_t res = soundFile.read(mem, size);
	ulog_str_num_BASE_str_L("size=0x", size, HEX,"\n");
	ulog_str_num_BASE_str_L("             res=0x", res, HEX,"\n");
	if (size != 0xFFFF && res == -1) {
		// try again.
		//multitimer_delay_coarse(2 MSEC);
		delay(150);
		res = soundFile.read(mem, size);
		ulog_str_num_BASE_str_L("size=0x", size, HEX,"\n");
		ulog_str_num_BASE_str_L("             res=0x", res, HEX,"\n");
	}
	if (size != 0xFFFF && res == -1) {
		// Some error occured.
		quit_with_error(sharedERRNO_read_minus_1);
	}
	else if (size != res) {
		size = res;
		mem[size] = '\n';
		res = 0xFFFF;
		soundFile.close();
	}
	
	ulog_str_L("BLOCK OF ");
	//dataFile.flush();
	log_num_BASE(size, DEC);
	log_str_L(" BYTES:\n");
	
	log_raw_data(mem, size+1);
	
	log_flush();
	
	if (res == 0xFFFF) {
		ulog_str_L("SD-Read-Write is DONE now as the source file has ended.\n");
		quit_with_error(sharedERRNO_EOF);
	}
}

// app_loop will be called repeatedly
void app_loop(uint8_t reason) {
	// There's no callback or other complicated timing so we ignore 'reason'.
	//ulog_str_L("App loop is here.\n");
	
	// Bit 7-4:  0: uninitialized
	//           1: Block size 1 B
	//           2: Block size 5 B
	//           3: Block size 64 B
	//           4: Block size 512 B
	//           5: Block size 1536 B
	// Bit 1: true -> write read data to ulog
	//        false -> undefined [In future maybe: write data to extra file]
	// Bit 0: true -> read+write 4 blocks
	//        false -> read+write 2 blocks
	static uint8_t select;
	
	if ((select & 0xF0) == 0) {
		// initialize
		select = 0x02;
	}
	select += 0x10;
	if (select >= ((4+1)<<4)) { // go up to 512 B
				//  512 may be not possible as the RAM (2048 B) is too small. (sdfatlib needs 512 B _per File_!)
				// 1536 isn't possible as the RAM (2048 B) is too small.
		select = 0x12 | (0x0F & select ^ 0x01);
	}
	
	uint8_t memblock [512+1]; // up to 512 B
	
	bool second = select & 0x01;
	ulog_str_num_BASE_str_L("select=0x", select, HEXBYTE, "\n");
	
	uint16_t bs = 1;
	if (select >= (5 << 4)) bs = 1536;
	else if (select >= (4 << 4)) bs = 512;
	else if (select >= (3 << 4)) bs = 64;
	else if (select >= (2 << 4)) bs = 5;
	ulog_str_num_BASE_str_L("Beginning transaction with ",
							bs, DEC, " byte blocks.\n");
	
	while(true) {
		
		memblock[bs] = '\n';
		rwcycle(memblock, bs);
		rwcycle(memblock, bs);
		
		if (second == false)
			break;
		second = false;
	}
	
	ulog_str_num_BASE_str_L("Transaction with ", 
				2+((select & 0x01) << 1), DEC, " blocks finished.\n");
	
	
	ulog_str_L("App loop: fin.\n");
	
	//multitimer_delay_coarse(10 MSEC);
	
	// Fast sofort weitermachen.
	//multitimer_wakemeup(CBNUM_first, 1000 USEC, 0xff/*No reason.*/);
}




