
#ifndef MULTIHW_C
#define MULTIHW_C

#define NUMBER_OF_TIMERS 4

// How many lines (1-255) are written without flush.
// 0 will lead to random behaviour
//#define ULOG_NUMBER_BASED_FLUSH 10  // disabled if undefined.

#if defined(BEAGLE) + defined(ARDU) + defined(SIMUL) != 1
#error   One of BEAGLE, ARDU or SIMUL must be defined!
#endif

#define _STR(macro) #macro
#define STR(macro) _STR(macro)


#include <stdint.h>
#ifndef NOCODE
#include <cassert.h>
#endif



// Needed for AODVroute <- müsste mal umprogrammiert werden!
#ifdef ARDU
// actually from /usr/include/i386-linux-gnu/bits/types/struct_timeval.h
struct timeval
{
  int16_t tv_sec;
  int16_t tv_usec;
};

#else // #ifdef ARDU
#include <sys/time.h>
#endif




#ifdef ARDU

//#include <utility/SdError.h>
#include <lib/SD/src/utility/SdError.h>

typedef long time_int;
#define MHI_ARG
#define MHI_ARG_single
#define MHI_USE_single
#define MHI_USE_arg
#define MHI_USE_access


#elif defined BEAGLE

typedef uint64_t time_int;
#define TIME_INVALID 0xFF00000000000000UL
#define MHI_ARG
#define MHI_ARG_single
#define MHI_USE_single
#define MHI_USE_arg
#define MHI_USE_access

#elif defined SIMUL

#include "station.h"
#include <stdio.h>
#include <stdlib.h> // malloc
#include <errno.h>
#include <string.h> // strerror
#include <pthread.h>

typedef uint64_t time_int;
#define TIME_INVALID 0xFF00000000000000UL
//#define MHI_ARG        struct multihw_global_struct *mgs,
//#define MHI_USE_access mgs->
#define MHI_ARG_single struct station * mh_station_arg
#define MHI_ARG        MHI_ARG_single,
#define MHI_USE_single mh_station_arg
#define MHI_USE_arg    MHI_USE_single,
#define MHI_USE_access mh_station_arg->mgs->

#ifndef NOCODE
#include "RFM69trancv/subobject.h"


#define NOCODE
#include "../Dosim/simulator.c"
#undef NOCODE
#else
#include "../Dosim/simulator.c"

#endif // -ifndef NOCODE -else

#endif // -ifdef ARDU -elif defined BEAGLE -elif defined simul



#ifndef NOCODE
#define NOCODE
#include "only.c"
#undef NOCODE
#endif


#ifdef SIMUL
#define get_systime get_simtime
#endif
time_int get_systime(MHI_ARG_single);
// In many cases you simply can use multitimer_delay(100 MSEC) or so.
// But there are some cases in which you will get wrong results because of
// the assumed integer size. Use delay(time(100 MSEC)) instead!
//#define time(tspec) ((time_int) tspec)




// Station name. Will be set by command line (BeagleBone) or implantation (Arduino)
#if defined(BEAGLE) || defined(ARDU)
#ifdef NOCODE
extern volatile char station_name;
#else // -ifdef NOCODE
volatile char station_name = '!';
#endif // -ifdef NOCODE -else
#else // -ifdef BEAGLE || defined(ARDU)
#define station_name mh_station_arg->name
#endif // -ifdef BEAGLE || defined(ARDU)

//#endif

struct app_global;


#ifndef SIMUL
#ifdef NOCODE
#define DEFINE_OR_NOT extern
#else
#define DEFINE_OR_NOT
#endif
#else // -ifndef SIMUL
#define DEFINE_OR_NOT
struct multihw_global_struct {
	struct station * station;
	
	FILE * ulog_file;
	
	
	pthread_attr_t threadattr;
	pthread_mutex_t auxiliary_mutex;
	pthread_mutex_t auxiliary_mutex2;
	pthread_cond_t startloop_sim2app;
	// piece_executed and please_resume could be melted to one condition.
	pthread_cond_t piece_executed_app2sim;
	pthread_cond_t please_resume_sim2app;
	pthread_t thethread;
	
	// S: Waiting for start impulse
	// W: Waiting even more (for start or resume impulse which will come soon)
	// k: Running ("kicked")
	// w: Waiting (having executed some piece of code)
	char phase;
	
	void (*loop)(MHI_ARG_single);
	
#ifdef SIMUL
	struct app_global * app;
#endif
	
	
#endif // -ifndef SIMUL else
	
	
	DEFINE_OR_NOT void (*timer_callback[NUMBER_OF_TIMERS])(MHI_ARG   uint8_t reason);
	DEFINE_OR_NOT uint8_t mode_of_callback[NUMBER_OF_TIMERS]; // 
	DEFINE_OR_NOT uint8_t reason_opaque[NUMBER_OF_TIMERS]; //
#ifdef ARDU
	DEFINE_OR_NOT long usecs_to_callback[NUMBER_OF_TIMERS]; // -1 = don't call.
#elif defined BEAGLE || defined SIMUL
	DEFINE_OR_NOT uint64_t nsecs_of_callback[NUMBER_OF_TIMERS]; // Relative to CLOCK_MONOTONIC_RAW
#endif



#ifdef SIMUL
}; // struct multihw_global_struct



#ifdef DEBUG_SPI
#ifdef NOCODE
// wXXh<-XXh\n   for write
// rXXh->XXh\n   for read
extern uint8_t spi_debug_line[10];
//extern time_int thetime;
#else // #ifdef NOCODE
uint8_t spi_debug_line[10]; // SIMUL: shared over all stations!
//time_int thetime;
#endif // -ifdef NOCODE -else
#endif // -ifdef DEBUG_SPI




// Will only store the pointer of station, and use the station name
// Rest of initialization is independend of station.
struct multihw_global_struct * init_multihw(struct station * station)
#ifndef NOCODE
{
	struct multihw_global_struct * retV = malloc (sizeof (struct  multihw_global_struct));
	if (retV == NULL) return NULL;
	
	retV->app = NULL;
	
	retV->station = station;
	
	char pathname[] = "ulog_?";
	pathname[5] = retV->station->name;
	
	retV->ulog_file = fopen(pathname, "a");
	if (retV->ulog_file == NULL) {
		printf("Coudln't initialize HAL(multihw): ulog file opening: errno %d (%s).\n", errno, strerror(errno));
		return NULL;
	}
	
	
#ifdef DEBUG_SPI
	spi_debug_line[0] = 'S';
	spi_debug_line[3] = 'h';
	spi_debug_line[4] = '-';
	spi_debug_line[5] = '-';
	spi_debug_line[8] = 'h';
	spi_debug_line[9] = '\n';
#endif
	
	return retV;
}
#endif // -ifndef NOCODE
;

#endif // ifdef SIMUL










/** UNIFIED LOGGING
What different logging facilities are available / make sense?
ulog_str_L      Log zero terminated string (string literal) (begins a new line)
ulog_str_P      Log zero terminated string (changing pointer) (begins a new line)
ulog_str_N      Log a string / bunch of data with known length (begins a new line)
ulog_num_BASE   Log a number (begins a new line)
log_num_BASE    Log a number (within a line)
ulog_fin_line   Finishes a logging line.
    Whenever ulog_str_L, ulog_str_P or ulog_str_N encounter an \n at the end of 
    the string, it is recognized as finishing request so the line is finished
    and no ulog_fin_line is needed.

How does a unified log line look like?
UL[systime][rtime]>[msg]>\n
with [systime] = @[timespec]   (example: @ms40033332 or @t12:32:58.945)
with [rtime]   = R[timespec]   (example: Rms40033332 or Rt12:32:58.945)
with [msg]     = (arbitrary message string.)
with [timespec] = @[timescope]s3921477     (Time in s, typical counted from 1.1.1970)
             or = @[timescope]ms40033332   (Time in ms, typical Arduino time)
             or = @[timescope]t[hh]:[mm]:[ss][.frac]
             or = @[timescope]t[hhmmss][.frac] 12:32:58.945
with [timescope] = U or S 
with \n = linebreak
where any of systime, rtime, msg, timescope and .frac are optional.

(U = Universal Time Coordinated (from GPS module), S = System Time)
(typically the GPS reference time will look like this: RUt123258)
(When simply synchronizing system clock and GPS, there will be some line like
 this: UL@s3921477RUt123258>>\n)
 */

#include <stdint.h>
#include <stdbool.h>

#ifndef SIMUL
#include "pindef.h"
#endif

void quit_with_error(MHI_ARG    uint8_t error);



#define HEX 16
#define HEXBYTE 116
#define DEC 10
#define DEC16 1016
#define TIME  8016
#define FLOAT 8111
#define CHAR 256


#ifdef ARDU

#include <SPI.h>
#include <SD_.h>

#ifdef NOCODE
extern File dataFile;
//extern bool dataFile_log_selected;
extern uint8_t cnt_ulog;
#else
File dataFile;
//bool dataFile_log_selected;
uint8_t cnt_ulog;
#endif

// Needs to be invoked _before_ the first call to u?log_.*
void ardu_ulog_start()
#ifndef NOCODE
{
	// SD Code: https://www.arduino.cc/en/Tutorial/Datalogger
	//          
	//          https://www.arduino.cc/en/Reference/FilePrintln
	//if (!SD.begin(SD_CHIP_SELECT)) {
	//if (!SD.begin(SPI_QUARTER_SPEED/*SPI_FULL_SPEED*/, SD_CHIP_SELECT)) {
	if (!SD.begin(5, SD_CHIP_SELECT)) {
		quit_with_error(ERRNO_SD);
	}
	
	//dataFile = SD.open("airNgps.txt", FILE_WRITE);
	// https://arduino.stackexchange.com/questions/28540/how-to-increase-sd-card-write-speed-in-arduino     ->
	// http://forum.arduino.cc/index.php?topic=49649.msg354701
	dataFile = SD.open("airNgps.txt", O_WRITE | O_CREAT | O_APPEND);
	if (!dataFile) {
		quit_with_error(sharedERRNO_SDOPEN);
	}
	//dataFile_log_selected = true;
}
#endif // ifndef NOCODE
;
#define ulog_start() ardu_ulog_start()

inline void arduino_try_write(uint8_t * data, int len) {
	if (dataFile.write(data, len) != len) {
		quit_with_error(sharedERRNO_SDWRITE);
	}
	//dataFile.flush(); // service_routine will flush datafile.
}
/*
void arduino_switch_to_ulog()
#ifndef NOCODE
{
	if (!dataFile_log_selected) {
		dataFile.flush();
		dataFile.close();
		dataFile = SD.open("airNgps.txt", FILE_WRITE);
		if (!dataFile) {
			quit_with_error(sharedERRNO_SDWRITE);
		}
		dataFile_log_selected = true;
	}
}
#endif
;*/

void ulog_prefix()
#ifndef NOCODE
{
	//arduino_switch_to_ulog();
	arduino_try_write("UL@ms", 5);
	dataFile.print(millis(), DEC);
	arduino_try_write(">", 1);
}
#endif // ifndef NOCODE
;
// 	
#define try_write(str, size) arduino_try_write(str, size);

#define log_raw_data(mem, size)  try_write(mem, size);
#define TIME  8016
#define CHAR 256
	// Arduino print /usr/share/arduino/hardware/arduino/avr/cores/arduino/Print.CPHA
	// only prints the minimum needed number of digits.
#define log_num_BASE(num, BASE)   \
	if (BASE == HEXBYTE) { \
		if (num < 16) \
			dataFile.write((uint8_t)0); \
		dataFile.print(num, 16); \
	} \
	else if (BASE == DEC16) \
		dataFile.print((uint16_t)num, 10); \
	else if (BASE == TIME) { \
		dataFile.print(num / 1 SEC, 10); \
		dataFile.print("s+"); \
		dataFile.print(num % 1 SEC); \
		dataFile.print("*16us"); \
	} \
	else if (BASE == FLOAT) \
		dataFile.print(num, 3); \
	else if (BASE == CHAR) \
		dataFile.print((char)num); \
	else /* BASE=HEX / BASE=DEC */ \
		dataFile.print(num, BASE)
#define log_flush() \
	dataFile.flush(); \
	//delay(100);

#ifdef ULOG_NUMBER_BASED_FLUSH
#define ulog_fin_line() \
	try_write(">\n", 2);\
	if (cnt_ulog > ULOG_NUMBER_BASED_FLUSH) {\
		log_flush(); \
		cnt_ulog = 0; \
	} \
	cnt_ulog++;
#endif // -ifdef ULOG_NUMBER_BASED_FLUSH


// To be called when nothing else happens.
void service_routine()
#ifndef NOCODE
{
	log_flush();
	//dataFile.flush();
}
#endif // ifndef NOCODE
;


//-ifdef ARDU
#elif defined(BEAGLE)

#include <stdio.h>

#define ulog_start() // Nothing to be done for BeagleBone (using stdout)

#define ulog_prefix() \
	printf("UL@ns");\
	printf("%'llu", get_systime()); \
	printf(">");

//#define log_str_N(str, size) write(stdout, str, size)
//#define log_str_N(str, size) fwrite(stdout, str, size)
#define try_write(str, size) fwrite(str, size, 1, stdout)



#define log_raw_data(mem, size)   fwrite(mem, size, 1, stdout)
#define log_num_BASE(num, BASE)   \
	do { \
		if (BASE == HEX) \
			printf("%llx", (uint64_t)num); \
		else if (BASE == HEXBYTE) \
			printf("%02x", (uint16_t)num); \
		else if (BASE == DEC) \
			printf("%llu", (uint64_t)num); \
		else if (BASE == DEC16) \
			printf("%d", (uint16_t)num); \
		else if (BASE == TIME) \
			printf("%'llu", (uint64_t)num); \
		else if (BASE == CHAR) \
			printf("%c", (uint16_t)num); \
		else if (BASE == FLOAT) \
			printf("%f", (double)(num)); \
	/*printf(str1 STR(BASE) str2, num)*/ \
	} while(false);
	//printf(STR(BASE), (uint64_t)(num))
	
	
	

#define log_flush() fflush(stdout)


void service_routine()
#ifndef NOCODE
{
}
#endif // ifndef NOCODE
;


#elif defined(SIMUL)  // ifdef ARDU elif defined (BEAGLE)



#define log_num_BASE(num, BASE) \
	if (BASE == DEC) { \
		printf("%lld", (int64_t)(num));  \
		fprintf(MHI_USE_access ulog_file, "%lld", (int64_t)(num)); \
	} \
	else if (BASE == DEC16) { \
		printf("%d", (uint16_t)num); \
		fprintf(MHI_USE_access ulog_file, "%d", (uint16_t)num); \
	} \
	else if (BASE == TIME) { \
		printf("%'llu", (int64_t)(num));  \
		fprintf(MHI_USE_access ulog_file, "%'llu", (int64_t)(num)); \
	} \
	else if (BASE == HEX) { \
		printf("%llx", (uint64_t)(num)); \
		fprintf(MHI_USE_access ulog_file, "%llx", (uint64_t)(num)); \
	} \
	else if (BASE == HEXBYTE) { \
		printf("%02llx", (uint64_t)(num)); \
		fprintf(MHI_USE_access ulog_file, "%02llx", (uint64_t)(num)); \
	} \
	else if (BASE == CHAR) { \
		printf("%c", (uint8_t)num); \
		fprintf(MHI_USE_access ulog_file, "%c", (uint8_t)num); \
	} \
	else if (BASE == FLOAT) { \
		printf("%f", (double)(num)); \
		fprintf(MHI_USE_access ulog_file, "%f", (double)(num)); \
	}

// ulog_start() is empty but the actual ulog initialization
// will be done by init_multihw, directly called by simulator and not
// by app.
#define ulog_start()
#define log_flush() fflush(MHI_USE_access ulog_file); fflush(stdout)

#define try_write(str, size) \
	fwrite(str, size, 1, stdout); \
	fwrite(str, size, 1, MHI_USE_access ulog_file)

#define log_raw_data(mem, size) \
	fwrite(mem, size, 1, stdout); \
	fwrite(mem, size, 1, MHI_USE_access ulog_file)
/*#define log_str_num_BASE_str_L(str1, num, BASE, str2) \
	printf(str1); \
	fprintf(MHI_USE_access ulog_file, str1); \
	log_num_BASE(num, BASE); \
	printf(str2); \
	fprintf(MHI_USE_access ulog_file, str2)*/

// You have to have MHI_ARG present!
#define ulog_prefix() \
	printf("%c:UL@ns%'llu>", MHI_USE_access station->name, get_simtime(MHI_USE_access station)); \
	fprintf(MHI_USE_access ulog_file, "UL@ns%'llu>", get_simtime(MHI_USE_access station))


void service_routine(MHI_ARG_single)
#ifndef NOCODE
{
}
#endif // ifndef NOCODE
;


#endif  // ifdef ARDU elif defined (BEAGLE) elif defined(SIMUL)




// Effective for all platforms:
#define  log_str_L(str)                          log_str_N(str, sizeof(str) - 1)
#define  log_str_P(str)                          log_str_N(str, strlen(str))
#define ulog_str_N(str, size)     ulog_prefix(); log_str_N(str, size);
#define ulog_str_L(str)                         ulog_str_N(str, sizeof(str) - 1)
#define ulog_str_P(str)                         ulog_str_N(str, strlen(str))
#define ulog_num_BASE(num, BASE) {ulog_prefix(); log_num_BASE(num, BASE);        }
#ifndef ulog_fin_line
#define ulog_fin_line() \
	try_write(">\n", 2)
#endif // -ifndef ulog_fin_line

// "size" denotes the size w/out the trailing zero.
#define log_str_N(str, size) \
	if ((str)[size-1] == '\n') { \
		try_write(str, size-1); \
		ulog_fin_line(); \
	} \
	else { \
		try_write(str, size); \
	}

#ifndef log_flush
#define log_flush ERROR! // (TODO: implement this.)
#endif
#ifndef log_str_num_BASE_str_L
#define log_str_num_BASE_str_L(str1, num, BASE, str2) \
	{try_write(str1, sizeof(str1)-1); log_num_BASE(num, BASE); log_str_L(str2);}
#endif
#ifndef log_str_num_BASE_str_P
#define log_str_num_BASE_str_P(str1, num, BASE, str2) \
	{try_write(str1, strlen(str1)); log_num_BASE(num, BASE); log_str_P(str2);}
#endif
#define ulog_str_num_BASE_str_L(str1, num, BASE, str2) \
		do {ulog_prefix(); log_str_num_BASE_str_L(str1, num, BASE, str2);} while(false)


// SPI

#ifdef ARDU
#define SPI_SETTINGS SPISettings
#define SPI_INSTANCE bool /*unused*/
#define MSBFIRST8 MSBFIRST
#include "SPI.h"
#elif defined(BEAGLE)
#define SPI_SETTINGS struct spi_ioc_transfer
#define SPI_INSTANCE int /*fd*/
#define MSBFIRST8 8
#include "linux/spi/spidev.h"
#endif

#ifndef SIMUL
#ifndef NOCODE
SPI_SETTINGS spi_rfm;
SPI_INSTANCE spi_inst;
#else
extern SPI_SETTINGS spi_rfm;
extern SPI_INSTANCE spi_inst;
#endif
#endif // -ifndef SIMUL



#ifdef ARDU


#define multihw_spi_settings(tgtvar, spi_inst, clk_Hz, bit_order, data_mode) \
	tgtvar = new SPISettings(clk_Hz, bit_order, data_mode)

/*void ardu_spi_write_single(SPI_SETTINGS spi_sett, SPI_INSTANCE spi_inst,
							uint16_t addr, uint16_t data) {
	SPI.beginTransaction(spi_sett);
	
	digitalWrite(PIN_SPI_NSS_RFM, LOW); \
	addr |= 0x80; // MSB = 1 for write access
	SPI.transfer(addr);
	addr &= 0x7f
	SPI.transfer(data);
	\
	digitalWrite(PIN_SPI_NSS_RFM, HIGH); \
	SPI.endTransaction();
}*/

// TODO: Find a way speed and mode etc. can be parametrized.
#define multihw_spi_write_single(spi_sett, spi_inst, addr, data) \
	SPI.beginTransaction(spi_sett); \
	\
	digitalWrite(PIN_SPI_NSS_RFM, LOW); \
	addr |= 0x80; /* MSB = 1 for write access*/ \
	SPI.transfer(addr); \
	addr &= 0x7f; /* Restore Address*/ \
	SPI.transfer(data); \
	\
	digitalWrite(PIN_SPI_NSS_RFM, HIGH); \
	SPI.endTransaction();


// TODO: Find a (better) way speed and mode etc. can be parametrized.
	// AH! Once we declare addr and retv as volatile the cbSensitivityBoost
	//  variable (and all the rest) will be transferred properly. 
volatile uint8_t ardu_spi_read_single(SPI_SETTINGS spi_sett, SPI_INSTANCE spi_inst,
							volatile uint16_t addr)
#ifndef NOCODE
{
	SPI.beginTransaction(spi_sett);
	bool state = digitalRead(PIN_SPI_NSS_SD);
	digitalWrite(PIN_SPI_NSS_SD, HIGH); // Deactivate SD CS (high active)
	digitalWrite(PIN_SPI_NSS_RFM, LOW); // Activate RFM CS (low active)
	addr &= 0x7f; // MSB = 0 for read access
	addr <<= 8;
	uint16_t retv = SPI.transfer16(addr);
	digitalWrite(PIN_SPI_NSS_RFM, HIGH);
	digitalWrite(PIN_SPI_NSS_SD, state);
	SPI.endTransaction();
	return (retv&0xFF);
}
#endif // ifndef NOCODE
;

#define multihw_spi_read_single(spi_sett, spi_inst, addr, retv) \
	retv = ardu_spi_read_single(spi_sett, spi_inst, addr)


#elif defined BEAGLE // -ifdef ARDU


#ifdef NOCODE
extern
#endif
char* spidevname; 


// See also https://github.com/tekkamanninja/linux-beagleboard/blob/master/Documentation/spi/spidev_test.c

//#include <stdlib.h>
//#include <unistd.h>
#include <fcntl.h> // open
#include <string.h> // memset, strlen
#include <sys/ioctl.h>



// data_mode must be a writeble var?
#define multihw_spi_settings(tgt_spi_sett, spi_inst, clk_Hz, bit_spec, data_mode) \
	memset(&tgt_spi_sett, 0, sizeof(tgt_spi_sett)); \
	tgt_spi_sett.speed_hz = clk_Hz; \
	tgt_spi_sett.bits_per_word = bit_spec; \
	tgt_spi_sett.cs_change = data_mode // Misusing the struct field.

// this macro is internal
#define beagle_spi_set_mode(spi_inst, mode_in, ret) \
	uint8_t mode = mode_in; \
	ret = ioctl(spi_inst, SPI_IOC_WR_MODE, &mode); \
	if (ret == -1) { \
		ulog_str_L("Error: can't set spi mode [0x"); \
		log_num_BASE(mode, HEX); \
		log_str_L("]\n"); \
		quit_with_error(sharedERRNO_SPI); \
	}

void beagle_spi_write_single(SPI_SETTINGS spi_sett, SPI_INSTANCE spi_inst,
							  uint8_t addr, uint8_t data)
#ifndef NOCODE
{
	int ret;
	beagle_spi_set_mode(spi_inst, spi_sett.cs_change, ret);
	spi_sett.cs_change = 1;
	if (ret != -1) {
		addr |= 0x80;
		uint8_t tx[] =  { addr, data };
		uint8_t rx[sizeof(tx)/sizeof(tx[0])];
		spi_sett.tx_buf = (unsigned long)tx;
		spi_sett.rx_buf = (unsigned long)rx;
		spi_sett.len = 2;
		int ret = ioctl(spi_inst, SPI_IOC_MESSAGE(1), &spi_sett);
		if (ret == -1) {
			ulog_str_L("Error: can't transfer [write] spi data! [Addr: 0x");
			log_num_BASE(addr, HEX);
			log_str_L(", Data: 0x");
			log_num_BASE(data, HEX);
			log_str_L("]\n");
		}
	}
	spi_sett.cs_change = mode; // Write mode back to the odd location.
}
#endif
;

// TODO: Error handling
uint8_t beagle_spi_read_single(SPI_SETTINGS spi_sett, SPI_INSTANCE spi_inst,
							uint8_t addr)
#ifndef NOCODE
{
	volatile int ret;
	beagle_spi_set_mode(spi_inst, spi_sett.cs_change, ret);
	spi_sett.cs_change = 1;
	if (ret == -1) {
		quit_with_error(sharedERRNO_SPI);
		return 0xAA;
	}
	
	ret = ioctl(spi_inst, SPI_IOC_WR_MAX_SPEED_HZ, &spi_sett.speed_hz);
	//spi_sett.speed_hz = 0;
	//printf("r%2d\n", ret);
	
	if (ret != -1) {
		addr &= 0x7f;
		uint8_t tx[] =  { addr, 0x00 };
		uint8_t rx[] =  { 0, 0};
		spi_sett.tx_buf = (unsigned long)tx;
		spi_sett.rx_buf = (unsigned long)rx;
		spi_sett.len = 2;
		ret = ioctl(spi_inst, SPI_IOC_MESSAGE(1), &spi_sett);
		if (ret == 1) {
			ulog_str_L("Error: can't transfer [read] spi data! [Addr: 0x");
			log_num_BASE(addr, HEX);
			log_str_L("]\n");
			ret = 0x55;
			//quit_with_error(sharedERRNO_SPI); // Don't quit for one single error.
		}
		else
			ret = rx[1];
	}
	/*printf("r%2d   spi_sett.rx_buf = %llx\n", ret, spi_sett.rx_buf);
	ret = ioctl(spi_inst, SPI_IOC_RD_BITS_PER_WORD, &spi_sett.bits_per_word);
	printf("r%2d   spi_sett.bits_per_word = %u\n", ret, spi_sett.bits_per_word);
	ret = ioctl(spi_inst, SPI_IOC_RD_MAX_SPEED_HZ, &spi_sett.speed_hz);
	printf("r%2d   spi_sett.speed_hz = %u\n", ret, spi_sett.speed_hz);
	ret = ioctl(spi_inst, SPI_IOC_RD_MODE, &mode);
	spi_sett.cs_change = mode; // Write mode back to the odd location.
	printf("r%2d   spi_sett.cs_change = %u\n", ret, spi_sett.cs_change);*/
	
	//printf("r%2d\n", ret);
	
	return ret;
}
#endif
;


#define multihw_spi_write_single(spi_sett, spi_inst, addr, data) \
	beagle_spi_write_single(spi_sett, spi_inst, addr, data)

#define multihw_spi_read_single(spi_sett, spi_inst, addr, retval) \
	retval = beagle_spi_read_single (spi_sett, spi_inst, addr)


#endif // -ifdef ARDU elif defined(BEAGLE)



#ifndef SIMUL
void multihw_spi_start(SPI_SETTINGS* spi_rfm, SPI_INSTANCE * spi_inst)
#ifndef NOCODE
{
	
	
#ifdef ARDU
	pinMode(PIN_SPI_NSS_RFM, OUTPUT);
	*spi_inst = true;
	*spi_rfm = SPISettings(SPI_RFM_CLK, SPI_RFM_BITSPEC, SPI_RFM_MODE);
	
	
	
#else // assuming BEAGLE:
	*spi_inst = open(spidevname, O_RDWR);
	if (*spi_inst < 0) {
		ulog_str_L("Error: can't open SPI device '");
		log_str_P(spidevname);
		log_str_L("'!\n");
		*spi_inst = 0;
	}
	else {
		multihw_spi_settings((*spi_rfm), (*spi_inst),
						SPI_RFM_CLK, SPI_RFM_BITSPEC, SPI_RFM_MODE);
	}
	
#endif // -ifdef ARDU -else
	
	
	
#ifdef DEBUG_SPI
	spi_debug_line[0] = 'S';
	spi_debug_line[3] = 'h';
	spi_debug_line[4] = '-';
	spi_debug_line[5] = '-';
	spi_debug_line[8] = 'h';
	spi_debug_line[9] = '\n';
#endif
	
}
#endif // -ifndef NOCODE
;
#endif // -ifndef SIMUL







#ifdef SIMUL
struct trancv_instance;
#define lTI_ARG struct trancv_instance * ti,
#else
#define lTI_ARG
#endif

#ifdef __cplusplus
extern "C" {
#endif
void rfm_spi_write_single(lTI_ARG  uint8_t addr, uint8_t data)
#ifndef NOCODE
{
	
#ifdef DEBUG_SPI
	struct station * mh_station_arg = ti->station;
	spi_debug_line[1] = hexleft(addr);
	spi_debug_line[2] = hexright(addr);
#endif
	
#ifdef SIMUL
	simulator_write_reg(ti, addr, data);
#else
	multihw_spi_write_single(spi_rfm, spi_inst, addr, data); // this is a macro!
#endif
	
#ifdef DEBUG_SPI
	spi_debug_line[0] = 'w';
	spi_debug_line[4] = '<';
	spi_debug_line[5] = '-';
	spi_debug_line[6] = hexleft(data);
	spi_debug_line[7] = hexright(data);
	//SDlog_dataline(spi_debug_line, sizeof(spi_debug_line));
	ulog_str_N(spi_debug_line, sizeof(spi_debug_line));
#endif
}
#endif // ifndef NOCODE
;



uint8_t rfm_spi_read_single(lTI_ARG  uint8_t addr)
#ifndef NOCODE
{
	uint8_t retval;
	
#ifdef SIMUL
	retval = simulator_read_reg(ti, addr);
#else
	multihw_spi_read_single(spi_rfm, spi_inst, addr, retval); // this is a macro!
#endif
	
#ifdef DEBUG_SPI
	struct station * mh_station_arg = ti->station;
	spi_debug_line[0] = 'r';
	spi_debug_line[1] = hexleft(addr);
	spi_debug_line[2] = hexright(addr);
	spi_debug_line[4] = '-';
	spi_debug_line[5] = '>';
	spi_debug_line[6] = hexleft(retval);
	spi_debug_line[7] = hexright(retval);
	//SDlog_dataline(spi_debug_line, sizeof(spi_debug_line));
	ulog_str_N(spi_debug_line, sizeof(spi_debug_line));
#endif
	return retval;
}
#endif // ifndef NOCODE
;
#ifdef __cplusplus
} // extern "C"
#endif









// ======= PINS ========

bool setup_pins();

#ifdef ARDU

#define FN_PINSPEC uint8_t fn_arduino_pin_number
#define FN_PINUSE fn_arduino_pin_number

#elif defined BEAGLE

#ifdef NOCODE
#include "../src_multihw/BBBIO/BBBiolib.h"
#else
#include "../src_multihw/BBBIO/BBBiolib.c"
#endif // ifdef NOCODE
#define digitalRead(PIN) is_high(PIN)
#define FN_PINSPEC uint8_t fn_pinspec_header_number, uint8_t fn_pinspec_pin_number
#define FN_PINUSE fn_pinspec_header_number, fn_pinspec_pin_number

#elif defined SIMUL
#define digitalRead(bla) 0 // TODO: rewrite this if GPS pin becomes relevant!

#endif // ifdef ARDU elif defined BEAGLE

#ifndef NOCODE
#ifdef ARDU
bool setup_pins() {
	pinMode(PIN_GPS_PULSE, INPUT);
	
	// LEDs
	pinMode(PIN_ERRORLED, OUTPUT);
	pinMode(PIN_LED2, OUTPUT);
	pinMode(PIN_LED3, OUTPUT);
	pinMode(PIN_HEARTBEAT, OUTPUT); // Prohibits PIN_ERRORLED
	
	pinMode(PIN_SPI_NSS_SD, OUTPUT);
	pinMode(PIN_SPI_NSS_RFM, OUTPUT);
	
	digitalWrite(PIN_ERRORLED, LOW);
	digitalWrite(PIN_SPI_NSS_RFM, HIGH);  // low active
	//digitalWrite(PIN_SPI_NSS_SD, LOW);    // high active
	digitalWrite(PIN_SPI_NSS_SD, HIGH);
	
	
	pinMode(SPI_MISO_PIN, INPUT);
	pinMode(SPI_MOSI_PIN, OUTPUT);
	pinMode(SPI_SCK_PIN, OUTPUT);
	
	return true;
}
#elif defined BEAGLE // ifdef ARDU


bool setup_pins() {
	
	int status = iolib_init();
	if (status)
		return false;
	
	iolib_setdir(PIN_GPS_PULSE, BBBIO_DIR_IN);
	
	
	// TODO: set pin directions. Works like this:
	//    iolib_setdir(8,11, DIR_IN);
	//    iolib_setdir(8,12, DIR_OUT);
	// Get state with this: is_high(8, 11) and
	// set state with this: pin_low(8, 12)
	return true;
}


#endif // ifdef ARDU elif defined BEAGLE
#endif // ifndef NOCODE





// Never returns!
void error_flash(uint8_t error)
#if defined ARDU && (0 == defined NOCODE)
{
	/*while (true) {
		digitalWrite(PIN_ERRORLED, HIGH);
		volatile uint8_t cycles = 140;
		while(cycles) cycles--;
		digitalWrite(PIN_ERRORLED, LOW);
		cycles = 50;
		while(cycles) cycles--;
	}*/
	
	if (error != ERRNO_SD && error != sharedERRNO_SDWRITE) {
		dataFile.print("error_flash(): ERROR ");
		dataFile.print(error, DEC);
		dataFile.print("!\n");
		dataFile.flush();
		dataFile.close(); // This is the only time we ever close the file...
	}
	
	
	if (error > 5) {
		error = 5;
		dataFile.print("error_flash(): @Developer: please don't use error numbers > 5. Thank You!\n");
	}
	
	while (1) {
		// timing results from pindef.h definition
		
		digitalWrite(PIN_ERRORLED, HIGH);
		//uint16_t cycles = 400 + 500*(uint16_t)error;
		
		//cycles *= 50; while (cycles) {cycles --;}
		delay(400 + 400*error - 30);
		//if (error == 4) while(1) {}
		
		digitalWrite(PIN_ERRORLED, LOW);
		//while(true);
		delay(2400 - 400*error - 30);
	}
}
#endif // #if defined ARDU && (0 == defined NOCODE)
;

#ifdef ARDU
void quit_with_error(uint8_t error)
#ifndef NOCODE
{
	if (error != ERRNO_SD) {
		
		// Prepare the message
		uint8_t buf[27];
		strncpy(buf, "SD-ERROR: raw >>????<<!\n+++", sizeof(buf));
		buf[16] =  SdErrorTrace & 0xFF;
		buf[17] = (SdErrorTrace & 0xFF00) >> 8;
		buf[18] = (SdErrorTrace & 0xFF0000) >> 16;
		buf[19] = (SdErrorTrace & 0xFF000000) >> 24;
		
		SD.implant_error_block(buf);
		delay(100);
		SD.implant_error_block(buf);
		error_flash(5);
	}
	error_flash(error);
}
#endif // ifndef NOCODE
;
#elif defined BEAGLE
extern bool beagle_global_error;
void quit_with_error(uint8_t error)
#ifndef NOCODE
{
	fprintf(stderr, "ERROR %d -> quitting.\n", error);
	beagle_global_error = true;
}
bool beagle_global_error
#endif // ifndef NOCODE
;

#elif defined SIMUL
void quit_with_error(MHI_ARG uint8_t eRroR)
#ifndef NOCODE
{
	fprintf(stderr, "ERROR %d -> quitting.\n", eRroR);
	error(MHI_USE_arg   "an error was signalled to multihw");
}
#endif // ifndef NOCODE
;



#endif




// ======== GPS ========

#ifdef ARDU
#define setup_gps_uart() Serial.begin(9600)  // Will use PIN "D 0" and "D 1"
	// TODO: configure GPS to faster speed -> NMEA is faster.
	//		(And configure it to only output one line)
#elif defined BEAGLE
#define setup_gps_uart()
	// TODO: That's not enough!
#elif defined SIMUL
#define setup_gps_uart()
#endif








// ======= TIMER =======


/** Main structure of the timing:
 * 
 * The instructions happening in one app loop are few, so that the
 * app loop is called very often. So the app loop can check the timers
 * in every iteration.
 * 
 * There are four timers which can be time driven (system clock) or 
 * pin driven (interrupt-like). There's no guarantee that a requested
 * time or pin "interrupt" will be propageted fastly.
 * In simulator, the times and interrupts will always be exactly in
 * time. On Arduino the TimerOne library will be used which is pretty
 * precise I guess. BeagleBone is no realtime
 * system, but you can assume that it has low delays as the code
 * execution speed should be low.
 * 
 * Following modes are available:
 * # TIME_PRECISE: Precise Time. If you use it make sure that the
 *   code which is called by interrupt will finish soon and does not
 *   quit the program with an inifinite loop.)
 * # TIME_LAZY: The code will be called at the time you specified,
 *   but can be little delayed if your software design is good or
 *   possibly be very delayed if the design is bad.
 * # PIN_LAZY: The code will be called when the pin state changes.
 *   Precision is similar to TIME_LAZY.
 * 
 * So far only TIME_LAZY is implemented. TIME_PRECISE and PIN_LAZY yet
 * have to be implemented.
 * 
 **/



#define TIME_PRECISE 0
#define TIME_LAZY    1
#define PIN_LAZY     3


#if defined(ARDU) || defined(BEAGLE)

// to mark function names which will later be redefined by a preprocessor
// macro and thus never used/defined:
#define FAKE extern
// TODO: remove this?
FAKE void multitimer_setpin(uint8_t cbnum, FN_PINSPEC, uint8_t reason);

#endif



void multitimer_start(MHI_ARG_single);
#ifndef NOCODE
static void multitimer_specific_start(MHI_ARG_single);
#endif

// Checks whether there are interrupts to be handled
// and which have been forgotten.
void multitimer_interrupt_check();
void multitimer_wakemeup(MHI_ARG   uint8_t cbnum, time_int usec, uint8_t reason);


void multitimer_delay_coarse(MHI_ARG  time_int delay);



#ifdef ARDU


#include <TimerOne.h> // This is not sufficient as only include.

// 16 us precision.

/*#define NSEC           / 16 / 1000
#define USEC           / 16
#define MSEC * 1000    / 16
#define SEC  * 1000000 / 16*/
// ... rethink
#define NSEC           / 16 / 1000
#define USEC           / 16
#define MSEC     * (time_int) 125 / 2   // For this to work with 16 bit types you can only use 9.5 out of the 19 h which would be available.
#define SEC      * (time_int) 100 * 625

#define SEC2ti(sec) (sec SEC)
#define ti2USEC(timeint) (timeint*16)


#ifndef NOCODE
// return unit is 16 us!
time_int get_systime()
{
	// multitimer does not interfere with millis().
	return millis() MSEC;
}
#endif // ifndef NOCODE


/*inline void setup_timer_hw_specific() {
	ulog_str_L("[Arduino]: Nothing to do to setup timer.\n");
}*/



#ifndef NOCODE
void multitimer_interrupt_check() {
	//bool ret = false;
	uint8_t timsk1_old = TIMSK1;
	TIMSK1 &= ~_BV(TOIE1);  // temporarily disables the timer overflow interrupt
	
	
	int i;
	for (i = 0; i < NUMBER_OF_TIMERS; i++) {
		if (!timer_callback[i]) continue;
		if (usecs_to_callback[i] == 0) {
			//ret = true;
			timer_callback[i](reason_opaque[i]);
			usecs_to_callback[i] = -1;
		}
	}
	
	
	TIMSK1 = timsk1_old;   // restores the timer overflow interrupt enable bit
	// Could with small probability cause an immediate interrupt.
	
	//return ret;
}



static void multitimer_specific_start() {
	Timer1.initialize(1000000);
}


static void multitimer_interrupt() {
	long elapsed = (long)Timer1.prescale_factor * (long)Timer1.pwmPeriod;
	
	uint8_t i_lowest = 0xFF;
	
	uint8_t i;
	for (i = 0; i < NUMBER_OF_TIMERS; i++) {
		if (usecs_to_callback[i] <= elapsed) {
			// TODO: error = elapsed - usecs_to_callback[i]
		//	timer_callback[i](reason_opaque[i]);
		//	usecs_to_callback[i] = -1;
			// TODO: DEACTIVATED THIS. Find some better interrupt concept.
			// TODO: Using this to notify multitimer_interrupt_check()
			usecs_to_callback[i] = 0;
			continue;
		}
		
		usecs_to_callback[i] -= elapsed;
		
		if (i_lowest == 0xFF || usecs_to_callback[i] < usecs_to_callback[i_lowest]) {
			i_lowest = i;
		}
		
		Timer1.attach_interrupt(NULL, usecs_to_callback[i_lowest]);
	}
}

void multitimer_wakemeup(uint8_t cbnum, time_int usec16, uint8_t reason) {
	
	if (TIMSK1 | _BV(TOIE1)) {
		usecs_to_callback[cbnum] = usec16 * 16;
		bool cnt_reset;
		long elapsed = Timer1.possibly_stop_earlier(usec16*16, cnt_reset);
		if (cnt_reset) {
			int i;
			for (i = 0; i < NUMBER_OF_TIMERS; i++) {
				if (usecs_to_callback[i] != -1) {
					if (usecs_to_callback[i] < elapsed) {
						// TODO: Something has gone wrong -> Warning message
						usecs_to_callback[i] = 0;
						// We shoud invoke the interrupt service routine
					}
					else
						usecs_to_callback[i] -= elapsed;
				}
			}
		}
	}
	else {
		usecs_to_callback[cbnum] = usec16*16;
		Timer1.attach_interrupt(multitimer_interrupt, usec16*16);
	}
}


void multitimer_delay_coarse(time_int requested_delay) {
	// *16/1000 = *2/125
	
	
	CASSERT((10*125)/2 == 10 MSEC, ___Definition_of_MSEC_changed___);
	time_int new_delay = (requested_delay *2)/125; //   / 1000;
	
	if (requested_delay && !new_delay) new_delay = 1;
	
	delay(new_delay);
}



#endif // ifndef NOCODE

#elif defined BEAGLE

#include <time.h>

time_int get_systime()
#ifndef NOCODE
{
	// https://stackoverflow.com/questions/3756323/how-to-get-the-current-time-in-milliseconds-from-c-in-linux
	// -> https://stackoverflow.com/a/17371925
	uint32_t  ns; // Nanoseconds
	time_t     s; // Seconds
	struct timespec spec;
	
	//clock_gettime(CLOCK_REALTIME, &spec);
	clock_gettime(CLOCK_MONOTONIC_RAW, &spec);
	
	s  = spec.tv_sec;
	ns = spec.tv_nsec;
	
	return (uint64_t)ns + 1000000000ULL * (uint64_t)s;
}
#endif // ifndef NOCODE
;


// Beaglebone counts in us.
#define NSEC 
#define USEC * 1000ULL
#define MSEC * 1000000ULL
#define  SEC * 1000000000ULL

#ifndef NOCODE
#define NOCODE
#include "beaglesupp/TimerOne.c"
#undef NOCODE
#else
#include "beaglesupp/TimerOne.c"
#endif


/*inline void setup_timer_hw_specific() {
}*/



#ifndef NOCODE

static void multitimer_specific_start() {
	struct timespec spec;
	
	clock_gettime(CLOCK_REALTIME, &spec);
	
	ulog_str_L("CLOCK_REALTIME is telling following time: ");
	
	log_num_BASE(spec.tv_sec, DEC);
	log_str_L(" s  + ");
	log_num_BASE(spec.tv_nsec, DEC);
	log_str_L(" ns since epoch 1970.\n");
}


void multitimer_delay_coarse(time_int delay) {
	struct timespec time;
	
	// On Debian IoT (BeagleBone) CLOCK_MONOTONIC_RAW is not supported.
	//clock_gettime(CLOCK_MONOTONIC_RAW, &time);
	clock_gettime(CLOCK_REALTIME, &time);
	
	//printf("Old time: %u.%9u s.\n", time.tv_sec, time.tv_nsec);
	
	uint64_t ns_tmp = time.tv_nsec;
	ns_tmp += delay;
	time.tv_sec += ns_tmp / 1000000000;
	time.tv_nsec = ns_tmp % 1000000000;
	
	//printf("Target time: %u.%9u s.\n", time.tv_sec, time.tv_nsec);
	//int ret = clock_nanosleep(CLOCK_MONOTONIC_RAW, TIMER_ABSTIME, &time, NULL);
	int ret = clock_nanosleep(CLOCK_REALTIME, TIMER_ABSTIME, &time, NULL);
	
	if (ret) {
		printf("Error sleeping %llu ns -> ret=%d, errno(str)=%s.\n", delay, ret, strerror(ret));
	}
	//printf("Slept %llu ns ? ret=%d, errno(str)=%s.\n", delay, ret, strerror(ret));
}

#endif // -ifndef NOCODE

#elif defined(SIMUL) // -ifdef ARDU -elif defined BEAGLE

#define RES_PROCESS(str, retv) \
	if (res) { \
		fprintf(stderr, str); \
		quit_with_error(MHI_USE_arg sharedERRNO_PTHREAD); \
		return retv; \
	}



/*void setup_timer_hw_specific(MHI_ARG_single)
#ifndef NOCODE
{
	ulog_str_L("[Simulation]: Nothing to do to setup timer.\n");
}
#endif 
;*/

//void setup(MHI_ARG_single);
//void loop(MHI_ARG_single);

#ifndef NOCODE

#define NOCODE
#include "common/app-user.h"
#undef NOCODE

static void* mh_simul_thread_fn(void* arg) {
	MHI_ARG_single = (struct station*) arg;
	int res;
	
	// Wait for start condition
	res = pthread_mutex_lock(&(MHI_USE_access auxiliary_mutex));
	RES_PROCESS("Mutex lock failed.\n", NULL)
	
	MHI_USE_access phase = 'S';
	
#ifdef DEBUG_SIMEVENT_DETAIL
	ulog_str_L("Station thread: waiting for start trigger.\n");
#endif
	
	res = pthread_cond_wait(
				&(MHI_USE_access startloop_sim2app),
				&(MHI_USE_access auxiliary_mutex));
	if (res) {
		fprintf(stderr, "pthread_cond_wait failed.\n");
		quit_with_error(MHI_USE_arg sharedERRNO_PTHREAD);
		pthread_mutex_unlock(&(MHI_USE_access auxiliary_mutex));
		return NULL;
	}
	
	// Now mutex is locked again.
	
	// TODO: Do this whith mutex2 locked
	if(MHI_USE_access phase != 'W') {
		ulog_str_num_BASE_str_L("mh_simul_thread_fn: phase should be 'W' but is '", MHI_USE_access phase, CHAR, "'!\n"
		);
	}
	MHI_USE_access phase = 'k';
	
	if (MHI_USE_access loop == NULL) {
		error(MHI_USE_arg   "no loop declared!\n");
	}
	
#ifdef DEBUG_SIMEVENT_DETAIL
	// loop runs with phase 'k' until it's waiting for something.
	ulog_str_L("Station thread: going into loop.\n");
#endif
	
	while (true) {
		// I would very appreciate it when loop would  invoke 
		// some delay, even a very tiny delay from time to time
		// would be sufficient.
		MHI_USE_access loop(MHI_USE_single);
		
	}
	
	return NULL; // TODO: some nicer value here?
}
#endif // -ifndef NOCODE


#ifndef NOCODE

// This will be called by simulator which calls setup() of only.c
static void multitimer_specific_start(MHI_ARG_single)
{
#ifdef DEBUG_SIMEVENT
	ulog_str_L("multitimer_specific_start.\n");
#endif
	// This is somewhat complicated here. We cannot come around threading
	// (Main problem is multitimer_delay_coarse.)
	int res;
	res = pthread_attr_init(&(MHI_USE_access threadattr));
	RES_PROCESS("Attr init failed.\n",)
	
#define xPTHREAD_MUTEX_INITIALIZER (pthread_mutex_t) PTHREAD_MUTEX_INITIALIZER
//#define xPTHREAD_MUTEX_INITIALIZER (pthread_mutex_t) PTHREAD_ERRORCHECK_MUTEX_INITIALIZER_NP
#define xPTHREAD_COND_INITIALIZER (pthread_cond_t) PTHREAD_COND_INITIALIZER
	MHI_USE_access auxiliary_mutex        = xPTHREAD_MUTEX_INITIALIZER;
	MHI_USE_access auxiliary_mutex2       = xPTHREAD_MUTEX_INITIALIZER;
	MHI_USE_access startloop_sim2app      = xPTHREAD_COND_INITIALIZER;
	MHI_USE_access please_resume_sim2app  = xPTHREAD_COND_INITIALIZER;
	MHI_USE_access piece_executed_app2sim = xPTHREAD_COND_INITIALIZER;
	
	MHI_USE_access phase = 0;
	
#ifdef DEBUG_SIMEVENT
	ulog_str_L("Creating thread.\n");
#endif
	res = pthread_create(
				&(MHI_USE_access thethread),
				&(MHI_USE_access threadattr),
				mh_simul_thread_fn,
				MHI_USE_single
						);
	RES_PROCESS("Thread creation failed.\n",)
	
	while (MHI_USE_access phase != 'S') {
		// Wait for thread to get ready.
	}
	
	// TODO: Now someone has to trigger the loop! (task for app.c)
}
#endif


#ifndef NOCODE
static void mh_execute_next_piece_of_app(MHI_ARG  char expected_state)
{
	int res;
	
	//ulog_str_L("mh_execute_next_piece_of_app: locking mutex.\n");
	log_flush();
	
	res = pthread_mutex_lock(&(MHI_USE_access auxiliary_mutex));
	RES_PROCESS("Mutex lock failed (starting loop).\n",)
	
	
	if(MHI_USE_access phase != expected_state) {
		ulog_str_num_BASE_str_L("mh_execute_next_piece_of_app: phase should be '", expected_state, CHAR, "' ");
		log_str_num_BASE_str_L("but is '", MHI_USE_access phase, CHAR, "'!\n");
	}
	MHI_USE_access phase = 'W'; // wait for kick
	
	if (expected_state == 'S')
		res = pthread_cond_signal(&(MHI_USE_access startloop_sim2app));
	else if (expected_state == 'w')
		res = pthread_cond_signal(&(MHI_USE_access please_resume_sim2app));
	else {RES_PROCESS("Please provide a known expected phase.\n",)}
	
	// Now the loop is triggered but still standing still
	// as mutex is locked.
	
	res = pthread_mutex_unlock(&(MHI_USE_access auxiliary_mutex));
	RES_PROCESS("Mutex unlock failed (starting loop).\n",)
	
	// ulog_str_L("mh_execute_next_piece_of_app: signalled thread.\n");
	
	// Idea to use another auxiliary_mutex(2) came when I saw this:
	// https://stackoverflow.com/questions/4544234/calling-pthread-cond-signal-without-locking-mutex
	res = pthread_mutex_lock(&(MHI_USE_access auxiliary_mutex2));
	RES_PROCESS("Mutex lock failed (fin-wait).\n",);
	
	if (MHI_USE_access phase == 'W' || MHI_USE_access phase == 'k') {
		// so MHI_USE_access phase != 'w' -> the thread execution is not yet done.
		
		// ulog_str_L("mh_execute_next_piece_of_app: Waiting for simulation break.\n");
		res = pthread_cond_wait(
					&(MHI_USE_access piece_executed_app2sim),
					&(MHI_USE_access auxiliary_mutex2));
		
		if (res) {
			fprintf(stderr, "pthread_cond_wait failed for fin-wait.\n");
			quit_with_error(MHI_USE_arg sharedERRNO_PTHREAD);
			pthread_mutex_unlock(&(MHI_USE_access auxiliary_mutex2));
			return;
		}
		
		res = pthread_mutex_unlock(&(MHI_USE_access auxiliary_mutex2));
		RES_PROCESS("Mutex unlock failed (fin-wait).\n",);
		
		// ulog_str_L("mh_execute_next_piece_of_app: Simulation has breaked.\n");
		//pthread_mutex_lock(&(MHI_USE_access auxiliary_mutex));
		//pthread_mutex_unlock(&(MHI_USE_access auxiliary_mutex));
		//ulog_str_L("Locked and unlocked.\n");
	}
	else {
		// We don't have to wait as the loop has paused execution
		// obviously.
		ulog_str_L("mh_execute_next_piece_of_app: Simulation did break fastly\n");
		res = pthread_mutex_unlock(&(MHI_USE_access auxiliary_mutex2));
		RES_PROCESS("Mutex unlock failed (after fastbreak).\n",);
	}
	
	// Now the first portion of loop is executed and the app thread is waiting.
	
	if(MHI_USE_access phase != 'w') {
		ulog_str_num_BASE_str_L("mh_execute_next_piece_of_app: phase really should be 'w' but is '", MHI_USE_access phase, CHAR, "'!\n"
		);
	}
}
#endif // -ifndef NOCODE

void mh_start_loop(MHI_ARG_single)
#ifndef NOCODE
{
	mh_execute_next_piece_of_app(MHI_USE_arg 'S');
}
#endif
;

void mh_resume_loop(MHI_ARG_single)
#ifndef NOCODE
{
	mh_execute_next_piece_of_app(MHI_USE_arg 'w');
}
#endif
;


#ifndef NOCODE


// TODO: create multitimer_feed_cpu_time(): like multitimer_delay_coarse()
// but doesn't neccessarily create an sim event (can be unprecise) and will
// forward time only in simulator but not in reality.
// Main purpose is the simulation sanity/efficiency/precision.




void multitimer_delay_coarse(MHI_ARG    time_int delay) {
	// Assuming mutex is locked.
	
	time_int time = get_simtime(MHI_USE_single);
	
	time += delay;
	
	MHI_USE_single -> app -> cl_delay = time;
	claim(MHI_USE_arg  time);
	
	//printf("multitimer_delay_coarse: Lock mutex2.\n");
	//fprintf(stderr, "multitimer_delay_coarse: Lock mutex2.\n");
	int res;
	//printf("auxiliary_mutex2 = %p.\n", &(MHI_USE_access auxiliary_mutex2));
	res = pthread_mutex_lock(&(MHI_USE_access auxiliary_mutex2));
	RES_PROCESS("Mutex lock failed (set-wait).\n",);
	
	
	MHI_USE_access phase = 'w'; // wait
	//printf("multitimer_delay_coarse: Signal break.\n");
	res = pthread_cond_signal(&(MHI_USE_access piece_executed_app2sim));
	
	res = pthread_mutex_unlock(&(MHI_USE_access auxiliary_mutex2));
	RES_PROCESS("Mutex unlock failed (set-wait).\n",);
	
	
	
	//printf("multitimer_delay_coarse: Wait for continue.\n");
	fflush (stdout);
	
	res = pthread_cond_wait(
				&(MHI_USE_access please_resume_sim2app),
				&(MHI_USE_access auxiliary_mutex));
	if (res) {
		fprintf(stderr, "pthread_cond_wait failed for delay.\n");
		quit_with_error(MHI_USE_arg sharedERRNO_PTHREAD);
		pthread_mutex_unlock(&(MHI_USE_access auxiliary_mutex));
		return;
	}
	
	res = pthread_mutex_lock(&(MHI_USE_access auxiliary_mutex2));
	RES_PROCESS("Mutex lock failed (set-wait II).\n",);
	
	
	MHI_USE_access phase = 'k'; // wait
	
	res = pthread_mutex_unlock(&(MHI_USE_access auxiliary_mutex2));
	RES_PROCESS("Mutex unlock failed (set-wait II).\n",);
	
	/*
	if(MHI_USE_access phase != 'k') {
		ulog_str_num_BASE_str_L("multitimer_delay_coarse: phase should be 'k' but is '", MHI_USE_access phase, CHAR, "'!\n"
		);
	}*/
	
	//printf("multitimer_delay_coarse: Continuing.\n");
	
	// Now mutex is locked again.
}


#endif // -ifndef NOCODE
;


#endif // -ifdef ARDU -elif defined BEAGLE -elif defined(SIMUL)



#if defined BEAGLE || defined SIMUL
#ifndef NOCODE
// Checks whether there are interrupts to be handled
// and which have been forgotten.
//
// This definition of the function is shared for beagle and simul
void multitimer_interrupt_check(MHI_ARG_single) {
	//bool ret = false;
	time_int time = get_systime(MHI_USE_single);
	//ulog_str_L("    Interrupt check.\n");
	int i;
	for (i = 0; i < NUMBER_OF_TIMERS; i++) {
		if (!MHI_USE_access timer_callback[i]) continue;
		//ulog_str_num_BASE_str_L("... ", i, DEC, "    ");
		//log_num_BASE(MHI_USE_access nsecs_of_callback[i], DEC);
		//log_str_L(" B\n");
		if (0LL <= MHI_USE_access nsecs_of_callback[i]
					&& MHI_USE_access nsecs_of_callback[i] <= time) {
			//ret = true;
			//ulog_str_num_BASE_str_L("... (",
			//			MHI_USE_access reason_opaque[i], DEC, ") C\n");
			MHI_USE_access nsecs_of_callback[i] = TIME_INVALID; // BEFORE actual callback 
							// (-> callback can request re-call.)
			MHI_USE_access timer_callback[i](
						MHI_USE_arg
						MHI_USE_access reason_opaque[i]);
//			ulog_str_L("...D\n");
		}
	}
//	ulog_str_L("...E\n");
	//return ret;
}


void multitimer_wakemeup(MHI_ARG   uint8_t cbnum, time_int nsec, uint8_t reason) {
	//printf("nsec = %'lu\n", nsec);
	MHI_USE_access nsecs_of_callback[cbnum] = TIME_INVALID;
	MHI_USE_access reason_opaque[cbnum] = reason;
	MHI_USE_access nsecs_of_callback[cbnum] = nsec + get_systime(MHI_USE_single);
#ifdef DEBUG_MULTITIMER
	printf("nsecs_of_callback (B) = %ld.\n", MHI_USE_access nsecs_of_callback[cbnum]);
#endif
}


#endif // -ifndef NOCODE
#endif // -ifdef BEAGLE || defined SIMUL




#ifndef NOCODE
void multitimer_start(MHI_ARG_single) {
	int i;
	for (i = 0; i < NUMBER_OF_TIMERS; i++) {
		MHI_USE_access timer_callback[i] = NULL;
	}
	//ulog_str_L("MTS Anfang.\n");
	multitimer_specific_start(MHI_USE_single);
	//setup_timer_hw_specific(MHI_USE_single);
	ulog_str_L("MTS\n");
}
#endif


// cbnum darf nur von 0 bis NUMBER_OF_TIMERS-1 lauten.
void multitimer_initialize_callback(MHI_ARG  uint8_t cbnum, void (*cb_routine)(MHI_ARG   uint8_t reason), uint8_t MODE)
#ifndef NOCODE
{
	MHI_USE_access timer_callback[cbnum] = cb_routine;
	MHI_USE_access mode_of_callback[cbnum] = MODE;
#ifdef ARDU
	usecs_to_callback[cbnum] = -1;
#elif defined BEAGLE || defined SIMUL
	MHI_USE_access nsecs_of_callback[cbnum] = TIME_INVALID;
#endif
}
#endif // ifndef NOCODE
;




// Needed for AODVroute

void get_systimeval(MHI_ARG   struct timeval* tv, void* tz)
#ifndef NOCODE 
{
	if (tv == NULL) return;
	time_int now = get_systime(MHI_USE_single);
	if (tz != NULL) {
		printf("ERROR: tz is not implemented!");
		return;
	}
	
	tv->tv_sec = now / (1 SEC);
	tv->tv_usec = ti2USEC(now - SEC2ti(tv->tv_sec));
}
#endif
;






#endif // #ifndef MULTIHW_C



