

// This file is named only.c for historic reasons. Actually it contains
// the hardware management / distribution functions.



#ifndef ONLY_C
#define ONLY_C

// There is no error handling in SPI.
/* https://www.arduino.cc/en/Reference/SPI says:
 MOSI: 11 or ICSP-4
 MISO: 12 or ICSP-1
 SCK:  13 or ICSP-3
 SS:   10 [Set to OUTPUT; is not needed.]
 */

#include <stdint.h>



//#define APP_GLOBAL_STRUCT_NAME  struct only_global
#define ONLY_GLOBAL_VARS(extern) \
	extern time_int thetime;





#ifndef NOCODE
#define NOCODE
#include "multihw.c"
#include "gps.c"
#undef NOCODE
#else // ifndef NOCODE
#include "multihw.c"
#include "gps.c"
#endif // ifndef NOCODE else




#ifndef ONLY_GLOBAL_VARS
#define ONLY_GLOBAL_VARS(extern)
#endif // -ifndef ONLY_GLOBAL_VARS
#ifndef APP_GLOBAL_VARS
#define APP_GLOBAL_VARS(extern)
#endif // -ifdef APP_GLOBAL_VARS

#ifdef SIMUL
struct app_global {
	ONLY_GLOBAL_VARS()
	APP_GLOBAL_VARS()
};
#define global(var) (MHI_USE_access app -> var)
#define gl MHI_USE_access app ->
#else
#ifdef NOCODE
ONLY_GLOBAL_VARS(extern)
APP_GLOBAL_VARS(extern)
#else
ONLY_GLOBAL_VARS()
APP_GLOBAL_VARS()
#endif
#define global(var) var
#define gl
#endif // -ifdef SIMUL -else





void app_setup(MHI_ARG_single);
void app_loop(MHI_ARG   uint8_t reason);

#include "pindef.h"

#define REASON_TIMEPULSE 0
#define REASON_TIMEEMULATION 1
#define REASON_UNKNOWN 3
#define BIGGEST_REASON 15 // No reason bigger than 15 to be implemented here.
		// (letting space for others)





#ifdef SIMUL
void setup(MHI_ARG_single);
#endif

#ifndef NOCODE
void setup(MHI_ARG_single)
{
	
	// Console code: https://www.arduino.cc/en/Tutorial/ReadASCIIString
	//  https://www.arduino.cc/reference/en/language/functions/communication/serial/
	
	setup_gps_uart();
	
	//ulog_str_L("AAA\n");
#ifdef ARDU
	setup_pins();
	//digitalWrite(PIN_ERRORLED, HIGH);
	
	// Clear the error message
	const uint8_t* str = "No SD-Error :-)\n+++       ";
	SD.implant_error_block(str);
	delay(100);
	SD.implant_error_block(str);
	SdErrorTrace = 0;
	
#elif defined BEAGLE
	bool status = setup_pins();
	//ulog_num_BASE(status, DEC);
	//log_str_L(" )) BBB\n");
	if (!status) {
		beagle_global_error = true;
		return;
	}
#endif
	
	
#ifndef SIMUL
	multihw_spi_start(&spi_rfm, &spi_inst);
	//quit_with_error(sharedERRNO_SPI);
#endif
	
	ulog_start();
	
	log_str_L("\n\n\n\n");
	ulog_str_L("Hello, here's station '");
	log_str_N((char*)&station_name, 1);
	log_str_L("'. :-)\n");
	
	
	multitimer_start(MHI_USE_single);
	
	
	
	ulog_str_L("Global startup done.\n");
	
	// expecting this to enable globals.
	app_setup(MHI_USE_single);
	
	// [SIMUL:] globals are available after app_setup
	global(thetime) = get_systime(MHI_USE_single);
	//printf("thetime=0x%llx\n", thetime);
	
}
;
#endif


static inline uint8_t hex(uint8_t byte) {
	if (byte < 10) {
		return '0' + byte;
		// 0x30 = '0'
	}
	return 'A' + byte - 10;
	// 0x41 = 'A'
	// 0x61 = 'a'
}
static inline uint8_t hexright(uint8_t byte) {
	return hex(byte & 0x0f);
}
static inline uint8_t hexleft(uint8_t byte) {
	return hex(byte >> 4);
}
/*
void serialEvent()
#ifndef NOCODE
{
	// directly "call" GPS
	serialEvent_happened();
}
#endif // ifndef NOCODE
;*/


// https://www.arduino.cc/en/Reference/Libraries
// Timing: https://playground.arduino.cc/Deutsch/HalloWeltMitInterruptUndTimerlibrary
// https://de.wikibooks.org/wiki/C-Programmierung_mit_AVR-GCC:_Timer_und_Counter
// https://www.arduino.cc/en/Tutorial/CurieTimer1Interrupt

static inline uint64_t add64(uint64_t a, uint64_t b) {
	uint64_t al = a & 0xffffffff;
	uint64_t ah = a >> 32;
	uint64_t bl = b & 0xffffffff;
	uint64_t bh = b >> 32;
	
	al += bl;
	ah += bh;
	if (al >= 0x100000000) {
		al -= 0x100000000;
		al ++;
	}
	return al | (ah << 32);
}

#if (defined NOLOOP) || (defined SIMUL)
void loop(MHI_ARG_single);
#endif


#ifndef NOLOOP

#ifndef NOCODE
void loop(MHI_ARG_single)
{
	//while (true)
	//	app_loop(0);
	bool ret;
	/*ret = */multitimer_interrupt_check();
	static bool pulsestate; // = 0 (C standard)
	static time_int service_time = 0; // = 0 (C standard)
	
	//app_loop(0); return;
	
	if (pulsestate)
		pulsestate = digitalRead(PIN_GPS_PULSE);
	else {
		if (digitalRead(PIN_GPS_PULSE)) {
			ulog_str_L("GPS PIN raised.\n");
			global(thetime) = get_systime(MHI_USE_single);
			app_loop(MHI_USE_arg   REASON_TIMEPULSE);
			pulsestate = true;
			
		}
	}
	//app_loop(0); return;
	time_int time2 = get_systime(MHI_USE_single);
	ret = false;
	
	/*ulog_str_num_BASE_str_L("    thetime           = ", thetime            , DEC, "\n");
	ulog_str_num_BASE_str_L("            1200 MSEC = ",          1200 MSEC, DEC, "\n");
	ulog_str_num_BASE_str_L("OA; thetime+1200 MSEC = ", thetime + 1200 MSEC, DEC, "\n");*/
//	ulog_str_L("OA\n");
	if (time2 > global(thetime) + 1200 MSEC) {
		//printf("time2=%llu > thetime=%llu + 1200 MSEC !\n", time2, thetime);
		global(thetime) += 1000 MSEC;
		
		// This is the fallback just in case GPS is not present.
		app_loop(MHI_USE_arg   REASON_TIMEEMULATION);
		
		//quit_with_error(4);
		ulog_str_L("OB\n");
	}
	
	//else app_loop(9); return; if (1) {}
	else if (time2 > service_time + 8500 MSEC) {
		// Force service routine at least every ~ 8.5 s
		// independend whether there are other waiting facilities.
		service_routine(MHI_USE_single);
		service_time = get_systime(MHI_USE_single);
		ret = true;
		
		ulog_str_L("OC\n");
	}
	else {
		int i;
		for (i = 0; i < NUMBER_OF_TIMERS; i++) {
			if (!MHI_USE_access timer_callback[i]) continue;
#ifdef ARDU
			if (usecs_to_callback[i] >= 0) {
#elif defined BEAGLE || defined SIMUL
			if (MHI_USE_access nsecs_of_callback[i] >= 0) {
#endif
				// pending callback will lead to an automatic wait (?)
				ret = true;
			}
		}
		//quit_with_error(3);
		
//		ulog_str_L("OD\n");
		if (!ret)
			multitimer_delay_coarse(MHI_USE_arg 100 MSEC); // Kleine Verschnaufpause.
	}
	
//	ulog_str_L("OE\n");
	if ((!ret) && time2 > service_time + 2600 MSEC) {
		//ulog_str_L("OV\n");
		// Try to invoke the service routine every 2.6 s.
		service_routine(MHI_USE_single);
		service_time = get_systime(MHI_USE_single);
		
		ulog_str_L("OW\n");
	}
	
//	ulog_str_L("OX\n");
	
#if defined SIMUL || defined BEAGLE
	time_int wtime = 10 MSEC;
	int i;
	for (i = 0; i < NUMBER_OF_TIMERS; i++) {
		if (!MHI_USE_access timer_callback[i]) continue;
		if (MHI_USE_access nsecs_of_callback[i] == TIME_INVALID) continue;
		
		//printf("Callback -> %'lu\n", MHI_USE_access nsecs_of_callback[i]);
		if (MHI_USE_access nsecs_of_callback[i] - get_systime(MHI_USE_single) < wtime)
			wtime = MHI_USE_access nsecs_of_callback[i] - get_systime(MHI_USE_single);
	}
	//printf("wtime = %'lu\n", wtime);
	multitimer_delay_coarse(MHI_USE_arg   wtime);
#endif
}
#endif
#endif // -ifndef NOLOOP

#ifndef NOCODE
//bool beagle_global_error;
#ifdef BEAGLE

#include "getopt.h"

// Option parsing from spidev-test.c which seems to come somewhere from
// Linux kernel.

static void print_usage(const char *prog) {
	printf("Usage: %s [-DsbdlHOLC3]\n", prog);
	puts(
		"  -G --GPS       next argument is for GPS.\n"
		"  -S --SPI       next argument is for SPI.\n"
		"  -D --device   device to use (default /dev/spidev1.1)\n"
		"  -N --name     name of station (typically one char out of WBCDO)\n"
/*	     "  -s --speed    max speed (Hz)\n"
	     "  -d --delay    delay (usec)\n"
	     "  -b --bpw      bits per word \n"
	     "  -l --loop     loopback\n"
	     "  -H --cpha     clock phase\n"
	     "  -O --cpol     clock polarity\n"
	     "  -L --lsb      least significant bit first\n"
	     "  -C --cs-high  chip select active high\n"
	     "  -3 --3wire    SI/SO signals shared\n"*/
		);
}

static bool parse_opts(int argc, char* argv[]) {
	char type = 0; // 0 = unknown, 'G' = GPS, 'S' = SPI
	spidevname = STR(SPI_DEV);
	
	while (1) {
		static const struct option lopts[] = {
			{ "GPS",     0, 0, 'G' },
			{ "SPI",     0, 0, 'S' },
			{ "device",  1, 0, 'D' },
			{ "name",    1, 0, 'N' },
/*			{ "speed",   1, 0, 's' },
			{ "delay",   1, 0, 'd' },
			{ "bpw",     1, 0, 'b' },
			{ "loop",    0, 0, 'l' },
			{ "cpha",    0, 0, 'H' },
			{ "cpol",    0, 0, 'O' },
			{ "lsb",     0, 0, 'L' },
			{ "cs-high", 0, 0, 'C' },
			{ "3wire",   0, 0, '3' },*/
			{ NULL, 0, 0, 0 },
		};
		int c;

		//c = getopt_long(argc, argv, "D:s:d:b:lHOLC3", lopts, NULL);
		c = getopt_long(argc, argv, "GSD:N:", lopts, NULL);
		
		if (c == -1)
			break;
		
		switch (c) {
		case 'G':
			type = 'G';
			break;
		case 'S':
			type = 'S';
			break;
		case 'D':
			if (type == 'S') spidevname = optarg;
			else if (type == 'G')
				printf("%s: WARNING: GPS not implemented yet!\n", argv[0]);
			else {
				printf("%s: Error: You specified a device in command line without\n"
					"   prepending the type (shall be: -SD %s or \n"
					"   --SPI --device %s)\n",
					argv[0], optarg, optarg);
				print_usage(argv[0]);
				return false;
			}
			break;
		case 'N':
			station_name = optarg[0];
			break;
/*		case 's':
			speed = atoi(optarg);
			break;
		case 'd':
			delay = atoi(optarg);
			break;
		case 'b':
			bits = atoi(optarg);
			break;
		case 'l':
			mode |= SPI_LOOP;
			break;
		case 'H':
			mode |= SPI_CPHA;
			break;
		case 'O':
			mode |= SPI_CPOL;
			break;
		case 'L':
			mode |= SPI_LSB_FIRST;
			break;
		case 'C':
			mode |= SPI_CS_HIGH;
			break;
		case '3':
			mode |= SPI_3WIRE;
			break;*/
		default:
			print_usage(argv[0]);
			return false;
		}
	}
	return true;
}


int main(int argc, char* argv[]) {
	
	if (!parse_opts(argc, argv))
		return 1;
	
	beagle_global_error = false;
	setup();
	while(!beagle_global_error)
		loop();
	// TODO: Exit-Bedingung (beagle_global_error wird nirgendwo gesetzt).
	
	return 55;
}




#endif // ifdef BEAGLE
#endif // ifndef NOCODE

#endif // ifndef ONLY_C
