
// (c) Jonas Bechtel 2018

#ifndef PINDEF_H
#define PINDEF_H


#define COMMON_FREQUENCY 433400000
#define COMMON_DEVIATION 10000
//#define COMMON_DEVIATION 35000
//#define COMMON_DEVIATION 100000
#define COMMON_BITRATE   1200  // Pretty slow; it will very probably arrive.
//#define COMMON_BITRATE 12000




#ifdef ARDU

// PIN "D 0": Used as receiving pin for UART
// PIN "D 1": Used as transmitting pin for UART

#define SD_CHIP_SELECT 5 // FIXED
#define PIN_SPI_NSS_SD SD_CHIP_SELECT
// SD_CHIP_SELECT is fixed as it has this position on the shield.
// See https://www.arduino.cc/en/uploads/Main/arduino-wireless-SD-shield-schematic.pdf
// -> J6/J7 has a SD_CS pin which will be shifted to SEN3 which is
//    Pin 2 (CD/DAT3) of the SD card
// See also http://www.interfacebus.com/Secure_Digital_Card_Pinout.html

#define PIN_ERRORLED 6
#define PIN_LED2 7
#define PIN_LED3 8
#define PIN_HEARTBEAT LED_BUILTIN // Is this same as ERRORLED?

#define PIN_GPS_PULSE 9 // Pin "D 9"

#define PIN_SPI_NSS_RFM 10 // FIXED: D 10 (SPI.SS)
// forbidden: D 11 (SPI.MOSI)
// forbidden: D 12 (SPI.MISO)
// forbidden: D 13 (SPI.SCK)
// remark: we use SPI0.

#elif defined BEAGLE

#define PIN_GPS_PULSE 8,11 // Pin P8.11

#define SPI_DEV /dev/spidev1.0
#define SPI_DELAY 100 // 10 us wait after transaction

#define SPI_MODE0 SPI_MODE_0
//#define SPI_MODE0 3 // Don't know why.

#endif


// 10 MHz max SCK time
// SPI_MODE0 := CPOL=0, CPHA=0

#define SPI_RFM_CLK 200000
#define SPI_RFM_BITSPEC MSBFIRST8
#define SPI_RFM_MODE SPI_MODE0


// ERRNO = 0 ... 4
// So there are only 5 error numbers; we have to share them.
// 0 ->  400 ms light, 2000 ms dark
#define ERRNO_SD      0
// 1 ->  400 ms light, 1600 ms dark
#define sharedERRNO_SDOPEN  1
#define sharedERRNO_PTHREAD 1
// 2 -> 1200 ms light, 1200 ms dark
#define sharedERRNO_SDWRITE 2
#define sharedERRNO_MALLOC  2
// 3 -> 1600 ms light,  800 ms dark
#define ERRNO_TXSTART 3
#define ERRNO_RXSTART 3
#define sharedERRNO_2nd_fileopen  3 // SD-Read-Write
#define sharedERRNO_AODV          3
// 4 -> 2000 ms light,  400 ms dark
#define sharedERRNO_INTERN_ASSERT 4
#define sharedERRNO_SPI           4
#define sharedERRNO_read_minus_1  4 // SD-Read-Write
// 4 -> 2400 ms light,    0 ms dark
#define sharedERRNO_EOF           5 // SD-Read-Write
#define sharedERRNO_STARTUP       5 // FrameSwitch

#endif // ifndef PINDEF_H




// INFORMATIVE, copied from /usr/share/arduino/hardware/arduino/avr/variants/standard/pins_arduino.h :


// ATMEL ATMEGA8 & 168 / ARDUINO
//
//                  +-\/-+
//            PC6  1|    |28  PC5 (AI 5)
//      (D 0) PD0  2|    |27  PC4 (AI 4)
//      (D 1) PD1  3|    |26  PC3 (AI 3)
//      (D 2) PD2  4|    |25  PC2 (AI 2)
// PWM+ (D 3) PD3  5|    |24  PC1 (AI 1)
//      (D 4) PD4  6|    |23  PC0 (AI 0)
//            VCC  7|    |22  GND
//            GND  8|    |21  AREF
//            PB6  9|    |20  AVCC
//            PB7 10|    |19  PB5 (D 13)
// PWM+ (D 5) PD5 11|    |18  PB4 (D 12)
// PWM+ (D 6) PD6 12|    |17  PB3 (D 11) PWM
//      (D 7) PD7 13|    |16  PB2 (D 10) PWM
//      (D 8) PB0 14|    |15  PB1 (D 9) PWM
//                  +----+
//
// (PWM+ indicates the additional PWM pins on the ATmega168.)

