

// © 2018 Jonas Bechtel
// Licensed under the Apache License, Version 2.0; see file <prj>/meta/LICENSE


#ifndef READ_RIDE_H
#define READ_RIDE_H



#include <stdbool.h>
#include <stdint.h>

#include <assert.h> // Remove this when no assert() is needed any more.

// time is, as whole simulation, in nanoseconds. We need at least 48 bit.
// time format in _ride format is in milliseconds.

#include "sim-common.h" // MSEC, SEC



#define MAX_PATHS 10
#define DEFAULT_DURATION 1000 // 1 s
struct WBCDOline {
	uint64_t time;
	uint16_t duration; // in ms.
	
	struct S2S_channel ch[15];
	//	WWch, WBch, WCch, WDch, WOch,
	//	      BBch, BCch, BDch, BOch,
	//	            CCch, CDch, COch,
	//	                  DDch, DOch,
	//	                        OOch;
};

#define TIMELINE_VAL_INVALID 0xff000000
struct timeline {
	// if (from == TIMELINE_VAL_INVALID) this struct instance is invalid.
	uint32_t from;
	uint32_t to;
};


struct channel_in_time {
	struct WBCDOline * chtimespecs;
	int speclines;
	uint32_t total_time_s;
	struct timeline * mapper;
};


int get_chsel(int st_num1, int st_num2);


#ifndef NOCODE

#include <stdio.h>


#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h> // open()
#include <string.h> // strerror()
#include <errno.h> // errno
#include <unistd.h> // read()
#include <stdlib.h> // realloc()

#ifndef TARGET_INCREASE
#define TARGET_INCREASE 100
#endif

#include "station-names.c"


int get_chsel(int num1, int num2) {
	// TODO: generalize that for WBCDOlen != 5
#define DiagonalSize (  (  WBCDOlen * (WBCDOlen+1)  )/2  )
	
	// Ansatz: Großes Dreieck, davon ein kleines Dreieck abziehen und noch
	//    die nötige Anzahl nach links gehen.
	
	// Beispiel für num1 = 'C'(2), num2 = 'D'(3)
	
	//	WWch, WBch, WCch, WDch, WOch,    0  1  2  3  4 
	//	      BBch, BCch, BDch, BOch,       5  6  7  8
	//	            CCch, CDch, COch,          9 10 11
	//	                  DDch, DOch,            12 13
	//	                       >OOch<              >14<
	
	// ->
	
	//	WWch, WBch, WCch, WDch, WOch,    0  1  2  3  4 
	//	      BBch, BCch, BDch, BOch,       5  6  7  8
	//	            CCch, CDch,>COch<          9 10>11<
	
	// -> 
	
	//	WWch, WBch, WCch, WDch, WOch,    0  1  2  3  4 
	//	      BBch, BCch, BDch, BOch,       5  6  7  8
	//	            CCch,>CDch<                9>10<
	
	
	return DiagonalSize-1 - (((WBCDOlen+1-(num1+1))-1)*(WBCDOlen+1-(num1+1)))/2 - (WBCDOlen-(num2+1));
}

#endif // ifndef NOCODE


struct channel_in_time read_ride(char* filename)
#ifndef NOCODE
{
	
	printf(__FILE__ ": Reading _ride file '%s'.\n", filename);
	
	struct channel_in_time retV;
	
	int fd = open(filename, O_RDONLY);
	if (fd == -1) {
		int myerrno = errno;
		fprintf(stderr, __FILE__ ":%d: Could not open file '%s': "
				"errno = %d/%s\n", __LINE__, filename, errno, strerror(myerrno));
		
		retV.speclines = 0;
		retV.total_time_s = 0;
		retV.chtimespecs = NULL;
		retV.mapper = NULL;
		return retV;
	}
	
	
	
	// STATE MACHINE
	int state = 0;
		// 0 = header (searching for bridge line)
		// 1 = header (searching for bridge line end)
		// 3 = >
		// 4 = time
		// 5 = <
		// 6 = channels: name, delay, strength
		// 9 = next channel or line.
	
	const char* search = "\nDIREKT JETZT BEGINNEN DIE DATEN:";
	int slen = strlen(search) - 1;
	
	// TARGET BUFFER
	int target_lines = 0;
	//int target_pos = 0;
	//struct WBCDOline* itgt = NULL;
//Cheap refactoring ...
#define target_pos (retV.speclines)
#define itgt (retV.chtimespecs)
	itgt = NULL;
	retV.total_time_s = 0;
	retV.speclines = 0;
	
	int chsel = 0;
	uint16_t last_time_read = 0;
	uint32_t last_time_ms = 0;
	int path_num = 0;
	
	// FILE READ BUFFER
	int brpos = 0;
	int bwpos = 0;
	char buf[1024];
	
	// last byte
	bool last_byte_written = false;
	
	ssize_t in;
	while (true) {
		
		// instead of bwpos - 1.5*slen it could be .5*sizeof(buf)
		if (brpos >= bwpos - (1.5*slen) && !last_byte_written) {
			off_t start = (off_t)&buf;
			size_t diff = bwpos - brpos;
			if (brpos < bwpos) {
				memmove((void*) start, (void*)(start + brpos), diff);
				start += diff;
			}
			
			size_t request = sizeof(buf) - diff;
			in = read(fd, (void*)start, request);
			if (in < request) {
				// Add one newline in order to get the last WBCDOline saved.
				buf[diff+in] = '\n';
				//printf("buf[diff+in=%d+%d] = %c.\n", diff, in, buf[diff+in]);
				in = in + 1;
				last_byte_written = true;
		//		printf("Extra byte written.\n");
			}
			brpos = 0;
			bwpos = in + diff;
			if (in == -1) {
				int myerrno = errno;
				fprintf(stderr, __FILE__ ": Error when reading: %d/%s.\n",
						errno, strerror(myerrno));
				break;
			}
		}
		//printf("bwpos=%d   brpos=%d\n", bwpos, brpos);
		
		if (brpos >= bwpos)
			break;
		
		//printf("state%2d.\n", state);
		
		if (state == 0) {
			int i;
			
			for(i = brpos; i < bwpos - slen; i++) {
				//printf ("i = %d.\n", i);
				//printf("_%c", buf[i]);
				int res = strncmp(search, &(buf[i]), slen);
				if (res == 0) {
					state = 1;
					
					i += slen;
					// -> brpos = i + slen;
					
					break;
				}
			}
			
			brpos = i;
			
			//printf ("state=%d, brpos=%d\n", state, brpos);
			continue;
		}
		
		if (state == 1) {
			// copied to state=9
			int i;
			for(i = brpos; i < bwpos; i++) {
				// printf("%c", buf[i]);
				if (buf[i] == '\n') {
					i += 1;
					//brpos = i+1;
					state = 3;
	//				printf("Commentary header skipped.\n");
					break;
				}
					
			}
			brpos = i;
			continue;
		}
		
		if (state == 3) {
			//printf("buf[%d..%d] = %c%c%c%c%c%c\n", brpos-6, brpos-1, buf[brpos-6], buf[brpos-5], buf[brpos-4], buf[brpos-3], buf[brpos-2], buf[brpos-1]);
		//	printf("buf[brpos=%d] = '%c'.\n", brpos, buf[brpos]);
			//printf("buf[%d..%d] = %c%c%c%c%c%c\n", brpos+1, brpos+6, buf[brpos+1], buf[brpos+2], buf[brpos+3], buf[brpos+4], buf[brpos+5], buf[brpos+6]);
			if (buf[brpos] == '\n') {
				// ignore empty line.
				brpos++;
				continue;
			}
			
			if (buf[brpos] != '>') {
				printf("Error in file format: missing '>' at begin of line.\n");
				goto fail;
			}
			
			// A new line has been opened.
			if (target_pos == target_lines) {
				target_lines += (TARGET_INCREASE);
				struct WBCDOline* ttmp;
				ttmp = realloc(itgt, sizeof(struct WBCDOline)* target_lines);
				if (ttmp)
					itgt = ttmp;
				else {
					int myerrno = errno;
					printf(__FILE__ ":%d: Error when allocating memory:"
						" %d/%s.\n", __LINE__, myerrno, strerror(myerrno));
					goto fail;
				}
			}
			
#define INIT_WBCDOline(access) \
			(access).duration = DEFAULT_DURATION; \
			int i; \
			for (i = 0; i < 15; i++) { \
				(access).ch[i].dist = 0xFF; \
			}
			INIT_WBCDOline(itgt[target_pos]);
			
			brpos++; state = 4;
			continue;
		}
		
		if (state == 4) {
			uint16_t pretime = ((uint16_t)(buf[brpos+1]) << 8)
			                  | (uint8_t)(buf[brpos]);
			//printf("pretime=%x %x.\n", (uint16_t)(buf[brpos+1]) << 8,
			//			(uint16_t)((uint8_t)buf[brpos]));
			
			if ((!(pretime & 0x80)) || pretime & 0x8000) {
				printf("Error in number format: time '0x%04X' is not a "
						 "plimb.\n", pretime);
				goto fail;
			}
			
			pretime = ((pretime & 0x7F00) >> 1) | (pretime & 0x007F);
			//printf("   pretime = %5d.   ", pretime);
			
			last_time_ms = last_time_ms + pretime - last_time_read;
			
			if (pretime < last_time_read) {
				// unfold
				last_time_ms += 0x4000;
			}
			last_time_read = pretime;
			
			itgt[target_pos].time = (uint64_t)last_time_ms * 1000000ULL; // ms -> ns; * 10^6
			//printf(" time/ms = %8d.\n", last_time_ms);
			
			brpos += 2;
			state = 5;
			if(target_pos == 0)
				continue;
			
			// Finalize former entry.
			int i;
			for(i = target_pos - 1 - path_num; i < target_pos; i++) {
				itgt[i].duration = last_time_ms - (itgt[i].time / (1 MSEC));
				uint64_t total_time = itgt[i].time/(1 SEC) + itgt[i].duration/1000 + 1; // +1 wg. Abrundungsfehler
				if (retV.total_time_s < total_time)
					retV.total_time_s = total_time;
				//printf("last_time_ms = %u, linetime = %llu/%llu (%llu), lineduration = %d (%d), total_time = %d.\n", last_time_ms, (itgt[i].time / (1 MSEC)), itgt[i].time/(1 SEC), itgt[i].time, itgt[i].duration*(1 SEC)/(1 MSEC), itgt[i].duration, total_time);
			}
			// TODO: There are still some entries not finalized, am I right?
			path_num = 0;
			
			continue;
		}
		
		if (state == 5) {
			if (buf[brpos] != '<') {
				printf("Error in file format: missing '<' after time.\n");
				goto fail;
			}
			
			
			
			brpos++; state = 6;
			continue;
		}
		
		if (state == 6) {
			int num1 = get_WBCDO_idx(buf[brpos]);
			int num2 = get_WBCDO_idx(buf[brpos+1]);
			
			if (num1 > 4 || num2 > 4) {
				printf("Error: the input '%c%c' does not correspond to known\n"
						"station names which could be selected from %s.\n",
						buf[brpos], buf[brpos+1], WBCDO);
			}
			
			if (num1 > num2) {
				int itmp = num1;
				num1 = num2; num2 = itmp;
			}
			
			chsel = get_chsel(num1, num2);
			chsel = 14 - (((6-(num1+1))-1)*(6-(num1+1)))/2 - (4-num2);
			assert(chsel == get_chsel(num1, num2));
			
			if (itgt[target_pos].ch[chsel].dist != 0xFF) {
				// There is already a channel set, so we discover here a 
				// reflected or otherwise happening second channel.
				// TODO: Yet there's no strategy to cope with that. <- TODO from when is this comment? we do cope with that!
				printf("Reflection dataset.\n");
				
				path_num++;
				target_pos++;
				if (target_lines == target_pos) {
					// nerviges Speichermanagement nötig
					struct WBCDOline* ttmp;
					ttmp = realloc(itgt,
							sizeof(struct WBCDOline) * ++target_lines);
					if (ttmp)
						itgt = ttmp;
					else {
						int myerrno = errno;
						printf(__FILE__ ":%d: Error when allocating memory:"
							" %d/%s.\n", __LINE__, myerrno, strerror(myerrno));
						goto fail;
					}
				}
				
				INIT_WBCDOline(itgt[target_pos])
				
				itgt[target_pos].time = itgt[target_pos-1].time;
			}
			
			
			brpos += 2;
			
			itgt[target_pos].ch[chsel].dist = buf[brpos];
			brpos ++;
			
			// Repeated in air.c: Damp really means damp. So if damp == 2 then
			// a wave sent out with  1 mW  will be received  @ 0.5 mW.
			itgt[target_pos].ch[chsel].damp = ((float*)&buf[brpos])[0];
			brpos += 4;
			
			if (buf[brpos] != '.') {
				printf("Error in file format: missing '.' at "
						"end of channel %c-%c.\n", WBCDO[num1], WBCDO[num2]);
				goto fail;
			}
			brpos ++;
			
			state = 8;
			
			continue;
		}
		
		if (state == 8) {
			char c = buf[brpos]; // peek
			if (('A' <= c && c <= 'Z') || ('a' <= c && c <= 'z') ) {
				state = 6; // next channel
				continue;
			}
			state = 9;
			continue;
		}
		
		if (state == 9) {
			// copied from state=1
			int i;
			for(i = brpos; i < bwpos; i++) {
				if (buf[i] == '\n') {
					target_pos ++; // So next line.
					
					i += 1;
					// -> brpos = i+1;
					state = 3; // next line;
					break; // out of for search
				}
			}
			brpos = i;
			continue; // state machine
			
		}
		
		fprintf(stderr, __FILE__ ": Error: wrong state number.\n");
		goto fail;
		
	}
	
	close(fd);
	
	// printf("Read %d lines.\n", target_pos);
	
	
	//*tgt = itgt;
	retV.speclines = target_pos;
	
	
	
	return retV;
	
fail:
	close(fd);
	retV.speclines = 0;
	retV.total_time_s = 0;
	retV.chtimespecs = NULL;
	retV.mapper = NULL;
	return retV;
}
#endif // ifndef NOCODE
;




// Create an array which knows which channelline is valid for a given second,
// where the second is the array index. (And some end-Array information has
// to be present)
// (array is stored in simulator?)
// (TODO: move this function to read_ride.c)
void map_channeltimes(struct channel_in_time * src)
#ifndef NOCODE
{
	// Allocate a big bunch of memory, deallocate the unused rest in the end.
	struct timeline * tl;
	
	int arraysize = src->total_time_s;
	
	tl = malloc (sizeof(struct timeline)*arraysize);
	if (tl == NULL) {
		printf("ERROR: couldn't allocate memory for timeline!\n");
		src->mapper = NULL;
		return;
	}
	
	int i;
	for (i = 0; i < arraysize; i++)
		tl[i].from = TIMELINE_VAL_INVALID;
	
	
	for (i = 0; i < src->speclines; i++) {
		// Iterate over every specline and put the entry into tl.
		
		uint32_t lfrom = src->chtimespecs[i].time / 1000000000ULL;
		uint32_t lto = (src->chtimespecs[i].time + 1000000ULL * (uint64_t)src->chtimespecs[i].duration - 1)/ 1000000000ULL; // -1 here in order to get the exact fitting value floored.
		
		assert(lfrom < arraysize);
		assert(lto < arraysize);
		
		int j;
		for (j = lfrom; j <= lto; j++) {
			if (tl[j].from == TIMELINE_VAL_INVALID) {
				// Noch kein anderes Paket war diese Sekunde da -> ganz neu schreiben.
				tl[j].from = i;
				tl[j].to = i;
				continue;
			}
			if (tl[j].from > i)
				tl[j].from = i;
			else if (tl[j].to < i)
				tl[j].to = i;
		}
	}
	
	src->mapper = tl;
}
#endif // ifndef NOCODE
;



#endif // ifndef READ_RIDE_H



