

// © 2018 Jonas Bechtel
// Licensed under the Apache License, Version 2.0; see file <prj>/meta/LICENSE


#ifndef SIMEVENT_H
#define SIMEVENT_H

struct event;


struct wave_seq_segment;


typedef struct event {
	uint64_t time;
#define TIME_INVALID 0xFF00000000000000UL
#define CLAIM_INVALID TIME_INVALID
#define TIME_MAX     0xfeffffffffffffffUL
	
	// type = 
	//   '1' - Startup
	//   '0' - Quit all simulation
	//   'E' - Quit all simulation because there was an error
	//   'C' - You had a claim for now = time  (says simulator to station)
	//   'C' - I have a claim for then = time  (says station to simulator)
	//   '%' - One segment of your wave_sequence is transmitted.  (says simulator to tranceiver of station)
	//   '$' - Your whole wave_sequence is transmitted.  (says simulator to tranceiver of station)
	//   'S' - packet sent from routing
	//   'R' - wave received from air (says simulator to tranceiver of station)
	char type;
	// 0 and E type events cause an immediate exit and usually have time 0.
	
	// Station name or 0 = '\0' for generic event
	// Is only used for type = '$'.
	char station;
	
	
	union {
		struct wave_seq_segment * wssp; // type = 'R'
	};
	
	//// In case of S/R id is identical to the packet ID
	//// Otherwise it set in some other way.
	//pid id;
	
} event;

/*
#include <stdlib.h> // malloc
inline struct event * new_event() {
	struct event * ev = malloc(sizeof(struct event));
	if (!ev) {
		printf(__FILE__ ": Error: couldn't malloc memory for event!");
		return NULL;
	}
	ev->station = 0;
	return ev;
}*/

#ifndef MAX_SIMEVENTS
#define MAX_SIMEVENTS 100
#endif

typedef struct simevent_for_sglib {
	struct event events[MAX_SIMEVENTS+1]; // +1 for the exchange element.
	uint64_t dummy[10];
	int i;
} simevent;

/*
inline void sglib_simevent_my_delete(simevent * e, int i) {
	e->events[i].time = time_invalid;
	
	if (e->i > 0) {
		if (e->events[i-1].time == time_invalid)
			e->i--;
	}
	
	USE THIS: sglib_simevent_delete_nth(e, i);
}*/


#endif // ifndef SIMEVENT_H



