

% © 2018 Jonas Bechtel
% Licensed under the Apache License, Version 2.0; see file <prj>/meta/LICENSE



% This is to be sourced by Genride.m



% Explanation for station and other abbreviations in /Hardware/STATIONSNAMEN.txt
% Front<- W B C D O <-End
% Policies: K P (defined below in DYNAMICS)
% Surroundings: U Y (Urban, countrYside)



% TODO: (Slow and Fast) Fading has to be respected
% TODO: Make up a overview on the different models and what they predict for
%	different situations.


% SURROUNDINGS

% (Not taken into consideration: typical electrical situation at traffic lights,% at breaks, ...)


% We assume following situation in U region: 
% - reflections by urban canyon (signal keeps strength in straight streets)
% - sometimes 
% - proposed loss (in dB): L = 46.614 + 31.635 log_10 (f) + 35.224 log_10 (d)
% We assume following situation in Y region: 
% - direct transmission
% - no dispersion
% - Free Space Path Loss (in dB): L_FSPL = 20 log_10 (d) + 20 log_10 (f) + 92.45
%	(Okumura-Modell nach "Radio propagation modeling on 433 MHz")
% - Flat Earth plus Directed Model (in dB): L_system = L_p.flat + L_rest
%	With L_p.flat = 120 + 40 log10(R_km) - 20 log10(h1) - 20 log10(h2)  | with h1=h2=1m
%	     L_p.flat = 120 + 40 log10(R_km)

% We use 433.92 MHz as f for all loss estimations. This is the center of the 
% 433 MHz ISM band (433.05 ... 434.79)


% Very simple aproach: Both formulas are calculated for the way between two
%	stations and the ratio between U and Y is used to give the respective
%	weight to the U and Y formula.
% (Because on the first view I couldn't find a good way to combine the single
%  parts of the formulas.)

% -> L_combined = dU/(dU+dY)*L(dU+dY, f) + dY((dU+dY)*L_p.flat(dU+dY)
%  =  dU/(dU+dY)*(46.614 + 31.635 log_10 (433.92) + 35.224 log_10 (dU+dY)) + dY/(dU+dY)*(40 log_10 (dU+dY) + 120)
%  = (dU * (130.0485 + 35.224 log_10(dU+dY)) + dY * (40 log_10(dU+dY) + 120))/(dU+dY)
%  = 120 + (log_10(dU+dY)*(35.224 dU + 40 dY) + 10.0485 dU) / (dU+dY)

% CAUTION: The Paper which we take the U formula from has presented the wrong formula!!
% It gets better if the coefficients are exchanged -> 
%  L* = 35.224+31.635*log10(f)+46.614*log10(d)
% (It is an interesting fact that this is pretty near to the L_p.flat model)

% => L_combined* = 120 + (log_10(dU+dY)*(46.614 dU + 40 dY) - 1.34 dU) / (dU+dY)


% CAUTION: Such path loss estimations can be very inaccurate and can vary from city to
% city. "Radiowave Propagation" book (Levis, Johnson, Teixeira) recommends to obtain the
% parameters empirically.

% Lee model would be great to use but seems not to be good for distances < 1.6 km.


% Keep in mind: the physical distance of the stations in the convoi is
% different to the actual route distance of the stations as in case of
% curves the path can be abbreviated.

% route distance = "tourkm"

% TODO: In addition to the proposed models from the papers, some noise has to
%	be assumed [Or isn't that a Genride but rather a Sim task?]



% st1, st2 = "W"/"B"/"C"/"D"/"O"
function [L_combined, t_on_air_100ns, L_refl, tp_refl] ...
			= determine_S2S_path_loss(LandR, rasterpairs, stpos, st1, st2, time)
	global RASTERSIZE;
	
	[Usum, Ysum, reflection] = ...
				classify_S2S_path(LandR, rasterpairs, stpos, st1, st2, time);
	
	Usum = Usum * RASTERSIZE;
	Ysum = Ysum * RASTERSIZE;
	
	L_combined = determine_classified_path_loss(Usum, Ysum);
	
	% This assumes LOS and therefore is unprecise but doesn't matter anyway as
	% the time is too low to affect simulation.
	speed_of_light = 299792458; % m/s  https://de.wikipedia.org/wiki/Lichtgeschwindigkeit
	t_on_air_100ns = (Usum + Ysum) * 10000000 / speed_of_light;
	
	
	if(reflection)
		L_refl=determine_classified_path_loss(reflection(1,1), reflection(1,2));
		tp_refl=sum(reflection);
	else
		L_refl = 0;
		tp_refl = 0;
	endif
	
endfunction


function L_combined = determine_classified_path_loss(Urban_meters, countrY_meters)
	
	% Die oben beschriebene Interimsformel ...
%	L_combined* = 120 + (log_10(dU+dY)*(46.614 dU + 40 dY) - 1.34 dU) / (dU+dY)
	% Achtung, das elektrische Modell ist da noch nicht so geeignet.
	
	
	Usum = Urban_meters;
	Ysum = countrY_meters;
	%totaldiff = sqrt(Usum^2 + Ysum^2);
	totaldiff = Usum + Ysum;
	
	if (totaldiff <= 10)
		% TODO: Find a good characteristic here which is "stetig".
		L_combined = 1;
		return;
	endif
	
	L_combined = 120 ...
		+ 1/totaldiff * ( (46.614*Usum + 40*Ysum)*log10(totaldiff/1000) - 1.34*Usum);
	
	% TODO: Fading (Railay? Fading)
endfunction



