

% © 2018 Jonas Bechtel
% Licensed under the Apache License, Version 2.0; see file <prj>/meta/LICENSE



1; % to be sourced by Genride.m

% TODO: Find straight streets -> channel with better conditions.


% Read in the "csv" file = octave variable file


% Bedeutung der Werte:
% 50 = Freie Landschaft
% 51 = Wald (wird im Projekt nicht verwendet)
% 52 = Dorf
% 53 = Stadt (wird im Projekt nicht verwendet)
% 55 = Reflektionsidikator (eine Verbindung, die hierüber geht, wird ein wenig
%		gedämpft (Leistungsfaktor 0,5) und ein Weg, der 400 m länger ist, wird
%       hinzugefügt.)
% 80 = Hier fährt die Tour in der freien Landschaft
% 82 = Hier fährt die Tour in der Stadt


% Every box denotes an area of 50m*50m.

% Every route box must have a defined surrounding (so the eight cells around
%  must be fully accessible by Octave.) -> don't make a route at the map border.


% Latitude = geogr. Breite = Zählt, wie weit man nördlich vom Äquator ist.
% Hier wird der Begriff nur zur klaren Trennung innerhalb des Projekts 
% verwendet, mit einer Kugeloberfläche hat er nichts zu tun.
% (Gilt auch für longitude)


% Start ist per Definitionem die am weitesten linke Stelle, die einen Wert
% höher gleich 80 hat.
% lon = number of the box, counted from left to right
% lat = number of the box, counted from top to bottom(!)
% Short: lon |||||     lat =
function [lon, lat] = find_start(LandR)
	
	for i = 1:size(LandR)(2)
		lat = find(LandR(:,i) >= 80, 1);
		if size(lat)(1) > 0
			lon = i;
			break;
		endif
	endfor
	
endfunction


function freebool = testfree(LandR, visited, pla, plo)
	freebool = 0;
	if (LandR(pla, plo) >= 80)
		if (visited(pla, plo) == 0)
			freebool = 1;
		endif
	endif
endfunction

function [rasterpairs] = find_route(LandR, start_lon, start_lat)
	
	visited = zeros(size(LandR));
	visited (start_lat, start_lon) = 1;
	pla = start_lat; npla = pla;
	plo = start_lon; nplo = plo;
	
	rasterpairs = [plo, pla];
	
	while 1
		if     testfree(LandR,visited,pla-1,plo  ); npla = pla-1; nplo = plo  ;
		elseif testfree(LandR,visited,pla+1,plo  ); npla = pla+1; nplo = plo  ;
		elseif testfree(LandR,visited,pla  ,plo+1); npla = pla  ; nplo = plo+1;
		elseif testfree(LandR,visited,pla  ,plo-1); npla = pla  ; nplo = plo-1;
		elseif testfree(LandR,visited,pla-1,plo+1); npla = pla-1; nplo = plo+1;
		elseif testfree(LandR,visited,pla+1,plo+1); npla = pla+1; nplo = plo+1;
		elseif testfree(LandR,visited,pla-1,plo-1); npla = pla-1; nplo = plo-1;
		elseif testfree(LandR,visited,pla+1,plo-1); npla = pla+1; nplo = plo-1;
		endif
		
		if (nplo == plo && npla == pla)
			break;
		endif
		rasterpairs(size(rasterpairs)(1)+1,1:2) = [nplo, npla];
		visited(npla, nplo) = 1;
		
		plo = nplo;
		pla = npla;
	endwhile
	
endfunction




function rasterpairs = annotate_tourmeter(rasterpairs_in)
	global RASTERSIZE
	boxwidth = RASTERSIZE; % = boxheight
	
	% Assuming that the route goes from one center to the next center
	
	totalnum = size(rasterpairs_in)(1);
	routerow = -1.* ones(totalnum, 1);
	routerow(1,1) = 0;
	
	for i = 2:totalnum
		routerow(i,1) = routerow(i-1,1) ...
			+ sqrt(sum(sum((rasterpairs_in(i-1:i,:).*[1;-1]),1).^2,2)) ...
				 * RASTERSIZE;
	endfor
	
	rasterpairs = rasterpairs_in;
	rasterpairs(:,3) = routerow(:,1);
endfunction


% Here we need the annotated version of the rasterpairs.
function list = YUlist(LandR, rasterpairs)	
	list = zeros(0,2);
	
	type = "U"; % We start in the city.
	
	for i = 1:size(rasterpairs)(1)
		qualifier = mod(LandR(rasterpairs(i,2),rasterpairs(i,1)), 10);
		if (qualifier == 0 && type != "Y")
			% landscape begins here.
			list(size(list)(1)+1, 1) = mean(rasterpairs(i-1:i,3));
			type = "Y";
		elseif (qualifier == 2 && type != "U")
			% city begins here.
			list(size(list)(1)  , 2) = mean(rasterpairs(i-1:i,3));
			type = "U";
		endif
	endfor
	
	if (type == "Y")
		list(size(list)(1),2) = rasterpairs(size(rasterpairs)(1),3)+1000;
	endif
endfunction






function b = is_urban(LandR, lon, lat)
	b = 0;
	
	qualifier = LandR(lat, lon);
	if (mod(qualifier, 10) == 2)
		b = 1;
	elseif (mod(qualifier, 10) == 5)
		b = 2;
	endif
endfunction




% rasterpairs = [lon,lat,tourmeter; lon,lat,tourmeter; ...]
%
% returns positions as LandR raster indices.
function pos = get_position(rasterpairs, stpos, stn, time)
	
	stnpos1 = stpos(floor(time), stn);
	stnpos2 = stpos(ceil(time), stn);
	
	factor1 = mod(time, 1);
	
	tourmeter = factor1 * stnpos1 + (1-factor1) * stnpos2;
	
	idx = find (rasterpairs(:,3) >= tourmeter, 1);
	idx = max(2, idx); % Access safety for the start
	% TODO: Access safety for the end
	proceeding = (tourmeter - rasterpairs(idx-1, 3)) ...
			/ (rasterpairs(idx,3)-rasterpairs(idx-1,3));
	assert(proceeding >= 0);
	assert(proceeding <= 1);
	
	% Short: lon |||||     lat =
	xpos = rasterpairs(idx-1,1) ...
		+ proceeding * (rasterpairs(idx,1) - rasterpairs(idx-1,1));
	ypos = rasterpairs(idx-1,2) ...
		+ proceeding * (rasterpairs(idx,2) - rasterpairs(idx-1,2));
	
	pos = [xpos, ypos];
endfunction





% st1, st2 = "W"/"B"/"C"/"D"/"O"
% Usum: amount of meters in urban space
% Ysum: amount of meters in rural space
% reflection: [Usum, Ysum]. Is != [0, 0] if there's a reflection.
function [Usum, Ysum, reflection] = classify_S2S_path(LandR, rasterpairs, stpos, st1, st2, time)
	
	reflection = [0, 0];
	
	st1 = index("WBCDO", st1);
	st2 = index("WBCDO", st2);
	
	% Expand time to 2D position (LandR indices):
	st1pos = get_position(rasterpairs, stpos, st1, time);
	st2pos = get_position(rasterpairs, stpos, st2, time);
	
	londiff = st2pos(1,1) - st1pos(1,1);
	latdiff = st2pos(1,2) - st1pos(1,2);
	
	cdiff = sqrt(londiff^2 + latdiff^2); % Actual distance in 2D layer
	% In terms of numerical stability this could be calculated later (inline)
	
	
	if(abs(londiff) >= abs(latdiff))
		% direction: mainly left<->right
		% -> in every column at most 2 boxes can occur
		
		if (londiff == 0)
			Usum = 0;
			Ysum = 0;
			return;
		endif
		
		1; % Leave everything as it is.
	else
		% direction: top<->bottom
		% -> in every row at most 2 boxes can occur
%		assert(0);
		
		% There could be two aproaches for the top<->bottom direction:
		% A) duplicate code from if->true branch (dangerous)
		% B) make an x,y exchange trick (dangerous)
		
		% As we don't want to duplicate code (too much is duplicated)
		% we do turn the map around 90°
		
		% TODO: Here ist the point reached when extensive testing
		% would really be required.
		
		tmp = st1pos(1,1);
		st1pos = [st1pos(1,2), tmp];
		
		tmp = st2pos(1,1);
		st2pos = [st2pos(1,2), tmp];
		
		LandR = LandR';

		tmp = londiff;
		londiff = latdiff;
		latdiff = tmp;
		
		clear tmp;
	endif
	
		
	% At first strip the possibly odd first and last
	% column.
	
	% Sort st1 and st2 by left-right-order
	if (londiff < 0)
		stx = st1; st1 = st2; st2 = stx;
		stxpos = st1pos; st1pos = st2pos; st2pos = stxpos;
		clear stx stxpos;
		londiff = -londiff;
		latdiff = -latdiff;
	endif
	
	% st1 is on the left, st2 is on the right
	
	lonstep_1st = 1 - mod (st1pos(1,1)+.5, 1);
	lonstart = st1pos(1,1)+.5 + lonstep_1st;
	ratio_1st = lonstep_1st / londiff;
	latstep_1st = ratio_1st * latdiff;
	latstart = st1pos(1,2) + latstep_1st;
	
	lonstep_last = mod (st2pos(1,1)+0.5, 1);
	lonexceeded = st2pos(1,1)+0.5 - lonstep_last;
	ratio_last = lonstep_last / londiff;
	latstep_last = ratio_last * latdiff;
	latexceeded = st2pos(1,2) - latstep_last;
	
	% TODO: Test accurate box selection for stXpos = *.5!
	
	down = st1pos(1,2) < st2pos(1,2);
	
	
	% Now we do similar things for three times:
	% We decide whether the line goes throung one or two
	%	boxes.
	% This decision is taken on
	%  - the first box
	%  - the last box
	%  - all the boxes in between
	
	
	% First box/es
	latstart_shifted = st1pos(1,2)+.5;
	lattmp_shifted = latstart_shifted + latstep_1st; % Will be reused.
	
	if (floor(latstart_shifted) == floor(lattmp_shifted))
		% Same box
		urban = is_urban(LandR, lonstart-1, floor(latstart_shifted));
		if (urban == 2); reflection = [1,0]; urban = 1; endif
		Usum = ratio_1st*cdiff * urban;
		Ysum = ratio_1st*cdiff * (1-urban);
	else
		% Two boxes
		
		% Get border
		if (down)
			% going down (-> South) (-> higher indices)
			latmed = floor(lattmp_shifted);
			ratio_N =   (latmed - latstart_shifted) ...
			          / (latstep_1st);    % pos/pos = pos
		else
			% going up (-> North) (-> lower indices)
			latmed = floor (latstart_shifted);
				ratio_N =   (lattmp_shifted - latmed) ...
				          / (latstep_1st);    % neg/neg = pos
		endif
		
		idx_S = latmed;
		idx_N = latmed - 1;
		
		urban = is_urban(LandR, lonstart-1, idx_S);
		if (urban == 2); reflection = [1,0]; urban = 1; endif
		Usum = (1-ratio_N) * ratio_1st*cdiff * urban;
		Ysum = (1-ratio_N) * ratio_1st*cdiff * (1-urban);
		
		urban = is_urban(LandR, lonstart-1, idx_N);
		if (urban == 2); reflection = [1,0]; urban = 1; endif
		Usum = Usum + ratio_N * ratio_1st*cdiff * urban;
		Ysum = Ysum + ratio_N * ratio_1st*cdiff * (1-urban);
		
	endif
	latstart_shifted = lattmp_shifted; % For use by in-between-routine
	
	% Last box/es
	latend_shifted = st2pos(1,2)+0.5;
	lattmp_shifted = latend_shifted - latstep_last; % reuse
	
	if (floor(latend_shifted) == floor(lattmp_shifted))
		% Same box
		urban = is_urban(LandR, lonexceeded, floor(latend_shifted));
		if (urban == 2); reflection = [1,0]; urban = 1; endif
		Usum = Usum + ratio_last*cdiff * urban;
		Ysum = Ysum + ratio_last*cdiff * (1-urban);
	else
		% Two boxes
		
		% Get border
		if (down)
			% going down (-> South) (-> higher indices)
			latmed = floor(latend_shifted);
			ratio_N =   (latmed - lattmp_shifted) ...
			          / (latstep_last);    % pos/pos = pos
		else
			% going up (-> North) (-> lower indices)
			latmed = floor(lattmp_shifted);
			ratio_N =   (latend_shifted - latmed) ...
			          / (latstep_last);    % neg/neg = pos
		endif
		
		idx_S = latmed;
		idx_N = latmed - 1;
		
		urban = is_urban(LandR, lonexceeded, idx_S);
		if (urban == 2); reflection = [1,0]; urban = 1; endif
		Usum = Usum + (1-ratio_N) * ratio_last*cdiff * urban;
		Ysum = Ysum + (1-ratio_N) * ratio_last*cdiff * (1-urban);
		
		urban = is_urban(LandR, lonexceeded, idx_N);
		if (urban == 2); reflection = [1,0]; urban = 1; endif
		Usum = Usum + ratio_N * ratio_last*cdiff * urban;
		Ysum = Ysum + ratio_N * ratio_last*cdiff * (1-urban);
		
	endif
	
	
	% "normal" operation between first and last box
	onelatdiff = latdiff * (1/londiff); % onelondiff = 1 per definitionem
	onecdiff = cdiff * (1/londiff);
	
	% Compensation if the stations are all in one box (pair)
	if (lonexceeded < lonstart)
		% We have added too much
		%    Left border   ----   st1   ----   st2   ----    Right border
		% Start:                    xxxxxxxxxxxxxxxxxxxxxxxxxx
		% End:         xxxxxxxxxxxxxxxxxxxxxxxxx
		% Correct:                  xxxxxxxxxxxx
		% Wrong:       xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
		
		% So let's subtract the wrong (überschüssig) part.
		onecdiff = -onecdiff;
		
		% Some further adjustment needed.
		tmp = lonstart;
		lonstart = lonexceeded;
		lonexceeded = tmp;
		latstart_shifted = lattmp_shifted; % -> Right index accessed.
		clear tmp;
	endif
	
	for i = lonstart : lonexceeded - 1
		
		if (floor(latstart_shifted) == floor(latstart_shifted + onelatdiff))
			% Same box
			urban = is_urban(LandR, i, floor(latstart_shifted));
			if (urban == 2); reflection = [1,0]; urban = 1; endif
			Usum = Usum + onecdiff * urban;
			Ysum = Ysum + onecdiff * (1-urban);
		else
			% Two boxes
			
			% Get border 
			if (down)
				% going down (-> South) (-> higher indices)
				idx_S = floor(latstart_shifted + onelatdiff);
				ratio_N = (1-mod(latstart_shifted,1)) / onelatdiff;
			else
				% going up (-> North) (-> lower indices)
				idx_S = floor(latstart_shifted);
				ratio_N = (mod(latstart_shifted+onelatdiff,1)-1) /onelatdiff;
						% neg/neg = pos
			endif
			
			urban = is_urban(LandR, i, idx_S);
			if (urban == 2); reflection = [1,0]; urban = 1; endif
			Usum = Usum + (1-ratio_N) * onecdiff * urban;
			Ysum = Ysum + (1-ratio_N) * onecdiff * (1-urban);
			
			urban = is_urban(LandR, i, idx_S - 1); % idx_S -1 = idx_N
			if (urban == 2); reflection = [1,0]; urban = 1; endif
			Usum = Usum + ratio_N * onecdiff * urban;
			Ysum = Ysum + ratio_N * onecdiff * (1-urban);
			
		endif
		
		latstart_shifted = latstart_shifted + onelatdiff;
	endfor
	
	
	Usum;
	Ysum;
	cdiff;  % Informative
	
	% Sonst ist da irgendwas falsch gelaufen.
	assert (abs(Usum + Ysum - cdiff) < 0.001);
	
	
	if (reflection(1,1))
		global REFLECTION_PATH_ADD;
		reflection(1,1) = Usum * (1 + REFLECTION_PATH_ADD/cdiff);
		reflection(1,2) = Ysum * (1 + REFLECTION_PATH_ADD/cdiff);
	endif
	
	
	
	
endfunction





