

% © 2018 Jonas Bechtel
% Licensed under the Apache License, Version 2.0; see file <prj>/meta/LICENSE



format short eng
1; % For this is not a single-function-file

% Explanation for station and other abbreviations in /Hardware/STATIONSNAMEN.txt
% Front<- W B C D O <-End
% Policies: K P (defined below in DYNAMICS)
% Surroundings: U Y (used in XXX TODO)



global RASTERSIZE = 50;  % (we calculate everything in km and the raster has a
                  %  size of 50 m.)
global OUTFILE = 'output.Gen._ride';

global REFLECTION_PATH_ADD = 400 / RASTERSIZE;


function a = typicalGenride()
	% WARNING: Code duplication! See below.
	
	load("Assumed_Land.csv"); % Load landscape and route combined data
	% -> LandR is created.
	% Tour begin must be within a city!
	
	[start_lon, start_lat] = find_start(LandR);
	rasterpairs = find_route(LandR, start_lon, start_lat);
	rasterpairs = annotate_tourmeter(rasterpairs);
	ctrYarr = YUlist(LandR, rasterpairs);
	
	tourmeter = rasterpairs(size(rasterpairs)(1),3);
	%ctrYarr = [2340, 6540; 12300, 23400];
	a = gen_dynamics_K(floor(tourmeter/3.5), ctrYarr);
	plotGenrideDynamics(ctrYarr, a);
endfunction

% Typical calls:

% [Usum, Ysum] = classify_S2S_path(LandR, rasterpairs, a, "W", "D",  10)
% [Usum, Ysum] = classify_S2S_path(LandR, rasterpairs, a, "W", "B", 850)





source Genride-Landscape.m
source Genride-Dynamics.m
source Genride-Electrical.m



function [] = Genride_generate()
	% WARNING: Code duplication! See above.
	
	landscapefile = 'Assumed_Land.csv';
	
	load(landscapefile); % Load landscape and route combined data
	
	[start_lon, start_lat] = find_start(LandR);
	rasterpairs = find_route(LandR, start_lon, start_lat);
	rasterpairs = annotate_tourmeter(rasterpairs);
	ctrYarr = YUlist(LandR, rasterpairs);
	
	tourmeter = rasterpairs(size(rasterpairs)(1),3);
	stpos = gen_dynamics_K(floor(tourmeter/3.5), ctrYarr);
	plotGenrideDynamics(ctrYarr, stpos);
	
	stoptime = size(stpos)(1);
	
	global OUTFILE;
	%assert (!exist(OUTFILE, "file"));
	
	fid = fopen(OUTFILE, "w");
	
	
	fwrite(fid, ["You see file '", OUTFILE, ...
			"' which easyly gets overriden.\n"]);
	fwrite(fid, ["So do save it if you need the results.\n\n\n"]);
	
	fwrite(fid, ["Landscape and tour data read from '", landscapefile, "'.\n"]);
	
	fwrite(fid, ["\n\nDIREKT JETZT BEGINNEN DIE DATEN: ",...
			"<Diese Zeile unverändert lassen>\n"]);
	mscount = 0;
	timeline = ">tt<";
	pathline = "XYd";
	stationlist = "WBCDO"
	stationmax = size(stationlist)(2);
	%t = 1;
	%while t <= stoptime
	for t = [1:stoptime]
		L_reflection = zeros(stationmax, stationmax);
		t_reflection = zeros(stationmax, stationmax);
		
		timeline(2) = 128 + mod(mscount, 2^7);
		timeline(3) = floor(mscount/2^7);
		fwrite(fid, timeline);
		
		
		for s = 1:stationmax
			st1 = stationlist(s);
			pathline(1) = st1;
			for u = s:stationmax
				st2 = stationlist(u);
				
				[L, tp, L_refl, tp_refl] = determine_S2S_path_loss(LandR, rasterpairs, stpos, st1, st2, t);
				pathline(2) = st2;
				pathline(3) = round(tp/100);
				c1 = fwrite(fid, pathline);
				c2 = fwrite(fid, 10^(L/10), "float32");
				%pathline(4:7) = mat2str() % Es gibt keine nutzbare Funktion,
				% um float32-Rohdaten in einem String zu erzeugen.
				c3 = fwrite(fid, ".");	
				assert(c1 == 3)
				assert(c2 == 1)
				assert(c3 == 1)
				
				if (tp_refl)
					t_reflection(s, u) = tp_refl;
					L_reflection(s, u) = L_refl;
				endif
			endfor
		endfor
		
		
		
		fwrite(fid, "\n");
		mscount = mscount + 1000;
		mscount = mod(mscount, 2^14);
		
		
		if (sum(sum(t_reflection)))
			t = t - 1;
			fwrite(fid, timeline);
			
			for s = 1:stationmax
				st1 = stationlist(s);
				pathline(1) = st1;
				for u = s:stationmax
					if (!t_reflection(s, u))
						continue; endif;
					st2 = stationlist(u);
					pathline(2) = st2;
					pathline(3) = round(t_reflection(s, u)/100);
					
					c1 = fwrite(fid, pathline);
					c2 = fwrite(fid, 10^(L_reflection(s, u)/10), "float32");
					c3 = fwrite(fid, ".");	
					assert(c1 == 3)
					assert(c2 == 1)
					assert(c3 == 1)
				endfor
			endfor
			
			fwrite(fid, "\n");
		endif
		
	endfor
	%	t = t + 1;
	%endwhile
	
	fclose(fid);
	
endfunction



