

// © 2018 Jonas Bechtel
// Licensed under the Apache License, Version 2.0; see file <prj>/meta/LICENSE


#ifndef READ_ULOG_C
#define READ_ULOG_C

// We want to extract this:
// 
// W:UL@ns2,081,671,091,448>    R   [   ]  5    5  >   AMR   >   _1_   1  #321  // W:UL@ns2,081,671,091,448>a50|321|8fb29eaab819824de8f05f48>




#ifndef NOCODE

#include <stdio.h>

#define NOCODE
#include "r-map.c"
#undef NOCODE

#else
#include "r-map.c"
#endif // -ifndef NOCODE


#define S_START 1
#define S_NAME 3
#define S_N 5
#define S_S 6
#define S_TIME 8
#define S_tick 10
#define S_SIG 12
#define S_SENDER 14
#define S_SEQUENCE 16
#define S_AMR 18

#include <stdlib.h>
bool read_next(struct rmap_frame * in_here, uint8_t st_name)
#ifndef NOCODE
{
	
	while (true) {
		uint8_t ch;
		bool next = false;
		while (true) {
			ch = getchar();
			//printf("ch=%2x (EOF=%2x)\n", ch, EOF);
			
			if (next && ch == st_name) {
				ch = getchar();
				//printf(" ch=%2x\n", ch);
				if (ch != ':') {
					if (st_name == 'U' && ch == 'L')
						break;
					next = false; // newline
					continue;
				}
				break;
			}
			if (next && ch == 'U' && getchar() == 'L') {
				break;
			}
			
			if (ch == '\n')
				next = true;
			
			if (ch == 0xFF) // EOF on my system
				return false;
		}
		
		//printf("ch=%2x\n", ch);
		
		if (getchar() != '@') continue;
		if (getchar() != 'n') continue;
		if (getchar() != 's') continue;
		
#define BUFSIZE 40
		char buf[BUFSIZE];
		uint8_t ptr = 0;
		
		while (true) {
			ch = getchar();
			
			if (ch == ',' || ch == '.')
				continue;
			
			if ('0' <= ch && ch <= '9') {
				buf[ptr++] = ch;
				continue;
			}
			
			if (ch == 0xFF) // EOF on my system
				return false;
			
			if (ptr + 1 >= BUFSIZE) {
				ptr = 0;
				break;
			}
			
			break;
		}
		
		if (ptr == 0) {
			fprintf(stderr, "Bad number format.\n");
			return false;
		}
		
		buf[ptr] = ' ';
		
		long long int time = strtoll(buf, NULL, 10);
		
		//printf("time = %llu\n", time);
		in_here->toa = time;
		
		if (ch != '>') continue;
		if (getchar() != 'a') continue;
		
		if (getchar() != '|')
			if (getchar() != '|')
				if (getchar() != '|')
					continue;
		
		buf[0] = getchar(); // 3,2,1,0
		buf[1] = getchar(); // f,d,e,d,c,b,a,9,8,7,6,5,4,3,2,1,0
		buf[2] = getchar(); // f,d,e,d,c,b,a,9,8,7,6,5,4,3,2,1,0
		buf[3] = ' ';
		
		long int seqno = strtol(buf, NULL, 16);
		
		if ((buf[0] | buf[1] | buf[2]) & 0x80) {
			fprintf(stderr, "Bad sequence number format.\n");
			return false;
		}
		
		//printf("seqno = %03x\n", (uint16_t)seqno);
		in_here->sequence_number = seqno; // This is a 10 bit sequence number!
		
		if (getchar() != '|')
			continue;
		
		ptr = 0xFF;
		while (++ptr < sizeof(in_here->data.data)) {
			buf[0] = getchar();
			buf[1] = getchar();
			buf[2] = ' ';
			if ((buf[0] | buf[1] | buf[2]) & 0x80) {
				fprintf(stderr, "Bad amr data format.\n");
				return false;
			}
			
			long int byte = strtol(buf, NULL, 16);
			
			//printf("%02x", (uint8_t)byte);
			in_here->data.data[ptr] = (uint8_t)byte;
		}
		
		if (getchar() != '>')
			continue; // Das muss jetzt sein.
		
		break;
	}
	
	/*
	printf("Data: ");
	for (int i = 0; i < 12; i++) {
		printf("%02x", in_here->data.data[i]);
	}
	printf("\n");*/
	
	
	return true;
}
#endif // ifndef NOCODE
;



#endif // -ifndef READ_ULOG_C

