

// © 2018 Jonas Bechtel
// Licensed under the Apache License, Version 2.0; see file <prj>/meta/LICENSE



#include <stdio.h>
#include <string.h>

#define DEFINE_OBJECT // DEFINE_OBJECT is to be included only once in 
#include "subintern.h"  //    all application's code.
#undef DEFINE_OBJECT

#include "assert.h"

#include "sim-event.h" // CLAIM_INVALID

#include "RFM69trancv/subobject.h"
// #include "AODVroute/subobject.h"


#define NOCODE
#include "multihw.c"
#include "FrameSwitch.c"
#undef NOCODE


// This is Frameswitch app

/*
void FS_main(MHI_ARG_single);
void FS_destroy(struct app_instance* ai);
*/
void FSsetup(MHI_ARG_single);


void system_startup_FS(struct app_instance* ai, uint64_t time) {
	struct station * mh_station_arg = ai->station;
	// The simulators' event loop is running now.
	
	printf("FRAMESWITCHapp-%c: got startup.\n", ai->station->name);
	
	// expecting that routing will be started by uC code:
	FSsetup(MHI_USE_single);
	
	mh_start_loop(MHI_USE_single);
}

struct trigger_corroboration notify_FS(
			struct app_instance * ai, event e) {
	struct station * mh_station_arg = ai->station; // Needed for ulog.
	
	struct trigger_corroboration answer;
	answer.do_delay = 0;
	bool taken = 0;
	
	uint64_t simtime = get_simtime(mh_station_arg); //e.time;
	
	uint64_t errtime = 0;
	char* errstr = NULL;
	if (ai->cl_delay < simtime) {
		// TODO: in case of received time this may be to strict?
		errtime = ai->cl_delay;
		errstr = "callback to continue after delay()";
	}
	
	if (errstr) {
		printf ("app.c: Error: a claim of us was not taken! "
			"(Waiting for '%s' @%'llu)\n", errstr, errtime);
	}
	
	
#ifdef SIMEVENT
	ulog_str_L("aEEEE\n");
#endif
	
	// TODO: How to coordinate the answer if multiple claims are hit at one time?
	
	if (ai->cl_delay == simtime) {
		ai->cl_delay = CLAIM_INVALID;
		
		// This will continue the old calculation which had requested to
		// be delayed.
		mh_resume_loop(MHI_USE_single);
		
		taken = true;
	}
	
	
	if (!taken) {
		printf("app.c: WARNING: claim triggered @%'llu but we didn't expect"
			" anything then.\n", simtime);
	}
	
	
	answer.type = nop;
	return answer;
}

int start_FRAMESWITCHapp(struct app_instance * ai) {
	
	struct station * mh_station_arg = ai->station;
	log_str_L("\n\n\n\n");
	ulog_str_L("Early start of station '");
	log_str_N((char*)&ai->station->name, 1);
	log_str_L("'.\n");
	
	
	
	ai->my_claim = CLAIM_INVALID;
	
	
	ai->system_startup = system_startup_FS;
	MHI_USE_access loop = FS_main; // = only.c:loop()
	ai->notify = notify_FS;
	
	ai->cl_delay = CLAIM_INVALID;
	
	
	ai->station->trancv = new_RFM69trancv(ai->station);
	if (!ai->station->trancv) {
		printf("ONLYSENDapp: Error: Couldn't instantiate RFM69trancv!\n");
		return OBJECT_START_FAIL;
	}
	
	int res = start_RFM69trancv(ai->station->trancv);
	if (res != OBJECT_START_OK)
		return res;
	
	
	ai->old_destroy = ai->destroy;
	ai->destroy = FS_destroy;
	
	
	return OBJECT_START_OK;
}


