

// © 2018 Jonas Bechtel
// Licensed under the Apache License, Version 2.0; see file <prj>/meta/LICENSE



#ifndef APP_USER_H
#define APP_USER_H

#include "object.h"

#ifdef OBJ_NAME
#undef OBJ_NAME
#endif

#define OBJ_NAME app


struct obj_instance(OBJ_NAME);

#include "sim-common.h"
#include "sim-event.h"
#include "station-app.h"


struct trigger_corroboration {
	// KEEP THIS STRUCT SMALL (will be direct function argument, no pointer)
	
	// How enum works http://en.cppreference.com/w/c/language/enum
	// This is an unnamed variant:
	enum {to_air, nop, Quit, Error} type;
	
	// Is zero if the reply is directly made at the trigger time.
	// Is non-zero if the reply is meant to happen later than trigger time.
	// Is ignored for Quit and Error events.
	uint32_t do_delay;
	
	union {
		struct wave_sequence* to_air; // this will be allocated by app and freed by simulator.
		const char* reason; // For quit/error
	};
};


// About trigger_claim():
// Reason:
//    0x01 = packet sent,
//    0x03 = segment sent,

//    0x00 = other reason.
#define TRIGGER_REASON_SENT 0x01
#define TRIGGER_REASON_SENT_SEGMENT 0x03
#define TRIGGER_REASON_OTHER 0x00




#define APP_INSTANCE_HEADER \
	struct station* station; \
	const char* thename; \
	void (*parle_lEtat)(struct app_instance*); \
	\
	/* Invoked when the simulator starts and the event queue is running. */ \
	void (*system_startup)(struct app_instance*, uint64_t simtime); \
	\
	/* Invoked when the simulator has reached the time the stations' claim points to. */ \
	/* Detailed information: see above */ \
	/* Remark: this method will NOT use the time member of event struct but */ \
	/*         look at the actual simtime in the simulator struct. */ \
	struct trigger_corroboration (*notify)(struct app_instance*, event); \
	\
	void (*destroy)(struct app_instance*); \
	\
	uint64_t my_claim; \
		/* When the thread pauses because it had requested a delay. */\
	uint64_t cl_delay; \
	\
	
	// destroy will be preset.
	// When you replace this, please take care that the old function (free) will be
	// called as the last action of the new function.


#ifndef SKIP_AI
	struct app_instance {
		APP_INSTANCE_HEADER
	};
#endif


#endif // #ifndef APP_USER_H










