#!/bin/sh


# © 2018 Jonas Bechtel
# Licensed under the Apache License, Version 2.0; see file <prj>/meta/LICENSE



rmap=../../Sim/Reception-Graph/r-map
mkdir -p autographic

plot_graphics() {
	infile=$1
	inbase=$2
	
	if [ "`stat -c %Y "$inbase--arrival.eps"`" -gt "`stat -c %Y $infile`" ];then
		return
	fi
	
	octave --no-gui <<END
infile="$infile"
inbase="$inbase"
source("quick_script.m")
END
}

process() {
	src=$1
	src_O=${src}_O
	src_S=${src}_$2
	out=$3
	number=$4
	
	if [ "`stat -c %Y $out`" -gt "`stat -c %Y $src_S`" ]; then
		# Does already exist -> [ returns 0
		return
	fi
	
	# [ returns 1 if $out is outdated or 2 if one of $out $src don't exist
	
	cat $src_S | $rmap > $out
	if [ \! $? = 0 ]; then
		rm $out
		continue
	fi
	printf '# name: station_number\n# type: scalar\n%d\n\n\n' $number >> $out
	
	cnt=`grep '>amrload: this is frame ' $src_O | tail -n 1 | awk '{print $5}'`
	scnt=`grep -E '> S      \[   \]  5   _5_ >   AMR   >' $src_O | wc -l`
	printf '# name: loaded_frame_count\n# type: scalar\n%d\n\n\n' $cnt >> $out
	printf '# name: sent_frame_count\n# type: scalar\n%d\n\n\n' $scnt >> $out
}


rrr=autographic/radio_reception_register
parameterset=
parametercount=`ls manual_simres | wc -l`
parameterline=


mv $rrr $rrr-old
printf '# name: rrr_setcount\n# type: scalar\n%d\n\n\n' $parametercount > $rrr
printf '# name: rrr_paramsets\n# type: string\n# elements: %d\n' $parametercount >> $rrr

for dir in `ls manual_simres`; do
	printf "Parameter set: %s\n" $dir;
	
	if [ ${dir%%_*} = bcloop ]; then
		# Extra: station D
		src=manual_simres/$dir/ulog
		out=autographic/amr_received_${dir}_D
		process $src D $out 4
		plot_graphics $out autographic/$dir--D
	fi
	
	src=manual_simres/$dir/ulog
	out=autographic/amr_received_${dir}_W
	process $src W $out 1
	plot_graphics $out autographic/$dir--W
	
	printf '# length: ' >> $rrr
	printf '%s' $dir | wc -c >> $rrr
	printf '%s\n' $dir >> $rrr
	
	#parametercount=$(($parametercount+1))
	#parameterset="$parameterset $dir"
	for S in W B C D O; do
		parameterline="$parameterline:`
				grep "Final stat of trancv:  collected" ${src}_$S \
				| awk '{print $8}'`"
	done
	parameterline="$parameterline "
done


printf '\n\n# name: rrr_data\n# type: matrix\n# rows: %d\n# columns: 5\n' $parametercount >> $rrr
tr ': ' ' \n' <<<"$parameterline" >>$rrr

diff $rrr $rrr-old >/dev/null 2>&1 && mv $rrr-old $rrr

if [ "`stat -c %Y autographic/rrr-overview.eps`" -gt "`stat -c %Y $rrr`" ];
then
	# Does already exist -> [ returns 0
	true
else
	# [ returns 1 if $out is outdated or 2 if one of $out $src don't exist
	octave --no-gui <<<'source("rrr.m")'
fi

true
