/*****************************************************************************
 *
 * Copyright (C) 2018 Jonas Bechtel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Authors: Jonas Bechtel, <post@jjjjbechtel.de> (nospam: use jbechtel.de)
 *          
 *
 *****************************************************************************/


#ifndef CHECKSUM_C
#define CHECKSUM_C

#include <stdint.h>
#include <stdio.h>

#define u8 uint8_t


u8 chk3(u8 * data, size_t len)
#ifndef NOCODE
{
	u8 chk = 1;
	while (len--) {
		chk +=  data[len] & 0x07;
		chk += (data[len] & 0x31) >> 3;
		chk += (data[len]       ) >> 6;
	}
	return chk & 0x07;
}
#endif // -ifndef NOCODE
;

u8 chkH(u8 d1, u8 d2, u8 d3, u8 d4)
#ifndef NOCODE
{
	u8 data[4] = {d1, d2, d3, d4};
	return chk3(data, 4);
}
#endif // -ifndef NOCODE
;


u8 chk5(u8* data, size_t len)
#ifndef NOCODE
{
	u8 chk = 2;
	while (len--) {
		chk +=  data[len] & 0x1F;
		chk += (data[len] & 0xF8) >> 3;
	}
	return chk & 0x1F;
}
#endif // -ifndef NOCODE
;

u8 chkA(u8* mr475)
#ifndef NOCODE
{
	// class A bits im AMR-NB mode MR475 make 5 B (42 bit) (ignoring two bit)
	return chk5(mr475, 5);
}
#endif // -ifndef NOCODE
;




u8 chk8(u8* data, size_t len)
#ifndef NOCODE
{
	u8 chk = 0xA3;
	while (len--) {
		chk ^=  data[len];
	}
	return chk;
}
#endif // -ifndef NOCODE
;



u8 chkB(u8* mr475)
#ifndef NOCODE
{
	// class B bits im AMR-NB mode MR475 make 7 B (53 bit)
	//  (additionally respecting two bit of class A bits)
	return chk8(&(mr475[5]), 7);
}
#endif // -ifndef NOCODE
;





#endif // -ifndef CHECKSUM_C
