/*****************************************************************************
 *
 * Copyright (C) 2018 Jonas Bechtel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Authors: Jonas Bechtel, <post@jjjjbechtel.de> (nospam: use jbechtel.de)
 *          
 *
 *****************************************************************************/


#ifndef CLI_DEV_C
#define CLI_DEV_C

#include <netinet/in.h>
#include <stdbool.h>

#include "defs.h"


#ifndef NOCODE
#include "cassert.h"
#include "assert.h"

#include <byteswap.h>
#include "debug.h"
#include "aodv_hello.h"
#include "aodv_rrep.h"
#include "aodv_rreq.h"
#include "aodv_socket.h"
#include "aodv_timeout.h"
//#include "all_packet.h"

#ifdef REMOTE_CTRL
#ifdef SIMUL
#include "common/route-user.h"
#define NOCODE
#include "FrameSwitch.c"
#undef NOCODE
//#include "multihw.c"
//#include "common/app-user.h"
#elif defined (BEAGLE)
// Expecting these two implemented by controlling application
#define MHI_ARG
void inbound_unpacked(MHI_ARG    uint64_t time, void* data, uint8_t len, uint16_t seq, uint8_t sender, uint8_t destination, float power_in_mW);
void outbound_raw    (MHI_ARG    uint8_t* data, uint8_t len, uint32_t tracking_mark);
#endif // -ifdef SIMUL -elif defined (BEAGLE)
#endif // -ifdef REMOTE_CTRL


#define NOCODE
#undef CLI_DEV_C
#include "cli-dev.c" // include ourselves.
#include "packetizer.c"
#include "routing_table_short.c"
#include "debug.c"
#include "aodv_rreq.c"
#include "ns-2/packet_queue.cc"
#undef NOCODE

#else // -ifndef NOCODE

#include "packetizer.c"

#endif // -ifndef NOCODE -else


#ifdef BEAGLE
#define u8 uint8_t
#endif



// no newline
void print_pkt(uint8_t* pkt, int len)
#ifndef NOCODE
{
	int i = 0;
	
	printf("P");
	while (i < len) {
		int little = 4;
		if (len-i < 4) little = len-i;
		int j;
		for (j = 0; j < little; j++) {
			printf("%02x", pkt[i + j]);
		}
		i += 4;
		if (i < len)
			printf("|");
	}
	printf (" .\n");
}
#endif
;


void cli_pack_and_send(hi hi, struct unpacked_L2_header* L2,
						const struct fieldbits** f)
#ifndef NOCODE
{
#ifdef SIMUL
	struct station * mh_station_arg = hi->multihw_instance->station;
#endif
	static uint8_t buf[100];
	uint8_t len = bitpack(buf, L2, f, sizeof(buf));
	uint32_t tracking_mark = 0xffffffff;
	
#if (defined REMOTE_CTRL)
	if (L2->top_type == TOP_TYPE_SEQ) {
		// Whith these AMR packets a unique number tracking can be provided.
		struct unpacked_seq_w_type * uswt = (struct unpacked_seq_w_type*) L2;
		if (uswt->seq_type == SEQ_TYPE_AMR) {
			struct unpacked_amrlong * ual = (struct unpacked_amrlong*) uswt;
			
			// TODO: Adding tracking mark here although with recent state 
			// we will never get some acknowledgement for the packets.
			
			// -> TODO find a mechanism to acknowledge tracking marks to FrameSwitch app and to read out foreign acknowledgements!
			
			tracking_mark = (0x20 << 24)
						  | (ual->ori_addr << 20) | (ual->dst_addr << 16)
						  | (ual->data_seq_msb << 8)
						  |  ual->data_seq_lsb;
			
			struct rt_table * rt = rt_table_find_short(hi, ual->dst_addr);
			if (rt) {
				/*struct timeval now;
				gettimeofday(&now, NULL);
				struct timeval old = rt->last_update;
				timeval_add_msec(old, DISCOVERY_INTERVAL);
				if (timeval_diff(&old, &now) < 0)
					rout discovery
				*/
			}
			// Added later
			else if(ual->dst_addr == PACKED_BROADCAST) {
				// No need for route
			}
			else {
				ulog_str_L("This is very improbable!\n");
			}
		}
	}
#endif
	
	
#if (defined REMOTE_CTRL) && (defined SIMUL)
#ifdef DEBUG_FOR_VALGRIND
	u8 i = 0;
	while (i < len)
		fprintf(stderr, "cd: %x \n", buf[i++]);
#endif
	hi->multihw_instance->app->rti->outbound_raw(
		hi->multihw_instance->app->rti, buf, len, tracking_mark);
#elif (defined REMOTE_CTRL) && (defined BEAGLE)
	outbound_raw(buf, len, tracking_mark);
#else // This is not REMOTE_CTRL
	print_pkt(buf + 4, len);
#endif // -ifdef REMOTE_CTRL -else
}
#endif // -ifndef NOCODE
;



#ifndef NOCODE
/**
 * hex2int - taken and modified from https://stackoverflow.com/a/39052987  via
 *   https://stackoverflow.com/questions/10156409/convert-hex-string-char-to-int
 * take a hex string and convert it to a 8 bit number (2 hex digits)
 */
uint8_t hex2int(uint8_t *hex) {
	uint8_t val = 0;
	bool one = false;
	do {
		// get current character then increment
		uint8_t byte = *hex++; 
		//printf("byte=0x%x.\n", byte);
		// transform hex character to the 4bit equivalent number, using the ascii table indexes
		if (byte >= '0' && byte <= '9') byte = byte - '0';
		else if (byte >= 'a' && byte <='f') byte = byte - 'a' + 10;
		else if (byte >= 'A' && byte <='F') byte = byte - 'A' + 10;    
		// shift 4 to make space for new digit, and add the 4 bits of the new digit 
		val = (val << 4) | (byte & 0xF);
		one = one ? false : true;
	} while (one);
	return val;
}
#endif


void cli_receive(hi hi, int fd) 
#ifndef NOCODE
{
	int res = 1;
	uint8_t ch = 0;
	
	uint8_t pos_old = hi->recv_buf_pos;
	while (res >= 0 && ch != 0x0a && hi->recv_buf_pos < SZ_RCV) {
		res = read(fd, &ch, 1);
		if (res < 0) {
			printf ("Error reading stdin.\n");
			break;
		}
		
		if (ch == 0) {
			debug(hi, LOG_DEBUG, 0, "Console: got NULL char, quitting.");
			hi->run = 0;
			return;
		}
		
		hi->recv_buf[hi->recv_buf_pos++] = ch;
		
		//printf("_%d:%02x", res, ch);
	}
	//printf("\n");
	
	if (hi->recv_buf[hi->recv_buf_pos-1] == 0x0a) {
		hi->recv_buf_pos--;
	}
	
	uint8_t len = hi->recv_buf_pos - pos_old;
	//debug(hi, LOG_DEBUG, 0, "   Received %d chars.", len);
	
	if (len == 0)
		return;
	
	bool quit = false;
	if (hi->recv_buf_pos >= 4) {
		if (strncmp((char*)&hi->recv_buf, "quit", 4) == 0) {
			quit = true;
		}
	}
	if (hi->recv_buf[0] == 'Q')
		quit = true;
	
	if (quit) {
		debug(hi, LOG_DEBUG, 0, "Console: got \"quit\" or 'Q', quitting.");
		hi->run = 0;
		return;
	}
	
	
	// Done. Most probably there's nothing left. (As we read char-by-char. If
	//  it is supposed to be efficient, a whole block of up to 100 bytes should
	//  be read.)
	hi->recv_buf_pos = 0;
	pos_old = 1;
	uint8_t direction = hi->recv_buf[0];
	
	while (pos_old < len) {
		if (hi->recv_buf[pos_old] == '|') {
			pos_old++;
			continue;
		}
		
		hi->recv_buf[hi->recv_buf_pos++] = hex2int(&hi->recv_buf[pos_old]);
		
		pos_old += 2;
		//printf ("read: 0x%02x.\n", hi->recv_buf[hi->recv_buf_pos-1]);
	}
	
	
	
	if(direction == 'A') {
		// This is a manually put AMR packet to send.
		// if there's a route.
		// otherwise it has to wait till route is discovered.
		
		put_amr_packet(hi, hi->recv_buf+1, hi->recv_buf_pos-1,
					   hi->recv_buf[0] & 0x0F);
		//cli_send(msg, hi->recv_buf_pos);
		hi->recv_buf_pos = 0;
		//rt_table_print(hi);
		
		return;
	}
	
	if (direction != 'P') {
		debug(hi, LOG_WARNING, 0, "Don't understand this packet.\n");
		return;
	}
	
	// Unpack AODV message
	
	process_received_packet(hi, hi->recv_buf, hi->recv_buf_pos - 1, 0.0);
	
	hi->recv_buf_pos = 0;
	
	//rt_table_print(hi);
	
	return;
	
	
}
#endif
;


// TODO: Limit Sequence numbers!

void put_aodv_packet(hi hi, AODV_msg* data, uint8_t len, uint8_t ttl, uint8_t dst)
#ifndef NOCODE
{
	CASSERT(sizeof(struct unpacked_rreq) >= sizeof (struct unpacked_rrep),
			___Memory_sharing_broken_because_of_odd_struct_sizes);
			// This is the place where you should exchange rrep and rreq.
	
#if (defined DEBUG_FLOW) && (defined SIMUL)
	//struct station * mh_station_arg = hi->multihw_instance->station;
#endif
	
	struct unpacked_rreq upp;
	
	get_addr_byte(upp.L2.station_addr, 3, hi->devs[0].ipaddr.s_addr);
	upp.L2.ttl = ttl;
	upp.L2.top_type = TOP_TYPE_AODV;
	
	const struct fieldbits ** type = NULL; // = NULL for compiler happiness
	
	if (data->type == AODV_RREQ) {
		RREQ* rreq = (RREQ*)data;
		type = hi->bits_rreq;
		// Always broadcast. The nexthop field is reused for other purpose here.
		get_addr_byte(upp.L2.nexthop_addr, 3, rreq->dest_addr);
		upp.AODV_type = AODV_TYPE_RREQ;
		upp.flags = (rreq->j<<4) + (rreq->r<<3) + (rreq->g<<2) + (rreq->d<<1)
					+ (0 << 0); // U flag seems to be absent in aodv-uu.
		upp.hop_count = rreq->hcnt;
		CASSERT(RREQ_ID_OVERFLOW <= 0x40, __must_fit_into_6_bit__); // must fit into 6 bit
		assert(ntohl(rreq->rreq_id) < RREQ_ID_OVERFLOW);
		upp.rreq_id = ntohl(rreq->rreq_id);
		
		upp.dst_seq_msb = (rreq->dest_seqno&0x0F00) >> 8;
		upp.dst_seq_lsb = rreq->dest_seqno&0xFF;
		get_addr_byte(upp.ori_addr, 3, rreq->orig_addr);
		upp.ori_seq_msb = (rreq->orig_seqno&0x0F00) >> 8;
		upp.ori_seq_lsb = rreq->orig_seqno&0xFF;
#ifdef DEBUG_FLOW
		pretty_rreQ(hi, " prep  ", &upp, OUTBOUND);
		//printf("      1    _2_ ->  RREQ  ->        5     ");
#endif
		//printf("rreq->dest_seqno = %d.\n", ntohl(rreq->dest_seqno));
	}
	else if (data->type == AODV_RREP) {
		RREP* rrep = (RREP*) data;
		struct unpacked_rrep * upr = (struct unpacked_rrep*) &upp;
		upr->L2.nexthop_addr = dst;
		type = hi->bits_rrep;
		upr->AODV_type = AODV_TYPE_RREP;
		upr->flags = (rrep->r<<1) + (rrep->a<<0);
		if (rrep->prefix)
			debug(hi, LOG_WARNING, 0, "Deleting prefix.");
		upr->hop_count = rrep->hcnt;
		get_addr_byte(upr->dst_addr, 3, rrep->dest_addr);
		upr->dst_seq_msb = (ntohl(rrep->dest_seqno)&0x0F00) >> 8;
		upr->dst_seq_lsb =  ntohl(rrep->dest_seqno)&  0xFF;
		//printf("rrep->dest_seqno = %d.\n", ntohl(rrep->dest_seqno));
		get_addr_byte(upr->ori_addr, 3, rrep->orig_addr);
		
		// TODO: When redirecting, provide a mechanism that garantuees a
		// decrease of lifetime on resending.
		uint32_t lifetime = ntohl(rrep->lifetime);
		//printf("%d\n", lifetime);
		lifetime = (lifetime + 19) / 20; // Aufrunden
		uint16_t faktor = 1;
		for (upr->lifetime_exp = 0; upr->lifetime_exp <= 3; upr->lifetime_exp++) {
			// TODO: There are situations where a finer granularity could
			// be achieved.
			if (lifetime <= 32*faktor) {
				upr->lifetime_val = lifetime/faktor;
				break;
			}
			faktor *= 10;
		}
		//printf("val: %d\n", upr->lifetime_val);
		if (upr->lifetime_exp > 3) {
			debug(hi, LOG_WARNING, 0,
						"Lifetime too big. [This doesn't happen normally]");
			upr->lifetime_exp = 3;
			upr->lifetime_val = 31;
		}
		if (upr->lifetime_val > 31)
			upr->lifetime_val = 31;
		
#ifdef DEBUG_FLOW
		pretty_rRep(hi, " prep  ", upr, OUTBOUND);
		//printf("      1     2  <-  RREP  <-  _4_   5     ");
#endif
	}
	else if (data->type == AODV_RERR) {
		RERR* rerr = (RERR*) data;
		struct unpacked_rerr * upr = (struct unpacked_rerr*)&upp;
		upr->L2.nexthop_addr = dst;
		upr->AODV_type = AODV_TYPE_RERR;
		upr->flags = rerr->n;
		
		RERR_udest * r1 = RERR_UDEST_FIRST(rerr);
		RERR_udest * r2;
		RERR_udest ru = *r1;
		u8 cnt = rerr->dest_count;
		while (--cnt) {
			r2 = RERR_UDEST_NEXT(r1);
			*r1 = *r2;
			printf("1st dst: %d\n", RERR_UDEST_FIRST(rerr)->dest_addr);
			printf("nth dst: %d\n", r1->dest_addr);
		}
		
		get_addr_byte(upr->dst_addr, 3, ru.dest_addr);
		
		if (ntohl(ru.dest_seqno) >= 0x1000)
			debug(hi, LOG_WARNING, 0, "Cann't handle big sequence numbers! CUTTING!");
		upr->dst_seq_msb = (ntohl(ru.dest_seqno) >> 8) & 0xF;
		upr->dst_seq_lsb =  ntohl(ru.dest_seqno)      & 0xFF;
		
		type = hi->bits_rerr;
		
		
		if (rerr->dest_count-- > 1) {
			// Resolve that recursively
			
			upp.L2.header_chksum = 0x7; // TODO (also below)
			cli_pack_and_send(hi, &upp.L2, type);
			
			// Recursion:
			put_aodv_packet(hi, data, len, ttl, dst);
		}
		
#ifdef DEBUG_FLOW
		pretty_rErr(hi, " prep  ", upr, OUTBOUND);
#endif
	}
	else if (data->type == AODV_RREP_ACK) {
		upp.L2.nexthop_addr = dst;
		type = hi->bits_rrep_ack;
		upp.AODV_type = AODV_TYPE_RREP_ACK;
#ifdef DEBUG_FLOW
		pretty_rRep_A(hi, " prep  ", (struct unpacked_rrep_ack*)&upp, OUTBOUND);
#endif
	}
	
	
	upp.L2.header_chksum = 0x7; // TODO (also at RERR repeat)
	cli_pack_and_send(hi, &upp.L2, type);
	return;
}
#endif // -ifndef NOCODE
;


void amr_prettyprint(hi hi, uint8_t* amrbytes,
			uint16_t seq, uint8_t ori_addr)
#ifndef NOCODE
{
#ifdef SIMUL
	struct station * mh_station_arg = hi->multihw_instance->station; // ulog
#endif
	
	// TODO: Check checksum!
	
	ulog_str_num_BASE_str_L("a", ori_addr << 4, HEXBYTE, "|");
	log_num_BASE(seq >> 8, HEX);
	log_num_BASE(seq & 0xFF, HEXBYTE);
	//printf("a%02x|%03x", ori_addr << 4, seq);
	
	int i;
	for (i = 0; i < 12; i++) {
		if (!i%4) {
			log_str_L("|");
			//printf("|");
		}
		//printf("%02x", amrbytes[i]);
		log_num_BASE(amrbytes[i], HEXBYTE);
	}
	log_str_L("\n");
}
#endif
;




void final_amr_receive(hi hi, struct unpacked_amrlong* ual)
#ifndef NOCODE
{
	
	// TODO: Check checksum!
	
	uint16_t seqno = (ual->data_seq_msb<<8) + ual->data_seq_lsb;
	amr_prettyprint(hi, ual->amrbytes, seqno, ual->ori_addr);
}
#endif
;



void final_receive(hi hi, struct unpacked_L2_header* upl, float pwr_mW)
#ifndef NOCODE
{
	
	if (upl->top_type != TOP_TYPE_SEQ) {
		debug(hi, LOG_WARNING, 0, "Warning: cann't process non-seq packet!");
		return;
	}
	
	struct unpacked_seq_w_type *  uswt = (struct unpacked_seq_w_type*) upl;
	if (uswt->seq_type != SEQ_TYPE_AMR) {
		debug(hi, LOG_WARNING, 0, "Warning: cann't process non-amr packet!");
		return;
	}
	
#ifdef REMOTE_CTRL
	struct unpacked_amrlong * ual = (struct unpacked_amrlong*) upl;
	uint16_t seq = (ual->data_seq_msb<<8) + ual->data_seq_lsb;
#ifdef SIMUL
	struct station * mh_station_arg = hi->multihw_instance->station;
	hi->multihw_instance->app->rti->inbound_unpacked(
			hi->multihw_instance->app->rti,
			get_systime(MHI_USE_single), ual->amrbytes, 12,
			seq, ual->ori_addr, ual->dst_addr, pwr_mW);
#else // This is BEAGLE:
	inbound_unpacked(0, ual->amrbytes, 12,
			seq, ual->ori_addr, ual->dst_addr, pwr_mW);
#endif
#else // -ifdef REMOTE_CTRL
	final_amr_receive(hi, (struct unpacked_amrlong*) uswt);
#endif // -ifdef REMOTE_CTRL -else
	
}
#endif
;


// This method does following:
// 1. Unpack L2 header
// 2. Data packets (TOP_TYPE_SEQ, typ. AMR): 
//    - forward packets for other destination
//    - forward broadcast packets
//    - unpack and provide broadcast packets for us
//    - unpack and provide packets for us locally
// 3. AODV packets (TOP_TYPE_AODV):
//    - unpack and provide packets for us locally
void process_received_packet(hi hi, void* packed, uint8_t len, float pwr_mW)
#ifndef NOCODE
{
#ifdef REMOTE_CTRL
	timer_age_queue(hi);
#endif
	
	// What happens here is similar to pretty_packed().
	// Kind of code duplication.
	
// 1. Unpack L2 header
	struct unpacked_seq_w_type uswt;
	bitunpack(&uswt.L2, packed, hi->bits_seq_w_type);
	
	
// 3. AODV packets (TOP_TYPE_AODV):
//    - unpack and provide packets for us locally
	if (uswt.L2.top_type == TOP_TYPE_AODV) {
		debug(hi, LOG_DEBUG, 0, "This is AODV.");
		
		// rrep ack contains exactly what we want to query for first examination
		struct unpacked_rrep_ack upra; 
		bitunpack(&upra.L2, packed, hi->bits_rrep_ack);
		
		CASSERT(sizeof(RREQ) > sizeof(RREP), ___Memory_sharing_broken_because_of_odd_struct_sizes);
		RREQ * msg = malloc(sizeof(RREQ));
		//uint8_t len = 255; // To make the compiler happy.
		struct in_addr addr_template;
		set_addr_byte(&addr_template, 0, 192);
		set_addr_byte(&addr_template, 1, 168);
		set_addr_byte(&addr_template, 2, 1);
		
		
		msg->type = upra.AODV_type;
		
		
		if (upra.AODV_type == AODV_TYPE_RREQ) {
			RREQ* rreq = (RREQ*) msg;
#ifdef DEBUG_FOR_VALGRIND
			printf("Payload inbound len %d / %d\n", len, get_packed_size(hi->bits_rreq));
#endif
			if (len < get_packed_size(hi->bits_rreq)) {
				// reject too short packets
				free(msg);
				return;
			}
			len = sizeof (RREQ);
			
			struct unpacked_rreq upr;
			bitunpack(&upr.L2, packed, hi->bits_rreq);
			
#ifdef DEBUG_FLOW
			pretty_rreQ(hi, "    R  ", &upr, INBOUND);
#endif
			
			rreq->j = upr.flags >> 3;
			rreq->r = (upr.flags >> 2) & 1;
			rreq->g = (upr.flags >> 1) & 1;
			rreq->d = (upr.flags >> 0) & 1;
			//rreq->u = 0; // aodv-uu has no implementation for U.
			
			rreq->hcnt = upr.hop_count;
			rreq->rreq_id = htonl(upr.rreq_id);
			
			set_addr_byte(&addr_template, 3, upr.L2.nexthop_addr); // TRICK
			rreq->dest_addr = addr_template.s_addr;
			assert(upr.dst_seq_msb < 0x10);
			rreq->dest_seqno = htonl((upr.dst_seq_msb << 8) + upr.dst_seq_lsb);
			//printf("rreq->dest_seqno = %d.\n", rreq->dest_seqno);
			
			set_addr_byte(&addr_template, 3, upr.ori_addr);
			rreq->orig_addr = addr_template.s_addr;
			assert(upr.ori_seq_msb < 0x10);
			rreq->orig_seqno = htonl((upr.ori_seq_msb << 8) + upr.ori_seq_lsb);
			//printf("rreq->orig_seqno = %d.\n", rreq->orig_seqno);
			
		}
		else if (upra.AODV_type == AODV_TYPE_RREP) {
			RREP* rrep = (RREP*) msg;
			if (len < get_packed_size(hi->bits_rrep)) {
				// reject too short packets
				free(msg);
				return;
			}
			len = sizeof (RREP);
			struct unpacked_rrep upr;
			bitunpack(&upr.L2, packed, hi->bits_rrep);
			
#ifdef DEBUG_FLOW
			pretty_rRep(hi, "    R  ", &upr, INBOUND);
#endif
			
			rrep->a = upr.flags &  1;
			rrep->r = upr.flags >> 1;
			rrep->prefix = 0;
			
			rrep->hcnt = upr.hop_count;
			
			set_addr_byte(&addr_template, 3, upr.dst_addr);
			rrep->dest_addr = addr_template.s_addr;
			rrep->dest_seqno = htonl((upr.dst_seq_msb << 8) + upr.dst_seq_lsb);
			//printf("rrep->dest_seqno = %d.\n", rrep->dest_seqno);
			
			set_addr_byte(&addr_template, 3, upr.ori_addr);
			rrep->orig_addr = addr_template.s_addr;
			uint32_t lifetime = upr.lifetime_val * 20;
			while (upr.lifetime_exp--) lifetime *= 10;
			rrep->lifetime = htonl(lifetime);
			
		}
		else if (upra.AODV_type == AODV_TYPE_RERR) {
			RERR* rerr = (RERR*) msg;
			if (len < get_packed_size(hi->bits_rerr)) {
				// reject too short packets
				free(msg);
				return;
			}
			len = sizeof (RERR);
			
			struct unpacked_rerr upr;
			bitunpack(&upr.L2, packed, hi->bits_rerr);
			
#ifdef DEBUG_FLOW
			pretty_rErr(hi, "    R  ", &upr, INBOUND);
#endif
			
			rerr->dest_count = 1;
			set_addr_byte(&addr_template, 3, upr.dst_addr);
			rerr->dest_addr = addr_template.s_addr;
			
			rerr->dest_seqno = htonl((upr.dst_seq_msb << 8) + upr.dst_seq_lsb);
			//printf("rerr->dest_seqno = %d.\n", rerr->dest_seqno);
			rerr->n = upr.flags;
			
		}
		else if (upra.AODV_type == AODV_TYPE_RREP_ACK) {
			msg->type = AODV_RREP_ACK; // 0 -> 4
			if (len < get_packed_size(hi->bits_rrep_ack)) {
				// reject too short packets
				free(msg);
				return;
			}
			len = sizeof (RREP_ACK);
			
#ifdef DEBUG_FLOW
			pretty_rRep_A(hi, "    R  ", &upra, INBOUND);
#endif
		}
		
		struct in_addr src, dst;
		src = addr_template;
		set_addr_byte(&src, 3, uswt.L2.station_addr);
		if (uswt.L2.nexthop_addr == PACKED_BROADCAST)
			dst.s_addr = AODV_BROADCAST;
		else {
			dst = addr_template;
			set_addr_byte(&dst, 3, uswt.L2.nexthop_addr);
		}
		
		uint8_t ifindex = 0; // Only one interface.
		aodv_socket_process_packet(hi, (AODV_msg *) msg, len,
					src, dst, uswt.L2.ttl - 1, ifindex);
		
		free(msg);
		
		//upp.L2.header_chksum = 0x7; // TODO: Check checksum
		return;
		
	}
	
	
// 2. Data packets (TOP_TYPE_SEQ, typ. AMR): 
//    - forward packets for other destination
//    - forward broadcast packets
//    - unpack and provide broadcast packets for us
//    - unpack and provide packets for us locally
	
	
	
	if (uswt.L2.top_type == TOP_TYPE_SEQ) {
		debug(hi, LOG_DEBUG, 0, "This is SEQ(~AMR).");
		
		if (uswt.seq_type == SEQ_TYPE_AMR) {
			struct unpacked_amrlong ual;
			uint16_t len = bitunpack(&ual.L2, packed, hi->bits_amrlong);
			
#ifdef DEBUG_FLOW
			pretty_AMR(hi, "    R  ", &ual, INBOUND);
#endif
			
			ual.L2.ttl--;
			
			process_packet_short(hi, &ual.L2, len, PROCESS_INBOUND, pwr_mW);
			/*uint8_t us;
			get_addr_byte(us, 3, hi->devs[0].ip_addr);
			
			if (ual.dst_addr == us)
			*/
		}
		else {
			
#ifdef DEBUG_FLOW
			pretty_data(hi, "    R  ", &uswt, INBOUND);
#endif
			debug(hi, LOG_WARNING, 0, "Warning: Cann't handle this NON-AMR packet. DROPPING.\n");
		}
		return;
	}
	
	
	
	debug(hi, LOG_WARNING, 0, "Warning: Cann't handle this packet. DROPPING.");
#ifdef DEBUG_FOR_VALGRIND
	debug(hi, LOG_WARNING, 0, "Len: %d, Data: ", len);
	for(int i = 0; i < len; i++)
		debug(hi, LOG_WARNING, 0, "%02x ", ((uint8_t*)packed)[i]);
	debug(hi, LOG_WARNING, 0, "\n");
#endif
	
}
#endif // -ifndef NOCODE
;


// Cann't distinguish between long and short format so far.
// Returns the new sequence number.
// If it returns 0xffff this can be a sequence number or more probable an error.
uint16_t put_amr_packet(hi hi, void* data, uint8_t len, uint8_t destination)
#ifndef NOCODE
{
#ifdef REMOTE_CTRL
	timer_age_queue(hi);
#endif
	struct unpacked_amrlong ual;
	
	if (len != 12)
		debug(hi, LOG_WARNING, 0, "Packet length unexpected: actual %d, expected 12.\n", len);
	
	// s_addr is in network order (first seen number is first byte)
	if (memcmp(&hi->devs[0].ipaddr.s_addr, &(uint8_t[]){192,168,1}, 3) != 0) {
		debug(hi, LOG_ERR, 0, "Interface 0 doesn't have the right prefix!\n");
		return 0xffff;
	}
	
	get_addr_byte(ual.ori_addr, 3, hi->devs[0].ipaddr.s_addr);
	if (ual.ori_addr & 0xf0) {
		debug(hi, LOG_ERR, 0, "Last part of IP Address must be less than 16!\n");
		return 0xffff;
	}
	if (ual.ori_addr == PACKED_BROADCAST) {
		debug(hi, LOG_ERR, 0, "Dont want to use a broadcast address as origin!\n");
		return 0xffff;
	}
	
	
	if (ual.ori_addr & 0xf0) {
		debug(hi, LOG_WARNING, 0, "Our address is cut to less than 16!\n");
		ual.ori_addr &= 0x0f;
	}
	
	
	ual.L2.station_addr = ual.ori_addr;
	ual.L2.nexthop_addr = 0xFF; // Route to find. This is task of process_packet
	ual.L2.ttl = TTL_START_AMR;
	ual.L2.header_chksum = 0; // TODO
	ual.L2.top_type = TOP_TYPE_SEQ;
	ual.seq_type = SEQ_TYPE_AMR;
	ual.dst_addr = destination;
	
	if (ual.dst_addr & 0xf0) {
		debug(hi, LOG_WARNING, 0, "Destination address is cut to less than 16!\n");
		ual.dst_addr &= 0x0f;
	}
	if (ual.dst_addr == PACKED_BROADCAST) {
		debug(hi, LOG_INFO, 0, "Broadcasting AMR.");
		//ual.L2.nexthop_addr = 0xC;
	}
	
	// Start or resume connection
	
	int i;
	int inval = -1;
	int use = -1;
	for (i = NR_AMR_CONNECTIONS - 1; i > 0; i--) {
		if (!amr_connection_valid(hi->ac[i])) {
			inval = i;
			continue;
		}
		if (hi->ac[i].origin == ual.ori_addr) {
			use = i;
			break;
		}
	}
	
	if (use == -1 && inval == -1) {
		debug(hi, LOG_ERR, 0, "No local memory for AMR connections!\n");
		return 0xffff;
	}
	
	if (use == -1) {
		use = inval;
		hi->ac[use].origin = ual.ori_addr;
		hi->ac[use].inbound_seqnr = 0;
		hi->ac[use].outbound_seqnr = 0;
	}
	
	ual.data_seq_msb = hi->ac[use].outbound_seqnr >>   8;
	ual.data_seq_lsb = hi->ac[use].outbound_seqnr & 0xFF;
	
	ual.chk_amrA = 0x1F; // TODO
	for (i = 0; i < 12; i++) ual.amrbytes[i] = ((uint8_t*)data)[i];
	ual.chk_amrB = 0xFF; // TODO
	
	// debug(hi, LOG_INFO, 0, "ual.ori_addr = 0x%x.", ual.ori_addr);
	
	process_packet_short(hi, &ual.L2, sizeof(ual), PROCESS_OUTBOUND, 0.0);
	
	return hi->ac[use].outbound_seqnr ++;
}
#endif // -ifndef NOCODE
;

// TODO: refactor: split put_data_packet in
// 1. put_non_aodv_packet()        // not so similar to processPacket()
// 2. process_incoming_non_aodv()  // similar to processPacket()
// TODO: Implement originator's route next_hop lifetime update. (See
//       section 6.2 of RFC 3561). In the end there should be 4x lifetimes
//       updated for each received/forwarded packet.
// TODO: Implement lifetime update even when redirecting a broadcast packet.
// TODO: when forwarding a packet in process_incoming_non_aodv():
//	- Make two variants: 1. 
// TODO: Variant: Synchronize time and calculate routes based on real processing times?
		// (of course that which-I-hear-first is not bad either)
		// DIPLOMA thesis: thoughts about the need (pro-con) of time
		//   synchronization for routing.
// TODO: Variant: Respect RSSI value for decisions




// copied from packet_input.cc:processPacket()
// AODV packets get distributed to aodv_socket_process_packet()
// processPacket() is intended for all packets except AODV packets.
// In processPacket() there's no explizit link source (=last hop) and link
//    destination (=us) processing.
// This has to implications:
// 1. Either no intermediate hops are defined in packet, the same packet is
//    simply sent unchanged (except TTL) by all relays which receive it. But I
//    guess this breaks function of the network.
// 2. Data layer makes the link processing. This circumvents many problems, but
//    in one point this implementation is wrong: when an incoming packet is 
//    processed, the originator's route is updated without checking for the
//    right next_hop.



#define PROCESS_INBOUND true
#define PROCESS_OUTBOUND false


void process_packet_short(hi hi, struct unpacked_L2_header * data, int unpacked_length, bool inbound, float pwr_mW)
#ifndef NOCODE
{
	rt_table_t *fwd_rt, *rev_rt;
	struct in_addr dest_addr, src_addr;
    u_int8_t rreq_flags = 0;
#ifdef CONFIG_GATEWAY
    int pkt_flags = 0;
#endif
	src_addr.s_addr = -1; // Compiler warning
	
    if (data == NULL) return; // Sanity check
	
	if (hi->nif < 1) {
		debug(hi, LOG_WARNING, 0, "No interfaces -> dropping packet!");
		return;
	}
	
	// TODO: Unpack data (many different packets defined in all-packet.h)
	// TODO: 
	
	const struct fieldbits** fields = get_type(hi, data);
	//debug(hi, LOG_INFO, 0, "fieldbits=%p", fields);
	
	
	uint8_t us;
	get_addr_byte(us, 3, hi->devs[0].ipaddr.s_addr);
	
	
	uint16_t dst_short = get_destination_address(data, fields);
	if (dst_short & 0xFF00) {
		debug(hi, LOG_ERR, 0, "Couldn't get destination from packet!");
		return;
	}
	uint16_t ori_short = get_origin_address(data, fields);
	if (ori_short & 0xFF00) {
		debug(hi, LOG_ERR, 0, "Couldn't get origin from packet!");
		return;
	}
	if (ori_short == PACKED_BROADCAST) {
		debug(hi, LOG_WARNING, 0, "Dropping packet which has no sender.");
		return;
	}
	
	
	if (inbound == true
				&& data->nexthop_addr != PACKED_BROADCAST
				&& data->nexthop_addr != us
				&& dst_short != us) {
		// We are not interested in this packet.
		debug(hi, LOG_DEBUG, 0, "Packet with L2 %x->%x is not for us.", data->station_addr, data->nexthop_addr);
		// TODO: That's wrong - 
		return;
	}
	if (inbound == true
				&& data->nexthop_addr == PACKED_BROADCAST
				&& data->station_addr == us) {
		// Packet returning
		debug(hi, LOG_WARNING, 0, "Packet with L2 %x->%x has come back.", data->station_addr, data->nexthop_addr);
		return;
	}
	
	
	
	debug(hi, LOG_DEBUG, 0, "src=0x%x  dst=0x%x  prev_hop=0x%x  ttl=%d.", ori_short, dst_short, data->station_addr, data->ttl);
	
	
	//debug(hi, LOG_DEBUG, 0,
	//	"Packet src=%08x %s    dst=%08x %s    prev_hop=%08x %s", 
	//	src_addr.s_addr, ip_to_str(src_addr),
	//	dest_addr.s_addr, ip_to_str(dest_addr), 
	//	prev_hop.s_addr, ip_to_str(prev_hop));
	//debug(hi, LOG_DEBUG, 0,
	//	"Packet src=%s   dst=%s   prev_hop=%s   ttl=%d", 
	//	ip_to_str(src_addr),
	//	ip_to_str(dest_addr), 
	//	ip_to_str(prev_hop), data->ttl);
	
#if DEBUG_PACKET
	
	set_addr_byte(&src_addr, 0, 192);
	set_addr_byte(&src_addr, 1, 168);
	set_addr_byte(&src_addr, 2, 1);
	set_addr_byte(&src_addr, 3, ori_short);
	
	dest_addr = src_addr;
	if (dst_short == PACKED_BROADCAST)
		dest_addr.s_addr = AODV_BROADCAST;
	else
		set_addr_byte(&dest_addr, 3, dst_short);
	
	struct in_addr prev_hop = src_addr;
	set_addr_byte(&prev_hop, 3, data->station_addr);
	
	
    if (ch->direction() == hdr_cmn::DOWN) {
	    debug(hi, LOG_DEBUG, 0,
		  "Packet DOWN src=%s dst=%s prev_hop=%s", 
		  ip_to_str(src_addr), ip_to_str(dest_addr), 
		  ip_to_str(prev_hop));
    } else if (ch->direction() == hdr_cmn::UP) {
	    debug(hi, LOG_DEBUG, 0,
		  "Packet UP src=%s dst=%s prev_hop=%s", 
		  ip_to_str(src_addr), ip_to_str(dest_addr), 
		  ip_to_str(prev_hop));
    }
#endif
#if 0
	/* If this is a TCP packet and we don't have a route, we should
	   set the gratuituos flag in the RREQ. */
	if (ch->ptype() == PT_TCP) {
		rreq_flags |= RREQ_GRATUITOUS;
	}
#endif
	
	if (ori_short != us) {
		/* Find the entry of the neighboring node and the destination
		 *  (if any). */
		rev_rt = rt_table_find_short(hi, ori_short);
		
		if (!rev_rt) {
			// Some packet arrived but is not from us and we don't know the
			//	route -> add an auxiliary route.
			
			if (data->station_addr != PACKED_BROADCAST) {
				struct rt_table * hop_rt = rt_table_find_short(hi, data->station_addr);
				if (!hop_rt) {
					debug(hi, LOG_DEBUG, 0, "neighbor %x RECONSTRUCTED!", 
							data->station_addr);
					rev_rt = rt_table_insert_short(hi,
							data->station_addr, data->station_addr, 1, 0,
							hi->active_route_timeout, VALID, 0, 0);
				}
				else {
				rt_table_update_timeout(hi, hop_rt, hi->active_route_timeout);
				}
			}
			else {
				debug(hi, LOG_WARNING, 0, "Don't handle broadcast as destination!");
			}
			
			if (   ori_short != data->station_addr
			    && ori_short != PACKED_BROADCAST) {
				debug(hi, LOG_DEBUG, 0, "far station %x RECONSTRUCTED!", ori_short);
				rev_rt = rt_table_find_short(hi, ori_short);
				if (!rev_rt)
					rev_rt = rt_table_insert_short(hi,
							ori_short, data->station_addr, 2*NET_DIAMETER, 0,
							hi->active_route_timeout, VALID, 0, 0);
				else if(rev_rt->hcnt >= 2*NET_DIAMETER) {
					rt_table_update_short(hi, rev_rt, data->station_addr,
						2*NET_DIAMETER, 0, hi->active_route_timeout, VALID, 0);
				}
			}
		}
		else {
			rt_table_update_route_timeouts(hi, NULL, rev_rt);
		}
	}
	else {
		rev_rt = NULL;
	}
	
	/* If the packet is a broadcast packet we just let it go
	* through... */
	if (dst_short == PACKED_BROADCAST) {
		
		debug(hi, LOG_DEBUG, 0, "Broadcast packet src=0x%x dst=0x%x ttl=%d",
				ori_short, dst_short, data->ttl);
		
		bool send = false;
		if (ori_short != us && data->station_addr != us) {
			if (!rev_rt) {
				debug(hi, LOG_ERR, 0, "WARNING: broadcast packet from %x (ttl %d) has no reverse route! (This cann't happen)", ori_short, data->ttl);
				// This cann't happen.
				return;
			}
			
			debug(hi, LOG_DEBUG, 0, "rev_rt->ttl_broadcast.ttl_seen=%d.",
							rev_rt->ttl_broadcast.ttl_seen);
			
			// Lookup the last ttl value.
			if (rev_rt && rev_rt->ttl_broadcast.ttl_seen <= data->ttl) {
				// All packets with same or higher ttl will be
				// re-broadcast.
				set_ttl_broadcast(hi, rev_rt, data->ttl);
				send = true;
			}
			else if (rev_rt && rev_rt->ttl_broadcast.ttl_seen == TTL_INVALID) {
				set_ttl_broadcast(hi, rev_rt, data->ttl);
				send = true;
			}
			else {
				// Packet has made some loop.
			}
		}
		
		if (send) {
			// Receive it ...
			final_receive(hi, data, pwr_mW);
		}
		
		if (inbound == PROCESS_OUTBOUND || send) {
			// And send it
			data->station_addr = us;
			data->nexthop_addr = PACKED_BROADCAST;
			cli_pack_and_send(hi, data, fields);
		}
		
		return;
	}
	fwd_rt = rt_table_find_short(hi, dst_short);

#ifdef CONFIG_GATEWAY
    /* Check if we have a route and it is an Internet destination (Should be
     * encapsulated and routed through the gateway). */
    if (fwd_rt && (fwd_rt->state == VALID) && 
	(fwd_rt->flags & RT_INET_DEST)) {
	/* The destination should be relayed through the IG */

	rt_table_update_timeout(hi, fwd_rt, hi->active_route_timeout);

	p = pkt_encapsulate(p, fwd_rt->next_hop);

	if (p == NULL) {
	    debug(hi, LOG_ERR, 0, "IP Encapsulation failed!");
	   return;	    
	}
	/* Update pointers to headers */
	ch = HDR_CMN(p);
	ih = HDR_IP(p);

	dest_addr = fwd_rt->next_hop;
	fwd_rt = rt_table_find(dest_addr);
	pkt_flags |= PKT_ENC;
    }
#endif /* CONFIG_GATEWAY */

    /* UPDATE TIMERS on active forward and reverse routes...  */

    rt_table_update_route_timeouts(hi, fwd_rt, /*rev_rt*/NULL);

	
	/* OK, the timeouts have been updated. Now see if either: 1. The
	packet is for this node -> ACCEPT. 2. The packet is not for this
	node -> Send RERR (someone want's this node to forward packets
	although there is no route) or Send RREQ. */
	
	/* If the packet is destined for this node, then just accept it. */
	if (dst_short == us) {
		
#ifdef CN_PORT
		final_receive(hi, data, pwr_mW);
#else
		ch->size() -= IP_HDR_LEN;    // cut off IP header size 4/7/99 -dam
		target_->recv(p, (Handler*)0);
		p = NULL;
#endif // -ifdev CN_PORT // else
		return;
		
	}
	
	//debug(hi, LOG_DEBUG, 0, "fwd_rt->hcnt = %d, fwd_rt->flags = 0x%x", fwd_rt->hcnt, fwd_rt->flags);
	
	if (!fwd_rt || fwd_rt->state == INVALID ||
		(fwd_rt->hcnt == 1 && (fwd_rt->flags & RT_UNIDIR))) {
		
		/* Check if the route is marked for repair or is INVALID. In
		* that case, do a route discovery. */
		if (fwd_rt && (fwd_rt->flags & RT_REPAIR))
			goto route_discovery;
		
		debug(hi, LOG_DEBUG, 0, "TODO: Put error packet when the route hasn't been found.");
		goto route_discovery;
		
		
		struct in_addr rerr_dest;
		
		debug(hi, LOG_DEBUG, 0,
				"No route, src=%s dest=%s - DROPPING!",
				ip_to_str(src_addr), ip_to_str(dest_addr));
	//TODO: Uncomment this:	
//		if (src_addr == hi->devs[0].ip_addr)
//			return;
		
		RERR *rerr;
		if (fwd_rt) {
			rerr = rerr_create(0, fwd_rt->dest_addr,
					fwd_rt->dest_seqno);
			
			rt_table_update_timeout(hi, fwd_rt, hi->delete_period);
		} else
			rerr = rerr_create(0, dest_addr, 0);
		
		debug(hi, LOG_DEBUG, 0, "Sending RERR to prev hop %s for unknown dest %s", ip_to_str(src_addr), ip_to_str(dest_addr));
		
		/* Unicast the RERR to the source of the data transmission
		* if possible, otherwise we broadcast it. */
		
		if (rev_rt && rev_rt->state == VALID)
			rerr_dest = rev_rt->next_hop;
		else
			rerr_dest.s_addr = AODV_BROADCAST;
		
		aodv_socket_send(hi, (AODV_msg *) rerr, rerr_dest,
				RERR_CALC_SIZE(rerr), 1, NULL);
		
		if (hi->wait_on_reboot) {
			debug(hi, LOG_DEBUG, 0, "Wait on reboot timer reset.");
			timer_set_timeout(hi, &hi->worb_timer, hi->delete_period);
		}
		
		/* debug(hi, LOG_DEBUG, 0, "Dropping pkt uid=%d", ch->uid()); */
		//drop(p, DROP_RTR_NO_ROUTE);
		
		return;
		
route_discovery:
		/* Buffer packets... Packets are queued by the ip_queue.o
		module already. We only need to save the handle id, and
		return the proper verdict when we know what to do... */
		if(false){}; // Fehler: eine Marke kann nur Teil einer Anweisung sein, und eine Deklaration ist keine Anweisung
		struct unpacked_L2_header * ul2h = malloc(unpacked_length);
		memcpy(ul2h, data, unpacked_length);
		packet_queue_add_short(hi, ul2h, dst_short, unpacked_length);
		
		if (fwd_rt && (fwd_rt->flags & RT_REPAIR))
			rreq_local_repair_short(hi, fwd_rt, ori_short, NULL);
		else
			rreq_route_discovery_short(hi, dst_short, rreq_flags, NULL);
		
		return;
		
	} else {
		// Route does exist.
		
		/* debug(hi, LOG_DEBUG, 0, "Sending pkt uid=%d", ch->uid()); */
		// TODO: Send
		debug(hi, LOG_DEBUG, 0, "Sending packet on known route.");
		uint8_t nexthop;
		get_addr_byte(nexthop, 3, fwd_rt->next_hop);
		
		data->station_addr = us;
		data->nexthop_addr = nexthop;
		//printf("NH=%d\n", data->nexthop_addr);
		cli_pack_and_send(hi, data, fields);
		
		//sendPacket(p, fwd_rt->next_hop, 0.0);
		
		/* When forwarding data, make sure we are sending HELLO messages */
		gettimeofday(&hi->fwd_time, NULL);
		
		if (!hi->llfeedback && hi->optimized_hellos)
			hello_start(hi);
	}
}
#endif
;



#endif // -ifndef CLI_DEV_C

