/*****************************************************************************
 *
 * Copyright (C) 2001 Uppsala University & Ericsson AB.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Authors: Erik Nordstrm, <erik.nordstrom@it.uu.se>
 *
 *****************************************************************************/
#ifndef _DEFS_H
#define _DEFS_H

#if (!defined REMOTE_CTRL) && ((defined BEAGLE) || (defined SIMUL))
#define REMOTE_CTRL
#endif

#if (defined REMOTE_CTRL) && !(defined CN_PORT)
#define CN_PORT // The non-CN parts cann't be simulated by Dosim/air
#endif

#define AODV_STRUCT_SIG 0xbadefaad


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>

#ifndef NS_PORT
#include <sys/signal.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netinet/ip.h>
#include <sys/ioctl.h>
#endif

#include <syslog.h>
#include <errno.h>
#include <string.h>
#include <fcntl.h>

#include "list.h"



#ifndef NS_PORT
#include "timer_queue.h"
#endif



#ifdef CN_PORT
#include "ns-2/packet_queue.h"

#if defined(REMOTE_CTRL) && (defined(BEAGLE) + defined(SIMUL) != 1)
#error   One of BEAGLE or SIMUL must be defined!
#endif

#ifdef SIMUL
#define gettimeofday(val, tz) get_systimeval(hi->multihw_instance->station, val, tz)
#endif // -ifdef SIMUL
#ifdef NOCODE
#include "multihw.c"
#else
#define NOCODE
#include "multihw.c"
#undef NOCODE
#endif

#endif // -ifdef CN_PORT



#include "params.h" // RREQ_RATELIMIT

#ifdef NS_PORT
#define NS_CLASS AODVUU::
#define NS_OUTSIDE_CLASS ::
#define NS_STATIC
#define NS_INLINE
/* NS_PORT: Using network device 0, with interface index 0. */
#define NS_DEV_NR 0
#define NS_IFINDEX NS_DEV_NR
#else
#define NS_CLASS
#define NS_OUTSIDE_CLASS
#define NS_STATIC static
//#define NS_INLINE inline
#define NS_INLINE // https://stackoverflow.com/questions/47657744/installing-aodv-protocol-on-raspberry-pi-https-github-com-erimatnor-aodv-uu
#endif

#define AODV_UU_VERSION "0.9"
#define DRAFT_VERSION "rfc3561"

#ifdef NS_PORT
/* NS_PORT: Log filename split into prefix and suffix. */
#define AODV_LOG_PATH_PREFIX "aodv-uu-"
#define AODV_RT_LOG_PATH_SUFFIX ".rtlog"
#define AODV_LOG_PATH_SUFFIX ".log"
#else
#define AODV_LOG_PATH "/var/log/aodvd.log"
#define AODV_RT_LOG_PATH "/var/log/aodvd.rtlog"
#endif				/* NS_PORT */

#define max(A,B) ( (A) > (B) ? (A):(B))

#define MINTTL 1		/* min TTL in the packets sent locally */

#define MAX_NR_INTERFACES 10
#define MAX_IFINDEX (MAX_NR_INTERFACES - 1)

#if !defined(IFNAMSIZ)
#define IFNAMSIZ 16
#endif

#if defined CN_PORT && defined CONFIG_GATEWAY
#warning aodv-uu cannot act as gateway when used in CN environment.
#undef CONFIG_GATEWAY
#endif
#ifdef CN_PORT
#define IP_BROADCAST 0xffffffff
#endif


/* Data for a network device */
struct dev_info {
    int enabled;		/* 1 if struct is used, else 0 */
    int sock;			/* AODV socket associated with this device */
#ifdef CONFIG_GATEWAY
    int psock;			/* Socket to send buffered data packets. */
#endif
    unsigned int ifindex;
    char ifname[IFNAMSIZ];
    struct in_addr ipaddr;	/* The local IP address */
    struct in_addr netmask;	/* The netmask we use */
    struct in_addr broadcast;
	
};


#define amr_connection_valid(ac) ((ac).origin != PACKED_BROADCAST)
struct amr_connection {
	uint8_t origin:4; // PACKED_BROADCAST [0x0c?] for invalid
	uint16_t inbound_seqnr:12;
	uint16_t outbound_seqnr:12;
};


/* The callback function */
#ifndef NS_PORT
struct host_info;
typedef void (*callback_func_t) (struct host_info*, int);
#endif
struct callback {
    int fd;
    callback_func_t func;
};


struct fieldbits;
struct routing_table;

struct host_info {
	
#ifdef REMOTE_CTRL
	// For multihw:
	struct multihw_global_struct * multihw_instance;
#endif
	
    u_int32_t seqno;		/* Sequence number */
    struct timeval bcast_time;	/* The time of the last broadcast msg sent */
    struct timeval fwd_time;	/* The time a data packet was last forwarded */
    u_int32_t rreq_id;		/* RREQ id */
    int nif;			/* Number of interfaces to broadcast on */
	struct dev_info devs[MAX_NR_INTERFACES+1]; /* Add +1 for returning as "error" in ifindex2devindex. */
	
	list_t TQ;
	
	
	/* "Global" Variables: */
	int log_to_file;
	int rt_log_interval;	/* msecs between routing table logging 0=off */
	int unidir_hack;
	int rreq_gratuitous;
	int expanding_ring_search;
	int internet_gw_mode;
	int local_repair;
	int receive_n_hellos;
	int hello_jittering;
	int ratelimit;		/* Option for rate limiting RREQs and RERRs. */
	int qual_threshold;
	int gw_prefix;
	char *progname;
	
	// Run state "global"
	uint8_t run;
	uint8_t wait_on_reboot;
	struct timer worb_timer;	/* Wait on reboot timer */
	
	/* Dynamic configuration values */
	int active_route_timeout;
	int delete_period;
#ifndef NS_PORT
	int ttl_start;
#endif
	int optimized_hellos;
	int llfeedback;
	
	// log
	uint8_t debug;
	int log_file_fd;
	int log_rt_fd;
	int log_nmsgs; // needed for what?
	struct timer rt_log_timer;
	
	
	// callbacks
	int nr_callbacks;
/* This will limit the number of handler functions we can have for
   sockets and file descriptors and so on... */
#define CALLBACK_FUNCS 5
	struct callback callbacks[CALLBACK_FUNCS];
	
	// Routing table (Link)
	struct routing_table * rt_tbl;
	
	// packetizer
	const struct fieldbits ** bits_rreq;
	const struct fieldbits ** bits_rrep;
	const struct fieldbits ** bits_rrep_ack;
	const struct fieldbits ** bits_rerr;
	const struct fieldbits ** bits_seq_w_type;
	const struct fieldbits ** bits_amrlong;
	const struct fieldbits ** bits_amrshort;
	
#define NR_AMR_CONNECTIONS 2
	struct amr_connection ac[NR_AMR_CONNECTIONS];
	
	struct timer hello_timer;
	
	
	LIST(seekhead);
	
	LIST(rreq_records);
	LIST(rreq_blacklist);
	
	// rate limit
	struct timeval rreq_ratel[RREQ_RATELIMIT - 1];
	struct timeval rerr_ratel[RERR_RATELIMIT - 1];
	int num_rreq;
	int num_rerr;
	
	
#ifdef CN_PORT
	struct packet_queue PQ;
#endif
	
	uint32_t signature;
	
#define SZ_RCV 100
	uint8_t recv_buf_pos;
	uint8_t recv_buf[SZ_RCV];
};
typedef struct host_info * hi;

#ifdef REMOTE_CTRL
// This is former main()
int AODV_start(struct host_info * hiP, int argc, char **argv);
#endif


/*
  NS_PORT: TEMPORARY SOLUTION: Moved the two variables into the AODVUU class,
  and placed the function definition after the AODVUU class definition.

  (This is to avoid running several passes through defs.h during the source
  code extraction performed by the AODVUU class.)

  TODO: Find some smarter way to accomplish this.
*/
#ifndef CN_PORT
#ifndef NS_PORT
/* This will point to a struct containing information about the host */
//struct host_info this_host;

/* Array of interface indexes */
unsigned int dev_indices[MAX_NR_INTERFACES];

/* Given a network interface index, return the index into the
   devs array, Necessary because ifindex is not always 0, 1,
   2... */
static inline unsigned int ifindex2devindex(unsigned int ifindex)
{
    int i;

    for (i = 0; i < this_host.nif; i++)
	if (dev_indices[i] == ifindex)
	    return i;

    return MAX_NR_INTERFACES;
}

static inline struct dev_info *devfromsock(int sock)
{
    int i;

    for (i = 0; i < this_host.nif; i++) {
	if (this_host.devs[i].sock == sock)
	    return &this_host.devs[i];
    }
    return NULL;
}

static inline int name2index(char *name)
{
    int i;

    for (i = 0; i < this_host.nif; i++)
	if (strcmp(name, this_host.devs[i].ifname) == 0)
	    return this_host.devs[i].ifindex;

    return -1;
}

#endif


/* Two macros to simplify retriving of a dev_info struct. Either using
   an ifindex or a device number (index into devs array). */
#define DEV_IFINDEX(hi, ifindex) (this_host.devs[ifindex2devindex(ifindex)])
#define DEV_NR(n) (this_host.devs[n])

#else

#define DEV_IFINDEX(hi, ifindex) (hi->devs[ifindex])
#define DEV_NR(hi, n) (hi->devs[n])

#endif // -ifndef CN_PORT

 /* Broadcast address according to draft (255.255.255.255) */
#define AODV_BROADCAST ((in_addr_t) 0xFFFFFFFF)

#define AODV_PORT 654

/* AODV Message types */
#define AODV_HELLO    0		/* Really never used as a separate type... */
#define AODV_RREQ     1
#define AODV_RREP     2
#define AODV_RERR     3
#define AODV_RREP_ACK 4
#define GENERIC_DATA  5  // This doesn't belong to AODV

#define RREQ_ID_OVERFLOW 0x40 // 6 bit
#define SEQNO_OVERFLOW 0x1000 // 12 bit

// REL: bigger >    or    bigger equal >=
#define is_seqno_a_REL_than_b(a, REL, b) ( \
		((a) REL (b)) || ((b) > (SEQNO_OVERFLOW>>2) \
		                    ? (a) + SEQNO_OVERFLOW REL (b) \
		                    : false \
		               ) \
	)
#define seq_cmp(a, REL, b) is_seqno_a_REL_than_b(a, REL, b)


/* A generic AODV packet header struct... */
#ifdef NS_PORT
struct AODV_msg {
#else
typedef struct {
#endif
    u_int8_t type;

/* NS_PORT: Additions for the AODVUU packet type in ns-2 */
#ifdef NS_PORT
    static int offset_;		// Required by PacketHeaderManager

    inline static int &offset() {
	return offset_;
    }
    inline static AODV_msg *access(const Packet * p) {
	return (AODV_msg *) p->access(offset_);
    }

    int size();
};

typedef AODV_msg hdr_aodvuu;	// Name convention for headers
#define HDR_AODVUU(p) ((hdr_aodvuu *) hdr_aodvuu::access(p))
#else
} AODV_msg;
#endif

/* AODV Extension types */
#define RREQ_EXT 1
#define RREP_EXT 1
#define RREP_HELLO_INTERVAL_EXT 2
#define RREP_HELLO_NEIGHBOR_SET_EXT 3
#define RREP_INET_DEST_EXT 4

/* An generic AODV extensions header */
typedef struct {
    u_int8_t type;
    u_int8_t length;
    /* Type specific data follows here */
} AODV_ext;

/* MACROS to access AODV extensions... */
#define AODV_EXT_HDR_SIZE sizeof(AODV_ext)
#define AODV_EXT_DATA(ext) ((char *)((char *)ext + AODV_EXT_HDR_SIZE))
#define AODV_EXT_NEXT(ext) ((AODV_ext *)((char *)ext + AODV_EXT_HDR_SIZE + ext->length))
#define AODV_EXT_SIZE(ext) (AODV_EXT_HDR_SIZE + ext->length)

#ifndef NS_PORT
extern int attach_callback_func(hi, int fd, callback_func_t func);
#endif

#endif				/* DEFS_H */
