/*****************************************************************************
 *
 * Copyright (C) 2001 Uppsala University and Ericsson AB.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Generafl Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Author: Erik Nordstrm, <erik.nordstrom@it.uu.se>
 *
 *****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
//#define _XOPEN_SOURCE 700
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <linux/sockios.h>
#include <linux/if.h>
#include <linux/wireless.h>
#include <getopt.h>
#include <ctype.h>

#include "defs.h"
#include "debug.h"
#include "timer_queue.h"
#include "params.h"
#include "aodv_socket.h"
#include "aodv_timeout.h"
#include "routing_table.h"
#include "aodv_hello.h"
#include "nl.h"
//#include "ri-imi.h"
#define NOCODE
#include "packetizer.c"
#undef NOCODE
#include "ns-2/packet_queue.h"

#ifdef LLFEEDBACK
#include "llf.h"
#endif



#ifndef CN_PORT
static void cleanup();
#endif

const struct option longopts[] = {
    {"interface", required_argument, NULL, 'i'},
    {"hello-jitter", no_argument, NULL, 'j'},
    {"log", no_argument, NULL, 'l'},
    {"n-hellos", required_argument, NULL, 'n'},
#ifndef CN_PORT
    {"daemon", no_argument, NULL, 'd'},
#endif
    {"force-gratuitous", no_argument, NULL, 'g'},
    {"opt-hellos", no_argument, NULL, 'o'},
    {"quality-threshold", required_argument, NULL, 'q'},
    {"log-rt-table", required_argument, NULL, 'r'},
    {"unidir_hack", no_argument, NULL, 'u'},
    {"gateway-mode", no_argument, NULL, 'w'},
    {"help", no_argument, NULL, 'h'},
    {"no-expanding-ring", no_argument, NULL, 'x'},
    {"no-worb", no_argument, NULL, 'D'},
    {"local-repair", no_argument, NULL, 'L'},
    {"rate-limit", no_argument, NULL, 'R'},
    {"version", no_argument, NULL, 'V'},
    {"llfeedback", no_argument, NULL, 'f'},
    {0}
};

void usage(int status, char* progname)
{
    if (status != 0) {
	fprintf(stderr, "Try `%s --help' for more information.\n", progname);
	exit(status);
    }

    printf
	("\nUsage: %s [-dghjlouwxLDRV] [-i if0,if1,..] [-r N] [-n N] [-q THR]\n\n"
#ifndef CN_PORT
	 "-d, --daemon            Daemon mode, i.e. detach from the console.\n"
#else
	 "-d, --daemon            NOT AVAILABLE.\n"
#endif
	 "-g, --force-gratuitous  Force the gratuitous flag to be set on all RREQ's.\n"
	 "-h, --help              This information.\n"
	 "-i, --interface         Network interfaces to attach to. Defaults to first\n"
	 "                        wireless interface.\n"
	 "-j, --hello-jitter      Toggle hello jittering (default ON).\n"
	 "-l, --log               Log debug output to %s.\n"
	 "-o, --opt-hellos        Send HELLOs only when forwarding data (experimental).\n"
	 "-r, --log-rt-table      Log routing table to %s every N secs.\n"
	 "-n, --n-hellos          Receive N hellos from host before treating as neighbor.\n"
	 "-u, --unidir-hack       Detect and avoid unidirectional links (experimental).\n"
	 "-w, --gateway-mode      Enable experimental Internet gateway support.\n"
	 "-x, --no-expanding-ring Disable expanding ring search for RREQs.\n"
	 "-D, --no-worb           Disable 15 seconds wait on reboot delay.\n"
	 "-L, --local-repair      Enable local repair.\n"
	 "-f, --llfeedback        Enable link layer feedback.\n"
	 "-R, --rate-limit        Toggle rate limiting of RREQs and RERRs (default ON).\n"
	 "-q, --quality-threshold Set a minimum signal quality threshold for control packets.\n"
	 "-V, --version           Show version.\n\n"
	 "Erik Nordstrm, <erik.nordstrom@it.uu.se>\n\n",
	 progname, AODV_LOG_PATH, AODV_RT_LOG_PATH);

    exit(status);
}

#ifndef CN_PORT
int set_kernel_options()
{
    int i, fd = -1;
    char on = '1';
    char off = '0';
    char command[64];

    if ((fd = open("/proc/sys/net/ipv4/ip_forward", O_WRONLY)) < 0)
	return -1;
    if (write(fd, &on, sizeof(char)) < 0)
	return -1;
    close(fd);
    /*
      Disabled for kernel 2.6.28.


    if ((fd = open("/proc/sys/net/ipv4/route/max_delay", O_WRONLY)) < 0)
	return -1;
    if (write(fd, &off, sizeof(char)) < 0)
	return -1;
    close(fd);

    if ((fd = open("/proc/sys/net/ipv4/route/min_delay", O_WRONLY)) < 0)
	return -1;
    if (write(fd, &off, sizeof(char)) < 0)
	return -1;
    close(fd);
    */

    /* Disable ICMP redirects on all interfaces: */

    for (i = 0; i < MAX_NR_INTERFACES; i++) {
	if (!DEV_NR(i).enabled)
	    continue;

	memset(command, '\0', 64);
	sprintf(command, "/proc/sys/net/ipv4/conf/%s/send_redirects",
		DEV_NR(i).ifname);
	if ((fd = open(command, O_WRONLY)) < 0)
	    return -1;
	if (write(fd, &off, sizeof(char)) < 0)
	    return -1;
	close(fd);
	memset(command, '\0', 64);
	sprintf(command, "/proc/sys/net/ipv4/conf/%s/accept_redirects",
		DEV_NR(i).ifname);
	if ((fd = open(command, O_WRONLY)) < 0)
	    return -1;
	if (write(fd, &off, sizeof(char)) < 0)
	    return -1;
	close(fd);
    }
    memset(command, '\0', 64);
    sprintf(command, "/proc/sys/net/ipv4/conf/all/send_redirects");
    if ((fd = open(command, O_WRONLY)) < 0)
	return -1;
    if (write(fd, &off, sizeof(char)) < 0)
	return -1;
    close(fd);


    memset(command, '\0', 64);
    sprintf(command, "/proc/sys/net/ipv4/conf/all/accept_redirects");
    if ((fd = open(command, O_WRONLY)) < 0)
	return -1;
    if (write(fd, &off, sizeof(char)) < 0)
	return -1;
    close(fd);

    return 0;
}

int find_default_gw(void)
{
    FILE *route;
    char buf[100], *l;

    route = fopen("/proc/net/route", "r");

    if (route == NULL) {
	perror("open /proc/net/route");
	exit(-1);
    }

    while (fgets(buf, sizeof(buf), route)) {
	l = strtok(buf, " \t");
	l = strtok(NULL, " \t");
	if (l != NULL) {
	    if (strcmp("00000000", l) == 0) {
		l = strtok(NULL, " \t");
		l = strtok(NULL, " \t");
		if (strcmp("0003", l) == 0) {
		    fclose(route);
		    return 1;
		}
	    }
	}
    }
    fclose(route);
    return 0;
}

/*
 * Returns information on a network interface given its name...
 */
struct sockaddr_in *get_if_info(char *ifname, int type)
{
    int skfd;
    struct sockaddr_in *ina;
    static struct ifreq ifr;

    /* Get address of interface... */
    skfd = socket(AF_INET, SOCK_DGRAM, 0);

    strcpy(ifr.ifr_name, ifname);
    if (ioctl(skfd, type, &ifr) < 0) {
	alog(LOG_ERR, errno, __FUNCTION__,
	     "Could not get address of %s ", ifname);
	close(skfd);
	return NULL;
    } else {
	ina = (struct sockaddr_in *) &ifr.ifr_addr;
	close(skfd);
	return ina;
    }
}
#endif // -ifndef CN_PORT

//#define CALLBACK_FUNCS 5
//static struct callback {
//    int fd;
//    callback_func_t func;
//} callbacks[CALLBACK_FUNCS];
//
//static int nr_callbacks = 0;

int attach_callback_func(hi hi, int fd, callback_func_t func)
{
    if (hi->nr_callbacks >= CALLBACK_FUNCS) {
	fprintf(stderr, "callback attach limit reached!!\n");
	exit(-1);
    }
    hi->callbacks[hi->nr_callbacks].fd = fd;
    hi->callbacks[hi->nr_callbacks].func = func;
    hi->nr_callbacks++;
    return 0;
}

/* Here we find out how to load the kernel modules... If the modules
   are located in the current directory. use those. Otherwise fall
   back to modprobe. */

void load_modules(char *ifname)
{
    struct stat st;
    char buf[1024], *l = NULL;
    int found = 0;
    FILE *m;

    memset(buf, '\0', 64);

    if (stat("./kaodv.ko", &st) == 0)
	sprintf(buf, "/sbin/insmod kaodv.ko ifname=%s &>/dev/null", ifname);
    else if (stat("./kaodv.o", &st) == 0)
	sprintf(buf, "/sbin/insmod kaodv.o ifname=%s &>/dev/null", ifname);
    else
	sprintf(buf, "/sbin/modprobe kaodv ifname=%s &>/dev/null", ifname);

    if (system(buf) == -1) {
        fprintf(stderr, "Could not load kaodv module\n");
        exit(-1);
    }

    usleep(100000);

    /* Check result */
    m = fopen("/proc/modules", "r");
    while (fgets(buf, sizeof(buf), m)) {
	l = strtok(buf, " \t");
	if (!strcmp(l, "kaodv"))
	    found++;
	if (!strcmp(l, "ipchains")) {
	    fprintf(stderr,
		    "The ipchains kernel module is loaded and prevents AODV-UU from functioning properly.\n");
	    exit(-1);
	}
    }
    fclose(m);

    if (found < 1) {
	fprintf(stderr,
		"A kernel module could not be loaded, check your installation... %d\n",
		found);
	exit(-1);
    }
}

void remove_modules(void)
{
	int ret;

	ret = system("/sbin/rmmod kaodv &>/dev/null");

	if (ret != 0) {
		fprintf(stderr, "Could not remove kernel module kaodv\n");
	}
}

// One single interface name!
void ri_if_init(hi hi, const char *ifname) {
	debug(hi, LOG_DEBUG, 0, "Initializing RI [fake] interface '%s'.", ifname);
	
	uint8_t number = 111;
	if (ifname == NULL)
		ifname = "noname";
	else if (strncmp ("ip", ifname, 2) == 0)
		number = ifname[2] - '0';
	
	
    /* Intitialize the local sequence number an rreq_id to zero */
    hi->seqno = 1;
    hi->rreq_id = 0;
    /* Zero interfaces enabled so far... */
    hi->nif = 0;

    gettimeofday(&hi->bcast_time, NULL);
	
	
	
	hi->devs[hi->nif].ifindex = hi->nif;
	
	//dev_indices[this_host.nif++] = ifreq.ifr_ifindex;
	
	strncpy(hi->devs[hi->nif].ifname, ifname, IFNAMSIZ);
	
	
	hi->devs[hi->nif].ipaddr.s_addr = inet_addr("192.168.1.1");
	//0xC0A80101;
	hi->devs[hi->nif].ipaddr.s_addr &= 0x00FFFFFF;
	hi->devs[hi->nif].ipaddr.s_addr += number << 24;
	
	// Netmask
	hi->devs[hi->nif].netmask.s_addr = inet_addr("255.255.0.0");
	//0xFFFF0000;
	
	hi->devs[hi->nif].broadcast.s_addr = inet_addr("192.168.0.255");
	//0xFFFFFFFF;
	
	hi->devs[hi->nif].enabled = 1;
	
	hi->nif++;
}


#ifndef CN_PORT
void host_init(char *ifname)
{
    struct sockaddr_in *ina;
    char buf[1024], tmp_ifname[IFNAMSIZ],
	ifnames[(IFNAMSIZ + 1) * MAX_NR_INTERFACES], *iface;
    struct ifconf ifc;
    struct ifreq ifreq, *ifr;
    int i, iw_sock, if_sock = 0;

    memset(&this_host, 0, sizeof(struct host_info));
    memset(dev_indices, 0, sizeof(unsigned int) * MAX_NR_INTERFACES);

    if (!ifname) {
	/* No interface was given... search for first wireless. */
	iw_sock = socket(PF_INET, SOCK_DGRAM, 0);
	ifc.ifc_len = sizeof(buf);
	ifc.ifc_buf = buf;
	if (ioctl(iw_sock, SIOCGIFCONF, &ifc) < 0) {
	    fprintf(stderr, "Could not get wireless info\n");
	    exit(-1);
	}
	ifr = ifc.ifc_req;
	for (i = ifc.ifc_len / sizeof(struct ifreq); i >= 0; i--, ifr++) {
	    struct iwreq req;

	    strcpy(req.ifr_name, ifr->ifr_name);
	    if (ioctl(iw_sock, SIOCGIWNAME, &req) >= 0) {
		strcpy(tmp_ifname, ifr->ifr_name);
		break;
	    }
	}
	/* Did we find a wireless interface? */
	if (!strlen(tmp_ifname)) {
	    fprintf(stderr, "\nCould not find a wireless interface!\n");
	    fprintf(stderr, "Use -i <interface> to override...\n\n");
	    exit(-1);
	}
	strcpy(ifreq.ifr_name, tmp_ifname);
	if (ioctl(iw_sock, SIOCGIFINDEX, &ifreq) < 0) {
	    alog(LOG_ERR, errno, __FUNCTION__,
		 "Could not get index of %s", tmp_ifname);
	    close(if_sock);
	    exit(-1);
	}
	close(iw_sock);

	ifname = tmp_ifname;

	alog(LOG_NOTICE, 0, __FUNCTION__,
	     "Attaching to %s, override with -i <if1,if2,...>.", tmp_ifname);
    }

    strcpy(ifnames, ifname);

    /* Intitialize the local sequence number an rreq_id to zero */
    this_host.seqno = 1;
    this_host.rreq_id = 0;
    /* Zero interfaces enabled so far... */
    this_host.nif = 0;

    gettimeofday(&this_host.bcast_time, NULL);


    /* Find the indices of all interfaces to broadcast on... */
    if_sock = socket(AF_INET, SOCK_DGRAM, 0);

    iface = strtok(ifname, ",");

    /* OK, now lookup interface information, and store it... */
    do {
		strcpy(ifreq.ifr_name, iface);
		if (ioctl(if_sock, SIOCGIFINDEX, &ifreq) < 0) {
			alog(LOG_ERR, errno, __FUNCTION__, "Could not get index of %s",
			iface);
			close(if_sock);
			exit(-1);
		}
		this_host.devs[this_host.nif].ifindex = ifreq.ifr_ifindex;

		dev_indices[this_host.nif++] = ifreq.ifr_ifindex;

		strcpy(DEV_IFINDEX(ifreq.ifr_ifindex).ifname, iface);

		/* Get IP-address of interface... */
		ina = get_if_info(iface, SIOCGIFADDR);

		if (ina == NULL)
			exit(-1);

		DEV_IFINDEX(ifreq.ifr_ifindex).ipaddr = ina->sin_addr;

		/* Get netmask of interface... */
		ina = get_if_info(iface, SIOCGIFNETMASK);

		if (ina == NULL)
			exit(-1);

		DEV_IFINDEX(ifreq.ifr_ifindex).netmask = ina->sin_addr;

		ina = get_if_info(iface, SIOCGIFBRDADDR);

		if (ina == NULL)
			exit(-1);

		DEV_IFINDEX(ifreq.ifr_ifindex).broadcast = ina->sin_addr;

		DEV_IFINDEX(ifreq.ifr_ifindex).enabled = 1;

		if (this_host.nif >= MAX_NR_INTERFACES)
			break;

    } while ((iface = strtok(NULL, ",")));

    close(if_sock);

    /* Load kernel modules */
    load_modules(ifnames);

    /* Enable IP forwarding and set other kernel options... */
    if (set_kernel_options() < 0) {
	fprintf(stderr, "Could not set kernel options!\n");
	exit(-1);
    }
}
#endif // -ifndef CN_PORT

/* This signal handler ensures clean exits */
void signal_handler(int type)
{

    switch (type) {
    case SIGSEGV:
	alog(NULL, LOG_ERR, 0, __FUNCTION__, "SEGMENTATION FAULT!!!! Exiting!!! "
	     "To get a core dump, compile with DEBUG option.");
    case SIGINT:
    case SIGHUP:
    case SIGTERM:
    default:
	exit(0);
    }
}

#ifdef REMOTE_CTRL
int AODV_start(struct host_info * hi, int argc, char **argv) {
	// After starting, we will copy hi to hi-> The method finishes and (old)
	// hi will vanish forever.
	// UPDATE: NO, WE WILL NOT COPY as there are timer pointers to the old place
	// then!
	
#else
int main(int argc, char **argv)
{
	struct host_info hi_data;
	struct host_info* hi = &hi_data;
#endif
	//fprintf(stderr, "sizeof hi: %ld.\n", sizeof(hi));
	//fprintf(stderr, "&hi = %p\n", &hi);
    static char *ifname = NULL;	/* Name of interface to attach to */
#ifndef CN_PORT
    int daemonize = 0;
#endif
#ifndef REMOTE_CTRL
    int n, nfds = 0, i;
    fd_set rfds, readers;
    struct timeval *timeout;
    struct timespec timeout_spec;
#endif
    struct sigaction sigact;
    sigset_t mask, origmask;

#ifdef REMOTE_CTRL
	struct multihw_global_struct * mhi = hi->multihw_instance;
#endif
	memset(hi, 0, sizeof(struct host_info));
#ifdef REMOTE_CTRL
	hi->multihw_instance = mhi;
	hi->signature = AODV_STRUCT_SIG; // Aus aktuellem Anlass
#endif
	
	list_init(&hi->TQ);
	list_init(&hi->seekhead);
	list_init(&hi->rreq_records);
	list_init(&hi->rreq_blacklist);
	
    /* Remember the name of the executable... */
    hi->progname = strrchr(argv[0], '/');

	if (hi->progname)
		hi->progname++;
	else
		hi->progname = argv[0];
	
	
	hi->wait_on_reboot = 1;
	
	/* Global Variables: */
	//printf("&log_to_file=%p\n", &hi->log_to_file);
	hi->log_to_file = 0;
	hi->rt_log_interval = 0;	/* msecs between routing table logging 0=off */
	hi->unidir_hack = 0;
	hi->rreq_gratuitous = 0;
	hi->expanding_ring_search = 1;
	hi->internet_gw_mode = 0;
	hi->local_repair = 0;
	hi->receive_n_hellos = 0;
	hi->hello_jittering = 1;
	hi->ratelimit = 1;		/* Option for rate limiting RREQs and RERRs. */
	hi->qual_threshold = 0;
	hi->gw_prefix = 1;
	
	
	/* Dynamic configuration values */
	hi->active_route_timeout = ACTIVE_ROUTE_TIMEOUT_HELLO;
	hi->delete_period = DELETE_PERIOD_HELLO;
	hi->ttl_start = TTL_START_HELLO;
	hi->optimized_hellos = 0;
	hi->llfeedback = 0;
	
    /* Use debug output as default */
    hi->debug = 1;
	hi->log_file_fd = -1;
	hi->log_rt_fd = -1;
	hi->log_nmsgs = 0;

    memset (&sigact, 0, sizeof(struct sigaction));
    sigact.sa_handler = signal_handler;
        
    /* This server should shut down on these signals. */
    sigaction(SIGTERM, &sigact, 0);
    sigaction(SIGHUP, &sigact, 0);
    sigaction(SIGINT, &sigact, 0);
    
	memset(&mask, 0, sizeof(sigset_t));
    sigaddset(&mask, SIGTERM);
    sigaddset(&mask, SIGHUP);
    sigaddset(&mask, SIGINT);
    /* Only capture segmentation faults when we are not debugging... */
#ifndef DEBUG
    sigaddset(&mask, SIGSEGV);
#endif

    /* Block the signals we are watching here so that we can
     * handle them in pselect instead. */
    sigprocmask(SIG_BLOCK, &mask, &origmask);

    /* Parse command line: */
    while (1) {
	int opt;

	opt = getopt_long(argc, argv, "i:fjln:dghoq:r:s:uwxDLRV", longopts, 0);

	if (opt == EOF)
	    break;

	switch (opt) {
	case 0:
	    break;
#ifndef CN_PORT
	case 'd':
	    debug = 0;
	    daemonize = 1;
	    break;
#endif
	case 'f':
	    hi->llfeedback = 1;
	    hi->active_route_timeout = ACTIVE_ROUTE_TIMEOUT_LLF;
	    break;
	case 'g':
	    hi->rreq_gratuitous = !hi->rreq_gratuitous;
	    break;
	case 'i':
	    ifname = optarg;
	    break;
	case 'j':
	    hi->hello_jittering = !hi->hello_jittering;
	    break;
	case 'l':
	    hi->log_to_file = !hi->log_to_file;
	    break;
	case 'n':
	    if (optarg && isdigit(*optarg)) {
		hi->receive_n_hellos = atoi(optarg);
		if (hi->receive_n_hellos < 2) {
		    fprintf(stderr, "-n should be at least 2!\n");
		    exit(-1);
		}
	    }
	    break;
	case 'o':
	    hi->optimized_hellos = !hi->optimized_hellos;
	    break;
	case 'q':
	    if (optarg && isdigit(*optarg))
		hi->qual_threshold = atoi(optarg);
	    break;
	case 'r':
	    if (optarg && isdigit(*optarg))
		hi->rt_log_interval = atof(optarg) * 1000;
	    break;
	case 'u':
	    hi->unidir_hack = !hi->unidir_hack;
	    break;
	case 'w':
	    hi->internet_gw_mode = !hi->internet_gw_mode;
	    break;
	case 'x':
	    hi->expanding_ring_search = !hi->expanding_ring_search;
	    break;
	case 'L':
	    hi->local_repair = !hi->local_repair;
	    break;
	case 'D':
	    hi->wait_on_reboot = !hi->wait_on_reboot;
	    break;
	case 'R':
	    hi->ratelimit = !hi->ratelimit;
	    break;
	case 'V':
	    printf
		("\nAODV-UU v%s, %s  Uppsala University & Ericsson AB.\nAuthor: Erik Nordstrm, <erik.nordstrom@it.uu.se>\n\n",
		 AODV_UU_VERSION, DRAFT_VERSION);
	    exit(0);
	    break;
	case '?':
	case ':':
	    exit(0);
	default:
	    usage(0, hi->progname);
	}
    }
    /* Check that we are running as root */
#ifndef CN_PORT
	if (geteuid() != 0) {
		fprintf(stderr, "must be root\n");
		exit(1);
	}
#endif // -ifndef CN_PORT

    /* Detach from terminal */
#ifndef CN_PORT
    if (daemonize) {
		if (fork() != 0)
			exit(0);
		// Close stdin, stdout and stderr... 
		//  close(0); 
		close(1);
		close(2);
		setsid();
	}
    /* Make sure we cleanup at exit... */
    atexit((void *) &cleanup);
#endif // -ifndef CN_PORT

    /* Initialize data structures and services... */
    rt_table_init(hi);
    log_init(hi);
    packet_queue_init(hi);
#ifdef CN_PORT
	ri_if_init(hi, ifname);
#else
    host_init(ifname);
#endif
    /*   packet_input_init(); */
    nl_init(hi);
#ifndef CN_PORT
    nl_send_conf_msg();
#endif
    aodv_socket_init(hi);
#ifdef LLFEEDBACK
    if (hi->llfeedback) {
	llf_init();
    }
#endif
	
	init_packetizer(hi);
	
	
	// Hm, what is if there would be some sockets which make sense in
	// simulation? With this directive we simply ignore them.
#ifndef REMOTE_CTRL
    /* Set sockets to watch... */
    FD_ZERO(&readers);
    for (i = 0; i < hi->nr_callbacks; i++) {
	FD_SET(hi->callbacks[i].fd, &readers);
	if (hi->callbacks[i].fd >= nfds)
	    nfds = hi->callbacks[i].fd + 1;
    }
#endif

    /* Set the wait on reboot timer... */
    if (hi->wait_on_reboot) {
	timer_init(hi, &hi->worb_timer, wait_on_reboot_timeout, NULL);
	timer_set_timeout(hi, &hi->worb_timer, hi->delete_period);
	alog(hi, LOG_NOTICE, 0, __FUNCTION__,
	     "In wait on reboot for %d milliseconds. Disable with \"-D\".",
	     hi->delete_period);
    }

    /* Schedule the first Hello */
	if (!hi->optimized_hellos && !hi->llfeedback)
		hello_start(hi);

	if (hi->rt_log_interval)
		log_rt_table_init();
	
	hi->run = 1;
#ifdef REMOTE_CTRL
	// Start is done. Main loop (which simply waits for some packet and then
	// updates timers etc.) is replaced by the direct calls when trancv did
	// receive a packet.
	
	//// Move the whole unit data to the routing class
	//memcpy(hiP, &hi, sizeof(struct host_info));
	// IS ALREADY on right place.
	
	// Quit with success.
	return 0;
#else
	
    while (hi->run) {
		//printf("A.\n");
	memcpy((char *) &rfds, (char *) &readers, sizeof(rfds));

		//printf("Done. wait_on_reboot = %d.\n", hi->wait_on_reboot);
		timeout = timer_age_queue(hi);
		//struct timespec tmp;
		
		// We want to wake every second.
		struct timeval tv;
		//printf("nfds: %d.\n", nfds);
		if (timeout) {
			tv = * timeout;
		}
		else {
			gettimeofday(&tv, NULL);
			tv.tv_sec++;
		}
		timeout_spec.tv_sec = tv.tv_sec;
		timeout_spec.tv_nsec = tv.tv_usec * 1000;
		if ((n = pselect(nfds, &rfds, NULL, NULL, &timeout_spec, &origmask)) < 0) {
			if (errno != EINTR)
				alog(hi, LOG_WARNING, errno, __FUNCTION__,
						"Failed select (main loop)");
			continue;
		}

	if (n > 0) {
	    for (i = 0; i < hi->nr_callbacks; i++) {
		if (FD_ISSET(hi->callbacks[i].fd, &rfds)) {
		    /* We don't want any timer SIGALRM's while executing the
		       callback functions, therefore we block the timer... */
		    (*hi->callbacks[i].func) (hi, hi->callbacks[i].fd);
		}
	    }
	}
		//printf ("B.\n");
    }				/* Main loop */
    nl_cleanup(hi);
    return 0;
#endif
} // int main() or AODV_start()

#ifndef CN_PORT
static void cleanup()
{
    debug(hi, LOG_DEBUG, 0, "CLEANING UP!");
#ifdef CN_PORT
	debug(hi, LOG_ERR, 0, "");
#else
    rt_table_destroy(hi);
    aodv_socket_cleanup();
#ifdef LLFEEDBACK
#ifndef CN_PORT
    if (llfeedback)
	llf_cleanup();
#endif
#endif
    log_cleanup();
    nl_cleanup();
    remove_modules();
#endif // -ifdef CN_PORT -else
}
#endif // -ifndef CN_PORT
