/*****************************************************************************
 *
 * Copyright (C) 2001 Uppsala University & Ericsson AB.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Authors: Erik Nordstrm, <erik.nordstrom@it.uu.se>
 *
 *****************************************************************************/
#ifndef _PACKET_QUEUE_H
#define _PACKET_QUEUE_H

#ifndef NS_NO_GLOBALS

#include <stdint.h>

#include <netinet/ip.h>
#include "../list.h"
#include "../timer_queue.h"

#define MAX_QUEUE_LENGTH 512
#define MAX_QUEUE_TIME 12000 /* Maximum time packets can be queued (ms) */
#define GARBAGE_COLLECT_TIME 1000 /* Interval between running the
				   * garbage collector (ms) */

/* Verdicts for queued packets: */
enum {
    PQ_DROP = 0,
    PQ_SEND = 1,
    PQ_ENC_SEND = 2
};

#ifdef CN_PORT
struct unpacked_L2_header;
#define Packet struct unpacked_L2_header
#endif

struct q_pkt {
    list_t l;
    struct in_addr  dest_addr;
    struct timeval q_time;
#ifdef CN_PORT
	uint16_t unpacked_length;
#endif
    Packet *p;
};

struct packet_queue {
    list_t head;
    unsigned int len;
    struct timer garbage_collect_timer;
};

#endif				/* NS_NO_GLOBALS */

#ifndef NS_NO_DECLARATIONS

#ifdef CN_PORT
struct host_info;
#define HI_ARG_single struct host_info* hi
#define HI_ARG HI_ARG_single,
#define HI_USE_access hi->
#define HI_USE_arg hi,
#define HI_or_NULL hi
#else
struct packet_queue PQ;
#define HI_ARG 
#define HI_ARG_single void
#define HI_USE_access
#define HI_USE_arg
#define HI_or_NULL NULL
#endif

void packet_queue_add(HI_ARG    Packet * p, struct in_addr dest_addr, uint16_t unpacked_length);
void packet_queue_init(HI_ARG_single);
void packet_queue_destroy();
int packet_queue_set_verdict(HI_ARG    struct in_addr dest_addr, int verdict);
int packet_queue_garbage_collect(HI_ARG_single);

#endif				/* NS_NO_DECLARATIONS */

#endif
