/*****************************************************************************
 *
 * Copyright (C) 2018 Jonas Bechtel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Authors: Jonas Bechtel, <post@jjjjbechtel.de> (nospam: use jbechtel.de)
 *          
 *
 *****************************************************************************/


#ifndef SMALL_PACKET_C
#define SMALL_PACKET_C

#ifndef NOCODE
#include <stdbool.h>
#include <string.h> // memcmp
#endif
#include "cassert.h"

#include "defs.h"

// TODO: This is unused:
struct transport_flags {
	// from the last acknowledged sequence number till now have been
	// x failed packets (failed = thrown away speech because of real-time)
	// failed = 0 ->  0  = x
	//          1 ->  1  = x
	//          2 ->  2 <= x <=  3 
	//          3 ->  4 <= x <=  6
	//          4 ->  7 <= x <= 10
	//          5 -> 11 <= x <= 15
	//          6 -> 16 <= x <= 21
	//          7 -> 22 <= x <= 28
	uint8_t failed:3; 
	
};


#define PACKED_BROADCAST 0x0c

#define get_addr_byte(tgt, num, src) \
	memcpy(&(tgt), &(((uint8_t*)(&(src)))[num]), 1);

// tgt: 32-bit addr in network order (struct in_addr.s_addr)
void set_addr_byte(struct in_addr* tgt, uint8_t num, uint8_t src)
#ifndef NOCODE
{
	uint8_t* ptr = (uint8_t*)(&tgt->s_addr);
	memcpy(&(ptr[num]), &src, 1);
}
#endif
;


// small_packet stands for generic packets
// Meant to be similar to transport layer (TCP/UDP)
struct small_packet {
	uint8_t src;
	uint8_t dst;
	
	union {
		AODV_msg aodv;
		struct {
	uint8_t type;
	// Known by fm module
	//uint8_t size; // Counting everything after dest_addr.
	
	// Checksum: see /Teile/Routing/Checksum_calculation.txt
	uint8_t chk_hdr:3; // header is everything inclusive dest_addr but no data.
	uint8_t chk_amrA:5;
	//uint8_t chk_amrB;
	uint8_t ttl;
	
	//uint8_t hop_count; // do we need this?
	struct transport_flags transport;
	
	uint32_t dest_addr;
	uint32_t seq_nr_dest;
	
	uint32_t orig_addr;
	uint32_t seq_nr_orig;
	
	uint8_t data[1];
		};
	};
};





struct fieldbits {
	const char* name;
	uint8_t num_bits;
};
#define glue(a, b) a ## b
#define str(a) #a
#define fieldbts(a, b) const struct fieldbits glue(bts_, a) = {str(a), b}


//const struct fieldbits bts_FIN = {NULL, 0};
#define bts_FIN NULL

#ifndef NOCODE
fieldbts(ADDRESS, 4);
fieldbts(TTL, 4);
fieldbts(HDR_CHK, 3);
fieldbts(TOP_TYPE, 2);

fieldbts(AODV_TYPE, 2);
fieldbts(AODV_RREQ_FLAGS, 5);
fieldbts(AODV_RREP_FLAGS, 2);
fieldbts(AODV_RERR_FLAGS, 1);
fieldbts(AODV_HOP_COUNT, 4);
CASSERT((RREQ_ID_OVERFLOW <= 0x40), __max_6_bits_for_RREQ_ID__);
fieldbts(AODV_RREQ_ID, 6);
fieldbts(ADDRESS_DST, 4);
fieldbts(ADDRESS_ORI, 4);
CASSERT(SEQNO_OVERFLOW <= 0x1000, __max_12_bits_for_SEQNO__);
fieldbts(AODV_SEQ_MSB, 4);
fieldbts(AODV_SEQ_LSB, 8);
fieldbts(AODV_PREFIX_SIZE, 0);
#if BITS_AODV_PREFIX_SIZE != 0
#error Prefix not implemented
#endif
fieldbts(AODV_LIFETIME_VAL, 6);
fieldbts(AODV_LIFETIME_EXP, 2);
fieldbts(AODV_DEST_COUNT, 0);
#if BITS_AODV_PREFIX_SIZE != 0
#error Destination Count not implemented this way.
#endif

fieldbts(SEQ_TYPE, 1);
fieldbts(SEQ_DATA_SEQ_MSB, 2);
fieldbts(SEQ_DATA_SEQ_LSB, 8);
fieldbts(SEQ_DATA_SEQ_FOLLOW, 2);

fieldbts(AMR_CHK_CL_A, 5);
fieldbts(AMR_BYTE, 8);
fieldbts(AMR_CHK_CL_B, 8);

#endif // -ifndef NOCODE

#define TOP_TYPE_AODV 0
#define TOP_TYPE_SEQ 1
#define TOP_TYPE_SEQ_FOLLOW 2
#define TOP_TYPE_ALOHA 3
#define AODV_TYPE_RREQ 1
#define AODV_TYPE_RREP 2
#define AODV_TYPE_RERR 3
#define AODV_TYPE_RREP_ACK 0 // = 4
#define SEQ_TYPE_AMR 0
#define SEQ_TYPE_GENERIC 1




struct unpacked_L2_header {
	uint8_t station_addr;
	uint8_t nexthop_addr;
	uint8_t ttl;
	uint8_t header_chksum;
	uint8_t top_type;
};


#ifndef NOCODE
const struct fieldbits * bits_rreq[] = {
	&bts_ADDRESS,
	&bts_ADDRESS_DST,
	&bts_TTL,
	&bts_HDR_CHK,
	&bts_TOP_TYPE,
	&bts_AODV_TYPE,
	&bts_AODV_RREQ_FLAGS,
	&bts_AODV_HOP_COUNT,
	&bts_AODV_RREQ_ID,
	// NOT &bts_ADDRESS,
	&bts_AODV_SEQ_MSB,
	&bts_AODV_SEQ_LSB,
	&bts_ADDRESS_ORI,
	&bts_AODV_SEQ_MSB,
	&bts_AODV_SEQ_LSB,
	bts_FIN
}
#endif // -ifndef NOCODE
;

struct unpacked_rreq {
	struct unpacked_L2_header L2;
	uint8_t AODV_type;
	uint8_t flags;
	uint8_t hop_count;
	uint8_t rreq_id;
	// NOT uint8_t dst_addr;
	uint8_t dst_seq_msb;
	uint8_t dst_seq_lsb;
	uint8_t ori_addr;
	uint8_t ori_seq_msb;
	uint8_t ori_seq_lsb;
};

#ifndef NOCODE
const struct fieldbits * bits_rrep[] = {
	&bts_ADDRESS,
	&bts_ADDRESS,
	&bts_TTL,
	&bts_HDR_CHK,
	&bts_TOP_TYPE,
	&bts_AODV_TYPE,
	&bts_AODV_RREP_FLAGS,
	// NO prefix size
	&bts_AODV_HOP_COUNT,
	&bts_ADDRESS,
	&bts_AODV_SEQ_MSB,
	&bts_AODV_SEQ_LSB,
	&bts_ADDRESS,
	&bts_AODV_LIFETIME_VAL,
	&bts_AODV_LIFETIME_EXP,
	bts_FIN
}
#endif // -ifndef NOCODE
;

struct unpacked_rrep {
	struct unpacked_L2_header L2;
	uint8_t AODV_type;
	uint8_t flags;
	uint8_t hop_count;
	uint8_t dst_addr;
	uint8_t dst_seq_msb;
	uint8_t dst_seq_lsb;
	uint8_t ori_addr;
	uint8_t lifetime_val;
	uint8_t lifetime_exp;
};

#ifndef NOCODE
const struct fieldbits * bits_rerr[] = {
	&bts_ADDRESS,
	&bts_ADDRESS,
	&bts_TTL,
	&bts_HDR_CHK,
	&bts_TOP_TYPE,
	&bts_AODV_TYPE,
	&bts_AODV_RERR_FLAGS,
	&bts_ADDRESS,
	&bts_AODV_SEQ_MSB,
	&bts_AODV_SEQ_LSB,
	// For the further destination errors there has to be some additional
	// mechanism.
	bts_FIN
}
#endif // -ifndef NOCODE
;

struct unpacked_rerr {
	struct unpacked_L2_header L2;
	uint8_t AODV_type;
	uint8_t flags;
	uint8_t dst_addr;
	uint8_t dst_seq_msb;
	uint8_t dst_seq_lsb;
};


#ifndef NOCODE
const struct fieldbits * bits_rrep_ack[] = {
	&bts_ADDRESS,
	&bts_ADDRESS,
	&bts_TTL,
	&bts_HDR_CHK,
	&bts_TOP_TYPE,
	&bts_AODV_TYPE,
	bts_FIN
}
#endif // -ifndef NOCODE
;

struct unpacked_rrep_ack {
	struct unpacked_L2_header L2;
	uint8_t AODV_type;
};

#define SEQ_W_TYPE \
	/* L2 */ \
	&bts_ADDRESS, \
	&bts_ADDRESS, \
	&bts_TTL, \
	&bts_HDR_CHK, \
	&bts_TOP_TYPE, \
	/* SEQ */ \
	&bts_SEQ_TYPE

#ifndef NOCODE
const struct fieldbits * bits_seq_w_type[] = {
	SEQ_W_TYPE,
	bts_FIN
};
#endif
struct unpacked_seq_w_type {
	struct unpacked_L2_header L2;
	uint8_t seq_type;
};

#ifndef NOCODE
const struct fieldbits * bits_amrlong[] = {
	SEQ_W_TYPE,
	&bts_ADDRESS_DST,
	&bts_ADDRESS_ORI,
	//&bts_AODV_HOP_COUNT,
	&bts_SEQ_DATA_SEQ_MSB,
	&bts_SEQ_DATA_SEQ_LSB,
	// AMR
	&bts_AMR_CHK_CL_A,
	&bts_AMR_BYTE,
	&bts_AMR_BYTE,
	&bts_AMR_BYTE,
	&bts_AMR_BYTE,
	&bts_AMR_BYTE,
	&bts_AMR_BYTE,
	&bts_AMR_BYTE,
	&bts_AMR_BYTE,
	&bts_AMR_BYTE,
	&bts_AMR_BYTE,
	&bts_AMR_BYTE,
	&bts_AMR_BYTE,
	&bts_AMR_CHK_CL_B,
	bts_FIN
};
#endif // -ifndef NOCODE

#undef SEQ_W_TYPE

struct unpacked_amrlong {
	struct unpacked_L2_header L2;
	uint8_t seq_type;
	uint8_t dst_addr;
	uint8_t ori_addr;
	//uint8_t hop_count;
	uint8_t data_seq_msb;
	uint8_t data_seq_lsb;
	uint8_t chk_amrA;
	uint8_t amrbytes[12];
	uint8_t chk_amrB;
};



#ifndef NOCODE
const struct fieldbits * bits_amrshort[] = {
	&bts_ADDRESS,
	&bts_ADDRESS,
	&bts_HDR_CHK,
	&bts_TOP_TYPE,
	// SEQ
	&bts_SEQ_DATA_SEQ_FOLLOW,
	// AMR
	&bts_AMR_CHK_CL_A,
	&bts_AMR_BYTE,
	&bts_AMR_BYTE,
	&bts_AMR_BYTE,
	&bts_AMR_BYTE,
	&bts_AMR_BYTE,
	&bts_AMR_BYTE,
	&bts_AMR_BYTE,
	&bts_AMR_BYTE,
	&bts_AMR_BYTE,
	&bts_AMR_BYTE,
	&bts_AMR_BYTE,
	&bts_AMR_BYTE,
	&bts_AMR_CHK_CL_B,
	bts_FIN
}
#endif // -ifndef NOCODE
;

struct unpacked_amrshort {
	struct unpacked_L2_header L2;
	uint8_t data_seq_follow;
	uint8_t chk_amrA;
	uint8_t amrbytes[12];
	uint8_t chk_amrB;
};


uint8_t get_packed_size(const struct fieldbits** fields)
#ifndef NOCODE
{
	int i = -1;
	uint8_t s = 0;
	while (fields[++i] != bts_FIN) {
		//printf("%d, == ? %c\n", i, fields[i] == bts_FIN ? 'Y':'N');
		s+=fields[i]->num_bits;
	}
	return (s+7)/8; // Aufrunden
}
#endif
;


// Returns 0xFF00 in case of error.
uint16_t get_destination_address(struct unpacked_L2_header* data, const struct fieldbits** fields)
#ifndef NOCODE
{
	int i = 0;
	while (fields[i] != NULL) {
		if (fields[i] == &bts_ADDRESS_DST) {
			return ((uint8_t*)data)[i];
		}
		i++;
	}
	return 0xFF00;
}
#endif // -ifndef NOCODE
;


// Returns 0xFF00 in case of error.
uint16_t get_origin_address(struct unpacked_L2_header* data, const struct fieldbits** fields)
#ifndef NOCODE
{
	int i = 0;
	while (fields[i] != NULL) {
		if (fields[i] == &bts_ADDRESS_ORI) {
			return ((uint8_t*)data)[i];
		}
		i++;
	}
	return 0xFF00;
}
#endif // -ifndef NOCODE
;


const struct fieldbits** get_type(hi hi, const struct unpacked_L2_header* data)
#ifndef NOCODE
{
	if (data->top_type == TOP_TYPE_AODV) {
		struct unpacked_rerr* p = (struct unpacked_rerr*) data;
		if (p->AODV_type == AODV_RREQ)
			return hi->bits_rreq;
		if (p->AODV_type == AODV_RREP)
			return hi->bits_rrep;
		if (p->AODV_type == AODV_RERR)
			return hi->bits_rerr;
		if (p->AODV_type == AODV_RREP_ACK)
			return hi->bits_rrep_ack;
		return NULL; // This cann't happen.
	}
	if (data->top_type == TOP_TYPE_SEQ) {
		struct unpacked_amrlong* p = (struct unpacked_amrlong*)data;
		if(p->seq_type == SEQ_TYPE_AMR)
			return hi->bits_amrlong;
		return NULL; // generic packets not supported yet.
	}
	if (data->top_type == TOP_TYPE_SEQ_FOLLOW) {
		return NULL; // Hm, what? This tells us only history. // TODO: Make a history database and load the appropriate packet type from there.
	}
	return NULL; // This is not implemented.
}
#endif // -ifndef NOCODE
;



void init_packetizer(hi hi)
#ifndef NOCODE
{
	hi->bits_rreq = bits_rreq;
	hi->bits_rrep = bits_rrep;
	hi->bits_rrep_ack = bits_rrep_ack;
	hi->bits_rerr = bits_rerr;
	hi->bits_seq_w_type = bits_seq_w_type;
	hi->bits_amrlong = bits_amrlong;
	hi->bits_amrshort = bits_amrshort;
	
#if NR_AMR_CONNECTIONS != 2
#error     This is unexpected.
#endif
	hi->ac[0].origin = PACKED_BROADCAST;
	hi->ac[1].origin = PACKED_BROADCAST;
}
#endif
;


// data must be readable for one additional bit!
// returns the length of the used bytes of the target_struct pointer.
uint16_t bitunpack(struct unpacked_L2_header* target_struct, void* data, const struct fieldbits** rules)
#ifndef NOCODE
{
	uint8_t * tgt = (uint8_t*) target_struct;
	int i = 0;
	int dcnt = 0;
	
	//fprintf(stderr, "rules = %17p   \n", rules);
	//fprintf(stderr, "bits_rreq = %17p   \n", bits_rreq);
	
	/*int  x = 0;
	while (x < sizeof(bits_rreq) / sizeof(void*)) {
		fprintf(stderr, "bits_rreq[%d] = %17p   ", x, bits_rreq[x]);
		if (bits_rreq[x] == NULL) break;
		fprintf(stderr, "bits_rreq[%d]->name = %17p\n", x, bits_rreq[x]->name);
		x++;
	}*/
	
	int loaded = 0;
	uint16_t buffer = 0;
	
	while (true) {
		if (rules[i] == NULL || rules[i]->name == NULL) 
			// We are done.
			break;
		//fprintf(stderr, "rules[%d] = %17p   ", i, rules[i]);
		//fprintf(stderr, "rules[%d]->name = %17p\n", i, rules[i]->name);
		
		if (loaded < rules[i]->num_bits) {
			buffer |= (uint16_t)(((uint8_t*)data)[dcnt]) << (8-loaded);
			//printf("Loaded: %02x\n", ((uint8_t*)data)[dcnt]);
			dcnt++;
			loaded += 8;
		}
		
		
		
		// We take the bits from the MSB first.
		tgt[i] = buffer >> (16 - rules[i]->num_bits);
		//printf ("%-15s: %dbit   0x%02x (%d)\n", rules[i]->name, rules[i]->num_bits, tgt[i], tgt[i]);
		buffer <<= rules[i]->num_bits;
		loaded -= rules[i]->num_bits;
		i++;
	}
	return i;
}
#endif // -ifndef NOCODE
;


uint8_t bitpack(void* target_data, struct unpacked_L2_header* unpacked, const struct fieldbits* rules[], uint8_t max_out_len)
#ifndef NOCODE
{
	uint8_t * src = (uint8_t*) unpacked;
	uint8_t * tgt = (uint8_t*) target_data;
	int i = 0;
	int dcnt = 0;
	
	
	int loaded = 0;
	uint16_t buffer = 0;
	
	while (true) {
		
		if (loaded >= 8) {
			tgt[dcnt++] = buffer >> 8;
			buffer <<=8;
			loaded -= 8;
		}
		
		if (rules[i] == NULL || rules[i]->name == NULL) {
			if (loaded) {
				tgt[dcnt++] = buffer >> 8;
			}
			// We are done.
			break;
		}
		
		uint8_t d = src[i] & ((1 << rules[i]->num_bits) - 1);
		buffer |= d << (16 - loaded - rules[i]->num_bits);
		loaded += rules[i]->num_bits;
		
#ifdef DEBUG_FOR_VALGRIND
		fprintf(stderr, "pa: src[%d] = 0x%02x\n", i, src[i]);
#endif
		//printf("loaded=%d. buffer=%04x\n", loaded, buffer);
		
		i++;
	};
	return dcnt;
}
#endif // -ifndef NOCODE
;



void test_unpack_pack(hi hi, void* data, uint8_t len)
#ifndef NOCODE
{
	
	
	// TODO: completely weired, but here for testing:
	
	uint8_t small_rreq[] = {
		0x45, //St.IP 4, DST.IP 5, 
		0xe2, //chk 7, Top Type 0, AODV Type 1, Flags 0
		0x04, //[Flags], Hop count 4,
		0xC8, // RREQ ID 0x32, DST SEQ 0x097
		0x25, // [DST SEQ]
		0xC5, // [DST SEQ], ORI IP 1, ORI SEQ 0x421
		0x08, // [ORI SEQ]
		0x40, // [ORI SEQ], 6xfill0
	};
	
	struct unpacked_rreq tgt;
	
	/* completely weired: it doesn't take the right struct.
	 * This is some strange compiler bug, does it happen accross all
	 * Fedora compilers?
	fprintf(stderr, "bits_rreq = %p.\n", bits_rreq);
	fprintf(stderr, "&bits_rreq = %p.\n", &bits_rreq);
	fprintf(stderr, "hi->bits_rreq = %p.\n", hi->bits_rreq);
	fprintf(stderr, "&hi->bits_rreq = %p.\n", &hi->bits_rreq);
	 */
	
	bitunpack((void*)&tgt, (void*)small_rreq, hi->bits_rreq);
	
	int i = 0;
	printf("0x...   ");
	while (i < sizeof(tgt)) {
		printf(" %02x", ((uint8_t*)&tgt)[i]);
		i++;
	}
	printf("\n");
	
	// And now unpack it againg...
	uint8_t packed[8];
	bitpack(packed, &tgt.L2, hi->bits_rreq, sizeof (packed));
	
	i = 0;
	printf("0x...   ");
	while (i < sizeof(packed)) {
		printf(" %02x", packed[i]);
		i++;
	}
	printf("\n");
	
	
	// TODO: Further
}
#endif // -ifndef NOCODE
;




#endif // -ifndef SMALL_PACKET_C

