/*****************************************************************************
 *
 * Copyright (C) 2018 Jonas Bechtel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Authors: Jonas Bechtel, <post@jjjjbechtel.de> (nospam: use jbechtel.de)
 *          
 *
 *****************************************************************************/

// (c) 2018 Jonas Bechtel


#ifndef ROUTING_TABLE_C
#define ROUTING_TABLE_C

#include <stdint.h>
#include "timer_queue.h"


struct TTLrec {
	uint8_t ori_addr; // This field won't be used. It would be useful for a different kind of list.
#define TTL_INVALID 0xff
	uint8_t ttl_seen;
	struct timer broadcast_invalidate;
	//struct timeval last_broadcast;
};


struct rt_table;

#ifndef NOCODE
#include "defs.h" // _Only_ as support for type AODV_msg in debug.h!
#include "debug.h"

#include "params.h"
#include "routing_table.h"
#define NOCODE
#include "packetizer.c"
#include "aodv_timeout.c"
#undef NOCODE
#endif


struct rt_table *NS_CLASS rt_table_insert_short(hi hi,
				uint8_t dst, uint8_t next,
				uint8_t hops, uint32_t seqno,
				uint32_t life, uint8_t state,
				uint16_t flags, unsigned int ifindex)
#ifndef NOCODE
{
	struct in_addr dest_addr, next_addr;
	
	set_addr_byte(&dest_addr, 0, 192);
	set_addr_byte(&dest_addr, 1, 168);
	set_addr_byte(&dest_addr, 2, 1);
	set_addr_byte(&dest_addr, 3, dst);
	
	next_addr = dest_addr;
	set_addr_byte(&next_addr, 3, next);
	
	return rt_table_insert(hi, dest_addr, next_addr, hops, seqno, life, state, flags, ifindex);
}
#endif
;


struct rt_table *NS_CLASS rt_table_update_short(hi hi,
				struct rt_table * rt, uint8_t next,
				uint8_t hops, uint32_t seqno,
				uint32_t life, uint8_t state,
				uint16_t flags)
#ifndef NOCODE
{
	struct in_addr next_addr;
	
	set_addr_byte(&next_addr, 0, 192);
	set_addr_byte(&next_addr, 1, 168);
	set_addr_byte(&next_addr, 2, 1);
	set_addr_byte(&next_addr, 3, next);
	
	return rt_table_update(hi, rt, next_addr, hops, seqno, life, state, flags);
}
#endif
;



struct rt_table *NS_CLASS rt_table_find_short(hi hi, uint8_t dest_addr)
#ifndef NOCODE
{
	
	struct in_addr long_dest;
	
	if (dest_addr == PACKED_BROADCAST) {
		debug(hi, LOG_WARNING, 0, "I think you are wrong searching the routing table for a broadcast address.");
		long_dest.s_addr = AODV_BROADCAST;
	}
	else {
		set_addr_byte(&long_dest, 0, 192);
		set_addr_byte(&long_dest, 1, 168);
		set_addr_byte(&long_dest, 2, 1);
		set_addr_byte(&long_dest, 3, dest_addr);
	}
	//printf("long_dest = %08x\n", long_dest.s_addr);
	return rt_table_find(hi, long_dest);
}
#endif // -ifndef NOCODE
;


// This is BC-TTL-REC
void set_ttl_broadcast(hi hi, struct rt_table* rt, uint8_t ttl)
#ifndef NOCODE
{
	if (!rt) return;
	
	//debug(hi, LOG_DEBUG, 0, "Set ttl to %d.\n", ttl);
	
	struct TTLrec* ttl_bc = &rt->ttl_broadcast;
	
	/*struct timeval new_timeout;
	gettimeofday(&new_timeout, NULL);
	timeval_add_msec(&new_timeout, TTL_BCLOOP_TIMEOUT);*/
	
	if (ttl_bc->ttl_seen != TTL_INVALID) {
		if (ttl == TTL_INVALID) {
			// Remove timeout
			timer_remove(&ttl_bc->broadcast_invalidate);
			//printf("Broadcast loop detection timer %p removed.\n", ttl_bc->broadcast_invalidate);
		}
		else {
			// Refresh timeout
			timer_set_timeout(hi, &ttl_bc->broadcast_invalidate, TTL_BCLOOP_TIMEOUT);
			//printf("Broadcast loop detection timer %p refreshed.\n", ttl_bc->broadcast_invalidate);
		}
	}
	else if (ttl != TTL_INVALID) {
		// Create timeout
		timer_init(hi,
				&ttl_bc->broadcast_invalidate,
				&bcloop_detection_timeout, ttl_bc);
		timer_set_timeout(hi, &ttl_bc->broadcast_invalidate, TTL_BCLOOP_TIMEOUT);
		//printf("Broadcast loop detection timer %p initialized.\n", ttl_bc->broadcast_invalidate);
	}
	
	
	ttl_bc->ttl_seen = ttl;
	return;
}
#endif
;
/*
// Makes some check.
uint8_t get_ttl_broadcast(hi hi, struct rt_table* rt) {
	if ()
}
*/

#endif // -ifndef ROUTING_TABLE_C



