#!/bin/sh
# Licence see below

# usage: give some bytes in stdin, you'll get the packet in stdout


# Generate following sample packets and print them as hex:
# (Format: Suuuuuuuu|uuuuuuuu|\n where u are hex bytes.)


# A=1, B=2, C=3

# A -> B: "Madonna"
# A -> C: "Santana"
# B -> A: "Beethoven"
# B -> C: "Wakeman"
# C -> A: "Ciccone"
# C -> B: "Olvera"


name2number() {
	if [ "_$1" = _A ]; then
		printf 0101a8c0
		# This is 192.168.1.1
	elif [ "_$1" = _B ]; then
		printf 0201a8c0
		# This is 192.168.1.2
	elif [ "_$1" = _C ]; then
		printf 0301a8c0
		# This is 192.168.1.3
	fi
}

gen_structure() {
	printf 06
	printf 55     # checksums
	printf AA     # checksum
	printf FF'|'  # flags
	name2number "$1"
	printf '|'00000000'|'
	name2number $weare
	printf '|'00000000'|'
}

hexpacket() {
	export LANG=C
	export LC_ALL=C
	p="`cat`"
	i=0
	while [ ${#p} -ge 4 ]; do
		part="`cat <<<"$p" | dd bs=4 count=1 2>/dev/null`"
		p="`cat <<<"$p" | dd bs=4 skip=1 2>/dev/null`"
	#	printf >&2 "part='%s', rest='%s'\n" "$part" "$p"
		xxd -g 4 <<<"$part" | sed -r 's/^0+: ([0-9a-f]{8}).*/\1|/' | tr -d '\n'
		#printf >&2 "   -> part='%s'\n" "$part"
		i=$((i+1))
	done
	xxd -g ${#p} <<<"$p" | sed -r 's/^0+: ([0-9a-f]+) .*/\1/'
	echo $(($i*4+${#p})) > len.temporary
}

complete() {
	data="`hexpacket`"
	len=`cat len.temporary`
	
	#printf P
	printf S
	gen_structure $1 
	printf "%s\n" "$data"
	#printf "Packet of len %d: %s\n" $len "$data"
}

# A -> B: "Madonna"
# A -> C: "Santana"
# B -> A: "Beethoven"
# B -> C: "Wakeman"
# C -> A: "Ciccone"
# C -> B: "Olvera"

weare=A; echo From Station A:
echo Madonna   | complete B
echo Santana   | complete C
weare=B; echo From Station B:
echo Beethoven | complete A
echo Wakeman   | complete C
weare=C; echo From Station C:
echo Coccone   | complete A
echo Olvera    | complete B





#*****************************************************************************
#*
#* Copyright (C) 2018 Jonas Bechtel
#*
#* This program is free software; you can redistribute it and/or modify
#* it under the terms of the GNU General Public License as published by
#* the Free Software Foundation; either version 2 of the License, or
#* (at your option) any later version.
#*
#* This program is distributed in the hope that it will be useful,
#* but WITHOUT ANY WARRANTY; without even the implied warranty of
#* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#* GNU General Public License for more details.
#*
#* You should have received a copy of the GNU General Public License
#* along with this program; if not, write to the Free Software
#* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#*
#* Authors: Jonas Bechtel, <post@jjjjbechtel.de> (nospam: use jbechtel.de)
#*          
#*
#*****************************************************************************/

