

// © 2018 Jonas Bechtel
// Licensed under the Apache License, Version 2.0; see file <prj>/meta/LICENSE



#include <stdio.h>


#include "assert.h"

#define DEFINE_OBJECT // DEFINE_OBJECT is to be included only once in 
#include "AODVintern.h"  //    all routing algorithm's code.
#undef DEFINE_OBJECT

#define NOCODE
#include "cli-dev.c"
#undef NOCODE


void inbound_raw_AODV(struct route_instance* ri, uint64_t time, uint8_t* data, uint8_t len, float power_in_mW) {
	process_received_packet(&ri->aodvuudd, data, len, power_in_mW);
}


// Returns: new sequence number of to-be-processed packet.
uint16_t outbound_unpacked_AODV(struct route_instance* ri, uint8_t* data, uint8_t len, uint8_t destination) {
	return put_amr_packet(&ri->aodvuudd, data, len, destination);
}

void service_task_AODV(struct route_instance * ri) {
	timer_age_queue (&ri->aodvuudd);
}



int start_AODVroute(struct route_instance * ri) {
	
	ri->inbound_raw = inbound_raw_AODV;
	ri->outbound_unpacked = outbound_unpacked_AODV;
	ri->service_task = service_task_AODV;
	ri->inbound_unpacked = NULL; // To be set by application
	ri->outbound_raw = NULL; // To be set by application
	
	// TODO: allow different variants.
	// Now start the AODV-uu-dd implementation
	char iparg[4] = "ipX";
	if (ri->station->name == 'W')    iparg[2] = '0' + 1;
	if (ri->station->name == 'B')    iparg[2] = '0' + 2;
	if (ri->station->name == 'C')    iparg[2] = '0' + 3;
	if (ri->station->name == 'D')    iparg[2] = '0' + 4;
	if (ri->station->name == 'O')    iparg[2] = '0' + 5;
#define NUM_ARGS 50
	char * argv[NUM_ARGS];
	int argc = 0;
	argv[argc++] = "//PROGNAME";
	//argv[argc++] = "-o";
	argv[argc++] = "-i";
	argv[argc++] = iparg;
	argv[argc++] = "-L"; // Local repair
	argv[argc++] = "-R"; // Toggle off rate limit.
	for (int i = 0; i < argc; i++) {
		
	}
	
#ifdef SIMUL
	ri->aodvuudd.multihw_instance = ri->station->mgs;
#endif
	int res = AODV_start(&ri->aodvuudd, argc-1, argv);
	assert2(ri->aodvuudd.signature == AODV_STRUCT_SIG, "YOUR_COMPILE_DEFINES_ARE_WRONG");
	//printf("&dev[0] = %p (correct)\n", &ri->aodvuudd.devs[0]);
	
	if (res != 0)
		// TODO: Cleanup/destroy
		return OBJECT_START_FAIL;
	
	
	// Now it goes on. Either put_amr_packet() or process_received_packet()
	// is to be called.
	
	return OBJECT_START_OK;
}


