

// © 2018 Jonas Bechtel
// Licensed under the Apache License, Version 2.0; see file <prj>/meta/LICENSE



//#define DEBUG_DEMOD
//#define DEBUG_DEMOD_DETAIL



#ifndef DIRECT_FUNCTION_INCLUDE
#error You cannot include or compile this. It belongs to trancv.c.
#endif

static void put_on_fifo(struct trancv_instance * ti, uint8_t data);



// Demodulation is not too easy.
// As there is the simulation possibility to read parts of packets this method
// can easily return NULL. Some bunch of intermediate data and temporary receiver
// state may be saved in ti and ensures the operation for subsequent calls to
// demodulate data.

	// TODO: Respect the filter settings; they may be too narrow and thus
	//  hinder a successful receive.
	
	// TODO: Emulate AFC
	// TODO: Emulate actual gain + AGC
	
// REMARK: demodulate_data will be called _after_ the wave segment has
//         been transmitted, so the starting time of the wave segment
//         is in the past.
static void demodulate_data_RFM69(struct trancv_instance * ti,
		struct wave_seq_segment inseg) {
	struct station * mh_station_arg = ti->station; // Needed for ulog.
	
	if (!inseg.waves) return;
	
	//printf("inseg.waves->refcount = %d.\n", inseg.waves->refcount);
	
	
	// Demodulate in following steps:
	//		1. Quantize (so make a zero/one out of the frequency)
	//		2. Bit syncronize
	//		3. Prozess (find preamble, sync word, etc.)
	
	
	
	
	
	if (inseg.airinfo.dist) {
		inseg.starttime += 100*inseg.airinfo.dist;
		inseg.airinfo.dist = 0;
	}
	
	
	
	
	
	// STEP 1: Quantize
	
	
	
	int i = 0;
	bool startbit; // true if the first bit is '1'.
	// bittime[2*x]   denote the times when bit is  startbit (x is a natural number)
	// bittime[2*x+1] denote the times when bit is ~startbit (x is a natural number)
	
	int array_cap = (inseg.waves->capacity - inseg.num_second_element)*3 + 1 + 1;
	// *3 as every single wave could possibly be random and thus lead to three
	// bittimes. [TODO: would be nice to have this allocated dynamically]
	
	uint32_t * bittime = malloc(sizeof(uint32_t) * (array_cap+1)); // + 1 as one never knows
	if (!bittime) {
		printf("demodulate_data_RFM69(): ERROR: no memory!\n");
		return;
	}
	
	//ulog_str_num_BASE_str_L("  inseg.num_second_element=", inseg.num_second_element, DEC, "\n");
	//ulog_str_num_BASE_str_L("  inseg.waves->capacity=", inseg.waves->capacity, DEC, "\n");
	
	
	
	uint32_t fr = get_frequency_silent(ti);
	uint32_t dev = get_deviation_silent(ti);
	
	// I don't know what the exact frequencies these devices take in order
	// to work properly. I assume it goes from 0.2*deviation to 2.0*deviation.
	
	// TODO: There's a frequency and a deviation filter, take them into account!
	// TODO: exchange '0' and '1'?
	uint32_t fr_0_min = fr -  2*dev;       // 2.0*deviation
	uint32_t fr_0_max = fr - (2*dev)/10;   // 0.2*deviation
	uint32_t fr_1_min = fr + (2*dev)/10;   // 0.2*deviation
	uint32_t fr_1_max = fr +  2*dev;       // 2.0*deviation
	
	struct waveblip recent_blip = inseg.first_element;
	if (isnan(recent_blip.amplitude)) {
		if (inseg.num_second_element >= inseg.waves->capacity) {
			
			clean_wave_seq_segment(&inseg);
			
			return;
		}
		recent_blip = inseg.waves->seq[inseg.num_second_element];
		inseg.num_second_element++;
			
	}
	
	if (fr_0_min <= recent_blip.frequency && recent_blip.frequency <= fr_0_max) {
		// This is zero.
		startbit = false;
	}
	else 
		// TODO: In case the frequency cann't be assigned we should use random
		// data here.
		startbit = true;
	
	bittime[0] = 0;
	
	uint32_t segment_duration = inseg.segment_duration;
	
	float rec_pwr_1 = recent_blip.amplitude;
	float rec_pwr_2 = recent_blip.amplitude;
	
	double total_dBms = 0;
	double exp_factor_precalc = ti->exp_factor_precalc;
	
	while (true) {
		assert (i < array_cap-1);
		
		if (isnan(recent_blip.amplitude)) {
//#ifdef DEBUG_DEMOD_DETAIL
			ulog_str_L("FIN (nan)\n");
//#endif
			break;
		}
		
#ifndef RFM_THRESHOLD
		//float threshold_dBm = -20; // very few packets arrive
		float threshold_dBm = -100;
		//float threshold_dBm = -1000; // many packet arrive
#else
		float threshold_dBm = RFM_THRESHOLD;
#endif
		int amp = 10.0 * log10f(recent_blip.amplitude);
		
		if (recent_blip.amplitude == 0) {
			// something is wrong.
#ifdef DEBUG_DEMOD_DETAIL
			ulog_str_L("demod.c: WARNING: Something is wrong. Got zero-amplitude wave.\n");
#endif
		}
		
		float dur = recent_blip.duration;
		dur /= (float)(1 SEC);
		// https://de.wikipedia.org/wiki/Exponentialfunktion#Basiswechsel
		total_dBms += dur * exp(recent_blip.amplitude * exp_factor_precalc);
		
#ifdef DEBUG_DEMOD
		ulog_str_num_BASE_str_L("blip.dur.=", recent_blip.duration, DEC, ", ");
		log_str_num_BASE_str_L("blip.freq.=", recent_blip.frequency, DEC, ", ");
		log_str_num_BASE_str_P("blip.amp.=", amp, DEC, amp < threshold_dBm ? "dBm!!" : "dBm\n");
#endif
		
		//float threshold = exp10f(threshold_dBm/10.0);
		if (amp < threshold_dBm) {
			recent_blip.amplitude = 0;
		}
		
		if (recent_blip.amplitude == 0) {
			// Here we receive absolute random waves from air.
			
			// Notify following code about the randomness:
			recent_blip.frequency = fr;
		}
		
		
		// This will be somewhat optimized by compiler.
		// (e. g. that fr_0_* < fr_1_* if low frequency = 0 and high frequency = 1)
		
		if (fr_0_min <= recent_blip.frequency && recent_blip.frequency <= fr_0_max) {
			// This is zero.
			if (!((i%2 == 0) ^ (startbit == false))) {
				// prolongue the duration
				bittime[i] += recent_blip.duration;
			}
			else {
				// different bit -> new bit
				i++;
				bittime[i] = recent_blip.duration;
			}
		}
		else if (fr_1_min <= recent_blip.frequency && recent_blip.frequency <= fr_1_max) {
			// This is one.
			
			if (!((i%2 == 0) ^ (startbit == true))) {
				// prolongue the duration
				bittime[i] += recent_blip.duration;
			}
			else {
				// different frequency -> new bit
				i++;
				bittime[i] = recent_blip.duration;
			}
		}
		else {
			
			// If we would simulate in correct manner, we would have to
			// respect that the frequency changes very often.
			
			// One "gentle" aproach to realize this while not destroying
			// the i odd/even order is to make 3 bits out of this blip.
			// [I guess even an even number of bits would work]
			// [For the sake of memory usage we only use 3 instead of 100
			// bits.]
			
			
			
			// TODO: As the frequency cann't be assigned we should use random
			// data here. But WARNING: This would make the simulation
			// undeterministic!
			bittime[i] += recent_blip.duration/2;
			bittime[i+1] = recent_blip.duration/4;
			//bittime[i+2] = recent_blip.duration/4;
			bittime[i+2] = recent_blip.duration
						- recent_blip.duration/2 - bittime[i+1];
			// To be exact and meet the nexttime.
			i+= 2;
		}
		
		if (inseg.num_second_element >= inseg.waves->capacity) {
			clean_wave_seq_segment(&inseg);
			break;
		}
		recent_blip = inseg.waves->seq[inseg.num_second_element];
		if (is_waveblip_invalid(recent_blip)) {
			clean_wave_seq_segment(&inseg);
			break;
		}
		inseg.num_second_element++;
		rec_pwr_2 = rec_pwr_1;
		rec_pwr_1 = recent_blip.amplitude;
		
#ifdef DEBUG_DEMOD
		ulog_str_num_BASE_str_L("    segment_duration=", inseg.segment_duration, DEC, "\n");
		ulog_str_num_BASE_str_L("    recent_blip.duration=", recent_blip.duration, DEC, "\n");
#endif 
		if(inseg.segment_duration <= recent_blip.duration) {
//#ifdef DEBUG_DEMOD_DETAIL
			ulog_str_num_BASE_str_L("FIN: remaining duration = ",
				inseg.segment_duration - recent_blip.duration, DEC, "\n");
//#endif
			break;
		}
		inseg.segment_duration -= recent_blip.duration;
	}
	
	
	ti->total_dBms += total_dBms;
	
	
	int num_bittimes = i + 1;
	// temporary value at the end which will in the end hold some waste
	bittime[i+1] = 0;
	
	
	// set rssi
	// (Datasheet says this power to be sampled for groups of two bits which we
	// try to approximate here. We don't know the exact hardware averaging
	// method.)
	float pwr_cum = rec_pwr_1;
	uint32_t pwr_duration = 2*ti->bit_length;
	if (pwr_duration > segment_duration)
		pwr_duration = segment_duration;
	if (num_bittimes > 1) {
		if (bittime[i] < pwr_duration) {
			uint32_t bittime_last_but_one = bittime[i - 1];
			if (bittime_last_but_one + bittime[i] > pwr_duration)
				bittime_last_but_one = pwr_duration - bittime[i];
			
			pwr_cum = rec_pwr_1 * bittime[i] + rec_pwr_2 * bittime_last_but_one;
			pwr_cum /= (bittime[i] + bittime_last_but_one);
		}
	}
	
	float rssi = 10.0 * log10f(pwr_cum);
	ti->rssi_queued = 0.0 - 2.0 * rssi; // Format in which this is stored.
	ti->rssi_queue_time = get_simtime(ti->station);
	
	
#ifdef DEBUG_DEMOD
	ulog_str_num_BASE_str_L("num_bittimes=", num_bittimes, DEC, ".\n");
#endif
#ifdef DEBUG_DEMOD_DETAIL
	for (i = 0; i < num_bittimes; i++) {
		printf("bittime[%3d] = %8d\n", i, bittime[i]);
	}
#endif
	
	// STEP 2: Synchronize
	
	
	
	// Rate matching must be better than 6.5% (datasheet, p. 31)
	uint32_t blength_min;
	uint32_t blength_max;
	// Condition for numerical stable way:
	//  .935 =  935/1000 = 187/200
	// 1.065 = 1065/1000 = 213/200
	if (ti->bit_length < 20000000) {
		// No calculation loss.
		blength_min = (ti->bit_length * 187)/200;
		blength_max = (ti->bit_length * 213)/200;
	}
	else {
		// This will NEVER happen because of bitrate > 488 Hz
		// No lower bitrate possible due to 16 bit register.
		// Tiny calculation loss.
		blength_min = (ti->bit_length / 200) * 187;
		blength_max = (ti->bit_length / 200) * 213;
	}
	
	
	// Rather a complicated task ... but we'll get it.
	//printf("array_cap=%d.\n", array_cap);
	int bitarr_size = array_cap + 1;
	bool * bitarr = malloc(sizeof(bool) * bitarr_size);
	if (!bitarr) {
		printf("Memory allocation error!\n");
		return;
	}
	
	time_int duration_cumulated = 0;
	
	// Load bit syncronizer state
	uint32_t start = ti->consumed_bittime;
	time_int now = get_simtime(ti->station);
	if (ti->consumed_fintime == now - segment_duration) {
		// Direkter Anschluss
#ifdef DEBUG_DEMOD_DETAIL
		printf("Loading %d.\n", ti->consumed_bitval);
#endif
		if (startbit == ti->consumed_bitval) {
#ifdef DEBUG_DEMOD_DETAIL
			printf("preconsumed: %d\n", ti->consumed_bittime);
#endif
			bittime[0] += ti->consumed_bittime;
			duration_cumulated -= ti->consumed_bittime;
			ti->consumed_bittime = 0;
			start = 0;
		}
	}
	//bitarr[0] = startbit;
	
	
	i = 0;
	int b = 0;//1;
	while (i < num_bittimes) {
#ifdef DEBUG_DEMOD
		ulog_str_num_BASE_str_L("start = ", start, DEC, ".");
		log_str_num_BASE_str_L(" bit_length_sync = ", ti->bit_length_sync, DEC, "\n");
		ulog_str_num_BASE_str_L("bittime[", i, DEC, "");
		log_str_num_BASE_str_L("]=", bittime[i], DEC, "\n");
		//ulog_str_num_BASE_str_L("duration_cumulated = ", duration_cumulated, DEC, ".\n");
#endif
		if (b + 2 >= bitarr_size) {
			bool * tmp = bitarr;
			bitarr = realloc(bitarr, sizeof(bool)*(bitarr_size + array_cap));
			if (!bitarr) {
				printf("Memory allocation error! Skipping some bits!\n");
				bitarr = tmp;
				break;
			}
			bitarr_size += array_cap;
		}
		
		if (bittime[i] + start > blength_max) {
			// No change happened -> two or more equal bits.
			bitarr[b] = (i % 2) ^ startbit;
			b++;
			
			
			bittime[i] -= ti->bit_length_sync - start;
			duration_cumulated += ti->bit_length_sync - start;
#ifdef DEBUG_DEMOD_DETAIL
			ulog_str_num_BASE_str_L("    long bit. Consumed ", ti->bit_length_sync - start, DEC, " .\n");
			ulog_str_num_BASE_str_L("    duration_cumulated=", duration_cumulated, DEC, " \n");
#endif
			
			uint32_t judge_time = blength_max - start;
			if (judge_time > start) {
				// Die Dauer etwas verlängern, da der Umschaltvorgang wohl
				// etwas weiter hinten als erwartet liegt.
				uint32_t distance = blength_max - ti->bit_length_sync;
				distance *= start;
				distance /= judge_time;
				//distance /= 3; // Das wird zu einem munteren Springen führen.
				// TODO: statt /= 3 dieses einsetzen:
				//distance *= start / (F*(blength_max - start));
				// Allerdings vorher numerisch stabil machen.
				// Ziel: kleine Werte von start führen zu einer schnellen
				// perfekten Anpassung, große Werte von start ziehen langsam
				// nach.
				
				ti->bit_length_sync += distance;
				//if (ti->bit_length_sync > blength_max) // Not needed
				//	ti->bit_length_sync = blength_max;
			}
			else { // Analog
				uint32_t distance = ti->bit_length_sync - blength_min;
				distance *= judge_time;
				distance /= start;
				//distance /= 3;
				ti->bit_length_sync -= distance;
			}
			
			start = 0;
			continue;
		}
		if (bittime[i] + start < blength_min) {
			// Change happened too fast -> ignored
			start += bittime[i];
			duration_cumulated += bittime[i];
#ifdef DEBUG_DEMOD_DETAIL
			ulog_str_num_BASE_str_L("    short bit. Consumed ", bittime[i], DEC, " .\n");
			ulog_str_num_BASE_str_L("    duration_cumulated=", duration_cumulated, DEC, " \n");
#endif
			i++;
			continue;
		}
		
		// Normal change in expected time window.
		
		bitarr[b] = (i % 2) ^ startbit;
		b++;
		
		// Don't know what the hardware does. Assume it is stuck to the 
		// old timing for 2/3 part and renews the timing for 1/3 part.
		ti->bit_length_sync = (bittime[i] + start + 2* ti->bit_length_sync) / 3  ;
		duration_cumulated += bittime[i];
#ifdef DEBUG_DEMOD_DETAIL
		ulog_str_num_BASE_str_L("    normal bit. Consumed ", bittime[i], DEC, " .\n");
		ulog_str_num_BASE_str_L("    duration_cumulated=", duration_cumulated, DEC, " \n");
#endif
		i++;
		start = 0;
		
		
		// Probably this is electrically not fully correct as I suppose some
		// capacity to be put in the circuit so that small glitches don't
		// matter.
		// This is a TODO in Step 1.
		
	}
	
	if (duration_cumulated != segment_duration) {
		ulog_str_num_BASE_str_L("WARNING: calculated segment duration of ", duration_cumulated, DEC, " which was not expected.\n");
		ulog_str_num_BASE_str_L("         (expected: ", segment_duration, DEC, " ).\n");
	}
	
	// Save bit syncronizer state
	ti->consumed_bittime = start;
		// This has only a meaning when a short bit was encountered recently.
	ti->consumed_bitval = ((i-1) % 2) ^ startbit;
	ti->consumed_fintime = now; // + duration_cumulated;
	//printf("Saving %d @ %'lu.\n", ti->consumed_bitval, ti->consumed_fintime);
	
	
	free(bittime);
	
	int num_bits = b;
	
	//ulog_str_L("trancv.c:demodulate_data(): step 1 + 2 done.\n");
	
#ifdef DEBUG_SEND_RECEIVE
	ulog_str_L("Received bits: ");
	
	for (i = 0; i < num_bits; i++) {
		if (bitarr[i])
			{log_str_L("1");}
		else
			{log_str_L("0");}
	}
	log_str_L(".\n");
#endif // -ifdef DEBUG_SEND_RECEIVE
	
	
	
	
	
	// STEP 3: Process
	
	
	uint8_t sync_length = sizeof(SYNC_WORD) - 1;
	b = 0;
	
	while(b < num_bits) {
		
		if (ti->state == STATE_STARTUP) {
			ti->reading_tmptime = 0;
			ti->state = STATE_PREAMBLE;
#ifdef DEBUG_DEMOD
			printf ("-> PREAMBLE.\n");
#endif
		}
		// END of if for direct go further
		
		if (ti->state == STATE_PREAMBLE) {
			// reading_tmptime denotes the count (! no time!) how many 0s/1s have
			// been read.
			
			if (ti->reading_tmptime != 0) {
				if (ti->last_bit == bitarr[b]) {
					// Two times the same bit.
					ti->reading_tmptime = 0;
					b++;
					continue;
				}
			}
			fflush(stdout);
			ti->last_bit = bitarr[b];
			b++;
			ti->reading_tmptime ++;
			
			// TODO: Magic number 12!
			if (ti->reading_tmptime >= 12) {
				ti->state = STATE_SYNC;
#ifdef DEBUG_DEMOD
				printf ("-> SYNC.\n");
#endif
				ti->reading_tmptime = 0;
			}
		}
		else if (ti->state == STATE_SYNC) {
			
			//ulog_str_num_BASE_str_L("reading_tmptime = ", ti->reading_tmptime, DEC, "\n");
			
			ti->reading_buffer[ti->reading_tmptime] = bitarr[b];
			ti->reading_tmptime++;
			b++;
			
			if (ti->reading_tmptime >= sync_length*8) {
				// Mathematically the most similar thing to convolution
				// (Faltung) in this project.
				
				bool wrong = false;
				
				for (i=0; i < sync_length*8; i++) {
					bool searching_sequence = (SYNC_WORD[i/8] >> (7-(i&7)))&1;
					//printf("ist: %d, soll: %d\n", ti->reading_buffer[i + ti->reading_tmptime - sync_length*8], (uint32_t)searching_sequence);
					if (ti->reading_buffer[i + ti->reading_tmptime - sync_length*8] != searching_sequence) {
						wrong = true;
						break;
					}
				}
				//printf("\n");
				if (wrong == false) {
#ifdef DEBUG_SEND_RECEIVE
					printf("Sync word found @bit %d - %d.\n",
							ti->reading_tmptime - sync_length*8,
							ti->reading_tmptime);
#endif
					
					if (REGFILE_SILENT_READ(ti, cbPacketFormat) ==
								PACKET_FORMAT_FIXEDlength) {
						// TODO: make this once?
						ti->packet_length
								= REGFILE_SILENT_READ(ti, cbPayloadLength);
						ti->state = STATE_DATA;
#ifdef DEBUG_DEMOD
						printf ("-> DATA.\n");
#endif
					}
					else {
						ti->state = STATE_LEN;
#ifdef DEBUG_DEMOD
						printf ("-> LEN.\n");
#endif
					}
					
					ti->reading_tmptime = 0;
				}
			}
			
			if (ti->reading_tmptime >= sync_length*16){
				for (i=0; i < sync_length*8; i++) {
					ti->reading_buffer[i] = ti->reading_buffer[i + sync_length*8];
				}
				ti->reading_tmptime -= sync_length*8;
			}
			
		}
		else if (ti->state == STATE_LEN) {
			ti->reading_buffer[ti->reading_tmptime] = bitarr[b];
			b++;
			ti->reading_tmptime++;
			
			if (ti->reading_tmptime == 8) {
				ti->packet_length = 0;
				for (i = 0; i < ti->reading_tmptime; i++) {
					ti->packet_length <<= 1;
					ti->packet_length |= ti->reading_buffer[i] ? 1 : 0;
				}
				// p. 54: The PayloadLength register is set to a value which
				// is greater than the maximum expected length of the received
				// packet.
				REGFILE_SILENT_WRITE(TI_USES, cbPayloadLength, ti->packet_length + 2);
				put_on_fifo(ti, ti->packet_length);
				//printf("len = %d.\n", ti->packet_length);
				ti->state = STATE_DATA;
#ifdef DEBUG_DEMOD
				printf ("-> DATA.\n");
#endif
				ti->reading_tmptime = 0;
				ti->data_count = 0;
				
				if (TI_USE always_receive) {
					ti->fifo_ptr = 0;
				}
			}
			assert(ti->reading_tmptime <= 8);
			
		}
		else if (ti->state == STATE_DATA) {
			// put it in the fifo.
			// reading_buffer is the "shift register" (datasheet p. 45)
			
			ti->reading_buffer[ti->reading_tmptime] = bitarr[b];
			b++;
			ti->reading_tmptime++;
			
			if (ti->reading_tmptime == 8) {
				uint8_t data = 0;
				for (i = 0; i < ti->reading_tmptime; i++) {
					data <<= 1;
					data |= ti->reading_buffer[i] ? 1 : 0;
				}
#ifdef DEBUG_DEMOD_DETAIL
				printf("data byte = %d, old FIFO size = %d.\n", data, ti->fifo_ptr);
#endif
				
				put_on_fifo(ti, data);
				
				
				ti->data_count++;
				if (ti->data_count == ti->packet_length) {
					ti->data_count = 0;
					
#ifdef DEBUG_SEND_RECEIVE
					ulog_str_num_BASE_str_L("RFM69: Packet received. "
							"BLOCK OF ~", ti->packet_length, DEC, " bytes:\n");
					for (i = ti->fifo_ptr - 1; i >= 0; i--) {
						//log_str_N(&(ti->fifo[i]), 1);
						log_num_BASE(ti->fifo[i], HEXBYTE);
						log_str_N(" ", 1);
					}
					// 2x \n for that the block doesn't get > appended;
					log_str_L("\n\n");
#endif
					
					
					REGFILE_SILENT_WRITE(TI_USES, cbPayloadReady, 1);
					
					// TODO: Receive and discard further bytes. If it was to be
					// correct: p. 54 "If the received length is greater than
					// the maximum length stored in PayloadLength register the
					// packet is discarded otherwise the complete packet is
					// received."
					ti->state = STATE_STARTUP;
#ifdef DEBUG_DEMOD
					printf ("-> STARTUP.\n");
#endif
					// No CRC implemented yet.
				}
				ti->reading_tmptime = 0;
			}
			
		}
		
	}
	
	free(bitarr);
	
}


static void put_on_fifo(struct trancv_instance * ti, uint8_t data) {
	
	bool overrun = false;
	assert (ti->fifo_ptr <= FIFO_SIZE);
	if (ti->fifo_ptr == FIFO_SIZE)
		overrun = true;
	
	// put it on the fifo.
	//for (i = FIFO_SIZE - 1; i > 0; i--) {
	for (int i = ti->fifo_ptr; i > 0; i--) {
		ti->fifo[i] = ti->fifo[i-1];
	}
	ti->fifo[0] = data;
	ti->fifo_ptr++;
	
	if (overrun)
		RFM69_fifo_overrun(ti);
	else
		RFM69_update_fifo_interrupts(ti);
	
}
