

// © 2018 Jonas Bechtel
// Licensed under the Apache License, Version 2.0; see file <prj>/meta/LICENSE



#ifndef DIRECT_FUNCTION_INCLUDE
#error You cannot include or compile this. It belongs to trancv.c.
#endif



// Do the actual work ...
// Remark: the bit-precise length of the data will be mapped to a byte
//    precise length. There are only whole bytes transmitted. (Unless
//    some manual sending mode is implemented.)
// Does free the data packet.
static struct wave_sequence* modulate_data_RFM69(struct trancv_instance * ti, struct digital_packet * data) {
	
	
	if (REGFILE_SILENT_READ(ti, cbMode) != RFM_MODE_TX) {
		printf("modulate_data_RFM69: WARNING: don't modulate the data because not in TX mode!\n");
		return NULL;
	}
	
	
	
	
	// 5.5.2.2. Variable Length Packet Format
	// Bits to be transferred:
	// * 0 to 65535 bytes:  Preamble (1010...)    -> Choose 2 (very short?)
	// * 0 to 8 bytes:      Sync Word             -> Choose 2: 'CN'
	// *                    Length Byte
	// * 0/1 bytes          Optional Address Byte -> choose 0
	// * 0 to 255 bytes:    Message
	// * 0/2 bytes          Optional CRC checksum -> choose 0 (TODO: Makes sense?)
	
	// TODO: Don't choose this here but read it from regfile!!
	uint16_t preamble_length = 2;
	uint8_t sync_length = sizeof(SYNC_WORD) - 1;
	uint8_t addr_length = 0;
	uint8_t data_length = (data->length + 7) / 8; // data->length is in bits
	
	// Don't use to much preamble.
	// Otherwise the length variables would overflow later.
	assert (preamble_length < 7500); 
	
	
	uint16_t extra_length = 8*preamble_length + 8*sync_length + 8 + 8*addr_length + 8*data_length + 0;
#ifdef DEBUG_SEND_RECEIVE
	printf("extra_length = %u\n", extra_length);
#endif
	struct wave_sequence * waves = malloc(sizeof (struct wave_sequence) + extra_length*sizeof(struct waveblip));
	if (waves == NULL) {
		printf("modulate_data_RFM69: ERROR: Couldn't allocate memory "
				"(pre/sync/data %d/%d/%d)\n",
				preamble_length, sync_length, data_length);
		ti->station->error(ti->station, "RFM69: Coudln't allocate memory.");
		return NULL;
	}
	//printf("modul: waves = %p \n", waves);
	waves->refcount = 0;
	waves->capacity = extra_length + 1;
	
	
	uint32_t bit_duration = get_bit_length_silent(TI_USES);
	//printf("bit_duration = %u\n", bit_duration);
	uint8_t phase = 0; // TODO: random?
	float_t amplitude = 10; // 10 mW
	
	uint32_t f0 = get_frequency_silent(ti) - get_deviation_silent(ti);
	uint32_t f1 = get_frequency_silent(ti) + get_deviation_silent(ti);
	
	
	
	// TODO: DC-free coding actually active? (Manchester coding or data whitening)
	
	// What about too many equal bits which could confuse the bit synchronizer)
	
	int i;
	
	for (i = 0; i < preamble_length * 8; i += 2) {
		waves->seq[i].amplitude = amplitude;
		waves->seq[i].duration = bit_duration;
		waves->seq[i].frequency = f1;
		waves->seq[i].phase = phase; // TODO: random per waveblip?
		waves->seq[i+1].amplitude = amplitude;
		waves->seq[i+1].duration = bit_duration;
		waves->seq[i+1].frequency = f0;
		waves->seq[i+1].phase = phase; // TODO: random per waveblip?
	}
	
/*#define MODSYNC 2
#define MODLENGTH 3
#define MODADDR 4
#define MODDATA 5
#define MODREST 6 // Flushing the last bits.*/
	
	uint8_t state = STATE_SYNC;
	uint8_t pos = 0;
	uint16_t tmp = 0x1234;
	uint8_t loaded = 0;
	
#define BITSWAP(in) \
		(	((in&0x80) >> 7) | \
			((in&0x40) >> 5) | \
			((in&0x20) >> 3) | \
			((in&0x10) >> 1) | \
			((in&0x08) << 1) | \
			((in&0x04) << 3) | \
			((in&0x02) << 5) | \
			((in&0x01) << 7)   )
	for (; i < waves->capacity; i++) {
		// When breaking this loop here please increase i.
		
		// At first load the bits
		if (loaded < 9 && state != STATE_TOOMUCH) {
			tmp &= 0xffff - (0xff << loaded);
			switch (state) {
			case STATE_SYNC:
				tmp |= BITSWAP(SYNC_WORD[pos]) << loaded; loaded += 8;
				//printf("Sync: 0x%02x/'%c'\n", SYNC_WORD[pos], SYNC_WORD[pos]);
				pos ++;
				if (pos >= sync_length) {
					pos = 0;
					if (REGFILE_SILENT_READ(TI_USES, cbPacketFormat) == PACKET_FORMAT_FIXEDlength) {
						if (addr_length)
							state = STATE_ADDR;
						else 
							state = STATE_DATA;
					}
					else {
						state = STATE_LEN;
					}
				}
				break;
			case STATE_LEN:
				tmp |= BITSWAP(data_length) << loaded; loaded += 8;
				//printf("data_length = 0x%02x\n", data_length);
				if (addr_length)
					state = STATE_ADDR;
				else
					state = STATE_DATA;
				break;
			case STATE_DATA:
				if (pos >= data->capacity || pos >= data_length) {
					state = STATE_TOOMUCH; // Diesen Zustand müssten wir gar nicht zuweisen.
					break;
				}
				// Uncomment to find unitialized variables.
				/*fprintf(stderr, "%d ", pos);
				fprintf(stderr, "%d ", data->capacity);
				fprintf(stderr, "%d ", data->length);
				fprintf(stderr, "%p ", data->data);
				fprintf(stderr, "%d\n", data->data[pos]);
				printf("data[%d] = 0x%02x/'%c' (cap %d, len %d)\n", pos,
					data->data[pos], data->data[pos], 
					data->capacity, data_length);*/
				tmp |= BITSWAP(data->data[pos]) << loaded; loaded += 8;
				pos++;
				break;
			case STATE_ADDR:
				printf("trancv.c: Error: Address byte not implemented.\n");
			case STATE_TOOMUCH:
			default:
				printf("trancv.c: Unknown byte load state. This cann't happen.\n");
				ti->station->error(ti->station, "Error when modulating.");
				// Will quit very soon as only up to 16 bits are stored temporarily.
				state = STATE_TOOMUCH;
			}
			
		}
		
		// Modulate bits, part 1: how many bits are equal?
		uint8_t count = 1;
		uint8_t bit = tmp & 1;
//		printf("tmp = 0x%04x (loaded: %d)\n", tmp, loaded);
		if ((tmp & 0xff) == (bit?0xff:0)) {
			count = 8;
			// TODO: Check for count > 8 (Maybe a performance improvement?)
		}
		else if((tmp & 0xf) == (bit?0xf:0)) {
			count = 4;
			if ((tmp & 0x30) == (bit?0x30:0)) {
				count = 6;
				if ((tmp & 0x40) == (bit?0x40:0))
					count = 7;
			}
			else if ((tmp & 0x10) == (bit?0x10:0))
				count = 5;
		}
		else if ((tmp & 0x3) == (bit?0x3:0)) {
			count = 2;
			if ((tmp & 0x7) == (bit?0x7:0)) {
				count = 3;
			}
		}
		
		if (count >= loaded) {
			count = loaded;
		}
//		printf("Bit %d for %d times.\n", bit, count);
		loaded -= count;
		// So finally we loaded bits from the right (LSB)
		tmp >>= count; // Move the remaining bits to LSB.
		// [Remark: as bits were swapped in fact the LSB will be sent out
		//   first as it is indicated by datasheet, too.]
		
		//printf("iiiii: i=%d\n", i);
		waves->seq[i].amplitude = amplitude;
		waves->seq[i].duration = bit_duration*count;
		waves->seq[i].frequency = bit?f1:f0;
		waves->seq[i].phase = phase; // TODO: random per waveblip?
		
		if (!loaded) {
			i++;
			break;
		}
		
		// When breaking this loop here please increase i.
	}
	
	free (data);
	
	if (i < waves->capacity) {
		waves->seq[i].amplitude = NAN;
	}
	
	waves->duration = extra_length * bit_duration;
	//printf("waves->duration = %'lu\n", waves->duration);
	waves->skip = 0;
	waves->refcount = 1;
	
	return waves;
}

