

// © 2018 Jonas Bechtel
// Licensed under the Apache License, Version 2.0; see file <prj>/meta/LICENSE




// There are two modes of this file:
// A) Simulator mode. The file will be compiled normally, defining
//    an RFM69transv object. -> #ifdef SIMUL
// B) "Real mode". This file will be compiled for AVR, with no debugging
//    output and no object orientation. (UPDATE: or it will be compiled
//    for ARM, so it can be used on BeagleBone) -> #ifdef BEAGLE/ARDU

#ifndef TRANCV_C
#define TRANCV_C

#ifdef __cplusplus
extern "C" {
#endif

#if defined (BEAGLE) || defined (ARDU)



#define TI_ARG
#define TI_ARGS 
#define TI_USE
#define TI_USE2
#define TI_USES

#include <stdint.h>

#ifndef NOCODE
uint8_t regfile_read(uint8_t addr);
uint8_t regfile_write(uint8_t addr, uint8_t data); // TODO: You can't set single bits without reading the other bits?
#include "RFM69-reg.h"
#include "RFM69-reghelper.h"
#endif // ifndef NOCODE

#else // -if defined BEAGLE || defined ARDU

#include <stdio.h>
#include <assert.h>
#define TI_ARGS struct trancv_instance* ti
#define TI_ARG  TI_ARGS ,
#define TI_USE  ti->
#define TI_USE2 ti,
#define TI_USES ti

#define SILENT_REG_ACCESS
#ifndef NOCODE
#define DEFINE_OBJECT // DEFINE_OBJECT is to be included only once in 
//    all routing algorithm's code.
#include "RFM69intern.h"
#include "RFM69-reghelper.h"
#undef DEFINE_OBJECT
#else
#include "common/trancv-user.h"
#include "RFM69-reg.h"
#include "RFM69-reghelper.h"
#include "subobject.h"
#endif

#ifndef NOCODE
#define NOCODE
#include "multihw.c"
#undef NOCODE
#else
#include "multihw.c"
#endif

#endif // -if defined BEAGLE || defined ARDU  -else

#ifndef NOCODE
#include "pindef.h"
uint8_t regfile_write_tmp; // requirement of RFM69-reg.h
#endif

#ifndef NOCODE
#if defined BEAGLE || defined ARDU
static inline void warn_bitrate_fdev(TI_ARG uint32_t bitrate, uint32_t fdev) {}
#else // if defined BEAGLE || defined ARDU
static void warn_bitrate_fdev(TI_ARG uint32_t bitrate, uint32_t fdev) {
	if (fdev + bitrate/2 > 500000) {
		printf("trancv.c: WARNING: assertion \"F_dev + Bitrate/2 <= 500 kHz!\" from\n"
		       "                   Datasheet not fulfilled!\n");
		ti->fdev_warned = true;
	}
	else if (ti->fdev_warned){
		printf("trancv.c: (Newly) Fdev and Bitrate OK.\n");
		ti->fdev_warned = false;
	}
	if (fdev < 600) {
		printf("trancv.c: WARNING: fdev too low. Datasheet, p. 20: deviation must exceed 600Hz\n");
	}
}
#endif // if defined BEAGLE || defined ARDU  #else
#endif // -ifndef NOCODE

uint32_t get_deviation(TI_ARGS);




#ifndef NOCODE
static uint32_t get_bitrate_universal(TI_ARG   bool silent)
{
	
	uint32_t BitRate;
#ifdef SIMUL
	if (silent) {
		BitRate =       REGFILE_SILENT_READ(TI_USES, cbBitrateLsb)
		            + ((REGFILE_SILENT_READ(TI_USES, cbBitrateMsb))<<8);
	}
	else {
#else
	if(true) {
#endif
		BitRate =       REGFILE_DO_READ(TI_USES, cbBitrateLsb)
		            + ((REGFILE_DO_READ(TI_USES, cbBitrateMsb))<<8);
	}
	if (BitRate == 0) // Gleitkomma-Ausnahme
		BitRate = 1;
	uint32_t bitps = FXOSC/BitRate;
	if (FXOSC - BitRate*bitps > BitRate*(bitps+1) - FXOSC)
		bitps++;
	
	return bitps;
}
#endif
;

uint32_t get_bitrate(TI_ARGS) 
#ifndef NOCODE
{
	return get_bitrate_universal(TI_USE2   false);
}
#endif
;

uint32_t get_bitrate_silent(TI_ARGS) 
#ifndef NOCODE
{
	return get_bitrate_universal(TI_USE2   true);
}
#endif
;


void set_bitrate(TI_ARG   uint32_t bitps)
#ifndef NOCODE
{
	
	uint32_t BitRate = FXOSC/bitps;
	if (FXOSC - BitRate*bitps > (BitRate+1)*bitps - FXOSC)
		BitRate++;
	
	// Setting Lsb as last (because this may be similar to frequency although
	// it is not noted on datasheet.)
	REGFILE_DO_WRITE(TI_USES, cbBitrateMsb, (BitRate&0xff00)>>8);
	REGFILE_DO_WRITE(TI_USES, cbBitrateLsb,  BitRate&0xff      );
//	printf("trancv.c: ti->regfile[%d,%d] = 0x %02x %02x\n",
//			RegBitrateMsb, RegBitrateLsb,
//			ti->regfile[RegBitrateMsb], ti->regfile[RegBitrateLsb]);
	
	warn_bitrate_fdev(TI_USE2 bitps, get_deviation(TI_USES));
	
}
#endif // -ifndef NOCODE
;

uint32_t get_bit_length(TI_ARGS)
#ifndef NOCODE
{
	// Expected Datarate: (in FXOSC units)
	uint32_t bit_value =   REGFILE_DO_READ(TI_USES, cbBitrateLsb)
	                   + ((REGFILE_DO_READ(TI_USES, cbBitrateMsb))<<8);
	
	// We get artificial: bit_value = FXOSC/bitrate = 32000000Hz/(bitrate bit/s)
	//                    [With FXOSC obviously with no unit -> scientific error
	//                     in datasheet. XX = no unit (normally Hz)]
	//           so    bitrate = 32000000Hz / bit_value *bit
	//                    
	// We need bit_length  =  1000000000 ns/s / bitrate        | bitrate = [...]
	//                     =  1000000000 ns/s / 32000000Hz * bit_value /bit
	//                     =  1000000000 / 32000000 * bit_value *ns /bit
	//                     =  1000/32 * bit_value *ns/bit
	
	return (125 * bit_value) / 4; // in ns
}
#endif
;

#ifdef SIMUL
uint32_t get_bit_length_silent(TI_ARGS)
#ifndef NOCODE
{
	uint32_t bit_value =   REGFILE_SILENT_READ(TI_USES, cbBitrateLsb)
	                   + ((REGFILE_SILENT_READ(TI_USES, cbBitrateMsb))<<8);
	return (125 * bit_value) / 4; // Calculation: see comments above
}
#endif
;
#endif

void set_frequency(TI_ARG uint32_t Hz)
#ifndef NOCODE
{
	// Carrier frequency F_RF = Fstep * Frf(23:0)
	
	// TODO: No TX/RX operation when setting frequency!
	
	uint32_t Frf = Hz/FSTEP;
	if (Hz - Frf*FSTEP > (Frf+1)*FSTEP - Hz)
		Frf++;
	
	// SET LSB AS VERY LAST VALUE! (Datasheet p. 17 Note)
	REGFILE_DO_WRITE(TI_USES, cbFrfMid, (Frf&0xff00  )>> 8);
	REGFILE_DO_WRITE(TI_USES, cbFrfMsb, (Frf&0xff0000)>>16);
	REGFILE_DO_WRITE(TI_USES, cbFrfLsb,  Frf&0xff         );
//	printf("trancv.c: ti->regfile[%d,%d,%d] = 0x %02x %02x %02x\n",
//			RegFrfMsb, RegFrfMid, RegFrfLsb,
//			ti->regfile[RegFrfMsb], ti->regfile[RegFrfMid],
//			ti->regfile[RegFrfLsb]);
}
#endif
;


uint32_t get_frequency(TI_ARGS)
#ifndef NOCODE
{
	uint32_t Frf =      REGFILE_DO_READ(TI_USES, cbFrfLsb)
	                + ((REGFILE_DO_READ(TI_USES, cbFrfMid))<< 8)
	                + ((REGFILE_DO_READ(TI_USES, cbFrfMsb))<<16);
	
	return Frf*FSTEP;
}
#endif
;

#ifdef SIMUL
#ifndef NOCODE
static uint32_t get_frequency_silent(TI_ARGS)
{
	uint32_t Frf =      REGFILE_SILENT_READ(TI_USES, cbFrfLsb)
	                + ((REGFILE_SILENT_READ(TI_USES, cbFrfMid))<< 8)
	                + ((REGFILE_SILENT_READ(TI_USES, cbFrfMsb))<<16);
	
	return Frf*FSTEP;
}
;
#endif // -ifndef NOCODE
#endif // -ifdef SIMUL


void set_deviation(TI_ARG uint32_t Hz)
#ifndef NOCODE
{
	// Frequency deviation F_dev = Fstep * Fdev(13:0)
	// F_dev + Bitrate/2 <= 500 kHz!
	
	uint32_t Fdev = Hz/FSTEP;
	if (Hz - Fdev*FSTEP > (Fdev+1)*FSTEP - Hz)
		Fdev++;
	
	
	// SET LSB AS LAST VALUE!
	REGFILE_DO_WRITE(TI_USES, cbFdevMsb, (Fdev&0x3f00)>>8);
	REGFILE_DO_WRITE(TI_USES, cbFdevLsb,  Fdev&0xff      );
//	printf("trancv.c: ti->regfile[%d,%d] = 0x %02x %02x\n",
//			RegFdevMsb, RegFdevLsb,
//			ti->regfile[RegFdevMsb], ti->regfile[RegFdevLsb]);
	
	warn_bitrate_fdev(TI_USE2 Fdev*FSTEP, get_bitrate(TI_USES));
}
#endif
;

uint32_t get_deviation(TI_ARGS)
#ifndef NOCODE
{
	uint32_t Fdev =      REGFILE_DO_READ(TI_USES, cbFdevLsb)
	                 + ((REGFILE_DO_READ(TI_USES, cbFdevMsb))<<8);
	return Fdev*FSTEP;
}
#endif
;

#ifdef SIMUL
#ifndef NOCODE
static uint32_t get_deviation_silent(TI_ARGS)
{
	uint32_t Fdev =      REGFILE_SILENT_READ(TI_USES, cbFdevLsb)
	                 + ((REGFILE_SILENT_READ(TI_USES, cbFdevMsb))<<8);
	return Fdev*FSTEP;
}
#endif // -ifndef NOCODE
#endif // -ifdef SIMUL


#ifdef SIMUL


void set_frequency_mistake(struct trancv_instance * ti, uint16_t Hz)
#ifndef NOCODE
{
	ti->frequency_mistake = Hz;
}
#endif
;
uint16_t get_frequency_mistake(struct trancv_instance * ti)
#ifndef NOCODE
{
	return ti->frequency_mistake;
}
#endif
;




#ifndef NOCODE

static void RFM69_fifo_overrun(TI_ARGS);
static void RFM69_update_fifo_interrupts(TI_ARGS);
#define DIRECT_FUNCTION_INCLUDE
#include "modul.c"
#include "demod.c"


// Checks for Fifo Level.
static void RFM69_dispatch_waves(struct trancv_instance * ti) {
	struct station * mh_station_arg = ti->station;
			// Needed for ulog, get_simtime.
	
	if (REGFILE_SILENT_READ(ti, cbMode) != RFM_MODE_TX) {
		return;
	}
	/*printf ("Start condition = %d, Threshold = %d.\n", 
		REGFILE_SILENT_READ(TI_USES, cbTxStartCondition),
		REGFILE_SILENT_READ(TI_USES, cbFifoThreshold)
	);*/
	
	if(REGFILE_SILENT_READ(TI_USES, cbTxStartCondition) == COND_FifoNotEmpty) {
		if (!ti->fifo_ptr)
			return;
	}
	
	if(REGFILE_SILENT_READ(TI_USES, cbTxStartCondition) == COND_FifoThreshold) {
		if (ti->fifo_ptr <= (REGFILE_SILENT_READ(TI_USES, cbFifoThreshold)))
			return;
	}
	
	// TODO: It is wasteful to pack digital_packet here and unpack it in
	// modulate_data(). This packaging should be done by modulate_data()
	// inherently.
	int capacity = ti->fifo_ptr;
	//fprintf(stderr, "capacity=%d.\n", capacity);
	struct digital_packet* dp = malloc(sizeof(struct digital_packet) + capacity);
	if (dp == NULL) {
		ulog_str_L("trancv:dispatch(): ERROR: couldn't malloc memory.\n");
		fprintf(stderr, "trancv:dispatch(): ERROR: couldn't malloc memory.\n");
		quit_with_error(MHI_USE_arg sharedERRNO_MALLOC);
	}
	
	// Fill in the fields.
	dp->starttime = get_simtime(MHI_USE_single);
	dp->capacity = capacity;
	
	int bias = 0;
	int length;
	if (REGFILE_SILENT_READ(ti, cbPacketFormat) == PACKET_FORMAT_FIXEDlength) {
		length = REGFILE_SILENT_READ(ti, cbPayloadLength);
	}
	else {
		length = ti->fifo[0];
		bias = 1;
	}
	if (ti->fifo_ptr < length + bias) {
		ulog_str_L("WARNING: You told me to send an packet which hasn't\n");
		ulog_str_L("         all bytes on FIFO; this is probably wrong!\n");
		ulog_str_L("         (Anyway - it won't work in this simulator)\n");
		
		ulog_str_num_BASE_str_L("(Requested length was ", length, DEC, ".)\n");
		if (bias) {
			ulog_str_num_BASE_str_L("(Have 1+", ti->fifo_ptr-1, DEC, " Bytes on FIFO.)\n");
		}
		else {
			ulog_str_num_BASE_str_L("(Have ", ti->fifo_ptr, DEC, " Bytes on FIFO.)\n");
		}
		
		length = ti->fifo_ptr - bias; // sanity II
	}
	
	dp->length = length * 8;
	
	
	//fprintf(stderr, "trancv.c: fifo_ptr = %d, length=%d, bias=%d\n", ti->fifo_ptr, length, bias);
	for (int i = 0; i < length && i + bias < ti->fifo_ptr; i++) {
		dp->data[i] = ti->fifo[i + bias];
	}
	for (int i = 0; i + length + bias < ti->fifo_ptr; i++) {
		ti->fifo[i] = ti->fifo[i + length + bias];
	}
	//fprintf(stderr, "trancv.c: fifo_ptr = %d, length=%d, bias=%d\n", ti->fifo_ptr, length, bias);
	ti->fifo_ptr -= length + bias;
	
	struct wave_sequence * waves = modulate_data_RFM69(ti, dp);
	if (waves == NULL) {
		ulog_str_L("ERROR: modulate_data_RFM69() failed!\n");
		fprintf(stderr, "ERROR: modulate_data_RFM69() failed!\n");
		quit_with_error(MHI_USE_arg   sharedERRNO_MALLOC);
		return;
	}
	
	waves->refcount = 1;
	
	//REGFILE_SILENT_WRITE(TI_USES, cbPacketSent, 1); // No, this is deferred.
	//printf("waves->duration = %'lu.\n", waves->duration);
	TI_USE packet_sent_at = get_systime(MHI_USE_single) + waves->duration;
	
	emit_wave_segment_to_air(MHI_USE_arg   waves, 0); // Will free waves.
	
	ti->state = STATE_STARTUP;
	
	
	//printf("waves->duration = %'lu,     systime = %'lu.\n", waves->duration,
	//	get_systime(MHI_USE_single));
#ifdef DEBUG_TRANCV
	ulog_str_num_BASE_str_L("trancv.c: packet_sent_at = ", TI_USE packet_sent_at, DEC, "\n");
#endif
	
#ifndef WE_CAN_SEND_MORE_THAN_1_PACKET_AT_ONCE
	// TODO: We can not send more than 1 packet for each fifo change.
	// So application has to reset and set the triggers.
	if (ti->fifo_ptr == 0)
#endif
		RFM69_update_fifo_interrupts(TI_USES);
}


static void RFM69_update_fifo_interrupts(struct trancv_instance * ti) {
#if (defined SIMUL) && (defined DEBUG_FIFO)
	struct station * mh_station_arg = ti->station; // Needed for ulog
#endif
	
	assert (ti->fifo_ptr <= FIFO_SIZE);
	if (ti->fifo_ptr == FIFO_SIZE) {
		REGFILE_SILENT_WRITE(TI_USES, cbFifoFull, 1);
	}
	else {
		REGFILE_SILENT_WRITE(TI_USES, cbFifoFull, 0);
	}
	
	if (ti->fifo_ptr) {
		REGFILE_SILENT_WRITE(TI_USES, cbFifoNotEmpty, 1);
	}
	else {
		REGFILE_SILENT_WRITE(TI_USES, cbFifoNotEmpty, 0);
		// Clear PayloadReady interrupt
		REGFILE_SILENT_WRITE(TI_USES, cbPayloadReady, 0);
	}
	
	uint8_t level = REGFILE_SILENT_READ(TI_USES, cbFifoThreshold);
	if (ti->fifo_ptr >= level) {
		REGFILE_SILENT_WRITE(TI_USES, cbFifoLevel, 1);
	}
	else {
		REGFILE_SILENT_WRITE(TI_USES, cbFifoLevel, 0);
	}
	
#ifdef DEBUG_FIFO
#ifdef SIMUL
	ulog_str_num_BASE_str_L("Updated FIFO interrupts. FIFO contains ", ti->fifo_ptr, DEC, " B.\n");
#else
	printf("Updated FIFO interrupts. FIFO contains %d B.\n", ti->fifo_ptr);
#endif
#endif
	
	// This will check for sending need:
	RFM69_dispatch_waves(TI_USES);
	
}



static void RFM69_fifo_overrun(struct trancv_instance * ti) {
	
	// Really force it (as trigger would catch it)
	ti->fifo[regfile_name(cbFifoOverrun)] |= 1 << regfile_bitshift(cbFifoOverrun);
	
	// Datasheet p. 45 states "Data is lost"
	// Datasheet p. 70 states "Flags and FIFO are cleared"
	ti->fifo_ptr = 0;
	
	RFM69_update_fifo_interrupts(ti);
}

#endif // -ifndef NOCODE

uint8_t simulator_read_fifo(struct trancv_instance * ti)
#ifndef NOCODE
{
#ifdef DEBUG_FIFO
	struct station * mh_station_arg = ti->station; // Needed for ulog.
#endif
	
	if (!ti->fifo_ptr)
		return ti->fifo[0];
	uint8_t retv = ti->fifo[ti->fifo_ptr-1];
	ti->fifo_ptr--;
	RFM69_update_fifo_interrupts(ti);
#ifdef DEBUG_FIFO
	ulog_str_num_BASE_str_L("RFM69: Fifo read by MCU: 0x", retv, HEXBYTE, "\n");
#endif
	return retv;
}
#endif
;

void simulator_write_fifo(struct trancv_instance* ti, uint8_t data)
#ifndef NOCODE
{
	struct station * mh_station_arg = ti->station; // Needed for ulog.
	
#ifdef DEBUG_FIFO
	ulog_str_num_BASE_str_L("RFM69: Fifo written by MCU: 0x", data, HEXBYTE, "\n");
#endif
	if (ti->fifo_ptr >= FIFO_SIZE) {
		ulog_str_L("RFM69: WARNING: Fifo is full. Rejecting SPI byte.\n");
		
		RFM69_fifo_overrun(TI_USES);
		
		return;
	}
	
	
	ti->fifo[ti->fifo_ptr] = data;
	ti->fifo_ptr++;
	RFM69_update_fifo_interrupts(TI_USES);
	
	return;
}
#endif
;



// DON'T use this! Only for testing.
void always_receive(TI_ARG   bool always_receive)
#ifndef NOCODE
{
	ti->always_receive = always_receive;
}
#endif
;


void RFM69_data_sent(struct trancv_instance* ti) 
#ifndef NOCODE
{
#ifdef DEBUG_SEND_RECEIVE
	printf("%c: Wave fully sent @%'luns.\n", ti->station->name, get_simtime(ti->station));
#endif
}
#endif
;

void RFM69_next_wave_segment(struct trancv_instance* ti, struct wave_seq_segment wss)
#ifndef NOCODE
{
	//struct station * mh_station_arg = ti->station; // uncomment for ulog
	
	if (!ti->always_receive)
		if (REGFILE_SILENT_READ(ti, cbMode) != RFM_MODE_RX) {
			// Hinsichtlich der Zeitspannen ist das nicht so genau - einen
			// Teil hätten wir ja empfangen können.
			// TODO: not so accurate in respect to time spans - we could
			// have received some part of the data.
			
	//		ulog_str_L("RFM69trancv: Not listening.\n");
			
			if (wss.waves)
				// THIS was the memory leak ...
				clean_wave_seq_segment(&wss);
				//unref_wave(wss.waves);
			
			return;
		}
	
	//ulog_str_L("RFM69trancv: Received wave segment.\n");
	//printf("%c: Received wave segment @%'luns.\n", ti->station->name, get_simtime(ti->station));
	
	demodulate_data_RFM69(ti, wss);
	
}
#endif
;

#ifndef NOCODE
// For every bit which has access mode "rwc" in datasheet
// returns true if interrupt bit is cleared.
static bool update_interrupt_c(struct trancv_instance * ti,
			uint32_t cbval, uint8_t changedbits) {
	
	if (REGFILE_SILENT_READ(TI_USES, cbval)) {
		if ((changedbits >> regfile_bitshift(cbval)) & 1) {
			// You cannot set this.
			ti->regfile[regfile_name(cbval)] &= 0xFF-(1>>regfile_bitshift(cbval));
			//REGFILE_FORCE_WRITE(TI_USES, cbval, 0);
		}
		else {
			// Writing 1 on 1 clears the interrupt
			// 		This is the 'c' operation of rwc
			ti->regfile[regfile_name(cbval)] &= 0xFF-(1>>regfile_bitshift(cbval));
			return true;
		}
	}
	else if ((changedbits >> regfile_bitshift(cbval)) & 1) {
		// You cannot reset this by writing 0.
		ti->regfile[regfile_name(cbval)] |= (1>>regfile_bitshift(cbval));
		//REGFILE_FORCE_WRITE(TI_USES, cbval, 1);
	}
	return false;
}
#endif


void RFM69_notify_reg(struct trancv_instance* ti, uint8_t addr,
					uint8_t changedbits)
#ifndef NOCODE
{
	if (changedbits == 0) return;
	
	// Needed for ulog, get_simtime.
	struct station * mh_station_arg = ti->station;
	
	
	
	if (addr == 0) {
		printf("trancv.c: This shall not happen.\n");
		assert(0); // This shall not happen.
	}
	
	
	//ulog_str_num_BASE_str_L("Regfile written: addr=0x", addr, HEXBYTE, ", changed=0x");
	//log_num_BASE(changedbits, HEXBYTE);
	//log_str_num_BASE_str_L(", new=0x", ti->regfile[addr], HEXBYTE, ".\n");
	
	if (addr == regfile_name(cbMode) && 
			changedbits & (regfile_bitmask(cbMode) << regfile_bitshift(cbMode))) {
		uint8_t old_mode = ti->regfile[addr] ^ changedbits;
		old_mode >>= regfile_bitshift(cbMode);
		old_mode &= regfile_bitmask(cbMode);
		
		if (old_mode == RFM_MODE_RX) {
			REGFILE_SILENT_WRITE(ti, cbRssi, 0); // p. 69
			ti->state = STATE_STARTUP;
		}
		
		if (old_mode == RFM_MODE_TX) {
			ti->fifo_ptr = 0; // clear FIFO
			ti->packet_sent_at = 0; // clear PacketSent
			REGFILE_SILENT_WRITE(ti, cbPacketSent, 0);
		}
		
		if (REGFILE_SILENT_READ(ti, cbMode) == RFM_MODE_TX) {
			if (old_mode == RFM_MODE_RX)
				ti->fifo_ptr = 0; // clear FIFO
			
			// TODO: realize a time gap with tx_ready_start
			// and a tranceiver specific event to simulator.
			REGFILE_SILENT_WRITE(ti, cbModeReady, 1);
			REGFILE_SILENT_WRITE(ti, cbRxReady, 0);
			REGFILE_SILENT_WRITE(ti, cbTxReady, 1);
#ifdef DEBUG_TRANCV
			ulog_str_L("trancv: switched to TX mode.\n");
#endif
		}
		else if (REGFILE_SILENT_READ(ti, cbMode) == RFM_MODE_RX) {
			if (old_mode == RFM_MODE_SLEEP || old_mode == RFM_MODE_STANDBY)
				ti->fifo_ptr = 0; // clear FIFO
			
			REGFILE_SILENT_WRITE(ti, cbModeReady, 1);
			REGFILE_SILENT_WRITE(ti, cbRxReady, 1);
			REGFILE_SILENT_WRITE(ti, cbTxReady, 0);
#ifdef DEBUG_TRANCV
			ulog_str_L("trancv: switched to RX mode.\n");
#endif
		}
		
		// More on FIFO: datasheet, p. 46
		RFM69_update_fifo_interrupts(ti);
	}
	
	
	if (addr == regfile_name(cbBitrateLsb) || addr == regfile_name(cbBitrateMsb)) {
		
		uint32_t bit_length = get_bit_length(TI_USES);
		if (bit_length != ti->bit_length) {
#ifdef DEBUG_TRANCV
			ulog_str_L("Changing bitrate.\n");
#endif
			ti->bit_length = bit_length;
			ti->bit_length_sync = bit_length;
			ti->consumed_bittime = 0;
		}
	}
	
	if (addr == regfile_name(cbFifoOverrun)) {
		
		if (update_interrupt_c(TI_USE2   cbFifoOverrun, changedbits)) {
			
			// Datasheet p. 45 states "Data is lost"
			// Datasheet p. 70 states "Flags and FIFO are cleared"
			//// Datasheet p. 45 states that "FIFO will also be cleared":
			for (int i = 0; i < FIFO_SIZE; i++) {
				ti->fifo[i] = 0;
			}
			
			// Code duplication (see above)
			ti->fifo_ptr = 0;
			
			RFM69_update_fifo_interrupts(ti);
		}
	}
	
	if (addr == regfile_name(cbFifoThreshold)) {
		// NOTHING here! "FifoLevel interrupt is updated only after a read or write operation on the FIFO. Thus the interrupt cannot be dynamically updated by only changing the FifoThreshold parameter" (p. 46)
		
		/*if (changedbits) {
			// Always true?
			RFM69_update_fifo_interrupts(ti);
		}*/
	}
	
	
	if (addr == regfile_name(cbRssi)) {
		update_interrupt_c(TI_USE2    cbRssi, changedbits);
	}
	
	if (addr == regfile_name (cbRssiStart)) {
		if (changedbits >> regfile_bitshift(cbRssiStart) & 1) {
			// p. 68: "always reads 0"
			REGFILE_SILENT_WRITE(TI_USES, cbRssiStart, 0); 
			
			// "RssiValue can only be read when it exceeds RssiThreshold" (p.28)
			// Is the rssi "above" threshold (so bit value lower than threshold)
			if (ti->rssi_queued < (REGFILE_SILENT_READ(TI_USES,cbRssiThreshold))){
				
				// Is the value not too ancient?
				//if (ti->rssi_queue_time > get_simtime(MHI_USE_single) - 5* ti->bit_length) {
				if (ti->rssi_queue_time > get_simtime(MHI_USE_single) - 500* ti->bit_length) {
					
					// So write the recent value
					REGFILE_SILENT_WRITE(TI_USES, cbRssi, 1);
					REGFILE_SILENT_WRITE(TI_USES, cbRssiValue, ti->rssi_queued);
				}
			}
			
			// TODO: In reality this would be 2 bits later
			// In simulation we are incredibly fast.
			REGFILE_SILENT_WRITE(TI_USES, cbRssiDone, 1);
		}
	}
	
	
	if (addr == RegFifoThresh) {
		// Do nothing for Threshold
		// As Datasheet p. 46 states "FifoLevel interrupt is updated only after a read or write operation"
		
		if ((changedbits >> regfile_bitshift(cbTxStartCondition)) & 1) {
			RFM69_dispatch_waves(ti);
		}
	}
	
	
	// TODO: There are many possibilities that something happenes.
	// For example on frequency changes the device timing must be recalled in
	// order to have a decent wait till the Pll is locked again etc.
	
	return;
}
#endif // -ifndef NOCODE
;


uint8_t simulator_read_reg(TI_ARG  uint8_t addr)
#ifndef NOCODE
{
	struct station * mh_station_arg = ti->station; // Needed for simtime.
	
	if (addr == 0)
		return simulator_read_fifo(TI_USES);
	
	if (addr == regfile_name(cbPacketSent)) {
		/*printf("trancv.c: packet_sent @%'lu ns, simtime is %'lu ns.\n",
				TI_USE packet_sent_at, get_simtime(MHI_USE_single));*/
		// timed read
		if (   TI_USE packet_sent_at
			&& TI_USE packet_sent_at < get_simtime(MHI_USE_single)) {
			
			REGFILE_SILENT_WRITE(TI_USES, cbPacketSent, 1);
			TI_USE packet_sent_at = 0;
		}
	}
	
	return ti->regfile[addr];
}
#endif
;

void simulator_write_reg(TI_ARG  uint8_t addr, uint8_t data) 
#ifndef NOCODE
{
	if (addr == 0)
		simulator_write_fifo(ti, data);
	else {
		uint8_t change = ti->regfile[addr] ^ data;
		ti->regfile[addr] = data;
		RFM69_notify_reg(ti, addr, change);
	}
}
#endif
;




int start_RFM69trancv(TI_ARGS)
#ifndef NOCODE
{
	ti->modulate_data = modulate_data_RFM69;
	ti->data_sent = RFM69_data_sent;
	ti->next_wave_segment = RFM69_next_wave_segment;
	ti->fdev_warned = false;
	
	// Init regfile
	ti->regfile[RegOpMode] = 0x04;
	ti->regfile[RegDataModul] = 0;
	ti->regfile[RegBitrateMsb] = 0x1A;
	ti->regfile[RegBitrateLsb] = 0x0B;
		ti->bit_length = 0;
		RFM69_notify_reg(TI_USES, RegBitrateMsb, 0xFF);
	ti->regfile[RegFdevMsb] = 0x00;
	ti->regfile[RegFdevLsb] = 0x52;
	ti->regfile[RegFrfMsb] = 0xE4;
	ti->regfile[RegFrfMid] = 0xC0;
	ti->regfile[RegFrfLsb] = 0x00;
//	ti->regfile[RegOsc1] = 0x41;
//	ti->regfile[RegAfcCtrl] = 0;
	
//	ti->regfile[RegListen1] = 0x92;
//	ti->regfile[RegListen2] = 0xF5;
//	ti->regfile[RegListen3] = 0x20;
//	ti->regfile[RegVersion] = 0x24;
	ti->regfile[RegPaLevel] = 0x9F;
//	ti->regfile[RegPaRamp] = 0x09;
	ti->regfile[RegOcp] = 0x1A;
	
	ti->regfile[RegLna] = 0x08;
	ti->regfile[RegRxBw] = 0x86; // Recommended: set to 0x55
//	ti->regfile[RegAfcBw] = 0x8A;
	// TODO, OokPeak, ... , OokFix
	ti->regfile[RegAfcFei] = 0x10;
	ti->regfile[RegAfcLsb] = 0x00;
	ti->regfile[RegAfcMsb] = 0x00;
	// TODO, FeiLsb, ... , RssiConfig
	ti->regfile[RegRssiValue] = 0xFF;
	ti->regfile[RegDioMapping1] = 0x00;
	ti->regfile[RegDioMapping2] = 0x05;
	ti->regfile[RegIrqFlags1] = 0x80;
	ti->regfile[RegIrqFlags2] = 0x00;
	ti->regfile[RegRssiThresh] = 0xFF;
	// TODO
//	ti->regfile[RegPreambleMsb] = 0x00;
//	ti->regfile[RegPreambleLsb] = 0x03;
	ti->regfile[RegSyncConfig] = 0x98;
	ti->regfile[RegSyncValue  ] = 0x00;
	ti->regfile[RegSyncValue+1] = 0x00;
	ti->regfile[RegSyncValue+2] = 0x00;
	ti->regfile[RegSyncValue+3] = 0x00;
	ti->regfile[RegSyncValue+4] = 0x00;
	ti->regfile[RegSyncValue+5] = 0x00;
	ti->regfile[RegSyncValue+6] = 0x00;
	ti->regfile[RegSyncValue+7] = 0x00;
	
	ti->regfile[RegPacketConfig1] = 0x10;
	ti->regfile[RegPayloadLength] = 0x40;
	ti->regfile[RegFifoThresh] = 0x8f;
	// TODO: RegPacketConfig2 , ...
	
	ti->regfile[RegTestLna] = 0x1B;
	ti->regfile[RegTestPa1] = 0x55;
	ti->regfile[RegTestPa2] = 0x70;
	
	// default: 915 MHz, even for 433 MHz version.
	set_frequency(ti, 915000000);  // -> 0xe4c000 in Regfile
	
	set_bitrate(ti, 4800); // Default value  -> 0x1a0b in Regfile
	
	set_deviation(ti, 5000); // Default value  -> 0x0052 in Regfile
	
	ti->state = STATE_STARTUP;
	ti->reading_tmptime = 0;
	//ti->data_count = 0; // will be reset with ti->state change.
	ti->consumed_fintime = TIME_INVALID;
	
	// We will set CONTINUOUS_MODE for measurement and PACKET_MODE for production.
	REGFILE_SILENT_WRITE(ti, cbDataMode, PACKET_MODE);
	
	ti->rssi_queued = 0xFF;
	ti->rssi_queue_time = 0;
	
	
	ti->always_receive = false; // We do respect cbMode.
	
	ti->packet_sent_at = 0;
	
	ti->fifo_ptr = 0;
	
	return OBJECT_START_OK;
}
#endif // -ifndef NOCODE
;

#endif // ifdef SIMUL



#ifdef __cplusplus
}
#endif



#endif // -ifndef TRANCV_C
